
/*
 * File Assertion.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <string.h>

#include "Assertion.hpp"
#include "Tracer.hpp"

#if VDEBUG

#include "Lib/Allocator.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Timer.hpp"
#include "Shell/Options.hpp"

using namespace Lib;
using namespace Debug;

bool Assertion::_violated = false;


void Assertion::violated (const char* file,int line,const char* cond)
{
  if (_violated) {
    return;
  }

  _violated = true;
  reportSpiderFail();
  if(outputAllowed()) {
    cout << "Condition in file " << file << ", line " << line
	<< " violated:\n" << cond << "\n"
	<< "----- stack dump -----\n";
    Tracer::printStack(cout);
    cout << "----- end of stack dump -----" << endl;
  }
} 


void Assertion::violatedStrEquality(const char* file,int line,const char* val1Str,
	  const char* val2Str, const char* val1, const char* val2)
{
  if (_violated) {
    return;
  }

  _violated = true;
  reportSpiderFail();
  if(outputAllowed()) {
    std::cout << "Condition for string equality "<<val1Str<<" == "<<val2Str
	<< " in file " << file << ", line " << line
	<< " was violated, as:\n" << val1Str<<" == \""<<val1 <<"\"\n"
	<< val2Str<<" == \""<<val2 << "\"\n"
	<< "----- stack dump -----\n";
    Tracer::printStack(cout);
    std::cout << "----- end of stack dump -----\n";
  }
}


void Assertion::checkType(const char* file,int line,const void* ptr, const char* assumed,
	const char* ptrStr)
{
  if (_violated) {
    return;
  }

  Allocator::Descriptor* desc = Allocator::Descriptor::find(ptr);

  if(!desc) {
    reportSpiderFail();
    if(outputAllowed()) {
      cout << "Type condition in file " << file << ", line " << line
	  << " violated:\n" << ptrStr << " was not allocated by Lib::Allocator.\n";
    }
  } else if( !USE_PRECISE_CLASS_NAMES && strcmp(assumed, desc->cls) ) {
    
    reportSpiderFail();
    if(outputAllowed()) {
      cout << "Type condition in file " << file << ", line " << line
	   << " violated:\n" << ptrStr << " was allocated as \"" << desc->cls
	   << "\" instead of \"" << assumed <<"\".\n";
    }
  } else if( !desc->allocated ) {
    reportSpiderFail();
    if(outputAllowed()) {
      cout << "Type condition in file " << file << ", line " << line
	   << " violated:\n" << ptrStr << " was allocated as \"" << desc->cls
	   << "\", but no longer is.\n";
    }
  } else {
    return;
  }

  _violated = true;

  if(outputAllowed()) {
    cout << "----- stack dump -----\n";
    Tracer::printStack(cout);
    cout << "----- end of stack dump -----\n";
  }
  throw Debug::AssertionViolationException(file,line);
} 

void Assertion::reportAssertValidException (const char* file,int line,const char* obj)
{
  if(outputAllowed()) {
    cout << "An exception was thrown by ASSERT_VALID on object " << obj
	<< " in file " << file << ", line " << line << ".\n";
  }
} 

AssertionViolationException::AssertionViolationException (const char* file,
							  int line)
  : _file(file),
    _line(line)
{
}

void AssertionViolationException::cry (ostream& str)
{
  str << "Assertion violation ";
  outputFileAndLine(str);
  str << '\n';
} 


void AssertionViolationException::outputFileAndLine (ostream& str) const
{
  str << "(file: '" << _file
      << "', line: " << _line << ")";
} 

#endif 
