
/*
 * File FiniteModelBuilderIncremental.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Kernel/Inference.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/SubstHelper.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/EqHelper.hpp"
#include "Kernel/Renaming.hpp"
#include "Kernel/Substitution.hpp"

#include "SAT/Preprocess.hpp"
#include "SAT/TWLSolver.hpp"
#include "SAT/LingelingInterfacing.hpp"
#include "SAT/MinisatInterfacing.hpp"
#include "SAT/BufferedSolver.hpp"

#include "Lib/Environment.hpp"
#include "Lib/Timer.hpp"
#include "Lib/List.hpp"
#include "Lib/Stack.hpp"
#include "Lib/System.hpp"

#include "Shell/UIHelper.hpp"
#include "Shell/Statistics.hpp"
#include "Shell/GeneralSplitting.hpp"

#include "ClauseFlattening.hpp"
#include "SortInference.hpp"
#include "CombinationsIterator.hpp"
#include "DefinitionIntroduction.hpp"
#include "FiniteModelBuilderIncremental.hpp"

#define VTRACE_FMB 0

namespace FMB 
{

Array<Term*> FiniteModelBuilderIncremental::_modelConstants;
unsigned FiniteModelBuilderIncremental::created=0;
unsigned FiniteModelBuilderIncremental::fchecked=0;

FiniteModelBuilderIncremental::FiniteModelBuilderIncremental(Problem& prb, const Options& opt)
: MainLoop(prb, opt), _maxSatVar(0), _groundClauses(0), _clauses(0), _functionDefinitionClauses(0),
                      _constantsCount(0), _isComplete(true), _maxModelSize(UINT_MAX)
{
  CALL("FiniteModelBuilderIncremental::FiniteModelBuilderIncremental");

  if(!opt.complete(prb)){
    _isComplete = false;
    return;
  }

  ASS(opt.fmbIncremental());
  _sortInference = opt.fmbSortInference();

  createSolver();
}

void FiniteModelBuilderIncremental::createSolver(){

  _maxSatVar = 0;
  _lookup.reset();
  _revLookup.reset();

  switch(_opt.satSolver()){
    case Options::SatSolver::VAMPIRE:
      _solver = new TWLSolver(_opt, true);
      break;
    case Options::SatSolver::LINGELING:
      _solver = new LingelingInterfacing(_opt, true);
      break;
    case Options::SatSolver::MINISAT:
      _solver = new MinisatInterfacing(_opt,true);
      break;
    default:
      ASSERTION_VIOLATION_REP(_opt.satSolver());
  }

}

Term* FiniteModelBuilderIncremental::getConstant(unsigned constant)
{
  CALL("FiniteModelBuilderIncremental::getConstant");
      while(constant >= created){
        vstring name;
        bool found=false;



        if(!found){
          name = "fmb_" + Lib::Int::toString(created);
        }
        _modelConstants[created] = Term::createConstant(name);
        created++;
      }
      return _modelConstants[constant];

}


void FiniteModelBuilderIncremental::init()
{
  CALL("FiniteModelBuilderIncremental::init");

  if(!_isComplete) return;

  
  
  DefinitionIntroduction cit = DefinitionIntroduction(_prb.clauseIterator());
  
  while(cit.hasNext()){
    Clause* c = cit.next();
#if VTRACE_FMB
    
#endif
    c = ClauseFlattening::flatten(c);
#if VTRACE_FMB
    
#endif
    ASS(c);

    if(isRefutation(c)){
      throw RefutationFoundException(c);
    }

    
    if(c->varCnt()==0){
#if VTRACE_FMB
      
#endif
      _groundClauses = _groundClauses->cons(c);    
    }else{
#if VTRACE_FMB
      
#endif
      _clauses = _clauses->cons(c);

      unsigned posEqs = 0;
      for(unsigned i=0;i<c->length();i++){
        Literal* l = (*c)[i];
        if(l->isTwoVarEquality() && l->isPositive() && 
           (*l->nthArgument(0))!=(*l->nthArgument(1))
          ){ posEqs++; }
        else break;
      }
      if(posEqs == c->length() && c->varCnt() < _maxModelSize){
#if VTRACE_FMB
        cout << "based on " << c->toString() << " setting _maxModelSize to " << _maxModelSize << endl;
#endif
        _maxModelSize = c->varCnt();
      }      
    }
  }

  
  GeneralSplitting splitter;
  {
    TimeCounter tc(TC_FMB_SPLITTING);
    splitter.apply(_clauses);
  }

  
  ClauseList::Iterator it(_clauses);
  while(it.hasNext()){
    Renaming n;
    Clause* c = it.next();

    
    for(unsigned i=0;i<c->length();i++){
      Literal* l = (*c)[i];
      n.normalizeVariables(l);
      (*c)[i] = n.apply(l);
    }
#if VTRACE_FMB
    
#endif

  }

  if(_sortInference){
    TimeCounter tc(TC_FMB_SORT_INFERENCE);
    ClauseIterator cit= pvi(getConcatenatedIterator(
                        ClauseList::Iterator(_clauses),
                        ClauseList::Iterator(_groundClauses)));    
    _sortedSignature = SortInference::apply(cit);
  }

  

  
  
  
  for(unsigned f = 0; f < env.signature->functions(); f++){

    Signature::Symbol* fun = env.signature->getFunction(f);
    ASS(fun);
    TermList y, z;
    y.makeVar(0); z.makeVar(1); 

    static DArray<TermList> args(8);
    args.ensure(fun->arity());
    for(unsigned i=0;i<fun->arity();i++){
      args[i].makeVar(i+2);
    }
    TermList fargs = TermList(Term::create(f,fun->arity(),args.array()));
    unsigned rSort = fun->fnType()->result(); 

    Clause* c = new(2) Clause(2,Unit::AXIOM, new Inference(Inference::FMB_FUNC_DEF));  
    Literal* l1 = Literal::createEquality(false,y,fargs,rSort);
    Literal* l2 = Literal::createEquality(false,z,fargs,rSort);
    (*c)[0] = l1;
    (*c)[1] = l2;
    
    
    _functionDefinitionClauses = _functionDefinitionClauses->cons(c);

    
    
    for(unsigned i=0;i<fun->arity();i++){
      args[i].makeVar(i);
    }
    _totalityFunctions.push(Term::create(f,fun->arity(),args.array()));

    if(fun->arity()==0){ _constantsCount++; }

    if(!_sortInference){
      
      if(fun->arity()==0){
        _unsortedConstants.push(Term::createConstant(f));
      }
      else{
        for(unsigned i=0;i<fun->arity();i++){
          args[i].makeVar(0);
        }
        _unsortedFunctions.push(Term::create(f,fun->arity(),args.array()));
      }
    }
  }

  if(_unsortedConstants.size()==0){
      unsigned fresh = env.signature->addFreshFunction(0,"fmbFreshConstant");
      _unsortedConstants.push(Term::createConstant(fresh));
  }

}

void FiniteModelBuilderIncremental::addGroundClauses()
{
  CALL("FiniteModelBuilderIncremental::addGroundClauses");

  
  if(!_groundClauses) return;

  ClauseList::Iterator cit(_groundClauses);


addGroundLoop:
  while(cit.hasNext()){

      Clause* c = cit.next();
      ASS(c);

      static SATLiteralStack satClauseLits;
      satClauseLits.reset();
#if VTRACE_FMB
      
#endif
      for(unsigned i=0;i<c->length();i++){
        Literal* lit = (*c)[i];
        
        if(EqHelper::isEqTautology(lit)){
#if VTRACE_FMB
          
#endif
          goto addGroundLoop;
        }
        
        if(lit->isEquality() && !lit->isPositive() &&
            (*lit->nthArgument(0))==(*lit->nthArgument(1))){
#if VTRACE_FMB
       
#endif
          continue;
        }
#if VTRACE_FMB
        
#endif
        SATLiteral slit = getSATLiteral(lit);
        satClauseLits.push(slit);
      }
#if VTRACE_FMB
      
#endif
      SATClause* satCl = SATClause::fromStack(satClauseLits);
      addSATClause(satCl);
  }
}

void FiniteModelBuilderIncremental::addNewInstances(unsigned size)
{
  CALL("FiniteModelBuilderIncremental::addNewInstances");

  ClauseList::Iterator cit(_clauses); 

  while(cit.hasNext()){
    Clause* c = cit.next();
    ASS(c);
#if VTRACE_FMB
    cout << "Instances of " << c->toString() << endl;
#endif

    unsigned fvars = c->varCnt();

    CombinationsIterator it(fvars,size);

instanceLoop:
    while(it.hasNext()){
      SubstCombination subst = it.next();

      static SATLiteralStack satClauseLits;
      satClauseLits.reset();

#if VTRACE_FMB
       cout << "Instance: ";
#endif
      
      for(unsigned i=0;i<c->length();i++){
        bool isTwoVar = ((*c)[i])->isTwoVarEquality();
        Literal* lit = SubstHelper::apply((*c)[i],subst);

        
        
        if(isTwoVar){
          bool equal = (*lit->nthArgument(0))==(*lit->nthArgument(1));
          if((lit->isPositive() && equal) || (!lit->isPositive() && !equal)){
#if VTRACE_FMB
              cout << "Skipped tautology" << endl;
#endif
              goto instanceLoop; 
          } 
          if((lit->isPositive() && !equal) || (!lit->isPositive() && equal)){
            
            continue;
          }
        }
#if VTRACE_FMB
        else{
          cout << lit->toString() << " | ";
        }
#endif
        SATLiteral slit = getSATLiteral(lit);
        satClauseLits.push(slit);
      }
#if VTRACE_FMB
      cout << endl;
#endif

      
      
      
      if(satClauseLits.isEmpty()){
        throw new RefutationFoundException(0);
      }

      SATClause* satCl = SATClause::fromStack(satClauseLits);
      addSATClause(satCl);
    }
  }

}

void FiniteModelBuilderIncremental::addNewFunctionalDefs(unsigned size)
{
  CALL("FiniteModelBuilderIncremental::addNewFunctionalDefs");

  
  
  
  
  
  
  
  

  ClauseList::Iterator cit(_functionDefinitionClauses);

  while(cit.hasNext()){
    Clause* c = cit.next();
    ASS(c);
    unsigned fvars = c->varCnt();
    ASS_G(fvars,0);

    CombinationsIterator it(fvars,size);

funDefLoop:
    while(it.hasNext()){
      SubstCombination subst = it.next();

      
      if(subst.apply(0) == subst.apply(1)){
        
        continue;
      }

      static SATLiteralStack satClauseLits;
      satClauseLits.reset();

      
      for(unsigned i=0;i<c->length();i++){
        Literal* lit = SubstHelper::apply((*c)[i],subst);
       if(EqHelper::isEqTautology(lit)){ goto funDefLoop; }
        
        if(lit->isEquality() && !lit->isPositive() &&
            (*lit->nthArgument(0))==(*lit->nthArgument(1))){
          continue;
        }
        SATLiteral slit = getSATLiteral(lit);
        satClauseLits.push(slit);
      }

      SATClause* satCl = SATClause::fromStack(satClauseLits);
      addSATClause(satCl);
    }
  }
  
}

void FiniteModelBuilderIncremental::addNewSymmetryAxioms(unsigned size,
                       Stack<Term*>& constants, Stack<Term*>& functions) 
{
  CALL("FiniteModelBuilderIncremental::addNewSymmetryAxioms");

  
  if(constants.size() < size){
    if(constants.size()==0) return;

#if VTRACE_FMB
    cout << "Added all constants, adding grounding of function" << endl; 
#endif

    
    
    
    unsigned n = constants.size();

    if(functions.size() <= (size/n)) return;

    Term* function = functions[size/n];

    
    
    int ci = (size%n);

    ASS(ci >= 0);
    Term* c = getConstant(ci+1);
    Substitution s;
    s.bind(0,c);
    Term* fc = SubstHelper::apply(function,s);
    unsigned sort = SortHelper::getResultSort(fc);

    static SATLiteralStack satClauseLits;
    satClauseLits.reset();

    for(unsigned i=0;i<size;i++){
      Term* c2 = getConstant(i+1); 
      Literal* l = Literal::createEquality(true,TermList(fc),TermList(c2),sort);
      SATLiteral sl = getSATLiteral(l);
      satClauseLits.push(sl);
    }
    
    SATClause* satCl = SATClause::fromStack(satClauseLits);
    addSATClause(satCl);

    
    return;
  }

  
  

  
  Term* c1 = constants[size-1]; 
  unsigned sort = SortHelper::getResultSort(c1);

#if VTRACE_FMB
    cout << "Adding symmetry constraint on constant " << c1->toString() << endl;
#endif

  static SATLiteralStack satClauseLits;
  satClauseLits.reset(); 
  for(unsigned i=0;i<size;i++){
    Term* c2 = getConstant(i+1); 
    Literal* l = Literal::createEquality(true,TermList(c1),TermList(c2),sort);
    SATLiteral sl = getSATLiteral(l);
    satClauseLits.push(sl);
  }
  SATClause* satCl = SATClause::fromStack(satClauseLits);
  addSATClause(satCl);

  
  
  
  
  
  if(size > 1){
    for(unsigned d=1;d<size;d++){
      satClauseLits.reset();

      Term* cd = getConstant(d+1); 
      Term* cdm = getConstant(d); 

      Literal* l = Literal::createEquality(false,TermList(c1),TermList(cd),sort); 
      satClauseLits.push(getSATLiteral(l));

      for(unsigned i=0;i<size-1;i++){
        Term* ci = constants[i]; 
        l = Literal::createEquality(true,TermList(ci),TermList(cdm),sort);
        satClauseLits.push(getSATLiteral(l));
      }

      addSATClause(SATClause::fromStack(satClauseLits));
    }
  }

}

unsigned FiniteModelBuilderIncremental::addNewTotalityDefs(unsigned size)
{
  CALL("FiniteModelBuilderIncremental::addNewTotalityDefs");

  unsigned domSizeVar = getNextSATVar();
  SATLiteral domSizeLit = SATLiteral(domSizeVar,false);

  Stack<Term*>::Iterator tit(_totalityFunctions);

  while(tit.hasNext()){
    Term* t = tit.next();
    ASS(t);
    

    unsigned rSort = SortHelper::getResultSort(t);
    unsigned fvars = t->arity();

    if(fvars==0){
      static SATLiteralStack satClauseLits;
      satClauseLits.reset();
      satClauseLits.push(domSizeLit);
      for(unsigned i=0;i<size;i++){
        Term* c = getConstant(i+1);
        Literal* lit = Literal::createEquality(true,TermList(t),TermList(c),rSort);
        SATLiteral slit = getSATLiteral(lit);
        satClauseLits.push(slit);
      }
      SATClause* satCl = SATClause::fromStack(satClauseLits);
      addSATClause(satCl); 
      continue;
    }

    CombinationsIterator it(fvars,size,true); 
    while(it.hasNext()){
      SubstCombination subst = it.next();
      

      Term* tsub = SubstHelper::apply(t,subst);

      static SATLiteralStack satClauseLits;
      satClauseLits.reset();

      satClauseLits.push(domSizeLit);

      
      for(unsigned i=0;i<size;i++){
        
        Term* c = getConstant(i+1);
        Literal* lit = Literal::createEquality(true,TermList(tsub),TermList(c),rSort);
        SATLiteral slit = getSATLiteral(lit);
        satClauseLits.push(slit);
      }    

      SATClause* satCl = SATClause::fromStack(satClauseLits);
      addSATClause(satCl);
    }
  }
  
  return domSizeVar;
}


unsigned FiniteModelBuilderIncremental::getNextSATVar()
{
  CALL("FiniteModelBuilderIncremental::getNextSATLiteral");
  
  _solver->ensureVarCount(++_maxSatVar);
  return _maxSatVar;
}

SATLiteral FiniteModelBuilderIncremental::getSATLiteral(Literal* lit)
{
  CALL("FiniteModelBuilderIncremental::getSATLiteral");
  bool polarity = lit->polarity();
  if(!polarity) lit = Literal::complementaryLiteral(lit);
  unsigned var;
  if(!_lookup.find(lit,var)){
    var = getNextSATVar();
#if VTRACE_FMB
    cout << "STORING " << var << " -> " << lit->toString() << endl;
#endif
    _lookup.insert(lit,var);
    _revLookup.insert(var,lit);
  }
  return SATLiteral(var,polarity);
}

void FiniteModelBuilderIncremental::addSATClause(SATClause* cl)
{
  CALL("FiniteModelBuilderIncremental::addSATClause");
  cl = Preprocess::removeDuplicateLiterals(cl);
  if(!cl){ return; }
#if VTRACE_FMB
  cout << "ADDING " << cl->toString() << endl;
#endif

  _clausesToBeAdded.push(cl);

}

MainLoopResult FiniteModelBuilderIncremental::runImpl()
{
  CALL("FiniteModelBuilderIncremental::runImpl");

  if(!_isComplete){
    
    return MainLoopResult(Statistics::UNKNOWN);
  }

  
  if(env.property->category()==Property::EPR){
    if(_constantsCount < _maxModelSize){
      _maxModelSize = _constantsCount; 
    }
  }

  unsigned modelSize = 1;
  int domSizeVar = -1;
  while(true){
#if VTRACE_FMB
    cout << "TRYING " << modelSize << endl;
#endif
    if(env.options->mode()!=Options::Mode::SPIDER){
      cout << "Trying model size " << modelSize << endl;
    }
    Timer::syncClock();
    if(env.timeLimitReached()){ return MainLoopResult(Statistics::TIME_LIMIT); }

    
    if(domSizeVar > -1){
      static SATLiteralStack satClauseLits;
      satClauseLits.reset();
      satClauseLits.push(SATLiteral(domSizeVar,false));
      addSATClause(SATClause::fromStack(satClauseLits));

      
      {
        TimeCounter tc(TC_FMB_SAT_SOLVING);
        _solver->simplify();
      }
    }

    {
    TimeCounter tc(TC_FMB_CONSTRAINT_CREATION);

    
#if VTRACE_FMB
    cout << "GROUND" << endl;
#endif
    if(modelSize == 1){
      addGroundClauses();
    }
#if VTRACE_FMB
    cout << "INSTANCES" << endl;
#endif
    addNewInstances(modelSize);
#if VTRACE_FMB
    cout << "FUNC DEFS" << endl;
#endif
    addNewFunctionalDefs(modelSize);
#if VTRACE_FMB
    cout << "SYM DEFS" << endl;
#endif
    addNewSymmetryAxioms(modelSize);
#if VTRACE_FMB
    cout << "TOTAL DEFS" << endl;
#endif
    domSizeVar = addNewTotalityDefs(modelSize);

    }

#if VTRACE_FMB
    cout << "SOLVING" << endl;
#endif
    
    
    {
      TimeCounter tc(TC_FMB_SAT_SOLVING);
      _solver->addClausesIter(pvi(SATClauseStack::ConstIterator(_clausesToBeAdded)));
    }

    _solver->addAssumption(SATLiteral(domSizeVar,true));

    SATSolver::Status satResult = SATSolver::UNKNOWN;
    {
      TimeCounter tc(TC_FMB_SAT_SOLVING);
      satResult = _solver->solve();
    }

    
    if(satResult == SATSolver::SATISFIABLE){
      onModelFound(modelSize);
      return MainLoopResult(Statistics::SATISFIABLE);
    }

    
    if(modelSize == UINT_MAX){
      return MainLoopResult(Statistics::UNKNOWN);
    }

    if(modelSize >= _maxModelSize){

      if(env.options->mode()!=Options::Mode::SPIDER) { 
        if(env.property->category()==Property::EPR){
          cout << "Checked all constants of an EPR problem" << endl;
        }
        else{
          cout << "All further models will be UNSAT due to variable constraint" << endl;
        }
      }

      
      Clause* empty = new(0) Clause(0,Unit::AXIOM,
         new Inference(Inference::MODEL_NOT_FOUND));
      return MainLoopResult(Statistics::REFUTATION,empty); 
    }

    TimeCounter tc(TC_FMB_SAT_SOLVING);
    _solver->retractAllAssumptions();

    
    _clausesToBeAdded.reset();

    modelSize++;
  }


  return MainLoopResult(Statistics::UNKNOWN);
}


void FiniteModelBuilderIncremental::onModelFound(unsigned modelSize)
{
 
 if(_opt.proof()==Options::Proof::OFF){ 
   return; 
 }

 
 if(UIHelper::cascMode) {
   env.beginOutput();
   env.out() << "% SZS status "<<( UIHelper::haveConjecture() ? "CounterSatisfiable" : "Satisfiable" )
       << " for " << _opt.problemName() << endl << flush;
   env.endOutput();
   UIHelper::satisfiableStatusWasAlreadyOutput = true;
 }
 if(_opt.mode()==Options::Mode::SPIDER){
   reportSpiderStatus('-');
 }

  
  Timer::setTimeLimitEnforcement(false);

 
 

        cout << "Found model of size " << modelSize << endl;
        for(unsigned i=1;i<=_maxSatVar;i++){
          Literal* lit;
          if(_revLookup.find(i,lit)){
            bool pol = _solver->trueInAssignment(SATLiteral(i,true));
            ASS(pol || !lit->isEquality() || lit->nthArgument(0)!=lit->nthArgument(1));
            if(!pol) lit = Literal::complementaryLiteral(lit);
            if(!lit->isEquality() || (pol && !EqHelper::isEqTautology(lit))){
              cout << lit->toString() << endl;
            }
          }
        }
}

}
