
/*
 * File LiteralSubstitutionTreeWithoutTop.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/TimeCounter.hpp"

#include "Kernel/Matcher.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/Term.hpp"

#include "LiteralSubstitutionTreeWithoutTop.hpp"

namespace Indexing
{

LiteralSubstitutionTreeWithoutTop::LiteralSubstitutionTreeWithoutTop()
: SubstitutionTree(0), _posRoot(0), _negRoot(0) 
{
  _nextVar=1;
}

void LiteralSubstitutionTreeWithoutTop::insert(Literal* lit, Clause* cls)
{
  CALL("LiteralSubstitutionTreeWithoutTop::insert");
  handleLiteral(lit,cls,true);
}

void LiteralSubstitutionTreeWithoutTop::remove(Literal* lit, Clause* cls)
{
  CALL("LiteralSubstitutionTreeWithoutTop::remove");
  handleLiteral(lit,cls,false);
}

void LiteralSubstitutionTreeWithoutTop::handleLiteral(Literal* lit, Clause* cls, bool insert)
{
  CALL("LiteralSubstitutionTreeWithoutTop::handleLiteral");
  
  Literal* normLit=Renaming::normalize(lit);

  Node** _root = lit->polarity() ? &_posRoot : &_negRoot;

  BindingMap svBindings;
  svBindings.insert(0,TermList(normLit));
  if(insert) {
    SubstitutionTree::insert(_root, svBindings, LeafData(cls, lit));
  } else {
    SubstitutionTree::remove(_root, svBindings, LeafData(cls, lit));
  }

}

SLQueryResultIterator LiteralSubstitutionTreeWithoutTop::getUnifications(Literal* lit,
	  bool complementary, bool retrieveSubstitutions)
{
  CALL("LiteralSubstitutionTreeWithoutTop::getUnifications");
  return getResultIterator<UnificationsIterator>(lit,
	  complementary, retrieveSubstitutions);
}

SLQueryResultIterator LiteralSubstitutionTreeWithoutTop::getGeneralizations(Literal* lit,
	  bool complementary, bool retrieveSubstitutions)
{
  CALL("LiteralSubstitutionTreeWithoutTop::getGeneralizations");

  SLQueryResultIterator res=

    getResultIterator<FastGeneralizationsIterator>(lit,
	  	  complementary, retrieveSubstitutions);


  return res;
}

SLQueryResultIterator LiteralSubstitutionTreeWithoutTop::getInstances(Literal* lit,
	  bool complementary, bool retrieveSubstitutions)
{
  CALL("LiteralSubstitutionTreeWithoutTop::getInstances");



  if(retrieveSubstitutions) {
#if VDEBUG
    NOT_IMPLEMENTED;
#endif
    return getResultIterator<InstancesIterator>(lit, complementary, true);
  }

  SLQueryResultIterator res=

      getResultIterator<FastInstancesIterator>(lit,
	  complementary, retrieveSubstitutions);


  return res;
}

struct LiteralSubstitutionTreeWithoutTop::SLQueryResultFunctor
{
  DECL_RETURN_TYPE(SLQueryResult);
  OWN_RETURN_TYPE operator() (const QueryResult& qr) {
    return SLQueryResult(qr.first->literal, qr.first->clause, qr.second);
  }
};

struct LiteralSubstitutionTreeWithoutTop::LDToSLQueryResultFn
{
  DECL_RETURN_TYPE(SLQueryResult);
  OWN_RETURN_TYPE operator() (const LeafData& ld) {
    return SLQueryResult(ld.literal, ld.clause);
  }
};

#define QRS_QUERY_BANK 0
#define QRS_RESULT_BANK 1

struct LiteralSubstitutionTreeWithoutTop::LDToSLQueryResultWithSubstFn
{
  LDToSLQueryResultWithSubstFn()
  {
    _subst=RobSubstitutionSP(new RobSubstitution());
  }
  DECL_RETURN_TYPE(SLQueryResult);
  OWN_RETURN_TYPE operator() (const LeafData& ld) {
    return SLQueryResult(ld.literal, ld.clause,
	    ResultSubstitution::fromSubstitution(_subst.ptr(),
		    QRS_QUERY_BANK,QRS_RESULT_BANK));
  }
private:
  RobSubstitutionSP _subst;
};

struct LiteralSubstitutionTreeWithoutTop::UnifyingContext
{
  UnifyingContext(Literal* queryLit)
  : _queryLit(queryLit) {}
  bool enter(SLQueryResult qr)
  {
    ASS(qr.substitution);
    RobSubstitution* subst=qr.substitution->tryGetRobSubstitution();
    ASS(subst);

    
    
    
    
    ALWAYS(subst->unifyArgs(_queryLit, QRS_QUERY_BANK, qr.literal, QRS_RESULT_BANK));

    return true;
  }
  void leave(SLQueryResult qr)
  {
    RobSubstitution* subst=qr.substitution->tryGetRobSubstitution();
    ASS(subst);
    subst->reset();
  }
private:
  Literal* _queryLit;
};


struct LiteralSubstitutionTreeWithoutTop::LeafToLDIteratorFn
{
  DECL_RETURN_TYPE(LDIterator);
  OWN_RETURN_TYPE operator() (Leaf* l) {
    return l->allChildren();
  }
};

struct LiteralSubstitutionTreeWithoutTop::PropositionalLDToSLQueryResultWithSubstFn
{
  PropositionalLDToSLQueryResultWithSubstFn()
  {
    _subst=ResultSubstitutionSP (new DisjunctQueryAndResultVariablesSubstitution()); 
  }
  DECL_RETURN_TYPE(SLQueryResult);
  OWN_RETURN_TYPE operator() (const LeafData& ld) {
    ASS_EQ(ld.literal->arity(),0);
    return SLQueryResult(ld.literal, ld.clause, _subst);
  }
private:
  ResultSubstitutionSP _subst;
};


SLQueryResultIterator LiteralSubstitutionTreeWithoutTop::getVariants(Literal* lit,
	  bool complementary, bool retrieveSubstitutions)
{
  CALL("LiteralSubstitutionTreeWithoutTop::getVariants");

  bool pol = complementary ? !lit->polarity() : lit->polarity();
  Node* _root = pol ? _posRoot : _negRoot; 

  if(_root==0) {
    return SLQueryResultIterator::getEmpty();
  }

  if(_root->isLeaf()) {
    LDIterator ldit=static_cast<Leaf*>(_root)->allChildren();
    if(retrieveSubstitutions) {
      
      return pvi( getMappingIterator(ldit,PropositionalLDToSLQueryResultWithSubstFn()) );
    } else {
      return pvi( getMappingIterator(ldit,LDToSLQueryResultFn()) );
    }
  }

  Literal* normLit=Renaming::normalize(lit);

  BindingMap svBindings;
  svBindings.insert(0,TermList(normLit));
  Leaf* leaf = findLeaf(_root,svBindings);
  if(leaf==0) {
    return SLQueryResultIterator::getEmpty();
  }

  LDIterator ldit=leaf->allChildren();
  if(retrieveSubstitutions) {
    return pvi( getContextualIterator(
	    getMappingIterator(
		    ldit,
		    LDToSLQueryResultWithSubstFn()),
	    UnifyingContext(lit)) );
  } else {
    return pvi( getMappingIterator(ldit,LDToSLQueryResultFn()) );
  }
}

SLQueryResultIterator LiteralSubstitutionTreeWithoutTop::getAll()
{
  CALL("LiteralSubstitutionTreeWithoutTop::getAll");

  return pvi( getMappingIterator(
      getMapAndFlattenIterator(
	  vi( new LeafIterator(this) ),
	  LeafToLDIteratorFn()),
      LDToSLQueryResultFn()) ) ;
}


struct LiteralSubstitutionTreeWithoutTop::EqualitySortFilter
{
  DECL_RETURN_TYPE(bool);

  EqualitySortFilter(Literal* queryLit)
  : _queryEqSort(SortHelper::getEqualityArgumentSort(queryLit)) {}

  bool operator()(const SLQueryResult& res)
  {
    CALL("LiteralSubstitutionTreeWithoutTop::EqualitySortFilter::operator()");
    ASS(res.literal->isEquality());

    unsigned resSort = SortHelper::getEqualityArgumentSort(res.literal);
    return resSort==_queryEqSort;
  }
private:
  unsigned _queryEqSort;
};

template<class Iterator>
SLQueryResultIterator LiteralSubstitutionTreeWithoutTop::getResultIterator(Literal* lit,
	  bool complementary, bool retrieveSubstitutions)
{
  CALL("LiteralSubstitutionTreeWithoutTop::getResultIterator");

  bool pol = complementary ? !lit->polarity() : lit->polarity();
  Node* _root = pol ? _posRoot : _negRoot;

  if(_root==0) {
    return SLQueryResultIterator::getEmpty();
  }

  if(_root->isLeaf()) {
    LDIterator ldit=static_cast<Leaf*>(_root)->allChildren();
    if(retrieveSubstitutions) {
      
      return pvi( getMappingIterator(ldit,PropositionalLDToSLQueryResultWithSubstFn()) );
    } else {
      return pvi( getMappingIterator(ldit,LDToSLQueryResultFn()) );
    }
  }

  if(lit->commutative()) {
    VirtualIterator<QueryResult> qrit1=vi(
  	    new Iterator(this, _root, lit, retrieveSubstitutions,false, true) );
    VirtualIterator<QueryResult> qrit2=vi(
  	    new Iterator(this, _root, lit, retrieveSubstitutions, true, true) );
    ASS(lit->isEquality());
    return pvi(
	getFilteredIterator(
	    getMappingIterator(
		getConcatenatedIterator(qrit1,qrit2), SLQueryResultFunctor()),
	    EqualitySortFilter(lit))
	);
  } else {
    VirtualIterator<QueryResult> qrit=VirtualIterator<QueryResult>(
  	    new Iterator(this, _root, lit, retrieveSubstitutions,false,true) );
    return pvi( getMappingIterator(qrit, SLQueryResultFunctor()) );
  }
}


}
