
/*
 * File SubstitutionTree_FastInst.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Allocator.hpp"
#include "Lib/Recycler.hpp"

#include "Kernel/Matcher.hpp"
#include "Kernel/SubstHelper.hpp"
#include "Kernel/TermIterators.hpp"

#include "SubstitutionTree.hpp"

namespace Indexing
{

class SubstitutionTree::InstMatcher
{
public:
  void reset()
  {
    _boundVars.reset();
    _bindings.reset();
    _derefBindings.reset();
  }

  CLASS_NAME(SubstitutionTree::InstMatcher);
  USE_ALLOCATOR(InstMatcher);

  struct TermSpec
  {
    TermSpec() {
    #if VDEBUG
      t.makeEmpty();
    #endif
    }
    TermSpec(bool q, TermList t)
    : q(q), t(t)
    {
      CALL("SubstitutionTree::InstMatcher::TermSpec::TermSpec");

      
      ASS(!q || !t.isTerm() || t.term()->shared());
      ASS(!q || !t.isSpecialVar());
    }

    vstring toString()
    {
      CALL("SubstitutionTree::InstMatcher::TermSpec::toString");
      return (q ? "q|" : "n|")+t.toString();
    }

   
    bool isFinal()
    {
      
      return q
  	? (t.isTerm() && t.term()->ground())
  	: (t.isOrdinaryVar() || (t.isTerm() && t.term()->shared()) );
    }

    bool q;
    TermList t;
  };

 
  void bindSpecialVar(unsigned var, TermList term)
  {
    CALL("SubstitutionTree::InstMatcher::bindSpecialVar");
    ASS_EQ(getBSCnt(), 0);

    ALWAYS(_bindings.insert(TermList(var,true),TermSpec(true,term)));
  }

  bool isSpecVarBound(unsigned specVar)
  {
    return _bindings.find(TermList(specVar,true));
  }

 
  TermSpec getSpecVarBinding(unsigned specVar)
  {
    TermSpec res=_bindings.get(TermList(specVar,true));

    return res;
  }

  bool findSpecVarBinding(unsigned specVar, TermSpec& res)
  {
    return _bindings.find(TermList(specVar,true), res);
  }

  bool matchNext(unsigned specVar, TermList nodeTerm, bool separate=true);

  void backtrack();
  bool tryBacktrack();
  ResultSubstitutionSP getSubstitution(Renaming* resultDenormalizer);

  int getBSCnt()
  {
    int res=0;
    TermStack::Iterator vsit(_boundVars);
    while(vsit.hasNext()) {
      if(vsit.next().isEmpty()) {
	res++;
      }
    }
    return res;
  }

  void onLeafEntered()
  {
    _derefBindings.reset();
  }

private:

  class Substitution;

  TermList derefQueryBinding(unsigned var);

  bool isBound(TermList var)
  {
    CALL("SubstitutionTree::InstMatcher::isBound");
    ASS(var.isVar());

    return _bindings.find(var);
  }
  void bind(TermList var, TermSpec trm)
  {
    CALL("SubstitutionTree::InstMatcher::bind");
    ASS(!var.isOrdinaryVar() || !trm.q); 

    ALWAYS(_bindings.insert(var, trm));
    _boundVars.push(var);
  }

  TermSpec deref(TermList var);

  typedef DHMap<TermList, TermSpec> BindingMap;
  typedef Stack<TermList> TermStack;

 
  TermStack _boundVars;

  BindingMap _bindings;

 
  DHMap<TermList,TermList> _derefBindings;

  struct DerefTask
  {
    DerefTask(TermList var) : var(var) { trm.t.makeEmpty(); }
    DerefTask(TermList var, TermSpec trm) : var(var), trm(trm) {}
    TermList var;
    TermSpec trm;
    bool buildDerefTerm() { return trm.t.isNonEmpty(); };
  };

  struct DerefApplicator
  {
    DerefApplicator(InstMatcher* im, bool query) : query(query), im(im) {}
    TermList apply(unsigned var)
    {
      CALL("SubstitutionTree::InstMatcher::DerefApplicator::apply");
      if(query) {
	return im->_derefBindings.get(TermList(var, false));
      }
      else {
	return TermList(var, false);
      }
    }
    TermList applyToSpecVar(unsigned specVar)
    {
      CALL("SubstitutionTree::InstMatcher::DerefApplicator::applyToSpecVar");
      ASS(!query);

      return im->_derefBindings.get(TermList(specVar, true));
    }
  private:
    bool query;
    InstMatcher* im;
  };
};

std::ostream& operator<< (ostream& out, SubstitutionTree::InstMatcher::TermSpec ts )
{
  CALL("operator<<(ostream&,SubstitutionTree::InstMatcher::TermSpec)");

  out<<ts.toString();
  return out;
}


class SubstitutionTree::InstMatcher::Substitution
: public ResultSubstitution
{
public:
  CLASS_NAME(SubstitutionTree::InstMatcher::Substitution);
  USE_ALLOCATOR(SubstitutionTree::InstMatcher::Substitution);
  
  Substitution(InstMatcher* parent, Renaming* resultDenormalizer)
  : _parent(parent), _resultDenormalizer(resultDenormalizer)
  {}
  ~Substitution()
  {
  }

  TermList applyToBoundQuery(TermList t)
  {
    CALL("SubstitutionTree::InstMatcher::Substitution::applyToBoundQuery");

    return SubstHelper::apply(t, *this);
  }

  TermList apply(unsigned var)
  {
    CALL("SubstitutionTree::InstMatcher::Substitution::apply");

    TermList normalized=_parent->derefQueryBinding(var);
    ASS_REP(!normalized.isTerm() || normalized.term()->shared(), normalized);
    return _resultDenormalizer->apply(normalized);
  }

  bool isIdentityOnResultWhenQueryBound() { return true; }
private:
  InstMatcher* _parent;
  Renaming* _resultDenormalizer;
};


ResultSubstitutionSP SubstitutionTree::InstMatcher::getSubstitution(Renaming* resultDenormalizer)
{
  CALL("SubstitutionTree::InstMatcher::getSubstitution");

  return ResultSubstitutionSP(
	  new Substitution(this, resultDenormalizer));
}

TermList SubstitutionTree::InstMatcher::derefQueryBinding(unsigned var)
{
  CALL("SubstitutionTree::InstMatcher::derefQueryBinding");

  TermList tvar0(var, false);
  TermList tvar=tvar0;

  TermSpec varBinding;
  {
    TermList val;
    if(_derefBindings.find(tvar, val)) {
      return val;
    }
    
    ALWAYS(_bindings.find(tvar, varBinding));

    if(varBinding.isFinal()) {
      ALWAYS(_derefBindings.insert(tvar, varBinding.t));
      return varBinding.t;
    }
  }
  static Stack<DerefTask> toDo;
  toDo.reset();

  for(;;) {
    while(!varBinding.isFinal() && !varBinding.t.isTerm()) {
      ASS(varBinding.t.isVar());
      ASS(!varBinding.q || !varBinding.t.isOrdinaryVar());


      TermList bvar=varBinding.t;
      TermList derefBoundTerm;

      if(_derefBindings.find(bvar, derefBoundTerm)) {
	ALWAYS(_derefBindings.insert(tvar, derefBoundTerm));
      }

      ALWAYS(_bindings.find(bvar,varBinding));
    }
    if(varBinding.isFinal()) {
      ALWAYS(_derefBindings.insert(tvar, varBinding.t));
      goto next_loop;
    }
    {
      ASS(varBinding.t.isTerm());
      toDo.push(DerefTask(tvar, varBinding));
      VariableIterator vit(varBinding.t);
      while(vit.hasNext()) {
	TermList btv=vit.next(); 
	if(varBinding.q || btv.isSpecialVar()) {
	  ASS(_bindings.find(btv));
	  if(!_derefBindings.find(btv)) {
	    toDo.push(DerefTask(btv));
	  }
	}
      }
    }
    next_loop:
    while(toDo.isNonEmpty() && toDo.top().buildDerefTerm()) {
      tvar=toDo.top().var;
      TermSpec tspec=toDo.pop().trm;
      DerefApplicator applicator(this, tspec.q);
      TermList derefTerm=SubstHelper::applySV(tspec.t, applicator);
      ASS_REP(!derefTerm.isTerm() || derefTerm.term()->shared(), derefTerm);
      ALWAYS(_derefBindings.insert(tvar, derefTerm));
    }
    if(toDo.isEmpty()) {
      break;
    }
    tvar=toDo.pop().var;
    ALWAYS(_bindings.find(tvar, varBinding));
  };
  return _derefBindings.get(tvar0);
}

SubstitutionTree::InstMatcher::TermSpec SubstitutionTree::InstMatcher::deref(TermList var)
{
  CALL("SubstitutionTree::InstMatcher::deref");
  ASS_REP(var.isVar(), var.tag());

#if VDEBUG
  int ctr=0;
#endif
  for(;;) {
    TermSpec res;
    if(!_bindings.find(var, res)) {
	return TermSpec(var.isOrdinaryVar() ? true : false, var);
    }
    if( res.t.isTerm() || (!res.q && res.t.isOrdinaryVar()) ) {
	return res;
    }
    ASS(!res.q || !res.t.isSpecialVar());
    var=res.t;
#if VDEBUG
    ctr++;
    ASS_L(ctr,1000000); 
#endif
  }
}

void SubstitutionTree::InstMatcher::backtrack()
{
  CALL("SubstitutionTree::InstMatcher::backtrack");

  for(;;) {
    TermList boundVar=_boundVars.pop();
    if(boundVar.isEmpty()) {
      break;
    }
    _bindings.remove(boundVar);
  }
}

bool SubstitutionTree::InstMatcher::tryBacktrack()
{
  CALL("SubstitutionTree::InstMatcher::tryBacktrack");

  while(_boundVars.isNonEmpty()) {
    TermList boundVar=_boundVars.pop();
    if(boundVar.isEmpty()) {
      return true;
    }
    _bindings.remove(boundVar);
  }
  return false;
}

bool SubstitutionTree::InstMatcher::matchNext(unsigned specVar, TermList nodeTerm, bool separate)
{
  CALL("SubstitutionTree::InstMatcher::matchNext");

  if(separate) {
    TermList sep;
    sep.makeEmpty();
    _boundVars.push(sep);
  }

#if VDEBUG
  {
    
    VariableIterator vit(nodeTerm);
    while(vit.hasNext()) {
      TermList var=vit.next();
      if(var.isSpecialVar()) {
	ASS(!isBound(var));
      }
    }
  }
#endif

  TermSpec tsNode(false, nodeTerm);

  TermSpec tsBinding;
  if(!findSpecVarBinding(specVar,tsBinding)) {
    bind(TermList(specVar,true), tsNode);
    return true;
  }

  if(tsBinding.q && tsBinding.t.isOrdinaryVar() && !isBound(tsBinding.t)) {
    bind(tsBinding.t, tsNode);
    return true;
  }

  bool success;

  if(nodeTerm.isTerm() && nodeTerm.term()->shared() && nodeTerm.term()->ground() &&
      tsBinding.q && tsBinding.t.isTerm() && tsBinding.t.term()->ground()) {
    success=nodeTerm.term()==tsBinding.t.term();
    goto finish;
  }

  static Stack<pair<TermSpec,TermSpec> > toDo;
  static DisagreementSetIterator dsit;

  toDo.reset();
  toDo.push(make_pair(tsBinding, tsNode));

  while(toDo.isNonEmpty()) {
    TermSpec ts1=toDo.top().first;
    TermSpec ts2=toDo.pop().second;


    dsit.reset(ts1.t, ts2.t, ts1.q!=ts2.q);
    while(dsit.hasNext()) {
      pair<TermList,TermList> disarg=dsit.next();
      TermList dt1=disarg.first;
      TermList dt2=disarg.second;

      bool dt1Bindable= !dt1.isTerm() && (ts1.q || !dt1.isOrdinaryVar());
      bool dt2Bindable= !dt2.isTerm() && (ts2.q || !dt2.isOrdinaryVar());

      if(!dt1Bindable && !dt2Bindable) {
	success=false;
	goto finish;
      }

      
      
      
      

      if(ts1.q && dt1.isOrdinaryVar() && !isBound(dt1)) {
	bind(dt1, TermSpec(ts2.q,dt2));
	continue;
      }
      if(ts2.q && dt2.isOrdinaryVar() && !isBound(dt2)) {
	bind(dt2, TermSpec(ts1.q,dt1));
	continue;
      }

      if(dt2.isSpecialVar() && !isBound(dt2)) {
	ASS(!ts2.q);
	bind(dt2, TermSpec(ts1.q,dt1));
	continue;
      }
      if(dt1.isSpecialVar() && !isBound(dt1)) {
	ASS(!ts1.q);
	bind(dt1, TermSpec(ts2.q,dt2));
	continue;
      }

      TermSpec deref1=TermSpec(ts1.q, dt1);
      TermSpec deref2=TermSpec(ts2.q, dt2);
      if(dt1Bindable) {
	ASS(isBound(dt1)); 
	deref1=deref(dt1);
      }
      if(dt2Bindable) {
	ASS(isBound(dt2));
	deref2=deref(dt2);
      }

      toDo.push(make_pair(deref1, deref2));
    }
  }
  success=true;

finish:
  if(!success) {
    
    
    
    if(separate) {
      
      backtrack();
    }
  }
  return success;
}


SubstitutionTree::FastInstancesIterator::FastInstancesIterator(SubstitutionTree* parent, Node* root,
	Term* query, bool retrieveSubstitution, bool reversed, bool withoutTop)
: _literalRetrieval(query->isLiteral()), _retrieveSubstitution(retrieveSubstitution),
  _inLeaf(false), _ldIterator(LDIterator::getEmpty()), _tree(parent),  _root(root),
  _alternatives(64), _specVarNumbers(64), _nodeTypes(64)
{
  CALL("SubstitutionTree::FastInstancesIterator::FastGeneralizationsIterator");
  ASS(root);
  ASS(!root->isLeaf());

#if VDEBUG
  _tree->_iteratorCnt++;
#endif

  Recycler::get(_subst);
  _subst->reset();


  if(withoutTop){
    _subst->bindSpecialVar(0,TermList(query));
  }else{
    if(reversed) {
      createReversedInitialBindings(query);
    } else {
      createInitialBindings(query);
    }
  }
}

SubstitutionTree::FastInstancesIterator::~FastInstancesIterator()
{
#if VDEBUG
  _tree->_iteratorCnt--;
#endif
  Recycler::release(_subst);

}


void SubstitutionTree::FastInstancesIterator::createInitialBindings(Term* t)
{
  CALL("SubstitutionTree::FastInstancesIterator::createInitialBindings");

  TermList* args=t->args();
  int nextVar = 0;
  while (! args->isEmpty()) {
    unsigned var = nextVar++;
    _subst->bindSpecialVar(var,*args);
    args = args->next();
  }
}

void SubstitutionTree::FastInstancesIterator::createReversedInitialBindings(Term* t)
{
  CALL("SubstitutionTree::FastInstancesIterator::createReversedInitialBindings");
  ASS(t->isLiteral());
  ASS(t->commutative());
  ASS_EQ(t->arity(),2);

  _subst->bindSpecialVar(1,*t->nthArgument(0));
  _subst->bindSpecialVar(0,*t->nthArgument(1));
}

bool SubstitutionTree::FastInstancesIterator::hasNext()
{
  CALL("SubstitutionTree::FastInstancesIterator::hasNext");

  while(!_ldIterator.hasNext() && findNextLeaf()) {}
  return _ldIterator.hasNext();
}

#undef LOGGING
#define LOGGING 0

SubstitutionTree::QueryResult SubstitutionTree::FastInstancesIterator::next()
{
  CALL("SubstitutionTree::FastInstancesIterator::next");

  while(!_ldIterator.hasNext() && findNextLeaf()) {}
  ASS(_ldIterator.hasNext());
  LeafData& ld=_ldIterator.next();

  if(_retrieveSubstitution) {
    _resultDenormalizer.reset();
    bool ground=_literalRetrieval
	? ld.literal->ground()
	: (ld.term.isTerm() && ld.term.term()->ground());
    if(!ground) {
      Renaming normalizer;
      if(_literalRetrieval) {
	normalizer.normalizeVariables(ld.literal);
      } else {
	normalizer.normalizeVariables(ld.term);
      }
      _resultDenormalizer.makeInverse(normalizer);
    }

    return QueryResult(&ld,
	    _subst->getSubstitution(&_resultDenormalizer));
  } else {
    return QueryResult(&ld, ResultSubstitutionSP());
  }
}
#undef LOGGING
#define LOGGING 0

bool SubstitutionTree::FastInstancesIterator::findNextLeaf()
{
  CALL("SubstitutionTree::FastInstancesIterator::findNextLeaf");

  Node* curr;
  bool sibilingsRemain;
  if(_inLeaf) {
    if(_alternatives.isEmpty()) {
      return false;
    }
    _subst->backtrack();
    _inLeaf=false;
    curr=0;
  } else {
    if(!_root) {
      
      
      return false;
    }
    curr=_root;
    _root=0;
    sibilingsRemain=enterNode(curr);
  }
  for(;;) {
main_loop_start:
    unsigned currSpecVar;

    if(curr) {
      if(sibilingsRemain) {
	ASS(_nodeTypes.top()!=UNSORTED_LIST || *static_cast<Node**>(_alternatives.top()));
	currSpecVar=_specVarNumbers.top();
      } else {
	currSpecVar=_specVarNumbers.pop();
      }
    }
    
    while(curr==0 && _alternatives.isNonEmpty()) {
      void* currAlt=_alternatives.pop();
      if(!currAlt) {
	
	_nodeTypes.pop();
	_specVarNumbers.pop();
	if(_alternatives.isNonEmpty()) {
	  _subst->backtrack();
	}
	continue;
      }

      NodeAlgorithm parentType=_nodeTypes.top();

      
      
      
      if(parentType==UNSORTED_LIST) {
	Node** alts=static_cast<Node**>(currAlt);
	curr=*(alts++);
	if(*alts) {
	  _alternatives.push(alts);
	  sibilingsRemain=true;
	} else {
	  sibilingsRemain=false;
	}
      } else {
	ASS_EQ(parentType,SKIP_LIST)
	NodeList* alts=static_cast<NodeList*>(currAlt);
	ASS(alts);

	curr=alts->head();
	if(alts->tail()) {
	  _alternatives.push(alts->tail());
	  sibilingsRemain=true;
	} else {
	  sibilingsRemain=false;
	}
      }

      if(sibilingsRemain) {
	currSpecVar=_specVarNumbers.top();
      } else {
	_nodeTypes.pop();
	currSpecVar=_specVarNumbers.pop();
      }
      ASS(curr);
      break;
    }
    if(!curr) {
      
      return false;
    }
    if(!_subst->matchNext(currSpecVar, curr->term, sibilingsRemain)) {	
      
      curr=0;
      if(!sibilingsRemain && _alternatives.isNonEmpty()) {
	_subst->backtrack();
      }
      continue;
    }
    while(!curr->isLeaf() && curr->algorithm()==UNSORTED_LIST && static_cast<UArrIntermediateNode*>(curr)->_size==1) {
      
      unsigned specVar=static_cast<UArrIntermediateNode*>(curr)->childVar;
      curr=static_cast<UArrIntermediateNode*>(curr)->_nodes[0];
      ASS(curr);
      ASSERT_VALID(*curr);
      if(!_subst->matchNext(specVar, curr->term, false)) {
	
	
	if(sibilingsRemain || _alternatives.isNonEmpty()) {
	  
	  
	  
	  _subst->backtrack();
	}
        curr=0;
        goto main_loop_start;
      }
    }
    if(curr->isLeaf()) {
      
      _ldIterator=static_cast<Leaf*>(curr)->allChildren();
      _inLeaf=true;
      _subst->onLeafEntered(); 
      return true;
    }

    
    sibilingsRemain=enterNode(curr);
    if(curr==0 && _alternatives.isNonEmpty()) {
      _subst->backtrack();
    }
  }
}

bool SubstitutionTree::FastInstancesIterator::enterNode(Node*& curr)
{
  CALL("SubstitutionTree::FastInstancesIterator::enterNode");
  ASSERT_VALID(*curr);
  ASS(!curr->isLeaf());

  IntermediateNode* inode=static_cast<IntermediateNode*>(curr);
  NodeAlgorithm currType=inode->algorithm();

  TermList query;
  InstMatcher::TermSpec querySpec;
  
  
  if(_subst->findSpecVarBinding(inode->childVar, querySpec)) {
    query=querySpec.t;
  }
  else {
    query.makeVar(0);
  }

  curr=0;

  if(currType==UNSORTED_LIST) {
    Node** nl=static_cast<UArrIntermediateNode*>(inode)->_nodes;
    ASS(*nl); 
    bool noAlternatives=false;
    if(query.isTerm()) {
      unsigned bindingFunctor=query.term()->functor();
      
      while(*nl && (!(*nl)->term.isTerm() || (*nl)->term.term()->functor()!=bindingFunctor)) {
        nl++;
      }

      if(*nl) {
	
	ASS_EQ((*nl)->term.term()->functor(),bindingFunctor);
        curr=*nl;
        noAlternatives=true; 
      }
    } else {
      ASS(query.isVar());
      
      curr=*nl;
      nl++;
    }

    if(curr) {
      _specVarNumbers.push(inode->childVar);
    }
    if(*nl && !noAlternatives) {
      _alternatives.push(nl);
      _nodeTypes.push(currType);
      return true;
    }
  } else {
    NodeList* nl;
    ASS_EQ(currType, SKIP_LIST);
    nl=static_cast<SListIntermediateNode*>(inode)->_nodes.toList();
    ASS(nl); 
    if(query.isTerm()) {
      
      Node** byTop=inode->childByTop(query, false);
      if(byTop) {
	curr=*byTop;
      }
      nl=0;
    }
    else {
      ASS(query.isVar());
      
      curr=nl->head();
      nl=nl->tail();
    }

    if(curr) {
      _specVarNumbers.push(inode->childVar);
    }
    if(nl) {
      _alternatives.push(nl);
      _nodeTypes.push(currType);
      return true;
    }
  }
  return false;
}


}
