
/*
 * File IGAlgorithm.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __IGAlgorithm__
#define __IGAlgorithm__

#include "Forwards.hpp"

#include "Lib/DHMap.hpp"
#include "Lib/DHSet.hpp"
#include "Lib/RatioKeeper.hpp"
#include "Lib/ScopedPtr.hpp"
#include "Lib/SmartPtr.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/LiteralSelector.hpp"
#include "Kernel/MainLoop.hpp"
#include "Kernel/Ordering.hpp"
#include "Kernel/RCClauseStack.hpp"

#include "Indexing/ClauseVariantIndex.hpp"
#include "Indexing/IndexManager.hpp"
#include "Indexing/LiteralIndex.hpp"
#include "Indexing/LiteralSubstitutionTreeWithoutTop.hpp"

#include "Inferences/GlobalSubsumption.hpp"
#include "Inferences/InferenceEngine.hpp"
#include "Inferences/TautologyDeletionISE.hpp"
#include "Inferences/URResolution.hpp"
#include "Inferences/DistinctEqualitySimplifier.hpp"

#include "SAT/SATSolver.hpp"

#include "Saturation/AWPassiveClauseContainer.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"
#include "Shell/EqualityProxy.hpp"

#include "Kernel/Grounder.hpp"

namespace InstGen {

using namespace Kernel;
using namespace Inferences;
using namespace Indexing;
using namespace SAT;
using namespace Saturation;
using namespace Shell;

class IGAlgorithm : public MainLoop {
public:
  CLASS_NAME(IGAlgorithm);
  USE_ALLOCATOR(IGAlgorithm);  
  
  typedef Statistics::TerminationReason TerminationReason;

  IGAlgorithm(Problem& prb, const Options& opt);
  ~IGAlgorithm();

  GroundingIndex& getGroundingIndex() { return *_groundingIndex.ptr(); }

  ClauseIterator getActive();

protected:
  virtual void init();
  virtual MainLoopResult runImpl();
private:

  bool addClause(Clause* cl);

  void doInprocessing(RCClauseStack& clauses);

  void restartWithCurrentClauses();
  void restartFromBeginning();


  void wipeIndexes();

  void processUnprocessed();
  void activate(Clause* cl, bool wasDeactivated=false);

  void deactivate(Clause* cl);
  void doImmediateReactivation();
  void doPassiveReactivation();

  void selectAndAddToIndex(Clause* cl);
  void removeFromIndex(Clause* cl);

  void tryGeneratingInstances(Clause* cl, unsigned litIdx);

  bool startGeneratingClause(Clause* orig, ResultSubstitution& subst, bool isQuery, Clause* otherCl,Literal* origLit, LiteralStack& genLits, bool& properInstance);
  void finishGeneratingClause(Clause* orig, ResultSubstitution& subst, bool isQuery, Clause* otherCl,Literal* origLit, LiteralStack& genLits);

  bool isSelected(Literal* lit);

  Clause* getFORefutation(SATClause* satRefutation);


  void onResolutionClauseDerived(Clause* cl);
  void doResolutionStep();


  MainLoopResult onModelFound();

 
  bool _doingSatisfiabilityCheck;

  RatioKeeper _instGenResolutionRatio;


  SATSolver* _satSolver;
  ScopedPtr<IGGrounder> _gnd;

 
  ScopedPtr<GroundingIndex> _groundingIndex;
  ScopedPtr<GlobalSubsumption> _globalSubsumption;

  Options _saturationOptions;
  ScopedPtr<IndexManager> _saturationIndexManager;
  ScopedPtr<Problem> _saturationProblem;
  ScopedPtr<SaturationAlgorithm> _saturationAlgorithm;

  OrderingSP _ordering;
  ScopedPtr<LiteralSelector> _selector;







 
  RCClauseStack _unprocessed;
 
  AWClauseContainer _passive;
 
  RCClauseStack _active;

 
  ClauseStack _deactivated;
  DHSet<Clause*> _deactivatedSet;

  RCClauseStack _inputClauses;

  bool _use_hashing;
  ClauseVariantIndex* _variantIdx;

  LiteralSubstitutionTree* _selected;

  DuplicateLiteralRemovalISE _duplicateLiteralRemoval;
  TrivialInequalitiesRemovalISE _trivialInequalityRemoval;
  TautologyDeletionISE _tautologyDeletion;
  DistinctEqualitySimplifier _distinctEqualitySimplifier;

  bool _use_niceness;
  bool _use_dm;

 
  struct DismatchingContraints {
    typedef DHMap<Literal*,DismatchingLiteralIndex*> Lit2Index;

    Lit2Index lit2index;

    void add(Literal* orig, Literal* inst) {
      DismatchingLiteralIndex* index;
      if (!lit2index.find(orig,index)) {
        LiteralIndexingStructure * is = new LiteralSubstitutionTreeWithoutTop();
        index = new DismatchingLiteralIndex(is); 
        ALWAYS(lit2index.insert(orig,index));
      }
      index->addLiteral(inst);
    }

    bool shouldBlock(Literal* orig, Literal* inst) {
      DismatchingLiteralIndex* index;
      
      return lit2index.find(orig,index) && index->getGeneralizations(inst,false,false).hasNext();
    }

    ~DismatchingContraints() {
      Lit2Index::Iterator iit(lit2index);
      while(iit.hasNext()){
        DismatchingLiteralIndex* index = iit.next();
        delete index;
      }
    }
  };

  typedef DHMap<Clause*,DismatchingContraints*> DismatchMap;

  DismatchMap _dismatchMap;

 
  EqualityProxy* _equalityProxy;
};

}

#endif 
