
/*
 * File BDD.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <utility>

#include "Lib/Cache.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Exception.hpp"
#include "Lib/Deque.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/DHSet.hpp"
#include "Lib/Int.hpp"
#include "Lib/List.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Timer.hpp"
#include "Lib/TimeCounter.hpp"

#include "Formula.hpp"
#include "Signature.hpp"
#include "Term.hpp"

#include "SAT/Preprocess.hpp"
#include "SAT/SingleWatchSAT.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "BDD.hpp"


namespace Kernel {

using namespace Lib;
using namespace SAT;





bool BDDNode::isTrue() const
{
  CALL("BDDNode::isTrue");
  return BDD::instance()->isTrue(this);
}

bool BDDNode::isFalse() const
{
  CALL("BDDNode::isFalse");
  return BDD::instance()->isFalse(this);
}






BDD* BDD::instance()
{
  CALL("BDD::instance");

  static BDD* inst=0;
  if(!inst) {
    inst=new BDD();
  }
  return inst;
}

BDD::BDD()
: _trueNode(0,0,0), _falseNode(0,0,0),
  _bddEvalPredicate(0), _nextNodeNum(1),
  _allowDefinitionOutput(true), _newVar(1)
{
  _trueNode._depth=0;
  _falseNode._depth=0;
}

BDD::~BDD()
{
  CALL("BDD::~BDD");

  ASS_REP(_allowDefinitionOutput, "Definition output has to be allowed at the BDD object destruction");

  NodeSet::Iterator nit(_nodes);
  while(nit.hasNext()) {
    BDDNode* node=nit.next();
    delete node;
  }
}

int BDD::getNewVar(unsigned pred)
{
  CALL("BDD::getNewVar(unsigned)");
  ASS_EQ(env.signature->predicateArity(pred), 0);

  int res=getNewVar();
  _predicateSymbols.insert(res, pred);
  return res;
}

vstring BDD::getPropositionalPredicateName(int var)
{
  CALL("BDD::getPropositionalPredicateName");

  vstring name;
  if(!getNiceName(var, name)) {
    vstring prefix(BDD_PREDICATE_PREFIX);
    prefix+=env.options->namePrefix();
    name = prefix + Int::toString(var);

    
    
    
    ASS(!env.signature->isPredicateName(name, 0));
  }

  return name;
}

bool BDD::getNiceName(int var, vstring& res)
{
  CALL("BDD::getNiceName");

  unsigned pred;
  bool found=_predicateSymbols.find(var, pred);
  if(found) {
    res=env.signature->predicateName(pred);
  }
  return found;
}

Signature::Symbol* BDD::getSymbol(int var)
{
  CALL("BDD::getSymbol");

  unsigned pred;
  if(_predicateSymbols.find(var, pred)) {
    return env.signature->getPredicate(pred);
  }
  return 0;
}

BDDNode* BDD::getAtomic(int varNum, bool positive)
{
  CALL("BDD::getAtomic");
  ASS_G(varNum,0);

  if(varNum>=_newVar) {
    _newVar=varNum+1;
  }

  if(positive) {
    return getNode(varNum, getTrue(), getFalse());
  } else {
    return getNode(varNum, getFalse(), getTrue());
  }
}

bool BDD::parseAtomic(BDDNode* node, unsigned& var, bool& positive)
{
  CALL("BDD::parseAtomic");

  if(isConstant(node)) {
    return false;
  }
  if(!isConstant(node->_pos) || !isConstant(node->_neg)) {
    return false;
  }
  if(isTrue(node->_pos)!=isFalse(node->_neg)) {
    return false;
  }

  var=node->_var;
  positive=isTrue(node->_pos);
  return true;
}

bool BDD::findTrivial(BDDNode* n, bool& areImplied, Stack<BDDNode*>& acc)
{
  CALL("BDD::findTrivial");
  ASS(acc.isEmpty());
  
  ASS(!n->isAtomic());

  if(n->isConst()) {
    return false;
  }

  
  static Deque<BDDNode*> que;
  que.reset();

  que.push_back(n);
  que.push_back(0);

  bool foundSome = false;

  bool haveTrueAside = false;
  bool haveFalseAside = false;

  unsigned nextVar = n->getVar();

  while(que.size()>1) {
    unsigned currVar = nextVar;
    nextVar = 0;

    bool canBeImpliedTrue = !haveTrueAside;
    bool canBeImpliedFalse = !haveTrueAside;
    bool canBeImplying = !haveFalseAside;
    bool canNegBeImplying = !haveFalseAside;

    
    

    while(que.front()!=0) {
      BDDNode* curr = que.pop_front();
      ASS(curr);
      if(curr->getVar()==currVar) {
	BDDNode* pos = curr->getPos();
	BDDNode* neg = curr->getNeg();

	if(pos->isFalse() || neg->isFalse()) {
	  haveFalseAside = true;
	}
	if(pos->isTrue() || neg->isTrue()) {
	  haveTrueAside = true;
	}

	canBeImpliedTrue &= neg->isFalse();
	canBeImpliedFalse &= pos->isFalse();
	canBeImplying &= neg->isTrue();
	canNegBeImplying &= pos->isTrue();

	if(!pos->isConst()) {
	  if(nextVar<pos->getVar()) { nextVar = pos->getVar(); }
	  que.push_back(pos);
	}
	if(!neg->isConst()) {
	  if(nextVar<neg->getVar()) { nextVar = neg->getVar(); }
	  que.push_back(neg);
	}
      }
      else {
	ASS_L(curr->getVar(),currVar);

	canBeImpliedTrue = false;
	canBeImpliedFalse = false;
	canBeImplying = false;
	canNegBeImplying = false;

	if(nextVar<curr->getVar()) { nextVar = curr->getVar(); }
	que.push_back(curr);
      }
    }
    ALWAYS(que.pop_front()==0);
    que.push_back(0);

    ASS(~canBeImpliedTrue || ~canBeImpliedFalse);
    if(canBeImpliedTrue) {
      ASS(!foundSome || areImplied);
      foundSome = true;
      areImplied = true;
      acc.push(getAtomic(currVar, true));
    }
    if(canBeImpliedFalse) {
      ASS(!foundSome || areImplied);
      foundSome = true;
      areImplied = true;
      acc.push(getAtomic(currVar, false));
    }
    ASS(~canBeImplying || ~canNegBeImplying);
    if(canBeImplying) {
      ASS(!foundSome || !areImplied);
      foundSome = true;
      areImplied = false;
      acc.push(getAtomic(currVar, true));
    }
    if(canNegBeImplying) {
      ASS(!foundSome || !areImplied);
      foundSome = true;
      areImplied = false;
      acc.push(getAtomic(currVar, false));
    }
  }
  ASS(!foundSome || assignValue(n, acc.top()->getVar(), !acc.top()->getPos()->isTrue())->isConst());
  return foundSome;
}

BDDNode* BDD::conjunction(BDDNode* n1, BDDNode* n2)
{
  CALL("BDD::conjunction");
  BDDNode* res=getBinaryFnResult(n1,n2, ConjunctionFn(this));

  if(isRefuted(n1) && isRefuted(n2)) {
    markRefuted(res);
  }

  return res;
}

BDDNode* BDD::disjunction(BDDNode* n1, BDDNode* n2)
{
  CALL("BDD::disjunction");
  BDDNode* res=getBinaryFnResult(n1,n2, DisjunctionFn(this));

  if(isRefuted(n1) || isRefuted(n2)) {
    markRefuted(res);
  }

  return res;
}

BDDNode* BDD::xOrNonY(BDDNode* x, BDDNode* y)
{
  CALL("BDD::xOrNonY");
  return getBinaryFnResult(x,y, XOrNonYFn(this));
}

BDDNode* BDD::assignValue(BDDNode* n, unsigned var, bool value)
{
  CALL("BDD::assignValue");

  BDDNode* asgnArg = getAtomic(var, value);
  return getBinaryFnResult(asgnArg, n, AssignFn(this));
}

bool BDD::isXOrNonYConstant(BDDNode* x, BDDNode* y, bool resValue)
{
  CALL("BDD::isXOrNonYConstant");
  if(resValue) {
    bool res=hasConstantResult<true>(x,y, XOrNonYFn(this));

    if(res && isRefuted(y)) {
      markRefuted(x);
    }

    return res;
  }
  else {
    return hasConstantResult<false>(x,y, XOrNonYFn(this));
  }
}

template<class BinBoolFn>
BDDNode* BDD::getBinaryFnResult(BDDNode* n1, BDDNode* n2, BinBoolFn fn)
{
  CALL("BDD::getBinaryFnResult");
  ASS(n1);
  ASS(n2);

  TimeCounter tc(TC_BDD);

  int counter=0;

  static Stack<BDDNode*> toDo(8);
  
  
  
  
  
  
  
  
  
  
  static Stack<BDDNode*> results(8);
  
  
  static Stack<int> vars(8);

  static Cache<pair<BDDNode*,BDDNode*>, BDDNode*, PtrPairSimpleHash > cache;
  
  cache.resetEvictionCounter();

  for(;;) {
    counter++;
    if(counter==50000) {
      counter=0;
      
      if(env.timeLimitReached()) {
	throw TimeLimitExceededException();
      }
    }
    if(BinBoolFn::commutative) {
      if(n1>n2) {
	swap(n1,n2);
      }
    }
    BDDNode* res=fn(n1,n2);
    if(res || cache.find(make_pair(n1, n2), res)) {
      while(results.isNonEmpty() && results.top()!=0) {
	BDDNode* pos=results.pop();
	BDDNode* neg=res;
	unsigned var=vars.pop();
	if(pos==neg) {
	  res=pos;
	} else {
	  res=getNode(var, pos, neg);
	}
	ASS_EQ(results.top(),0);
	results.pop();
	BDDNode* arg1=results.pop();
	BDDNode* arg2=results.pop();
	cache.insert(make_pair(arg1, arg2), res);

	if(BinBoolFn::op==DISJUNCTION) {
	  if(isRefuted(arg1) || isRefuted(arg2)) {
	    markRefuted(res);
	  }
	}

      }
      results.push(res);
    } else {
      
      unsigned splitVar=max(n1->_var, n2->_var);
      ASS_G(splitVar,0);
      toDo.push((n2->_var==splitVar) ? n2->_neg : n2);
      toDo.push((n1->_var==splitVar) ? n1->_neg : n1);
      toDo.push((n2->_var==splitVar) ? n2->_pos : n2);
      toDo.push((n1->_var==splitVar) ? n1->_pos : n1);
      results.push(n2);
      results.push(n1);
      results.push(0);
      
      
      vars.push(splitVar);
    }

    if(toDo.isEmpty()) {
      break;
    }
    n1=toDo.pop();
    n2=toDo.pop();
  }

  ASS(toDo.isEmpty());
  ASS_EQ(results.length(),1);
  return results.pop();
}

template<bool ResValue, class BinBoolFn>
bool BDD::hasConstantResult(BDDNode* n1, BDDNode* n2, BinBoolFn fn)
{
  CALL("BDD::hasConstantResult");
  ASS(n1);
  ASS(n2);

  TimeCounter tc(TC_BDD);

  int counter=0;

  static Stack<BDDNode*> toDo(8);
  toDo.reset();

  static Stack<pair<BDDNode*,BDDNode*> > current(8);
  current.reset();

  static Cache<pair<BDDNode*,BDDNode*>, EmptyStruct, PtrPairSimpleHash > examined;
  examined.resetEvictionCounter();

  for(;;) {
    counter++;
    if(counter==50000) {
      counter=0;
      
      if(env.timeLimitReached()) {
	throw TimeLimitExceededException();
      }
    }
    BDDNode* res=fn(n1,n2);
    if(res) {
      if( (ResValue && !isTrue(res)) ||
	      (!ResValue && !isFalse(res))) {
	return false;
      }
    }
    else {
      if(!examined.find(make_pair(n1, n2)))
      {
	current.push(make_pair(n1, n2));
	toDo.push(0);

	
	unsigned splitVar=max(n1->_var, n2->_var);
	ASS_G(splitVar,0);
	toDo.push((n2->_var==splitVar) ? n2->_neg : n2);
	toDo.push((n1->_var==splitVar) ? n1->_neg : n1);
	toDo.push((n2->_var==splitVar) ? n2->_pos : n2);
	toDo.push((n1->_var==splitVar) ? n1->_pos : n1);
      }
    }
    while(!toDo.isEmpty() && toDo.top()==0) {
      toDo.pop();
      ASS(current.isNonEmpty());
      if(counter%4) {
	examined.insert(current.top());
      }
      current.pop();
    }
    if(toDo.isEmpty()) {
      break;
    }
    n1=toDo.pop();
    n2=toDo.pop();
    ASS(n1);
    ASS(n2);
  }

  return true;
}

BDDNode* BDD::ConjunctionFn::operator()(BDDNode* n1, BDDNode* n2)
{
  if(_parent->isFalse(n1) || _parent->isFalse(n2)) {
    return _parent->getFalse();
  }
  if(_parent->isTrue(n1)) {
    return n2;
  }
  if(_parent->isTrue(n2)) {
    return n1;
  }
  if(n1==n2) {
    return n1;
  }
  return 0;
}

BDDNode* BDD::DisjunctionFn::operator()(BDDNode* n1, BDDNode* n2)
{
  if(n1==n2) {
    return n1;
  }
  if(_parent->isTrue(n1) || _parent->isTrue(n2)) {
    return _parent->getTrue();
  }
  if(_parent->isFalse(n1)) {
    return n2;
  }
  if(_parent->isFalse(n2)) {
    return n1;
  }
  return 0;
}

BDDNode* BDD::XOrNonYFn::operator()(BDDNode* n1, BDDNode* n2)
{
  if(n1==n2) {
    return _parent->getTrue();
  }
  if(_parent->isTrue(n1) || _parent->isFalse(n2)) {
    return _parent->getTrue();
  }
  if(_parent->isTrue(n2)) {
    return n1;
  }
  return 0;
}

BDDNode* BDD::AssignFn::operator()(BDDNode* n1, BDDNode* n2)
{
  CALL("BDD::AssignFn::operator ()");

  if(n1->isConst() || n2->isConst()) {
    
    return n2;
  }
  if(n1->getVar()!=n2->getVar()) {
    
    ASS_L(n1->getVar(),n2->getVar());
    return 0;
  }
  unsigned var;
  bool asgn;
  ALWAYS(_parent->parseAtomic(n1, var, asgn));
  ASS_EQ(var, n2->getVar());
  if(asgn) {
    return n2->getPos();
  }
  else {
    return n2->getNeg();
  }
}

BDDNode* BDD::getNode(int varNum, BDDNode* pos, BDDNode* neg)
{
  CALL("BDD::getNode");
  ASS_G(varNum,0);
  ASS_L(varNum,_newVar);
  ASS_NEQ(pos,neg);

  
  
  static BDDNode* newNode=0;

  if(newNode==0) {
    newNode=new BDDNode();
    env.statistics->bddMemoryUsage += sizeof(BDDNode);
  }

  newNode->_var=varNum;
  newNode->_pos=pos;
  newNode->_neg=neg;

  BDDNode* res=_nodes.insert(newNode);
  if(res==newNode) {
    newNode=0;
    res->_depth=max(pos->depth(), neg->depth())+1;
  }
  return res;
}


vstring BDD::toString(BDDNode* node)
{
  return getDefinition(node);
}

vstring BDD::toTPTPString(BDDNode* node, vstring bddPrefix)
{
  if(isTrue(node)) {
    return "$true";
  } else if(isFalse(node)) {
    return "$false";
  } else if(isTrue(node->_pos) && isFalse(node->_neg)) {
    return bddPrefix+Int::toString(node->_var);
  } else if(isFalse(node->_pos) && isTrue(node->_neg)) {
    return "~"+bddPrefix+Int::toString(node->_var);
  } else {
    return vstring("( ( ")+bddPrefix+Int::toString(node->_var)+" => "+toTPTPString(node->_pos, bddPrefix)+
      ") & ( ~"+bddPrefix+Int::toString(node->_var)+" => "+toTPTPString(node->_neg, bddPrefix)+" ) )";
  }
}

vstring BDD::toTPTPString(BDDNode* node)
{
  if(isTrue(node)) {
    return "$true";
  }
  else if(isFalse(node)) {
    return "$false";
  }
  else {
    return vstring("( ( ")+getPropositionalPredicateName(node->_var)+" => "+toTPTPString(node->_pos)+
      ") & ( ~"+getPropositionalPredicateName(node->_var)+" => "+toTPTPString(node->_neg)+" ) )";
  }
}


vstring BDD::getDefinition(BDDNode* node)
{
  
  if(isTrue(node)) {
    return "$true";
  }
  if(isFalse(node)) {
    return "$false";
  }

  vstring name;
  if(_nodeNames.find(node, name)) {
    return name;
  }

  vstring propPred=getPropositionalPredicateName(node->_var);
  if(isTrue(node->_pos) && isFalse(node->_neg)) {
    return propPred;
  }
  else if(isFalse(node->_pos) && isTrue(node->_neg)) {
    return "~"+propPred;
  }
  else if(isTrue(node->_pos)) {
    return "("+propPred+" | "+getDefinition(node->_neg)+")";
  }
  else if(isFalse(node->_neg)) {
    return "("+propPred+" & "+getDefinition(node->_pos)+")";
  }
  else if(isFalse(node->_pos)) {
    return "(~"+propPred+" & "+getDefinition(node->_neg)+")";
  }
  else if(isTrue(node->_neg)) {
    return "(~"+propPred+" | "+getDefinition(node->_pos)+")";
  }
  else {
    vstring posDef=getDefinition(node->_pos); 
    vstring negDef=getDefinition(node->_neg); 
    return introduceName(node, "("+propPred+" ? "+posDef+" : "+negDef+")");
  }

}

vstring BDD::introduceName(BDDNode* node, vstring definition)
{
  ASS(!_nodeNames.find(node));
  vstring name="$bddnode"+Int::toString(_nextNodeNum++);
  vstring report="BDD definition: "+name+" = "+definition;
  outputDefinition(report);
  ALWAYS(_nodeNames.insert(node, name));

  return name;
}

void BDD::allowDefinitionOutput(bool allow)
{
  _allowDefinitionOutput=allow;
  if(allow && _postponedDefinitions.isNonEmpty()) {
    unsigned stLen=_postponedDefinitions.size();
    env.beginOutput();
    for(unsigned i=0;i<stLen;i++) {
      env.out()<<_postponedDefinitions[i]<<endl;
    }
    env.endOutput();
    _postponedDefinitions.reset();
  }
}

void BDD::outputDefinition(vstring def)
{
  if(_allowDefinitionOutput) {
    env.beginOutput();
    env.out()<<def<<endl;
    env.endOutput();
  }
  else {
    _postponedDefinitions.push(def);
  }

}


vstring BDD::getName(BDDNode* node)
{
  vstring name;
  if(!_nodeNames.find(node, name)) {
    vstring def=getDefinition(node);
    
    if(!_nodeNames.find(node, name)) {
      name=introduceName(node, def);
    }
  }
  return name;
}

TermList BDD::getConstant(BDDNode* node)
{
  TermList res;
  if(!_nodeConstants.find(node, res)) {
    vstring name=getName(node);
    unsigned func;
    bool added;

    func=env.signature->addFunction(name, 0, added);
    while(!added) {
      name+="_1";
      func=env.signature->addFunction(name, 0, added);
      if(added) {
        vstring report="Name collision, BDD node now uses name "+name;
	outputDefinition(report);
        _nodeNames.set(node, name);
      }
    }
    res=TermList(Term::create(func, 0, 0));
    _nodeConstants.insert(node, res);
  }
  return res;
}

bool BDD::equals(const BDDNode* n1,const BDDNode* n2)
{
  return n1->_var==n2->_var && n1->_pos==n2->_pos && n1->_neg==n2->_neg;
}
unsigned BDD::hash(const BDDNode* n)
{
  CALL("BDD::hash");

  unsigned res=Hash::hash(n->_var);
  res=Hash::hash(n->_pos, res);
  res=Hash::hash(n->_neg, res);
  return res;
}

Formula* BDD::toFormula(BDDNode* node)
{
  if(isTrue(node)) {
    static Formula* tf=new Formula(true);
    return tf;
  } else if(isFalse(node)) {
    static Formula* ff=new Formula(false);
    return ff;
  }

  if(!_bddEvalPredicate) {
    vstring name="$bddEval";
    bool added;
    _bddEvalPredicate=env.signature->addPredicate(name, 1, added);
    while(!added) {
      name+="_1";
      _bddEvalPredicate=env.signature->addPredicate(name, 1, added);
    }
    ASS(_bddEvalPredicate);
  }
  TermList bddConst=getConstant(node);
  Literal* lit=Literal::create(_bddEvalPredicate,1,true,false,&bddConst);
  return new AtomicFormula(lit);

}


}
