
/*
 * File BDD.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __BDD__
#define __BDD__

#include <iosfwd>

#include "Forwards.hpp"
#include "Lib/Allocator.hpp"
#include "Lib/Array.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/Hash.hpp"
#include "Lib/Int.hpp"
#include "Lib/List.hpp"
#include "Lib/Set.hpp"
#include "Lib/SkipList.hpp"
#include "Lib/Stack.hpp"
#include "Lib/VString.hpp"

#include "Kernel/Signature.hpp"

#define BDD_PREDICATE_PREFIX "$bdd"

namespace Kernel {

using namespace std;
using namespace Lib;
using namespace SAT;

class BDDConjunction;

class BDDNode
{
public:
  CLASS_NAME(BDDNode);
  USE_ALLOCATOR(BDDNode);

  unsigned depth() const { return _depth; }
  bool isConst() const { return _var==0; }
  bool isTrue() const;
  bool isFalse() const;
  bool isAtomic() const
  { return !isConst() && getPos()->isConst() && getNeg()->isConst(); }

  unsigned getVar() const { ASS(!isConst()); return _var; }
  BDDNode* getPos() const { ASS(!isConst()); return _pos; }
  BDDNode* getNeg() const { ASS(!isConst()); return _neg; }
private:
  BDDNode() : _refuted(false) {}
  BDDNode(unsigned var, BDDNode* pos, BDDNode* neg) :
      _refuted(false), _var(var), _pos(pos), _neg(neg) {}

  bool _refuted : 1;
  unsigned _var : 31;
  unsigned _depth;

  BDDNode* _pos;
  BDDNode* _neg;

  friend class BDD;
  friend class BDDConjunction;
  friend class BDDClausifier;
  friend class Shell::LaTeX;
};

class BDD
{
public:
  CLASS_NAME(BDD);
  USE_ALLOCATOR(BDD);
  
  static BDD* instance();

  BDD();
  ~BDD();

 
  int getNewVar() { return _newVar++; }
  int getNewVar(unsigned pred);

  vstring getPropositionalPredicateName(int var);
  bool getNiceName(int var, vstring& res);
  Signature::Symbol* getSymbol(int var);

 
  BDDNode* getTrue() { return &_trueNode; }
 
  BDDNode* getFalse() { return &_falseNode; }

  BDDNode* getAtomic(int varNum, bool positive);

  BDDNode* conjunction(BDDNode* n1, BDDNode* n2);
  BDDNode* disjunction(BDDNode* n1, BDDNode* n2);
  BDDNode* xOrNonY(BDDNode* x, BDDNode* y);
 
  BDDNode* negation(BDDNode* n)
  { return xOrNonY(getFalse(), n); }
  BDDNode* assignValue(BDDNode* n, unsigned var, bool value);

  bool isXOrNonYConstant(BDDNode* x, BDDNode* y, bool resValue);

 
  bool isTrue(const BDDNode* node) { return node==getTrue(); }
 
  bool isFalse(const BDDNode* node) { return node==getFalse(); }
 
  bool isConstant(const BDDNode* node) { return node->_var==0; }

  bool parseAtomic(BDDNode* node, unsigned& var, bool& positive);

  bool findTrivial(BDDNode* n, bool& areImplied, Stack<BDDNode*>& acc);

  static bool equals(const BDDNode* n1,const BDDNode* n2);
  static unsigned hash(const BDDNode* n);

  vstring toString(BDDNode* node);
  vstring toTPTPString(BDDNode* node, vstring bddPrefix);
  vstring toTPTPString(BDDNode* node);

  Formula* toFormula(BDDNode* node);

  vstring getDefinition(BDDNode* node);
  vstring getName(BDDNode* node);
  TermList getConstant(BDDNode* node);

  void allowDefinitionOutput(bool allow);

  void markRefuted(BDDNode* n) { n->_refuted=true; }
  bool isRefuted(BDDNode* n) { return n->_refuted; }

private:
  void outputDefinition(vstring def);
  vstring introduceName(BDDNode* node, vstring definition);

  BDDNode* getNode(int varNum, BDDNode* pos, BDDNode* neg);

  template<class BinBoolFn>
  BDDNode* getBinaryFnResult(BDDNode* n1, BDDNode* n2, BinBoolFn fn);

  template<bool ResValue, class BinBoolFn>
  bool hasConstantResult(BDDNode* n1, BDDNode* n2, BinBoolFn fn);

  enum Operation
  {
    CONJUNCTION,
    DISJUNCTION,
    X_OR_NON_Y,
    ASSIGNMENT
  };

 
  struct ConjunctionFn
  {
    static const Operation op=CONJUNCTION;
    static const bool commutative=true;

    ConjunctionFn(BDD* parent) : _parent(parent) {}
    inline BDDNode* operator()(BDDNode* n1, BDDNode* n2);
    BDD* _parent;
  };

 
  struct DisjunctionFn
  {
    static const Operation op=DISJUNCTION;
    static const bool commutative=true;

    DisjunctionFn(BDD* parent) : _parent(parent) {}
    inline BDDNode* operator()(BDDNode* n1, BDDNode* n2);
    BDD* _parent;
  };

 
  struct XOrNonYFn
  {
    static const Operation op=X_OR_NON_Y;
    static const bool commutative=false;

    XOrNonYFn(BDD* parent) : _parent(parent) {}
    inline BDDNode* operator()(BDDNode* n1, BDDNode* n2);
    BDD* _parent;
  };

 
  struct AssignFn
  {
    static const Operation op=ASSIGNMENT;
    static const bool commutative=false;

    AssignFn(BDD* parent) : _parent(parent) {}
    inline BDDNode* operator()(BDDNode* n1, BDDNode* n2);
    BDD* _parent;
  };




 
  BDDNode _trueNode;
 
  BDDNode _falseNode;

 
  typedef Set<BDDNode*,BDD> NodeSet;
 
  NodeSet _nodes;

 
  DHMap<int, unsigned> _predicateSymbols;

  DHMap<BDDNode*,vstring> _nodeNames;
  DHMap<BDDNode*,TermList> _nodeConstants;
  unsigned _bddEvalPredicate;

  int _nextNodeNum;
  bool _allowDefinitionOutput;
  Stack<vstring> _postponedDefinitions;

 
  int _newVar;
};

};

#endif
