
/*
 * File ColorHelper.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/DHMap.hpp"
#include "Lib/Environment.hpp"

#include "Saturation/SaturationAlgorithm.hpp"

#include "Shell/Options.hpp"

#include "Clause.hpp"
#include "EqHelper.hpp"
#include "Inference.hpp"
#include "Renaming.hpp"
#include "Signature.hpp"
#include "Term.hpp"
#include "TermIterators.hpp"

#include "ColorHelper.hpp"

namespace Kernel
{

bool ColorHelper::isTransparent(bool predicate, unsigned functor)
{
  CALL("ColorHelper::isTransparent");

  Signature::Symbol* sym;
  if (predicate) {
    sym = env.signature->getPredicate(functor);
  }
  else {
    sym = env.signature->getFunction(functor);
  }
  return sym->color()==COLOR_TRANSPARENT;
}

bool ColorHelper::hasColoredPredicates(Clause* c)
{
  CALL("ColorHelper::hasColoredPredicates");

  unsigned clen = c->length();
  for (unsigned i=0; i<clen; i++) {
    Literal* lit = (*c)[i];

    if (!isTransparent(true, lit->functor())) {
      return true;
    }
  }
  return false;
}

void ColorHelper::collectColoredConstants(Clause* c, Stack<Term*>& acc)
{
  CALL("ColorHelper::collectColoredConstants");

  unsigned clen = c->length();
  for (unsigned i=0; i<clen; i++) {
    Literal* lit = (*c)[i];

    NonVariableIterator tit(lit,true);
    while (tit.hasNext()) {
      Term* t = tit.next().term();
      if (t->color() == COLOR_TRANSPARENT) {
	tit.right();
	continue;
      }
      if (t->arity() == 0) {
	acc.push(t);
      }
    }
  }
} 

Clause* ColorHelper::skolemizeColoredConstants(Clause* c)
{
  CALL("ColorHelper::skolemizeColoredConstants");

  Stack<Term*> coloredConstants;
  collectColoredConstants(c,coloredConstants);
  if (coloredConstants.isEmpty()) {
    return 0;
  }

  unsigned clen = c->length();
  static LiteralStack resStack;
  resStack.reset();
  resStack.loadFromIterator(Clause::Iterator(*c));

  ASS_EQ(resStack.size(), clen);

  while (coloredConstants.isNonEmpty()) {
    TermList replaced = TermList(coloredConstants.pop());

    unsigned newFn = env.signature->addSkolemFunction(0);
    TermList newTrm = TermList(Term::create(newFn, 0, 0));

    for (unsigned i=0; i<clen; i++) {
      resStack[i] = EqHelper::replace(resStack[i], replaced, newTrm);
    }
  }
  Clause* res = Clause::fromStack(resStack, c->inputType(), new Inference1(Inference::COLOR_UNBLOCKING, c));
  return res;
}

void ColorHelper::ensureSkolemReplacement(Term* t, TermMap& map)
{
  CALL("ColorHelper::ensureSkolemReplacement");

  
  
  

  Term* norm = Renaming::normalize(t);
  if (map.find(norm)) {
    return;
  }
  unsigned varCnt = norm->getDistinctVars();
  unsigned newFn = env.signature->addSkolemFunction(varCnt, "CU");

  static Stack<TermList> argStack;
  argStack.reset();
  
  
  for (unsigned i=0; i<varCnt; i++) {
    argStack.push(TermList(i, false));
  }
  Term* replacement = Term::create(newFn, varCnt, argStack.begin());
  map.insert(norm, replacement);
}

void ColorHelper::collectSkolemReplacements(Clause* c, TermMap& map)
{
  CALL("ColorHelper::collectSkolemReplacements");

  unsigned clen = c->length();
  for (unsigned i=0; i<clen; i++) {
    Literal* lit = (*c)[i];

    NonVariableIterator tit(lit,true);
    while (tit.hasNext()) {
      Term* t = tit.next().term();
      if (!isTransparent(false, t->functor())) {
	ensureSkolemReplacement(t, map);
	
	tit.right();
      }
    }
  }
}

Term* ColorHelper::applyReplacement(Term* t, TermMap& map)
{
  CALL("ColorHelper::applyReplacement");

  Renaming r;
  r.normalizeVariables(t);
  Term* norm = r.apply(t);
  ASS(map.find(norm));
  Term* tgtNorm = map.get(norm);

  Renaming inv;
  inv.makeInverse(r);
  Term* tgt = inv.apply(tgtNorm);
  ASS(tgt->containsAllVariablesOf(t));
  ASS(t->containsAllVariablesOf(tgt));
  ASS_EQ(tgt->color(), COLOR_TRANSPARENT);
  return tgt;
}

Clause* ColorHelper::skolemizeColoredTerms(Clause* c)
{
  CALL("ColorHelper::skolemizeColoredTerms");

  static TermMap replMap;
  replMap.reset();

  collectSkolemReplacements(c, replMap);
  if (replMap.isEmpty()) {
    return 0;
  }

  static LiteralStack resStack;
  resStack.reset();

  unsigned clen = c->length();
  for (unsigned i=0; i<clen; i++) {
    Literal* lit = (*c)[i];
  start_replacing:
    NonVariableIterator nvit(lit);
    while (nvit.hasNext()) {
      Term* t = nvit.next().term();
      if (!isTransparent(false,t->functor())) {
	
	Term* newTrm = applyReplacement(t,replMap);
	lit = EqHelper::replace(lit,TermList(t),TermList(newTrm));
	goto start_replacing;
      }
    }
    resStack.push(lit);
  }

  ASS_EQ(resStack.size(), clen);
  Clause* res = Clause::fromStack(resStack, c->inputType(), new Inference1(Inference::COLOR_UNBLOCKING, c));
  return res;
}

void ColorHelper::tryUnblock(Clause* c, SaturationAlgorithm* salg)
{
  CALL("ColorHelper::tryUnblock");


  if (!hasColoredPredicates(c)) {

    Clause* unblocked = skolemizeColoredTerms(c);
    if (unblocked) {
      if (env.options->showBlocked()) {
	env.beginOutput();
	env.out()<<"Unblocking clause "<<unblocked->toString()<<endl;
	env.endOutput();
      }
      salg->addNewClause(unblocked);
    }
  }

}

} 
