
/*
 * File Formula.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __Formula__
#define __Formula__

#include <utility>

#include "Forwards.hpp"

#include "Lib/List.hpp"
#include "Lib/XML.hpp"

#include "Connective.hpp"
#include "Term.hpp"


namespace Kernel {

using namespace Lib;

class Formula
{
public:
  typedef List<int> VarList;
 
  explicit Formula (bool value)
    : _connective(value ? TRUE : FALSE)
  {
  } 

  
 
  Connective connective () const { return _connective; }

  TermList termLetLhs() const;
  TermList termLetRhs() const;
  Literal* formulaLetLhs() const;
  Formula* formulaLetRhs() const;
  Formula* letBody() const;
  Formula* condArg() const;
  Formula* thenArg() const;
  Formula* elseArg() const;
  const FormulaList* args() const;
  FormulaList* args();
  FormulaList** argsPtr();
  const Formula* left() const;
  Formula* left();
  const Formula* right() const;
  Formula* right();
  const Formula* qarg() const;
  Formula* qarg();
  const VarList* vars() const;
  VarList* vars();
  const Formula* uarg() const;
  Formula* uarg();
  const Literal* literal() const;
  Literal* literal();
  VarList* freeVariables () const;
  VarList* boundVariables () const;

  
  bool equals(const Formula*) const;
  void collectAtoms(Stack<Literal*>& acc);
  void collectPredicates(Stack<unsigned>& acc);
  void collectPredicatesWithPolarity(Stack<pair<unsigned,int> >& acc, int polarity=1);

  XMLElement toXML() const;

  
  vstring toString() const;
  vstring toStringInScopeOf(Connective con) const;
  static vstring toString(Connective con);
  bool parenthesesRequired(Connective outer) const;
  
  void destroy();

  unsigned weight() const;

  Color getColor();
  bool getSkip();

  static Formula* fromClause(Clause* cl, BDDNode* prop);

  static Formula* fromClause(Clause* cl);

  static Formula* quantify(Formula* f);

  static Formula* trueFormula();
  static Formula* falseFormula();

  
  CLASS_NAME(Formula);
  USE_ALLOCATOR(Formula);
protected:
 
  explicit Formula(Connective con)
    : _connective(con)
  {}

 
  Connective _connective;
  
  static vstring _connectiveNames[];
}; 

class AtomicFormula
  : public Formula
{
public:
 
  explicit AtomicFormula (Literal* lit)
    : Formula(LITERAL),
      _literal(lit)
  {}
 
  const Literal* getLiteral() const { return _literal; }
 
  Literal* getLiteral() { return _literal; }

  
  CLASS_NAME(AtomicFormula);
  USE_ALLOCATOR(AtomicFormula);
protected:
 
  Literal* _literal;
}; 


class QuantifiedFormula
  : public Formula
{
 public:
 
  QuantifiedFormula(Connective con, VarList* vs, Formula* arg)
    : Formula(con),
      _vars(vs),
      _arg(arg)
  {
    ASS(con == FORALL || con == EXISTS);
    ASS(vs);
  }

 
  const Formula* subformula () const { return _arg; }
 
  Formula* subformula () { return _arg; }
 
  const VarList* varList() const { return _vars; }
 
  VarList* varList() { return _vars; }

  
  CLASS_NAME(QuantifiedFormula);
  USE_ALLOCATOR(QuantifiedFormula);
 protected:
 
  VarList* _vars;
 
  Formula* _arg;
}; 

class NegatedFormula
  : public Formula
{
public:
 
  explicit NegatedFormula (Formula* f)
    : Formula(NOT),
      _arg(f)
  {}

 
  const Formula* subformula() const { return _arg; }
 
  Formula* subformula() { return _arg; }

  
  CLASS_NAME(NegatedFormula);
  USE_ALLOCATOR(NegatedFormula);
protected:
 
  Formula* _arg;
}; 


class BinaryFormula
  : public Formula
{
public:
 
  explicit BinaryFormula (Connective con,Formula* lhs,Formula* rhs)
    : Formula(con),
      _left(lhs),
      _right(rhs)
  {
    ASS(con == IFF || con == XOR || con == IMP);
  }

 
  const Formula* lhs() const { return _left; }
 
  Formula* lhs() { return _left; }
 
  const Formula* rhs() const { return _right; }
 
  Formula* rhs() { return _right; }

  
  CLASS_NAME(BinaryFormula);
  USE_ALLOCATOR(BinaryFormula);
protected:
 
  Formula* _left;
 
  Formula* _right;
}; 


class JunctionFormula
  : public Formula
{
 public:
  JunctionFormula (Connective con, FormulaList* args)
    : Formula(con),
      _args(args)
  {
    ASS(con == AND || con == OR);
    ASS_GE(args->length(),2);
  }

 
  void setArgs(FormulaList* args) { _args = args; }

 
  const FormulaList* getArgs() const { return _args; }
 
  FormulaList* getArgs() { return _args; }
 
  FormulaList** getArgsPtr() { return &_args; }

  static Formula* generalJunction(Connective c, FormulaList* args);

  
  CLASS_NAME(JunctionFormula);
  USE_ALLOCATOR(JunctionFormula);
 protected:
 
  FormulaList* _args;
}; 

class IteFormula
  : public Formula
{
 public:
  IteFormula(Formula* condarg, Formula* thenarg, Formula* elsearg)
    : Formula(ITE),
      _condarg(condarg),
      _thenarg(thenarg),
      _elsearg(elsearg)
  {
  }

 
  Formula* condArg() const { return _condarg; }
 
  Formula* thenArg() const { return _thenarg; }
 
  Formula* elseArg() const { return _elsearg; }

  
  CLASS_NAME(IteFormula);
  USE_ALLOCATOR(IteFormula);
 protected:
  Formula* _condarg;
  Formula* _thenarg;
  Formula* _elsearg;
}; 

class FormulaLetFormula
  : public Formula
{
 public:
 
  FormulaLetFormula (Literal* lhs, Formula* rhs, Formula* body)
    : Formula(FORMULA_LET),
      _lhs(lhs),
      _rhs(rhs),
      _body(body)
  {
    ASS(lhs->hasOnlyDistinctVariableArgs());
    ASS(lhs->shared());
  }

 
  Literal* lhs() const { return _lhs; }
 
  Formula* rhs() const { return _rhs; }
 
  Formula* body() const { return _body; }

  
  CLASS_NAME(FormulaLetFormula);
  USE_ALLOCATOR(FormulaLetFormula);
 protected:
  Literal* _lhs;
  Formula* _rhs;
  Formula* _body;
}; 

class TermLetFormula
  : public Formula
{
 public:
 
  TermLetFormula (TermList lhs, TermList rhs, Formula* body)
    : Formula(TERM_LET),
      _lhs(lhs),
      _rhs(rhs),
      _body(body)
  {
    ASS(lhs.isSafe());
    ASS(lhs.isVar() || lhs.term()->hasOnlyDistinctVariableArgs());
  }

 
  TermList lhs() const { return _lhs; }
 
  TermList rhs() const { return _rhs; }
 
  Formula* body() const { return _body; }

  
  CLASS_NAME(TermLetFormula);
  USE_ALLOCATOR(TermLetFormula);
 protected:
  TermList _lhs;
  TermList _rhs;
  Formula* _body;
}; 



inline
const Formula::VarList* Formula::vars() const
{
  ASS(_connective == FORALL || _connective == EXISTS);
  return static_cast<const QuantifiedFormula*>(this)->varList();
}
inline
Formula::VarList* Formula::vars()
{
  ASS(_connective == FORALL || _connective == EXISTS);
  return static_cast<QuantifiedFormula*>(this)->varList();
}

inline
const Formula* Formula::qarg() const
{
  ASS(_connective == FORALL || _connective == EXISTS);
  return static_cast<const QuantifiedFormula*>(this)->subformula();
}
inline
Formula* Formula::qarg()
{
  ASS(_connective == FORALL || _connective == EXISTS);
  return static_cast<QuantifiedFormula*>(this)->subformula();
}

inline
const Formula* Formula::uarg() const
{
  ASS(_connective == NOT);
  return static_cast<const NegatedFormula*>(this)->subformula();
}

inline
Formula* Formula::uarg()
{
  ASS(_connective == NOT);
  return static_cast<NegatedFormula*>(this)->subformula();
}

inline
const FormulaList* Formula::args() const
{
  ASS(_connective == AND || _connective == OR);
  return static_cast<const JunctionFormula*>(this)->getArgs();
}

inline
FormulaList* Formula::args()
{
  ASS(_connective == AND || _connective == OR);
  return static_cast<JunctionFormula*>(this)->getArgs();
}

inline
FormulaList** Formula::argsPtr()
{
  ASS(_connective == AND || _connective == OR);
  return static_cast<JunctionFormula*>(this)->getArgsPtr();
}

inline
const Literal* Formula::literal() const
{
  ASS(_connective == LITERAL);
  return static_cast<const AtomicFormula*>(this)->getLiteral();
}

inline
Literal* Formula::literal()
{
  ASS(_connective == LITERAL);
  return static_cast<AtomicFormula*>(this)->getLiteral();
}

inline
const Formula* Formula::left() const
{
  ASS(_connective == IFF || _connective == XOR || _connective == IMP);
  return static_cast<const BinaryFormula*>(this)->lhs();
}
inline
Formula* Formula::left()
{
  ASS(_connective == IFF || _connective == XOR || _connective == IMP);
  return static_cast<BinaryFormula*>(this)->lhs();
}

inline
const Formula* Formula::right() const
{
  ASS(_connective == IFF || _connective == XOR || _connective == IMP);
  return static_cast<const BinaryFormula*>(this)->rhs();
}
inline
Formula* Formula::right()
{
  ASS(_connective == IFF || _connective == XOR || _connective == IMP);
  return static_cast<BinaryFormula*>(this)->rhs();
}
inline
Formula* Formula::condArg() const {
  ASS(_connective == ITE);
  return static_cast<const IteFormula*>(this)->condArg();
}
inline
Formula* Formula::thenArg() const {
  ASS(_connective == ITE);
  return static_cast<const IteFormula*>(this)->thenArg();
}
inline
Formula* Formula::elseArg() const {
  ASS(_connective == ITE);
  return static_cast<const IteFormula*>(this)->elseArg();
}
inline
Formula* Formula::letBody() const {
  if(_connective == TERM_LET) {
    return static_cast<const TermLetFormula*>(this)->body();
  }
  else {
    ASS(_connective == FORMULA_LET)
    return static_cast<const FormulaLetFormula*>(this)->body();
  }
}
inline
Literal* Formula::formulaLetLhs() const {
  ASS(_connective == FORMULA_LET)
  return static_cast<const FormulaLetFormula*>(this)->lhs();
}
inline
Formula* Formula::formulaLetRhs() const {
  ASS(_connective == FORMULA_LET)
  return static_cast<const FormulaLetFormula*>(this)->rhs();
}
inline
TermList Formula::termLetLhs() const {
  ASS(_connective == TERM_LET);
  return static_cast<const TermLetFormula*>(this)->lhs();
}
inline
TermList Formula::termLetRhs() const {
  ASS(_connective == TERM_LET);
  return static_cast<const TermLetFormula*>(this)->rhs();
}



std::ostream& operator<< (ostream& out, const Formula& f);
std::ostream& operator<< (ostream& out, const Formula* f);

}

#endif 
