
/*
 * File FormulaVarIterator.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "FormulaVarIterator.hpp"
#include "Formula.hpp"
#include "Term.hpp"
#include "TermIterators.hpp"

using namespace Lib;
using namespace Kernel;

FormulaVarIterator::FormulaVarIterator (const Formula* f)
  : _found(false)
{
  _instructions.push(FVI_FORMULA);
  _formulas.push(f);
} 

FormulaVarIterator::FormulaVarIterator (const Term* trm)
  : _found(false)
{
  CALL("FormulaVarIterator::FormulaVarIterator(const Literal*)");
  _instructions.push(FVI_TERM);
  _terms.push(trm->args());
} 

int FormulaVarIterator::next ()
{
  CALL("FormulaVarIterator::next");

  ASS(_found);
  _found = false;
  return _nextVar;
} 

bool FormulaVarIterator::hasNext()
{
  CALL("FormulaVarIterator::hasNext");

  if (_found) return true;
  
  while (! _instructions.isEmpty()) {
    switch (_instructions.pop()) {
    case FVI_FORMULA:
      {
	const Formula* f = _formulas.pop();
	switch (f->connective()) {
	case LITERAL:
	  _instructions.push(FVI_TERM);
	  _terms.push(f->literal()->args());
	  break;

	case AND:
	case OR:
	  {
	    FormulaList::Iterator fs(f->args());
	    while (fs.hasNext()) {
	      _instructions.push(FVI_FORMULA);
	      _formulas.push(fs.next());
	    }
	  }
	  break;

	case IMP:
	case IFF:
	case XOR:
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->left());
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->right());
	  break;

	case NOT:
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->uarg());
	  break;

	case FORALL:
	case EXISTS:
	  {
	    IntList::Iterator vs(f->vars());
	    while (vs.hasNext()) {
	      _bound.inc(vs.next());
	    }
	    _instructions.push(FVI_UNBIND);
	    _formulas.push(f);
	    _instructions.push(FVI_FORMULA);
	    _formulas.push(f->qarg());
	  }
	  break;
	  
	case ITE:
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->condArg());
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->thenArg());
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->elseArg());
	  break;

	case TRUE:
	case FALSE:
	  break;

	case FORMULA_LET:
	case TERM_LET:
	{
	  _instructions.push(FVI_FORMULA);
	  _formulas.push(f->letBody());
	  _instructions.push(FVI_UNBIND);
	  _formulas.push(f);
	  countFormulaLetLhsVars(f, true);
	  if(f->connective()==FORMULA_LET) {
	    _instructions.push(FVI_FORMULA);
	    _formulas.push(f->formulaLetRhs());
	  }
	  else {
	    TermList rhs = f->termLetRhs();
	    if(rhs.isTerm()) {
	      _instructions.push(FVI_TERM);
	      _terms.push(rhs.term()->args());
	    }
	    else {
	      if(suggestNextVar(rhs.var())) {
		return true;
	      }
	    }
	  }
	}
	}
      }
      break;
    case FVI_TERM:
      {
	const TermList* ts = _terms.pop();
	if (ts->isEmpty()) break;
	_instructions.push(FVI_TERM);
	_terms.push(ts->next());
	if (ts->isVar()) {
	  unsigned v = ts->var();
	  if(suggestNextVar(v)) {
	    return true;
	  }
	}
	else {
	  const Term* trm = ts->term();
	  if(trm->isSpecial()) {
	    if(processSpecialTerm(ts)) {
	      return true;
	    }
	  }
	  else {
	    _instructions.push(FVI_TERM);
	    _terms.push(trm->args());
	  }
	}
      }
      break;
    case FVI_UNBIND:
      {
	const Formula* f = _formulas.pop();
	Connective con = f->connective();
	if(con==FORMULA_LET || con==TERM_LET) {
	  countFormulaLetLhsVars(f, false);
	}
	else {
	  ASS(con==EXISTS || con==FORALL)
	  IntList::Iterator vs(f->vars());
	  while (vs.hasNext()) {
	    _bound.dec(vs.next());
	  }
	}
      }
      break;
    case FVI_UNBIND_TERM_LET:
      {
	const TermList* t = _terms.pop();
	countTermLetLhsVars(t->term(), false);
      }
      break;
    }
  }
  return false;
} 

bool FormulaVarIterator::processSpecialTerm(const TermList* ts)
{
  CALL("FormulaVarIterator::processSpecialTerm");

  const Term* t = ts->term();
  ASS(t->isSpecial())

  const Term::SpecialTermData* sd = t->getSpecialData();
  if(t->functor()==Term::SF_TERM_ITE) {
    _instructions.push(FVI_FORMULA);
    _formulas.push(sd->getCondition());
    _instructions.push(FVI_TERM);
    _terms.push(t->args());
  }
  else {
    _instructions.push(FVI_TERM);
    _terms.push(t->args());
    _instructions.push(FVI_UNBIND_TERM_LET);
    _terms.push(ts);
    countTermLetLhsVars(t, true);
    if(t->functor()==Term::SF_LET_FORMULA_IN_TERM) {
      _instructions.push(FVI_FORMULA);
      _formulas.push(sd->getRhsFormula());
    }
    else {
      TermList rhs = sd->getRhsTerm();
      if(rhs.isTerm()) {
	_instructions.push(FVI_TERM);
	_terms.push(rhs.term()->args());
      }
      else {
	if(suggestNextVar(rhs.var())) {
	  return true;
	}
      }
    }
  }
  return false;
}

bool FormulaVarIterator::suggestNextVar(unsigned v)
{
  CALL("FormulaVarIterator::suggestNextVar");

  if (_free.get(v)) return false;
  if (_bound.get(v)) return false;
  _nextVar = v;
  _found = true;
  _free.inc(v);
  return true;
}

void FormulaVarIterator::countFormulaLetLhsVars(const Formula* f, bool inc)
{
  CALL("FormulaVarIterator::countFormulaLetLhsVars");

  static VariableIterator vit;
  if(f->connective()==FORMULA_LET) {
    vit.reset(f->formulaLetLhs());
  }
  else {
    ASS_EQ(f->connective(), TERM_LET);
    if(f->termLetLhs().isTerm()) {
      vit.reset(f->termLetLhs().term());
    }
    else {
      INVALID_OPERATION("Let expressions with variable lhs are currently not handled");
    }
  }
  while(vit.hasNext()) {
    TermList var = vit.next();
    if(inc) {
      _bound.inc(var.var());
    }
    else {
      _bound.dec(var.var());
    }
  }
}

void FormulaVarIterator::countTermLetLhsVars(const Term* t, bool inc)
{
  CALL("FormulaVarIterator::countTermLetLhsVars");

  static VariableIterator vit;
  const Term::SpecialTermData* sd = t->getSpecialData();
  if(t->functor()==Term::SF_LET_FORMULA_IN_TERM) {
    vit.reset(sd->getLhsLiteral());
  }
  else {
    ASS_EQ(t->functor(),Term::SF_LET_TERM_IN_TERM);
    if(sd->getLhsTerm().isTerm()) {
      vit.reset(sd->getLhsTerm().term());
    }
    else {
      INVALID_OPERATION("Let expressions with variable lhs are currently not handled");
    }
  }
  while(vit.hasNext()) {
    TermList var = vit.next();
    if(inc) {
      _bound.inc(var.var());
    }
    else {
      _bound.dec(var.var());
    }
  }
}
