
/*
 * File InferenceStore.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */



#ifndef __InferenceStore__
#define __InferenceStore__

#include <utility>
#include <ostream>

#include "Forwards.hpp"

#include "Lib/Allocator.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/DHMultiset.hpp"
#include "Lib/Stack.hpp"
#include "Lib/VString.hpp"

#include "Kernel/Signature.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"

namespace Kernel {

using namespace Lib;

class SplittingRecord;

struct UnitSpec
{
  UnitSpec() {}
  explicit UnitSpec(Unit* u, bool ignoreProp=false) : _unit(u){tag=u->number();}
  bool operator==(const UnitSpec& o) const { return _unit==o._unit;}
  bool operator!=(const UnitSpec& o) const { return !(*this==o); }

  static unsigned hash(const UnitSpec& o)
  {
    ASS(!o.isEmpty());
    return Hash::hash(o._unit);
  }
  static bool equals(const UnitSpec& left, const UnitSpec& right){ return left==right; }

  bool isEmpty() const { return _unit==0; }
  bool isClause() const { ASS(!isEmpty()); return _unit->isClause(); }

  Clause* cl() const
  {
    ASS(!isEmpty());
    ASS(_unit->isClause());
    return static_cast<Clause*>(_unit);
  }
  Unit* unit() const { ASS(!isEmpty()); return _unit; }

  vstring toString() const
  {
    if(isClause()) {
	return cl()->toString();
    }
    else {
	return unit()->toString();
    }
    
  }
private:
  Unit* _unit;
  int tag;
};

typedef VirtualIterator<UnitSpec> UnitSpecIterator;


class InferenceStore
{
public:
  CLASS_NAME(InferenceStore);
  USE_ALLOCATOR(InferenceStore);
  
  static InferenceStore* instance();

  typedef List<int> IntList;

  struct FullInference
  {
    FullInference(unsigned premCnt) : csId(0), premCnt(premCnt) { }

    void* operator new(size_t,unsigned premCnt)
    {
      size_t size=sizeof(FullInference)+premCnt*sizeof(UnitSpec);
      size-=sizeof(UnitSpec);

      return ALLOC_KNOWN(size,"InferenceStore::FullInference");
    }

    size_t occupiedBytes()
    {
      size_t size=sizeof(FullInference)+premCnt*sizeof(UnitSpec);
      size-=sizeof(UnitSpec);
      return size;
    }

    void increasePremiseRefCounters();

    int csId;
    unsigned premCnt;
    Inference::Rule rule;
    UnitSpec premises[1];
  };


  
  class SplittingRecord
  {
  public:
    SplittingRecord(Clause* splitClause) : namedComps(1), premise(UnitSpec(splitClause)) {}

    Stack<pair<int,Clause*> > namedComps;
    UnitSpec premise;
    UnitSpec result;


    CLASS_NAME(InferenceStore::SplittingRecord);
    USE_ALLOCATOR(SplittingRecord);
  };

  void recordInference(UnitSpec unit, FullInference* inf);

  
  void recordSplittingNameLiteral(UnitSpec us, Literal* lit);

  void recordIntroducedSymbol(Unit* u, bool func, unsigned number);

  void outputProof(ostream& out, Unit* refutation);
  void outputProof(ostream& out, UnitList* units);

  UnitSpecIterator getParents(UnitSpec us, Inference::Rule& rule);
  UnitSpecIterator getParents(UnitSpec us);

  vstring getUnitIdStr(UnitSpec cs);
  vstring getClauseIdSuffix(UnitSpec cs);

  bool findInference(UnitSpec cs, FullInference*& finf)
  {
    return _data.find(cs,finf);
  }

  bool findSplitting(UnitSpec cs, SplittingRecord*& srec)
  {
    return _splittingRecords.find(cs,srec);
  }


private:
  InferenceStore();

  struct ProofPrinter;
  struct TPTPProofPrinter;
  struct ProofCheckPrinter;

  ProofPrinter* createProofPrinter(ostream& out);

 
  DHMap<UnitSpec, FullInference*, UnitSpec> _data;
  DHMultiset<Clause*, PtrIdentityHash> _nextClIds;

  DHMap<UnitSpec, SplittingRecord*, UnitSpec> _splittingRecords;

  DHMap<UnitSpec, Literal*> _splittingNameLiterals;


  DHMap<Clause*, IntList*> _bddizeVars;

 
  typedef pair<bool,unsigned> SymbolId;
  typedef Stack<SymbolId> SymbolStack;
  DHMap<unsigned,SymbolStack> _introducedSymbols;

  
};


};

#endif
