
/*
 * File Number.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */

 
#if GNUMP
#include <stdlib.h>
#include <cmath>
#include <limits>
#include <gmp.h>

#include "Lib/Exception.hpp"
#include "Lib/Hash.hpp"
#include "Lib/Int.hpp"
#include "Lib/Random.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Array.hpp"
#include "Number.hpp"
#include "Rational.hpp"



namespace Kernel
{
 
namespace __Aux_Number
{


bool nativeEqual(const NativeNumber& n1, const NativeNumber& n2)
{
  CALL("nativeEqual");

  NativeNumber diff=fabs(n1-n2);

  if(n1>1) {
    return diff<=(fabs(n1)*1E-11);
  }
  else {
    return diff<=1E-11;
  } 
}

void reduceNumbers(size_t cnt, NativeNumber** vals)
{
  CALL("reduceNumbers");

  size_t i = 0;
  while(i<cnt && *vals[i]==0) { i++; }

  if(i==cnt) {
    return;
  }

  long double absVal = fabs(*vals[i]);
  long double smallest = absVal;
  long double largest = absVal;
  for(; i<cnt; i++) {
    absVal = fabs(*vals[i]);
    if(absVal==0) {
      continue;
    }
    if(smallest>absVal) {
      smallest = absVal;
    }
    if(largest<absVal) {
      largest = absVal;
    }
  }
  ASS_GE(largest,smallest);
  long double divisor;
  if(smallest>4) {
    divisor = smallest;
  }
  else if(largest<.25) {
    divisor = largest;
  }
  else {
    return;
  }
  ASS_G(divisor,0);
  for(size_t i=0; i<cnt; i++) {
    *vals[i]/=divisor;
  }
}

long long intGcd(long long a, long long b)
{
  long long tmp;
  while (b) {
    tmp = b;
    b = a % b;
    a = tmp;
  }
  return a;
}

bool getIntFromDouble(NativeNumber val, NativeNumber& res)
{
  CALL("getIntFromDouble");

  if(floorl(val)==val && val<numeric_limits<long long>::max()
      && val>numeric_limits<long long>::min()) {
    res = static_cast<long long>(val);
    return true;
  }
  return false;
}

void reduceIntNumbers(size_t cnt, NativeNumber** vals)
{
  CALL("reduceIntNumbers");

  size_t i = 0;
  while(i<cnt && *vals[i]==0) { i++; }

  if(i==cnt) {
    return;
  }

  NativeNumber ival;
  if(!getIntFromDouble(*vals[i], ival)) {
    return;
  }
  long long gcd_val = ::abs(ival);

  for(; i<cnt; i++) {
    if(!getIntFromDouble(*vals[i], ival)) {
      return;
    }
    if(ival==0) {
      continue;
    }
    gcd_val = intGcd(gcd_val, ::abs(ival));
  }

  if(gcd_val==1) {
    return;
  }
  ASS_G(gcd_val,1);
  for(size_t i=0; i<cnt; i++) {
    *vals[i]/=gcd_val;
  }
}

int getDecimalPlaces(double dbVal)
{
  dbVal = fmod(dbVal, 1);
  static const int MAX_DP = 24;
  double THRES = pow(0.1, MAX_DP);
  if (dbVal == 0.0)
    return 0;
  int nDecimal = 0;

 
  while (dbVal - floor(dbVal) > THRES &&
	 ceil(dbVal) - dbVal > THRES &&
	 nDecimal < MAX_DP)
    {
      dbVal *= 10.0;
      THRES *= 10.0;
      nDecimal++;
    }
  return nDecimal;
}

void doubleToRational(double dbVal, long long* num, long long* den){
	int noDec;
	noDec = getDecimalPlaces(dbVal);
	
	
	if (noDec == 0){
		*den = 1;
		*num = (long long)dbVal;
		return;
	}
	
	
	*den = pow(10,noDec);
	dbVal = dbVal * (*den);
	
	long double temp;
	modf(dbVal, &temp);
	
	*num = (long long)temp;
}

long double getDoubleNumber(double lhs, double rhs) {
	long long numLhs, numRhs, denLhs, denRhs;
	
	doubleToRational(lhs, &numLhs, &denLhs);
	doubleToRational(rhs, &numRhs, &denRhs);
	
	if ( nativeEqual(rhs,0) || nativeEqual(lhs,0) ){
		return 0.0;
	}
	
	if( nativeEqual(lhs,rhs)){
		return lhs;
	}
	
	if(denLhs==1 || denRhs==1){
		return (long double) (lhs+rhs)/2;
	}

	int noIntegerParts = 0;
	
	Array<long long> res;
	
	long long tempA, tempDenA, tempB, tempDenB;
	do {
		tempA = numLhs / denLhs;
		tempDenA = numLhs % denLhs;
		tempB = numRhs / denRhs;
		tempDenB = numRhs % denRhs;
		numLhs = denLhs;
		denLhs = tempDenA;
		numRhs = denRhs;
		denRhs = tempDenB;
		res[noIntegerParts++] = tempA;
	} while (tempA == tempB && denLhs != 0 && denRhs !=0);
	if (denLhs == 0 || denRhs==0){
		return (long double)(lhs+rhs)/2;
	}
	long long den = 1, num = 1;
	
	
	if (noIntegerParts==1){
		return (long double)(rhs+lhs)/2;
	}
	
	if (tempA > tempB) {
		
		den = tempB + den;
	} else {
		
		den = tempA + den;
	}

	long double result;
	
	for (int i = noIntegerParts - 2; i > 0; i--) {
		long long tempDen = den;
		den = (res[i] * den) + num;
		num = tempDen;
	}
	
	num = res[0] * den + num;
	
	result = (long double) num / den;
	return result;
}


}

using namespace Lib;
using namespace __Aux_Number;


bool CommonNumberBase::_usePrecise = false;
bool CommonNumberBase::_useRational = false;

NativeNumber CommonNumberBase::parseString(vstring str)
{
  CALL("CommonNumberBase::parseString");

  double dbl;
  ALWAYS(Int::stringToDouble(str.c_str(), dbl));
  return dbl;
}

unsigned CoeffNumber::hash(const CoeffNumber& n)
{
  CALL("CoeffNumber::hash");

  if(usePrecise()) {
    vstring str = n.precise().get_str(16);
    return Hash::hash(str);
  }
  if(useRational()){
	  if(n.isZero()){
		  return 12345;
	  }
	  float lowerPrec = static_cast<float>(n.rational().toDouble());
	  return Hash::hash(lowerPrec);
  }
  else {
    if(n.isZero()) {
      return 1234; 
    }
    
    
    
    float lowerPrec = static_cast<float>(n.native());
    return Hash::hash(lowerPrec);
  }
}

void CoeffNumber::reduceNumbers(size_t cnt, CoeffNumber** vals, bool allowDecimal)
{
  CALL("CoeffNumber::reduceNumbers");

  if(usePrecise()) {
    size_t i = 0;
    while(i<cnt && vals[i]->precise()==0) { i++; }

    if(i==cnt) {
      return;
    }

    mpz_class gcd_val = ::abs(vals[i]->precise());

    for(; i<cnt; i++) {
      if(vals[i]->precise()==0) {
	continue;
      }
      mpz_class absNum = ::abs(vals[i]->precise());
      mpz_gcd(gcd_val.get_mpz_t(), gcd_val.get_mpz_t(), absNum.get_mpz_t());
    }

    if(gcd_val==1) {
      return;
    }
    ASS_G(gcd_val,1);
    for(size_t i=0; i<cnt; i++) {
      vals[i]->precise()/=gcd_val;
    }
  }
  else if(useRational()){
;
  }
  else {
    static Stack<NativeNumber*> numPtrs;
    numPtrs.reset();
    for(size_t i=0; i<cnt; i++) {
      numPtrs.push(&(vals[i]->native()));
    }
    if(allowDecimal) {
      __Aux_Number::reduceNumbers(cnt, numPtrs.begin());
    }
    else {
      __Aux_Number::reduceIntNumbers(cnt, numPtrs.begin());
    }
  }
}

BoundNumber BoundNumber::getRandomValue(const BoundNumber& min, const BoundNumber& max)
{
  CALL("BoundNumber::getRandomValue");

  ASS_L(min,max);
  if(usePrecise()) {
    static const unsigned randomDivisor = 64;
    Precise diff = max.precise()-min.precise();
    Precise part = (diff*(Random::getInteger(randomDivisor-2)+1))/randomDivisor;
    return BoundNumber(min.precise()+part);
  }
  else if(useRational()){
	  double den, num, den1, num1;
	  den = (min.rational().Denomination());
	  den1 = (max.rational().Denomination());
	  num = (min.rational().Numerator());
	  num1 = (max.rational().Numerator());
	  if(den == den1 )
		  return BoundNumber(Rational(Random::getDouble(num, num1),den));
	  else
	  {
		  double cden = den* den1;
		  double min = num*cden;
		  double max = num1*cden;
		  return BoundNumber(Rational(Random::getDouble(min, max), cden));
	  }
  }
  else {
    return BoundNumber(Random::getDouble(min.native(), max.native()));
  }
}

BoundNumber BoundNumber::getMagicNumber(BoundNumber& rhs){
	CALL("BoundNumber::getMagicNumber");
	if (this->usePrecise() || this->useRational()){
		
		if(this->usePrecise() && mpq_cmp( this->getPrecise().__get_mp(), rhs.getPrecise().__get_mp()))
		{	
			return rhs;
		}
		if(this->useRational() && this->getRational() == rhs.getRational())
		{	
			return rhs;
		}
	
		mpz_class numLhs, numRhs, denLhs, denRhs;
		if(this->useRational()){
			numLhs = (unsigned long)this->getRational().Numerator();
			numRhs = (unsigned long)rhs.getRational().Numerator();
			denLhs = (unsigned long)this->getRational().Denomination();
			denRhs = (unsigned long)rhs.getRational().Denomination();
		}
		else
			{
			numLhs = (getPrecise().get_num());
			numRhs = (rhs.getPrecise().get_num());
			denLhs = (getPrecise().get_den());
			denRhs = (rhs.getPrecise().get_den());
			}
		
	    if( !cmp(denLhs,1) || !cmp(denRhs,1) ){
	    	if(usePrecise()){
	    		mpq_class result(mpz_class(numLhs+numRhs),2);
	    		
	    		return BoundNumber(result);
	    	}
	    	else{
	    		
	    		Rational result = Rational((numLhs.get_ui()+numRhs.get_ui()),2);
	    		return BoundNumber(result);
	    	}
	    }

	    int noIntegerParts = 0;
	    Array<mpz_class>res;
	    
	    mpz_class tempA, tempDenA, tempB, tempDenB;

	    do{
	    	tempA = numLhs / denLhs;
	    	tempDenA = numLhs % denLhs;
	    	tempB = numRhs / denRhs;
	    	tempDenB = numRhs % denRhs;
	    	numLhs = denLhs;
	    	denLhs = tempDenA;
	    	numRhs = denRhs;
	    	denRhs = tempDenB;
	    	res[noIntegerParts++] = tempA;
	    	}while(!mpz_cmp(tempA.__get_mp(), tempB.__get_mp()) && !mpz_cmp(denLhs.__get_mp(), 0) && !mpz_cmp(denRhs.__get_mp(),0));
	    if(!cmp(denRhs,0) || !cmp(denLhs,0)){
	    	if(usePrecise()){
	    		
	    		return BoundNumber(mpq_class(getPrecise()+rhs.getPrecise())/2);
	    	}
	    	if(useRational()) {
	    		Rational a = getRational();
	    		Rational b = rhs.getRational();
	    		Rational result = (a + b) / Rational(2);
	    		
	    		return BoundNumber(result);
	    	}
	    	ASS(!useRational() || !usePrecise());
	    }
	    
		if (noIntegerParts == 1) {
			if (usePrecise()) {
				
				return BoundNumber(
						mpq_class(getPrecise() + rhs.getPrecise()) / 2);
			}
			if(useRational()){
				Rational a = getRational();
				Rational b = rhs.getRational();
				Rational result = (a+b)/Rational(2);
				
				return BoundNumber(result);
						
			}
			ASS(!useRational() || !usePrecise());
		}

	    mpz_class den(1),num(1);
	    if(mpz_cmp(tempA.__get_mp(),tempB.__get_mp())>0){
	    	
	    	den = tempB + den;
	    } else {
	    	
	    	den = tempA + den;
	    }

	    for (int i = noIntegerParts-2; i>0 ; i--) {
	    	mpz_class tempDen = den;
	    	den = (res[i] * den) + num;
	    	num = tempDen;
	    	}
	    num = res[0]*den + num;
	    if(usePrecise()){
	    	mpq_class result(num, den);
	    	
	    	return BoundNumber(result);
	    }
	    if(useRational()) {
	    	
	    	Rational result(num.get_ui() ,den.get_ui());
	    	return BoundNumber(result);
	    }
	    ASS(!useRational() || !usePrecise());
	}

	
	ASS(!usePrecise() || !useRational());
	NativeNumber left, right;
	left = getNative();
	right = rhs.getNative();

    return BoundNumber(getDoubleNumber((double)left,(double)right));
}


bool usingPreciseNumbers()
{
  CALL("usingPreciseNumbers");

  return CommonNumberBase::usePrecise();
}

bool usingRationalNumbers(){
	CALL("usingRationalNumbers()");

	return CommonNumberBase::useRational();
}
void switchToPreciseNumbers()
{
  CALL("switchToPreciseNumbers");
  ASS(!CommonNumberBase::usePrecise());
  CommonNumberBase::switchToPreciseNumbers();
  ASS(CommonNumberBase::usePrecise());
}

void switchToRationalNumbers(){
	CALL("switchToRationalNumbers");
	ASS(!CommonNumberBase::useRational());

	CommonNumberBase::switchToRationalNumbers();
	ASS(CommonNumberBase::useRational());
}

std::ostream& operator<< (ostream& out, const CoeffNumber& num)
{
  CALL("operator<<(ostream&,const CoeffNumber&)");

  if(CommonNumberBase::usePrecise()) {
    out<<num.precise();
  }
  else if(CommonNumberBase::useRational()){
	  Rational rat = num.rational();
	  out<<rat.toString();
  }
  else {
    out<<num.native();
  }
  return out;
}

std::ostream& operator<< (ostream& out, const BoundNumber& num)
{
  CALL("operator<<(ostream&,const BoundNumber&)");

  if(CommonNumberBase::usePrecise()) {
    out<<num.precise();
  }
  else if(CommonNumberBase::useRational()){
	  Rational rat = num.rational();
	  out<<rat.toString();
  }
  else {
    out<<num.native();
  }
  return out;
}

}
#endif
