
/*
 * File Number.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __Number__
#define __Number__
#if GNUMP
#include <iosfwd>
#include <gmpxx.h>

#include "Forwards.hpp"

#include "Lib/Exception.hpp"
#include "Lib/Int.hpp"
#include "Rational.hpp"

#include <cmath>


namespace Kernel {

namespace __Aux_Number
{

typedef double NativeNumber;

bool nativeEqual(const NativeNumber& n1, const NativeNumber& n2);

class CommonNumberBase
{
public:
  static bool usePrecise() { return _usePrecise; }
  static void switchToPreciseNumbers() { _usePrecise = true; _useRational=false; }

  static bool useRational() {return _useRational;}
  static void switchToRationalNumbers() {_useRational = true; _usePrecise=false; }
protected:

  static NativeNumber parseString(vstring str);

  static bool _usePrecise;
  static bool _useRational;
};

template<class T, class Rat, class PrecCl>
class NumberBase : protected CommonNumberBase
{
public:
  typedef PrecCl Precise;
  typedef Rat NativeRational;

 
  NumberBase() : _isPrecise(false),_isRational(false) {}

  explicit NumberBase(const NativeNumber val) : _isPrecise(false), _isRational(false) {
    CALL("NumberBase(Native)");
	  if(usePrecise()) {
      initPrecise();
      precise() = static_cast<double>(val);
    }
    else if(useRational()){
    	initRational();
    	rational() = val;
    }
    else {
      native() = val;
    }

  }
  explicit NumberBase(const Precise& val) : _isPrecise(false), _isRational(false) {
    CALL("NumberBase(Precise)");
    ASS(usePrecise());
    initPrecise();
    precise() = val;
  }
  explicit NumberBase(Rational val) : _isPrecise(false), _isRational(false){
	  ASS(useRational());
	  initRational();
	  rational() = val;
  }
  template<class T2, class Ratio, class Prec2>
  explicit NumberBase(const NumberBase<T2,Ratio,Prec2>& val) : _isPrecise(false),_isRational(false) {
    if(usePrecise()) {
      initPrecise();
      precise() = val.precise();
    }
    else if(useRational()){
    	initRational();
    	rational() = val.native();
    }
    else {
    	native() = val.native();
    }
  }
  NumberBase(const NumberBase& val) : _isPrecise(false),_isRational(false) {
    if(usePrecise()) {
      initPrecise();
      precise() = val.precise();
    }else if(useRational()){
    	initRational();
    	rational() = val.rational();
    }
    else {
      native() = val.native();
    }
  }

  ~NumberBase() {
 
  }

  NumberBase& operator=(const NumberBase& val) {
    if(usePrecise()) {
      if(!_isPrecise) {
    	  initPrecise();
      }
      precise() = val.precise();
    }
    else if(useRational()){
    	if(!_isRational){
    		initRational();
    	}
    	rational() = val.rational();
    }
    else {
      native() = val.native();
    }
    return *this;
  }

  
  T operator+(const T& o) const {
    if(usePrecise()) {
      return T(precise()+o.precise());
    }
    else if(useRational()){
    	return T(rational()+o.rational());
    }
    else {
      return T(native()+o.native());
    }
  }
  T operator-() const {
    if(usePrecise()) {
      return T(-precise());
    }
    else if(useRational()){
    	return T(-rational());
    }
    else {
      return T(-native());
    }
  }
  T operator-(const T& o) const {
    if(usePrecise()) {
      return T(precise()-o.precise());
    }
    else if(useRational()){
    	return T(rational()-o.rational());
    }
    else {
      return T(native()-o.native());
    }
  }
  T operator*(const T& o) const {
    if(usePrecise()) {
      return T(precise()*o.precise());
    }
    else if(useRational()){
    	return T(rational()*o.rational());
    }
    else {
      return T(native()*o.native());
    }
  }
  bool operator==(const T& o) const {
    if(usePrecise()) {
      return precise()==o.precise();
    }
    else if(useRational()){
    	return rational()==o.rational();
    }
    else {
      return nativeEqual(native(), o.native());
    }
  }
  bool operator>(const T& o) const {
    if(usePrecise()) {
      return precise()>o.precise();
    }
    else if(useRational()){
    	return rational()>o.rational();
    }
    else {
      return (*this)!=o && native()>o.native();
    }
  }
 
  T abs() const {
    if(usePrecise()) {
      return T(::abs(precise()));
    }
    else if(useRational()){
    	return T(rational().isNegative() ? (-rational()) : rational());

    }
    else {
      return T((native()<0) ? (-native()) : native());
    }
  }
 
  T ceil() const {
	  if(usePrecise()){
		  return T(::abs(::ceil((mpf_class)(precise()))));
	  }
	  else if(useRational()){
		  
		  return T(rational());
	  }
	  else {
		  ASS(!usePrecise());
		  return T(ceill(native()));
	  }
  }

 
  T floor() const {
	  if(usePrecise()){
	  		  return T(::abs(::floor((mpf_class)(precise()))));
	  	  }
	  else if(useRational()){
		  return T(rational());
	  }
	  else {
	  	  ASS(!usePrecise());
	  	  return T(floorl(native()));
	  }
  }

  
  static const T& zero() { static T res(0); return res; }
  static const T& one() { static T res(1); return res; }
  static const T& minusOne() { static T res(-1); return res; }
  static const T& two() { static T res(2); return res; }

  
  bool operator!=(const T& o) const { return !(static_cast<const T&>(*this)==o); }
  bool operator<(const T& o) const { return o>static_cast<const T&>(*this); }
  bool operator>=(const T& o) const { return !(o>static_cast<const T&>(*this)); }
  bool operator<=(const T& o) const { return !(static_cast<const T&>(*this)>o); }

  T& operator+=(const T& o) {
    if(usePrecise()) {
      precise()+=o.precise();
    }
    else if(useRational()){
    	rational()+=o.rational();
    }
    else {
      native()+=o.native();
    }
    return static_cast<T&>(*this);
  }
  T& operator-=(const T& o) {
    if(usePrecise()) {
      precise()-=o.precise();
    }
    else if(useRational()){
    	rational()-=o.rational();
    }
    else {
      native()-=o.native();
    }
    return static_cast<T&>(*this);
  }
  T& operator*=(const T& o) {
    if(usePrecise()) {
      precise()*=o.precise();
    }
    else if(useRational()){
    	rational()*=o.rational();
    }
    else {
      native()*=o.native();
    }
    return static_cast<T&>(*this);
  }

  bool isPositive() const { return (*this)>zero(); }
  bool isNegative() const { return (*this)<zero(); }
  bool isZero() const { return (*this)==zero(); }

  bool isPositiveAssumingNonzero() const {
    if(usePrecise()) {
      return precise()>0;
    }
    else if(useRational()){
    	return rational().isPositiveNonZero();
    }
    else {
      return native()>0;
    }
  }
  bool isNegativeAssumingNonzero() const {
    return !isPositiveAssumingNonzero();
  }


  void transitionNativeToPrecise()
  {
    CALL("NumberBase::transitionNativeToPrecise");
    ASS(usePrecise());
    if(!useRational()){
    	NativeNumber val=_native;
    	if(static_cast<double>(val)!=static_cast<double>(val)){
    		cout<<_native<<static_cast<double>(val);
    		cout<<_native<<endl;
    		ASSERTION_VIOLATION;
    	}
    	
    	initPrecise();
    	precise() = static_cast<double>(val); 
    }
    else {
    	double val = double(_rational);
    	initPrecise();
    	precise() = val;
    }
  }
  void transitionNativeToRational(){
	  CALL("NumberBase::transitionNativeToRational");
	  ASS(useRational());
	  NativeNumber val = _native;
	  initRational();
	  rational() = static_cast<double>(val);

  }

  NativeNumber native() const { ASS(!usePrecise()); return _native; }
  const Precise& precise() const { return const_cast<NumberBase&>(*this).precise(); }
  const Rational& rational() const {return const_cast<NumberBase&>(*this).rational();}
protected:
  NativeNumber& native() { ASS(!usePrecise()); return _native; }

  void initPrecise() {
    CALL("NumberBase::initPrecise");

    _isPrecise = true;
    _isRational = false;
    new(&precise()) Precise();
  }

  void initRational() {
	  CALL("NumberBase::initRational");

	  _isRational = true;
	  _isPrecise = false;

	  new(&rational()) NativeRational();

  }
  void destroyPrecise() {
    CALL("NumberBase::destroyPrecise")
    ASS(_isPrecise);

  
  }

  Precise& precise() {
    ASS(usePrecise());
    if(!_isPrecise) {
      transitionNativeToPrecise();
    }
    return _precise;
  }

  NativeRational& rational(){
	  ASS(useRational());
	  if(!_isRational){
		  initRational();
		  rational() = _native;
	  }
	  return _rational;
  }

private:
 
  NativeNumber _native;
  Precise _precise;
  NativeRational _rational;

  bool _isPrecise;
  bool _isRational;
};

typedef NumberBase<CoeffNumber, Rational,  mpz_class> CoeffNumberBase;
typedef NumberBase<BoundNumber, Rational, mpq_class> BoundNumberBase;

}

using namespace __Aux_Number;

class CoeffNumber : public CoeffNumberBase
{
public:
 
  CoeffNumber() {}
  explicit CoeffNumber(const NativeNumber val) : CoeffNumberBase(val) {}
  explicit CoeffNumber(const NativeRational val) : CoeffNumberBase(val) {}
  explicit CoeffNumber(const Precise& val) : CoeffNumberBase(val) {}
  explicit CoeffNumber(vstring val) : CoeffNumberBase(parseString(val)) {}

  static void reduceNumbers(size_t cnt, CoeffNumber** vals, bool allowDecimal=true);

  static unsigned hash(const CoeffNumber& n);
};

class BoundNumber : public BoundNumberBase
{
public:
  BoundNumber() {}
  explicit BoundNumber(const NativeNumber val) : BoundNumberBase(val) {}
  explicit BoundNumber(const NativeRational val) : BoundNumberBase(val){}
  explicit BoundNumber(const Precise& val) : BoundNumberBase(val) {}
  explicit BoundNumber(const CoeffNumber& val) : BoundNumberBase(val) {}
  explicit BoundNumber(vstring val) : BoundNumberBase(parseString(val)) {}

  static BoundNumber getRandomValue(const BoundNumber& min, const BoundNumber& max);
  BoundNumber getMagicNumber(BoundNumber& rhs);
  using BoundNumberBase::operator*;

  BoundNumber operator*(const CoeffNumber& o) const {
    if(usePrecise()) {
      return BoundNumber(precise()*o.precise());
    }
    else if(useRational()){
    	return BoundNumber(rational()*o.rational());
    }
    else {
      return BoundNumber(native()*o.native());
    }
  }

  BoundNumber operator/(const BoundNumber& o) const {
    CALL("BoundNumber::operator/(const BoundNumber&)");
    ASS(!o.isZero());
    if(usePrecise()) {
      return BoundNumber(precise()/o.precise());
    }
    else if(useRational()){
    	return BoundNumber(rational()/o.rational());
    }
    else {
      return BoundNumber(native()/o.native());
    }
  }
  BoundNumber operator/(const CoeffNumber& o) const {
    CALL("BoundNumber::operator/(const CoeffNumber&)");
    ASS(!o.isZero());

    if(usePrecise()) {
      return BoundNumber(precise()/o.precise());
    }
    else if(useRational()){
    	return BoundNumber(rational()/o.rational());
    }
    else {
      return BoundNumber(native()/o.native());
    }
  }
  BoundNumber& operator/=(const BoundNumber& o) {
    CALL("BoundNumber::operator/=");
    ASS(!o.isZero());
    if(usePrecise()) {
      precise()/=o.precise();
    }
    else if(useRational()){
    	rational()/=o.rational();
    }
    else {
      native()/=o.native();
    }
    return *this;
  }
  const NativeNumber getNative() const {ASS(!usePrecise() && !useRational()); return native();}
  const Precise& getPrecise() const {ASS(usePrecise() && !useRational()); return precise();}
  const Rational& getRational() const {ASS(useRational() && !usePrecise()); return rational();}
};

void switchToPreciseNumbers();
bool usingPreciseNumbers();
void switchToRationalNumbers();
bool usingRationalNumbers();

std::ostream& operator<< (ostream& out, const CoeffNumber& num);
std::ostream& operator<< (ostream& out, const BoundNumber& num);

}

#endif 
#endif 
