
/*
 * File Signature.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __Signature__
#define __Signature__

#include "Forwards.hpp"

#include "Debug/Assertion.hpp"

#include "Lib/Allocator.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Map.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/VString.hpp"

#include "Sorts.hpp"
#include "Theory.hpp"


namespace Kernel {

using namespace std;
using namespace Lib;

class Signature
{
 public:
 
  class Symbol {
  protected:
   
    vstring _name;
   
    unsigned _arity;
   
    unsigned _interpreted : 1;
   
    unsigned _introduced : 1;
   
    unsigned _protected : 1;
   
    unsigned _skip : 1;
   
    unsigned _cfName : 1;
   
    unsigned _swbName : 1;
   
    unsigned _equalityProxy : 1;
   
    unsigned _color : 2;
   
    unsigned _stringConstant : 1;
   
    unsigned _numericConstant : 1;
   
    unsigned _answerPredicate : 1;
   
    mutable BaseType* _type;
   
    List<unsigned>* _distinctGroups;

    ~Symbol();
  public:
   
    Symbol(const vstring& nm,unsigned arity, bool interpreted=false, bool stringConstant=false,bool numericConstant=false);
    void destroyFnSymbol();
    void destroyPredSymbol();

    void addColor(Color color);
   
    void markIntroduced() { _introduced=1; }
   
    void markProtected() { _protected=1; }
   
    void markSkip() { _skip=1; }
   
    void markCFName() { ASS_EQ(arity(), 0); _cfName=1; markProtected(); }
   
    void markAnswerPredicate() { _answerPredicate=1; markProtected(); }
   
    void markSWBName() { ASS_EQ(arity(), 0); _swbName=1; }
   
    void markEqualityProxy() { _equalityProxy=1; }
   
    bool skip() const { return _skip; }
   
    bool cfName() const { return _cfName; }
   
    bool swbName() const { return _swbName; }
   
    Color color() const { return static_cast<Color>(_color); }
   
    inline unsigned arity() const { return _arity; }
   
    inline const vstring& name() const { return _name; }
   
    inline bool interpreted() const { return _interpreted; }
   
    inline bool introduced() const { return _introduced; }
   
    inline bool protectedSymbol() const { return _protected; }
   
    inline bool stringConstant() const { return _stringConstant; }
   
    inline bool numericConstant() const { return _numericConstant; }
   
    inline bool answerPredicate() const { return _answerPredicate; }
   
    inline bool equalityProxy() const { return _equalityProxy; }

   
    inline bool integerConstant() const
    { return interpreted() && arity()==0 && fnType()->result()==Sorts::SRT_INTEGER; }
   
    inline bool rationalConstant() const
    { return interpreted() && arity()==0 && fnType()->result()==Sorts::SRT_RATIONAL; }
   
    inline bool realConstant() const
    { return interpreted() && arity()==0 && fnType()->result()==Sorts::SRT_REAL; }
          
   
    inline IntegerConstantType integerValue() const
    { ASS(integerConstant()); return static_cast<const IntegerSymbol*>(this)->_intValue; }
   
    inline RationalConstantType rationalValue() const
    { ASS(rationalConstant()); return static_cast<const RationalSymbol*>(this)->_ratValue; }
   
    inline RealConstantType realValue() const
    { ASS(realConstant()); return static_cast<const RealSymbol*>(this)->_realValue; }

    const List<unsigned>* distinctGroups() const { return _distinctGroups; }
   
    void addToDistinctGroup(unsigned group,unsigned this_number);

    void setType(BaseType* type);
    void forceType(BaseType* type);
    FunctionType* fnType() const;
    PredicateType* predType() const;

    CLASS_NAME(Signature::Symbol);
    USE_ALLOCATOR(Symbol);
  }; 

  
  class VarSymbol{
  protected:
   
    vstring _name;
  public: 
   
    VarSymbol(const vstring& nm);
    
   
    inline const vstring& name() const {return _name;}
    
    CLASS_NAME("Signature::VarSymbol");
    USE_ALLOCATOR(VarSymbol);
  };
  
  
  class InterpretedSymbol
  : public Symbol
  {
    friend class Signature;
    friend class Symbol;
  protected:
    Interpretation _interp;

  public:

    InterpretedSymbol(const vstring& nm, Interpretation interp)
    : Symbol(nm, Theory::getArity(interp), true), _interp(interp)
    {
      CALL("InterpretedSymbol");
    }

    InterpretedSymbol(const vstring& nm)
    : Symbol(nm, 0, true)
    {
      CALL("InterpretedSymbol");
    }
    CLASS_NAME(Signature::InterpretedSymbol);
    USE_ALLOCATOR(InterpretedSymbol);

   
    inline Interpretation getInterpretation() const { ASS(interpreted()); ASS_NEQ(arity(),0); return _interp; }
  };

  class IntegerSymbol
  : public Symbol
  {
    friend class Signature;
    friend class Symbol;
  protected:
    IntegerConstantType _intValue;

  public:
    IntegerSymbol(const IntegerConstantType& val)
    : Symbol(val.toString(), 0, true), _intValue(val)
    {
      CALL("IntegerSymbol");

      setType(new FunctionType(0, 0, Sorts::SRT_INTEGER));
    }
    CLASS_NAME(Signature::IntegerSymbol);
    USE_ALLOCATOR(IntegerSymbol);
  };

  class RationalSymbol
  : public Symbol
  {
    friend class Signature;
    friend class Symbol;
  protected:
    RationalConstantType _ratValue;

  public:
    RationalSymbol(const RationalConstantType& val)
    : Symbol(val.toString(), 0, true), _ratValue(val)
    {
      CALL("RationalSymbol");

      setType(new FunctionType(0, 0, Sorts::SRT_RATIONAL));
    }
    CLASS_NAME(Signature::RationalSymbol);
    USE_ALLOCATOR(RationalSymbol);
  };

  class RealSymbol
  : public Symbol
  {
    friend class Signature;
    friend class Symbol;
  protected:
    RealConstantType _realValue;

  public:
    RealSymbol(const RealConstantType& val)
    : Symbol(val.toNiceString(), 0, true), _realValue(val)
    {
      CALL("RealSymbol");

      setType(new FunctionType(0, 0, Sorts::SRT_REAL));
    }
    CLASS_NAME(Signature::RealSymbol);
    USE_ALLOCATOR(RealSymbol);
  };
    
  
  
  
  
 
  unsigned addVar(const vstring& name){
    CALL("Signature::addVar");
    bool added;
    return addVar(name, added);
  }
  
  const vstring& varName(Var number){
    return _vars[number]->name();
  }
  
 
  size_t vars() const {return _vars.length(); }
  
  
 
  inline VarSymbol* getVar(unsigned n){
    CALL("Signature::getVar");
    ASS(n<vars());
    return _vars[n];
  }
  
  unsigned addVar(const vstring& name, bool& added);
  
  
  

  unsigned addPredicate(const vstring& name,unsigned arity,bool& added);
  unsigned addFunction(const vstring& name,unsigned arity,bool& added);

 
  unsigned addPredicate(const vstring& name,unsigned arity)
  {
    bool added;
    return addPredicate(name,arity,added);
  }
 
  unsigned addFunction(const vstring& name,unsigned arity)
  {
    bool added;
    return addFunction(name,arity,added);
  }
 
  unsigned addStringConstant(const vstring& name);
  unsigned addFreshFunction(unsigned arity, const char* prefix, const char* suffix = 0);
  unsigned addIteFunction(unsigned arity, unsigned* argSorts, unsigned resSort);
  unsigned addSkolemFunction(unsigned arity,const char* suffix = 0);
  unsigned addFreshPredicate(unsigned arity, const char* prefix, const char* suffix = 0);
  unsigned addNamePredicate(unsigned arity);

  

  unsigned addInterpretedFunction(Interpretation itp, const vstring& name);
  unsigned addInterpretedPredicate(Interpretation itp, const vstring& name);

  unsigned addIntegerConstant(const vstring& number,bool defaultSort);
  unsigned addRationalConstant(const vstring& numerator, const vstring& denominator,bool defaultSort);
  unsigned addRealConstant(const vstring& number,bool defaultSort);

  unsigned addIntegerConstant(const IntegerConstantType& number);
  unsigned addRationalConstant(const RationalConstantType& number);
  unsigned addRealConstant(const RealConstantType& number);

  unsigned getInterpretingSymbol(Interpretation interp);

 
  bool haveInterpretingSymbol(Interpretation interp) const { return _iSymbols.find(interp); }

 
  bool anyInterpretedSymbols() const
  {
    CALL("Signature::anyInterpretedSymbols");
    ASS_G(_iSymbols.size(),0); 

    return _iSymbols.size()!=1;
  }

 
  const vstring& functionName(int number)
  {
    return _funs[number]->name();
  }
 
  const vstring& predicateName(int number)
  {
    return _preds[number]->name();
  }
 
  const unsigned functionArity(int number)
  {
    CALL("Signature::functionArity");
    return _funs[number]->arity();
  }
 
  const unsigned predicateArity(int number)
  {
    CALL("Signature::predicateArity");
    return _preds[number]->arity();
  }

  const bool predicateColored(int number)
  {
    return _preds[number]->color()!=COLOR_TRANSPARENT;
  }

  const bool functionColored(int number)
  {
    return _funs[number]->color()!=COLOR_TRANSPARENT;
  }

 
  bool isPredicateName(vstring name, unsigned arity)
  {
    vstring symbolKey = key(name,arity);
    unsigned tmp;
    return _predNames.find(symbolKey,tmp);
  }

 
  unsigned functions() const { return _funs.length(); }
 
  unsigned predicates() const { return _preds.length(); }

 
  inline Symbol* getFunction(unsigned n)
  {
    ASS(n < _funs.length());
    return _funs[n];
  } 
 
  inline Symbol* getPredicate(unsigned n)
  {
    ASS(n < _preds.length());
    return _preds[n];
  } 

  Signature();
  ~Signature();

  CLASS_NAME(Signature);
  USE_ALLOCATOR(Signature);

  bool functionExists(const vstring& name,unsigned arity) const;
  bool predicateExists(const vstring& name,unsigned arity) const;

  Unit* getDistinctGroupPremise(unsigned group);
  unsigned createDistinctGroup(Unit* premise = 0);
  void addToDistinctGroup(unsigned constantSymbol, unsigned groupId);
  bool hasDistinctGroups(){ return _distinctGroupsAddedTo; }
  void noDistinctGroupsLeft(){ _distinctGroupsAddedTo=false; }
  Stack<Stack<unsigned>*> getDistinctGroupMembers(){ return _distinctGroupMembers; }

  static vstring key(const vstring& name,int arity);

 
  unsigned strings() const {return _strings;}
 
  unsigned integers() const {return _integers;}
 
  unsigned rationals() const {return _rationals;}
 
  unsigned reals() const {return _reals;}

  static const unsigned STRING_DISTINCT_GROUP;
  static const unsigned INTEGER_DISTINCT_GROUP;
  static const unsigned RATIONAL_DISTINCT_GROUP;
  static const unsigned REAL_DISTINCT_GROUP;
  static const unsigned LAST_BUILT_IN_DISTINCT_GROUP;

private:
  static bool isProtectedName(vstring name);
  static bool symbolNeedsQuoting(vstring name, bool interpreted, unsigned arity);
  static bool charNeedsQuoting(char c, bool first);
 
  Stack<VarSymbol*> _vars;
 
  Stack<Symbol*> _funs;
 
  Stack<Symbol*> _preds;
 
  SymbolMap _funNames;
 
  SymbolMap _predNames;
 
  SymbolMap _arityCheck;
 
  int _nextFreshSymbolNumber;

 
  unsigned _skolemFunctionCount;

 
  SymbolMap _varNames;
  
  
  Stack<Unit*> _distinctGroupPremises;
  
  Stack<Stack<unsigned>*> _distinctGroupMembers;
  
  bool _distinctGroupsAddedTo;

 
  DHMap<Interpretation, unsigned> _iSymbols;
 
  unsigned _strings;
 
  unsigned _integers;
 
  unsigned _rationals;
 
  unsigned _reals;
}; 

}

#endif 
