
/*
 * File SortHelper.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Clause.hpp"
#include "FormulaUnit.hpp"
#include "Signature.hpp"
#include "Sorts.hpp"
#include "SubformulaIterator.hpp"
#include "Term.hpp"
#include "TermIterators.hpp"

#include "SortHelper.hpp"

using namespace Kernel;

BaseType& SortHelper::getType(Term* t)
{
  if (t->isLiteral()) {
    return *(env.signature->getPredicate(t->functor())->predType());
  }
  return *env.signature->getFunction(t->functor())->fnType();
} 

unsigned SortHelper::getResultSort(Term* t)
{
  CALL("SortHelper::getResultSort(Term*)");
  ASS(!t->isSpecial());
  ASS(!t->isLiteral());

  Signature::Symbol* sym = env.signature->getFunction(t->functor());
  return sym->fnType()->result();
}

bool SortHelper::tryGetResultSort(Term* t, unsigned& result)
{
  CALL("tryGetResultSort(Term*,unsigned&)");
  ASS(!t->isLiteral());

  TermList masterVar;
  return getResultSortOrMasterVariable(t, result, masterVar);
}

bool SortHelper::tryGetResultSort(TermList t, unsigned& result)
{
  CALL("tryGetResultSort(TermList,unsigned&)");
  if (t.isVar()) {
    return false;
  }
  return tryGetResultSort(t.term(), result);
}

unsigned SortHelper::getResultSort(TermList t, DHMap<unsigned,unsigned>& varSorts)
{
  CALL("SortHelper::getResultSort");

  unsigned res;
  TermList masterVar;
  if (!getResultSortOrMasterVariable(t, res, masterVar)) {
    ASS(masterVar.isOrdinaryVar());
    res = varSorts.get(masterVar.var());
  }
  return res;
}

bool SortHelper::getResultSortOrMasterVariable(Term* t, unsigned& resultSort, TermList& resultVar)
{
  CALL("SortHelper::getResultSortOrMasterVariable");

  if (!t->isSpecial()) {
    resultSort = getResultSort(t);
    return true;
  }
  
  static Stack<Term*> candidates;
  candidates.reset();

  TermList masterVar;
  masterVar.makeEmpty();

  for (;;) {
    switch(t->functor()) {
    case Term::SF_TERM_ITE:
    {
      TermList arg1 = *t->nthArgument(0);
      TermList arg2 = *t->nthArgument(1);
      if (arg1.isTerm()) {
	candidates.push(arg1.term());
      }
      else {
	masterVar = arg1;
      }
      if (arg2.isTerm()) {
	candidates.push(arg2.term());
      }
      else {
	masterVar = arg2;
      }
      break;
    }
    case Term::SF_LET_TERM_IN_TERM:
    case Term::SF_LET_FORMULA_IN_TERM:
    {
      TermList arg1 = *t->nthArgument(0);
      if (arg1.isTerm()) {
	candidates.push(arg1.term());
      }
      else {
	masterVar = arg1;
      }
      break;
    }
    default:
      ASS(!t->isSpecial());
      resultSort = getResultSort(t);
      return true;
    }
    if (candidates.isEmpty()) {
      ASS(masterVar.isVar());
      resultVar = masterVar;
      return false;
    }
    t = candidates.pop();
  }
} 

bool SortHelper::getResultSortOrMasterVariable(TermList t, unsigned& resultSort, TermList& resultVar)
{
  CALL("SortHelper::getResultSortOrMasterVariable");

  if (t.isVar()) {
    resultVar = t;
    return false;
  }
  return getResultSortOrMasterVariable(t.term(), resultSort, resultVar);
}

unsigned SortHelper::getArgSort(Term* t, unsigned argIndex)
{
  CALL("SortHelper::getArgSort(Term*,unsigned)");
  ASS_L(argIndex, t->arity());

  if (t->isLiteral() && static_cast<Literal*>(t)->isEquality()) {
    return getEqualityArgumentSort(static_cast<Literal*>(t));
  }

  return getType(t).arg(argIndex);
} 

unsigned SortHelper::getEqualityArgumentSort(const Literal* lit)
{
  CALL("SortHelper::getEqualityArgumentSort");
  ASS(lit->isEquality());

  if (lit->isTwoVarEquality()) {
    return lit->twoVarEqSort();
  }

  TermList arg1 = *lit->nthArgument(0);
  unsigned srt1;
  if (tryGetResultSort(arg1, srt1)) {
    return srt1;
  }

  TermList arg2 = *lit->nthArgument(1);
  unsigned srt2;
  ALWAYS(tryGetResultSort(arg2, srt2));
  return srt2;
} 

unsigned SortHelper::getTermSort(TermList trm, Literal* lit)
{
  CALL("SortHelper::getTermSort");

  if (trm.isTerm()) {
    return getResultSort(trm.term());
  }
  ASS(trm.isVar());
  return getVariableSort(trm, lit);
}

unsigned SortHelper::getVariableSort(TermList var, Term* t)
{
  CALL("SortHelper::getVariableSort(TermList,Term*)");

  unsigned res;
  ALWAYS(tryGetVariableSort(var, t, res));
  return res;
}

bool SortHelper::tryGetVariableSort(unsigned var, Formula* f, unsigned& res)
{
  CALL("SortHelper::tryGetVariableSort(unsigned,Formula*,unsigned&)");

  TermList varTerm(var, false);

  SubformulaIterator sfit(f);
  while (sfit.hasNext()) {
    Formula* sf = sfit.next();
    if (sf->connective() != LITERAL) {
      continue;
    }
    Literal* lit = sf->literal();

    if (!lit->containsSubterm(varTerm)) {
      continue;
    }
    res = getVariableSort(varTerm, lit);
    return true;
  }
  return false;
}

void SortHelper::collectVariableSorts(Term* t0, DHMap<unsigned,unsigned>& map)
{
  CALL("SortHelper::collectVariableSorts(Term*,...)");

  NonVariableIterator sit(t0,true);
  while (sit.hasNext()) {
    int idx = 0;
    Term* t = sit.next().term();
    if (t->shared() && t->ground()) {
      sit.right();
    }
    TermList* args = t->args();
    while (!args->isEmpty()) {
      if (args->isOrdinaryVar()) {
	unsigned varNum = args->var();
	unsigned varSort = getArgSort(t, idx);
	if (!map.insert(varNum, varSort)) {
	  ASS_EQ(varSort, map.get(varNum));
	}
      }
      idx++;
      args=args->next();
    }
  }
} 

void SortHelper::collectVariableSorts(Formula* f, DHMap<unsigned,unsigned>& map)
{
  CALL("SortHelper::collectVariableSorts(Formula*,...)");

  SubformulaIterator sfit(f);
  while (sfit.hasNext()) {
    Formula* sf = sfit.next();
    if (sf->connective() != LITERAL) {
      continue;
    }
    Literal* lit = sf->literal();

    collectVariableSorts(lit, map);
  }
}

void SortHelper::collectVariableSorts(Unit* u, DHMap<unsigned,unsigned>& map)
{
  CALL("SortHelper::collectVariableSorts(Unit*,...)");

  if (!u->isClause()) {
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    collectVariableSorts(fu->formula(), map);
    return;
  }

  Clause* cl = static_cast<Clause*>(u);
  Clause::Iterator cit(*cl);
  while (cit.hasNext()) {
    Literal* l = cit.next();
    collectVariableSorts(l, map);
  }
}

bool SortHelper::tryGetVariableSort(TermList var, Term* t0, unsigned& result)
{
  CALL("SortHelper::tryGetVariableSort");

  NonVariableIterator sit(t0,true);
  while (sit.hasNext()) {
    Term* t = sit.next().term();
    if (t->ground()) {
      sit.right();
      continue;
    }
    int idx = 0;
    TermList* args = t->args();
    while (!args->isEmpty()) {
      if (*args==var) {
	result = getArgSort(t, idx);
        return true;
      }
      idx++;
      args=args->next();
    }
  }
  return false;
} 

bool SortHelper::areImmediateSortsValid(Term* t)
{
  CALL("SortHelper::areImmediateSortsValid");

  if (t->isLiteral() && static_cast<Literal*>(t)->isEquality()) {
    Literal* lit = static_cast<Literal*>(t);
    unsigned eqSrt = getEqualityArgumentSort(lit);
    for (unsigned i=0; i<2; i++) {
      TermList arg = *t->nthArgument(i);
      if (!arg.isTerm()) { continue; }
      Term* ta = arg.term();
      unsigned argSort = getResultSort(ta);
      if (eqSrt != argSort) {
        return false;
      }
    }
    return true;
  }

  BaseType& type = getType(t);
  unsigned arity = t->arity();
  for (unsigned i=0; i<arity; i++) {
    TermList arg = *t->nthArgument(i);
    if (!arg.isTerm()) { continue; }
    Term* ta = arg.term();
    unsigned argSort = getResultSort(ta);
    if (type.arg(i) != argSort) {
      return false;
    }
  }
  return true;
}

bool SortHelper::areSortsValid(Clause* cl)
{
  CALL("SortHelper::areSortsValid");

  static DHMap<unsigned,unsigned> varSorts;
  varSorts.reset();

  unsigned clen = cl->length();
  for (unsigned i=0; i<clen; i++) {
    if (!areSortsValid((*cl)[i], varSorts)) {
      return false;
    }
  }
  return true;
}

bool SortHelper::areSortsValid(Term* t0, DHMap<unsigned,unsigned>& varSorts)
{
  CALL("SortHelper::areSortsValid");

  NonVariableIterator sit(t0,true);
  while (sit.hasNext()) {
    Term* t = sit.next().term();
    int idx = 0;
    TermList* args = t->args();
    while (!args->isEmpty()) {
      unsigned argSrt = getArgSort(t,idx);
      TermList arg = *args;
      if (arg.isVar()) {
	unsigned varSrt;
	if (!varSorts.findOrInsert(arg.var(), varSrt, argSrt)) {
	  
	  if (varSrt != argSrt) {
	    return false;
	  }
	}
      }
      else {
	if (argSrt != getResultSort(arg.term())) {
	  return false;
	}
      }
      idx++;
      args=args->next();
    }
  }
  return true;
} 
