
/*
 * File Sorts.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Sorts.hpp"

using namespace Kernel;










Sorts::Sorts()
{
  CALL("Sorts::Sorts");

  unsigned aux;

  aux = addSort("$i");
  ASS_EQ(aux, SRT_DEFAULT);

  aux = addSort("$o");
  ASS_EQ(aux, SRT_BOOL);

  aux = addSort("$int");
  ASS_EQ(aux, SRT_INTEGER);

  aux = addSort("$rat");
  ASS_EQ(aux, SRT_RATIONAL);

  aux = addSort("$real");
  ASS_EQ(aux, SRT_REAL);
  
  aux = addSort("$array1");
  ASS_EQ(aux, SRT_ARRAY1);
    
  aux = addSort("$array2");
  ASS_EQ(aux, SRT_ARRAY2);
  
  aux = addSort("$fus");
  ASS_EQ(aux,FIRST_USER_SORT);
    
 _hasSort = false;
} 

Sorts::~Sorts()
{
  CALL("Sorts::~Sorts");

  while(_sorts.isNonEmpty()) {
    delete _sorts.pop();
  }
} 

unsigned Sorts::addSort(const vstring& name)
{
  CALL("Sorts::addSort/1");
  bool dummy;
  return addSort(name, dummy);
} 

unsigned Sorts::addSort(const vstring& name, bool& added)
{
  CALL("Sorts::addSort/2");

  unsigned result;
  if (_sortNames.find(name,result)) {
    added = false;
    return result;
  }
  _hasSort = true;
  result = _sorts.length();
  _sorts.push(new SortInfo(name));
  _sortNames.insert(name,result);
  added = true;
  return result;
} 

bool Sorts::haveSort(const vstring& name)
{
  CALL("Sorts::haveSort");
  return _sortNames.find(name);
} 

bool Sorts::findSort(const vstring& name, unsigned& idx)
{
  CALL("Sorts::findSort");
  return _sortNames.find(name, idx);
} 

BaseType* BaseType::makeType(unsigned arity, const unsigned* domainSorts, unsigned rangeSort)
{
  CALL("BaseType::makeType");

  if (rangeSort==Sorts::SRT_BOOL) {
    return new PredicateType(arity, domainSorts);
  }
  else {
    return new FunctionType(arity, domainSorts, rangeSort);
  }
} 

BaseType* BaseType::makeType0(unsigned rangeSort)
{
  CALL("BaseType::makeType0");
  return makeType(0, 0, rangeSort);
} 

BaseType* BaseType::makeType1(unsigned arg1Sort, unsigned rangeSort)
{
  CALL("BaseType::makeType1");

  unsigned args[] = { arg1Sort };
  return makeType(1, args, rangeSort);
} 

BaseType* BaseType::makeType2(unsigned arg1Sort, unsigned arg2Sort, unsigned rangeSort)
{
  CALL("BaseType::makeType2");
  unsigned args[] = { arg1Sort, arg2Sort };
  return makeType(2, args, rangeSort);
} 

BaseType* BaseType::makeType3(unsigned arg1Sort, unsigned arg2Sort, unsigned arg3Sort, unsigned rangeSort)
{
  CALL("BaseType::makeType3");

  unsigned args[] = { arg1Sort, arg2Sort, arg3Sort };
  return makeType(3, args, rangeSort);
} 

BaseType* BaseType::makeTypeUniformRange(unsigned arity, unsigned argsSort, unsigned rangeSort)
{
  CALL("BaseType::makeTypeUniformRange");

  static Stack<unsigned> argSorts;
  argSorts.reset();
  for (unsigned i=0; i<arity; i++) {
    argSorts.push(argsSort);
  }
  return makeType(arity, argSorts.begin(), rangeSort);
} 

BaseType::BaseType(unsigned arity, const unsigned* sorts)
{
  CALL("BaseType::BaseType");

  if (!arity) {
    _args = 0;
    return;
  }

  _args = SortVector::allocate(arity);
  if (!sorts) {
    
    for (unsigned i = 0; i < arity; i++) {
      (*_args)[i] = Sorts::SRT_DEFAULT;
    }
    return;
  }
  
  for (unsigned i = 0; i < arity; i++) {
    (*_args)[i] = sorts[i];
  }
} 
 
BaseType::~BaseType()
{
  CALL("BaseType::~BaseType");

  if (_args) {
    _args->deallocate();
  }
} 
 
bool BaseType::isSingleSortType(unsigned srt) const
{
  CALL("BaseType::isAllDefault");

  unsigned len = arity();
  for (unsigned i = 0; i <len; i++) {
    if (arg(i) != srt) {
      return false;
    }
  }
  return true;
} 

bool BaseType::operator==(const BaseType& o) const
{
  CALL("BaseType::operator==");

  if (isFunctionType() != o.isFunctionType()) {
    return false;
  }
  if (isFunctionType()) {
    if (static_cast<const FunctionType&>(*this).result() != 
	static_cast<const FunctionType&>(o).result()) {
      return false;
    }
  }
  unsigned len = arity();
  if (len != o.arity()) {
    return false;
  }
  for (unsigned i=0; i < len; i++) {
    if (arg(i) != o.arg(i)) {
      return false;
    }
  }
  return true;
} 

vstring BaseType::argsToString() const
{
  CALL("BaseType::argsToString");

  vstring res = "(";
  unsigned ar = arity();
  ASS(ar);
  for (unsigned i = 0; i < ar; i++) {
    res += env.sorts->sortName(arg(i));
    if (i != ar-1) {
      res += " * ";
    }
  }
  res += ')';
  return res;
} 

vstring PredicateType::toString() const
{
  CALL("PredicateType::toString");
  return argsToString() + " > $bool";
} 

vstring FunctionType::toString() const
{
  CALL("FunctionType::toString");
  return argsToString() + " > " + env.sorts->sortName(result());
} 

FunctionType::FunctionType(unsigned arity)
 : BaseType(arity)
{
  CALL("FunctionType::FunctionType");
  _result = Sorts::SRT_DEFAULT;
} 

bool FunctionType::isSingleSortType(unsigned srt) const
{
  CALL("FunctionType::isSingleSortType");

  if (result() != srt) {
    return false;
  }
  return BaseType::isSingleSortType(srt);
} 
