
/*
 * File Theory.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <math.h>

#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"

#include "Lib/Environment.hpp"
#include "Lib/Int.hpp"

#include "Shell/Skolem.hpp"

#include "Signature.hpp"
#include "SortHelper.hpp"
#include "Sorts.hpp"
#include "Term.hpp"

#include "Theory.hpp"

namespace Kernel
{

using namespace Lib;





IntegerConstantType::IntegerConstantType(const vstring& str)
{
  CALL("IntegerConstantType::IntegerConstantType(vstring)");

  if (!Int::stringToInt(str, _val)) {
    
    throw ArithmeticException();
  }
}

IntegerConstantType IntegerConstantType::operator+(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator+");

  InnerType res;
  if (!Int::safePlus(_val, num._val, res)) {
    throw ArithmeticException();
  }
  return IntegerConstantType(res);
}

IntegerConstantType IntegerConstantType::operator-(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator-/1");

  InnerType res;
  if (!Int::safeMinus(_val, num._val, res)) {
    throw ArithmeticException();
  }
  return IntegerConstantType(res);
}

IntegerConstantType IntegerConstantType::operator-() const
{
  CALL("IntegerConstantType::operator-/0");

  InnerType res;
  if (!Int::safeUnaryMinus(_val, res)) {
    throw ArithmeticException();
  }
  return IntegerConstantType(res);
}

IntegerConstantType IntegerConstantType::operator*(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator*");

  InnerType res;
  if (!Int::safeMultiply(_val, num._val, res)) {
    throw ArithmeticException();
  }
  return IntegerConstantType(res);
}

IntegerConstantType IntegerConstantType::operator/(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator/");

  
  if (num._val==0) {
    throw ArithmeticException();
  }
  return IntegerConstantType(_val/num._val);
}

IntegerConstantType IntegerConstantType::operator%(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator%");

  
  if (num._val==0) {
    throw ArithmeticException();
  }
  return IntegerConstantType(_val%num._val);
}

bool IntegerConstantType::operator==(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator==");

  return _val==num._val;
}

bool IntegerConstantType::operator>(const IntegerConstantType& num) const
{
  CALL("IntegerConstantType::operator>");

  return _val>num._val;
}

IntegerConstantType IntegerConstantType::floor(RationalConstantType rat)
{
  CALL("IntegerConstantType::floor");

  IntegerConstantType numer = rat.numerator();
  IntegerConstantType denom = rat.denominator();
  ASS_REP(denom>0, denom.toString());

  if (numer>0) {
    return numer/denom;
  }

  IntegerConstantType numerAbs = (numer>=0) ? numer : -numer;
  IntegerConstantType absRes = numerAbs/denom;
  if (numer%denom!=0) {
    absRes = absRes+1;
  }
  return -absRes;
}

Comparison IntegerConstantType::comparePrecedence(IntegerConstantType n1, IntegerConstantType n2)
{
  CALL("IntegerConstantType::comparePrecedence");
  try {
    if (n1 == INT_MIN) {
      if (n2 == INT_MIN) {
        return EQUAL;
      } else {
        return GREATER;
      }
    } else {
      if (n2 == INT_MIN) {
        return LESS;
      } else {
        int an1 = abs(n1.toInt());
        int an2 = abs(n2.toInt());

        ASS_GE(an1,0);
        ASS_GE(an2,0);

        return an1 < an2 ? LESS : (an1 == an2 ? 
            static_cast<Comparison>(-Int::compare(n1.toInt(), n2.toInt()))
                              : GREATER);
      }
    }
  }
  catch(ArithmeticException) {
    ASSERTION_VIOLATION;
    throw;
  }
}

vstring IntegerConstantType::toString() const
{
  CALL("IntegerConstantType::toString");

  return Int::toString(_val);
}





RationalConstantType::RationalConstantType(InnerType num, InnerType den)
{
  CALL("RationalConstantType::RationalConstantType");

  init(num, den);
}

RationalConstantType::RationalConstantType(const vstring& num, const vstring& den)
{
  CALL("RationalConstantType::RationalConstantType");

  init(InnerType(num), InnerType(den));
}

void RationalConstantType::init(InnerType num, InnerType den)
{
  CALL("RationalConstantType::init");

  _num = num;
  _den = den;
  cannonize();

  
  if(_den.toInt()==0) throw ArithmeticException();
}

RationalConstantType RationalConstantType::operator+(const RationalConstantType& o) const
{
  CALL("RationalConstantType::operator+");

  if (_den==o._den) {
    return RationalConstantType(_num + o._num, _den);
  }
  return RationalConstantType(_num*o._den + o._num*_den, _den*o._den);
}

RationalConstantType RationalConstantType::operator-(const RationalConstantType& o) const
{
  CALL("RationalConstantType::operator-/1");

  return (*this) + (-o);
}

RationalConstantType RationalConstantType::operator-() const
{
  CALL("RationalConstantType::operator-/0");

  return RationalConstantType(-_num, _den);
}

RationalConstantType RationalConstantType::operator*(const RationalConstantType& o) const
{
  CALL("RationalConstantType::operator*");

  return RationalConstantType(_num*o._num, _den*o._den);
}

RationalConstantType RationalConstantType::operator/(const RationalConstantType& o) const
{
  CALL("RationalConstantType::operator/");

  return RationalConstantType(_num*o._den, _den*o._num);
}

bool RationalConstantType::isInt() const
{
  CALL("RationalConstantType::isInt");

  return _den==1;
}

bool RationalConstantType::operator==(const RationalConstantType& o) const
{
  CALL("IntegerConstantType::operator==");

  return _num==o._num && _den==o._den;
}

bool RationalConstantType::operator>(const RationalConstantType& o) const
{
  CALL("IntegerConstantType::operator>");

  return (_num*o._den)>(o._num*_den);
}


vstring RationalConstantType::toString() const
{
  CALL("RationalConstantType::toString");

  vstring numStr = _num.toString();
  vstring denStr = _den.toString();


  return numStr+"/"+denStr;
}

void RationalConstantType::cannonize()
{
  CALL("RationalConstantType::cannonize");

  int gcd = Int::gcd(_num.toInt(), _den.toInt());
  if (gcd!=1) {
    _num = _num/InnerType(gcd);
    _den = _den/InnerType(gcd);
  }
  if (_den<0) {
    _num = -_num;
    _den = -_den;
  }
  
  
  
  if(_num==0 && _den!=0){ _den=1; }
}

Comparison RationalConstantType::comparePrecedence(RationalConstantType n1, RationalConstantType n2)
{
  CALL("RationalConstantType::comparePrecedence");
  try {

    if (n1==n2) { return EQUAL; }

    bool haveRepr1 = true;
    bool haveRepr2 = true;

    IntegerConstantType repr1, repr2;

    try {
      repr1 = n1.numerator()+n1.denominator();
    } catch(ArithmeticException) {
      haveRepr1 = false;
    }

    try {
      repr2 = n2.numerator()+n2.denominator();
    } catch(ArithmeticException) {
      haveRepr2 = false;
    }

    if (haveRepr1 && haveRepr2) {
      Comparison res = IntegerConstantType::comparePrecedence(repr1, repr2);
      if (res==EQUAL) {
	res = IntegerConstantType::comparePrecedence(n1.numerator(), n2.numerator());
      }
      ASS_NEQ(res, EQUAL);
      return res;
    }
    if (haveRepr1 && !haveRepr2) {
      return LESS;
    }
    if (!haveRepr1 && haveRepr2) {
      return GREATER;
    }

    ASS(!haveRepr1);
    ASS(!haveRepr2);

    Comparison res = IntegerConstantType::comparePrecedence(n1.denominator(), n2.denominator());
    if (res==EQUAL) {
      res = IntegerConstantType::comparePrecedence(n1.numerator(), n2.numerator());
    }
    ASS_NEQ(res, EQUAL);
    return res;
  }
  catch(ArithmeticException) {
    ASSERTION_VIOLATION;
    throw;
  }
}






Comparison RealConstantType::comparePrecedence(RealConstantType n1, RealConstantType n2)
{
  CALL("RealConstantType::comparePrecedence");

  return RationalConstantType::comparePrecedence(n1, n2);
}

bool RealConstantType::parseDouble(const vstring& num, RationalConstantType& res)
{
  CALL("RealConstantType::parseDouble");

  try {
    vstring newNum;
    IntegerConstantType denominator = 1;
    bool haveDecimal = false;
    bool neg = false;
    size_t nlen = num.size();
    for(size_t i=0; i<nlen; i++) {
      if (num[i]=='.') {
	if (haveDecimal) {
	  return false;
	}
	haveDecimal = true;
      }
      else if (i==0 && num[i]=='-') {
	neg = true;
      }
      else if (num[i]>='0' && num[i]<='9') {
	if (newNum=="0") {
	  newNum = num[i];
	}
	else {
	  newNum += num[i];
	}
	if (haveDecimal) {
	  denominator = denominator * 10;
	}
      }
      else {
	return false;
      }
    }
    if (neg) {
      newNum = '-'+newNum;
    }
    IntegerConstantType numerator(newNum);
    res = RationalConstantType(numerator, denominator);
  } catch(ArithmeticException) {
    return false;
  }
  return true;
}

RealConstantType::RealConstantType(const vstring& number)
{
  CALL("RealConstantType::RealConstantType");

  RationalConstantType value;
  if (parseDouble(number, value)) {
    init(value.numerator(), value.denominator());
    return;
  }

  double numDbl;
  if (!Int::stringToDouble(number, numDbl)) {
    
    throw ArithmeticException();
  }
  InnerType denominator = 1;
  while(floor(numDbl)!=numDbl) {
    denominator = denominator*10;
    numDbl *= 10;
  }

  InnerType::InnerType numerator = static_cast<InnerType::InnerType>(numDbl);
  if (numerator!=numDbl) {
    
    throw ArithmeticException();
  }
  init(numerator, denominator);
}

vstring RealConstantType::toNiceString() const
{
  CALL("RealConstantType::toNiceString");

  if (denominator().toInt()==1) {
    return numerator().toString()+".0";
  }
  float frep = (float) numerator().toInt() /(float) denominator().toInt();
  return Int::toString(frep);
  
}





Theory Theory::theory_obj;  

Theory* theory = &Theory::theory_obj;

Theory* Theory::instance()
{
  return theory;
}

Theory::Theory()
  : _array1SkolemFunction(0), _array2SkolemFunction(0)
{

}

unsigned Theory::getArity(Interpretation i)
{
  CALL("Signature::InterpretedSymbol::getArity");
  ASS_LE(i,MAX_INTERPRETED_ELEMENT);

  switch(i) {
  case INT_IS_INT:
  case INT_IS_RAT:
  case INT_IS_REAL:
  case RAT_IS_INT:
  case RAT_IS_RAT:
  case RAT_IS_REAL:
  case REAL_IS_INT:
  case REAL_IS_RAT:
  case REAL_IS_REAL:

  case INT_TO_INT:
  case INT_TO_RAT:
  case INT_TO_REAL:
  case RAT_TO_INT:
  case RAT_TO_RAT:
  case RAT_TO_REAL:
  case REAL_TO_INT:
  case REAL_TO_RAT:
  case REAL_TO_REAL:

  case INT_SUCCESSOR:
  case INT_UNARY_MINUS:
  case RAT_UNARY_MINUS:
  case REAL_UNARY_MINUS:

  case INT_FLOOR:
  case INT_CEILING:
  case INT_TRUNCATE:
  case INT_ROUND:

  case RAT_FLOOR:
  case RAT_CEILING:
  case RAT_TRUNCATE:
  case RAT_ROUND:

  case REAL_FLOOR:
  case REAL_CEILING:
  case REAL_TRUNCATE:
  case REAL_ROUND:

    return 1;

  case EQUAL:

  case INT_GREATER:
  case INT_GREATER_EQUAL:
  case INT_LESS:
  case INT_LESS_EQUAL:
  case INT_DIVIDES:

  case RAT_GREATER:
  case RAT_GREATER_EQUAL:
  case RAT_LESS:
  case RAT_LESS_EQUAL:
  case RAT_DIVIDES:

  case REAL_GREATER:
  case REAL_GREATER_EQUAL:
  case REAL_LESS:
  case REAL_LESS_EQUAL:
  case REAL_DIVIDES:

  case INT_PLUS:
  case INT_MINUS:
  case INT_MULTIPLY:
  case INT_DIVIDE:
  case INT_MODULO:
  case INT_QUOTIENT_E:
  case INT_QUOTIENT_T:
  case INT_QUOTIENT_F:
  case INT_REMAINDER_E:
  case INT_REMAINDER_T:
  case INT_REMAINDER_F:

  case RAT_PLUS:
  case RAT_MINUS:
  case RAT_MULTIPLY:
  case RAT_DIVIDE:
  case RAT_QUOTIENT:
  case RAT_QUOTIENT_E:
  case RAT_QUOTIENT_T:
  case RAT_QUOTIENT_F:
  case RAT_REMAINDER_E:
  case RAT_REMAINDER_T:
  case RAT_REMAINDER_F:

  case REAL_PLUS:
  case REAL_MINUS:
  case REAL_MULTIPLY:
  case REAL_DIVIDE:
  case REAL_QUOTIENT:
  case REAL_QUOTIENT_E:
  case REAL_QUOTIENT_T:
  case REAL_QUOTIENT_F:
  case REAL_REMAINDER_E:
  case REAL_REMAINDER_T:
  case REAL_REMAINDER_F:

  case SELECT1_INT:
  case SELECT2_INT:
    return 2;
          
  case STORE1_INT:
  case STORE2_INT: 
    return 3;
          
  default:
    ASSERTION_VIOLATION;
  }
}

bool Theory::isFunction(Interpretation i)
{
  CALL("Signature::InterpretedSymbol::isFunction");
  ASS_LE(i,MAX_INTERPRETED_ELEMENT);

  switch(i) {
  case INT_TO_INT:
  case INT_TO_RAT:
  case INT_TO_REAL:
  case RAT_TO_INT:
  case RAT_TO_RAT:
  case RAT_TO_REAL:
  case REAL_TO_INT:
  case REAL_TO_RAT:
  case REAL_TO_REAL:

  case INT_SUCCESSOR:
  case INT_UNARY_MINUS:
  case RAT_UNARY_MINUS:
  case REAL_UNARY_MINUS:

  case INT_PLUS:
  case INT_MINUS:
  case INT_MULTIPLY:
  case INT_DIVIDE:
  case INT_MODULO:
  case INT_QUOTIENT_E:
  case INT_QUOTIENT_T:
  case INT_QUOTIENT_F:
  case INT_REMAINDER_E:
  case INT_REMAINDER_T:
  case INT_REMAINDER_F:
  case INT_FLOOR:
  case INT_CEILING:
  case INT_TRUNCATE:
  case INT_ROUND:

  case RAT_PLUS:
  case RAT_MINUS:
  case RAT_MULTIPLY:
  case RAT_DIVIDE:
  case RAT_QUOTIENT:
  case RAT_QUOTIENT_E:
  case RAT_QUOTIENT_T:
  case RAT_QUOTIENT_F:
  case RAT_REMAINDER_E:
  case RAT_REMAINDER_T:
  case RAT_REMAINDER_F:
  case RAT_FLOOR:
  case RAT_CEILING:
  case RAT_TRUNCATE:
  case RAT_ROUND:

  case REAL_PLUS:
  case REAL_MINUS:
  case REAL_MULTIPLY:
  case REAL_DIVIDE:
  case REAL_QUOTIENT:
  case REAL_QUOTIENT_E:
  case REAL_QUOTIENT_T:
  case REAL_QUOTIENT_F:
  case REAL_REMAINDER_E:
  case REAL_REMAINDER_T:
  case REAL_REMAINDER_F:
  case REAL_FLOOR:
  case REAL_CEILING:
  case REAL_TRUNCATE:
  case REAL_ROUND:
          
  case SELECT1_INT:
  case SELECT2_INT:          
  case STORE1_INT:
  case STORE2_INT:
    return true;

  case EQUAL:

  case INT_GREATER:
  case INT_GREATER_EQUAL:
  case INT_LESS:
  case INT_LESS_EQUAL:
  case INT_DIVIDES:

  case RAT_GREATER:
  case RAT_GREATER_EQUAL:
  case RAT_LESS:
  case RAT_LESS_EQUAL:
  case RAT_DIVIDES:

  case REAL_GREATER:
  case REAL_GREATER_EQUAL:
  case REAL_LESS:
  case REAL_LESS_EQUAL:
  case REAL_DIVIDES:

  case INT_IS_INT:
  case INT_IS_RAT:
  case INT_IS_REAL:
  case RAT_IS_INT:
  case RAT_IS_RAT:
  case RAT_IS_REAL:
  case REAL_IS_INT:
  case REAL_IS_RAT:
  case REAL_IS_REAL:
    return false;

  default:
    ASSERTION_VIOLATION;
  }
}

bool Theory::isInequality(Interpretation i)
{
  CALL("Signature::InterpretedSymbol::isInequality");
  ASS_LE(i,MAX_INTERPRETED_ELEMENT);

  switch(i) {
  case INT_GREATER:
  case INT_GREATER_EQUAL:
  case INT_LESS:
  case INT_LESS_EQUAL:
  case RAT_GREATER:
  case RAT_GREATER_EQUAL:
  case RAT_LESS:
  case RAT_LESS_EQUAL:
  case REAL_GREATER:
  case REAL_GREATER_EQUAL:
  case REAL_LESS:
  case REAL_LESS_EQUAL:
    return true;
  default:
    return false;
  }
  ASSERTION_VIOLATION;
}

bool Theory::hasSingleSort(Interpretation i)
{
  CALL("Theory::hasSingleSort");

  switch(i) {
  case EQUAL:  
  case INT_TO_RAT:
  case INT_TO_REAL:
  case RAT_TO_INT:
  case RAT_TO_REAL:
  case REAL_TO_INT:
  case REAL_TO_RAT:
    return false;
  default:
    return true;
  }
}

unsigned Theory::getOperationSort(Interpretation i)
{
  CALL("Theory::getOperationSort");

  
  ASS(hasSingleSort(i));
  ASS_LE(i,MAX_INTERPRETED_ELEMENT);

  switch(i) {
  case INT_GREATER:
  case INT_GREATER_EQUAL:
  case INT_LESS:
  case INT_LESS_EQUAL:
  case INT_DIVIDES:
  case INT_SUCCESSOR:
  case INT_UNARY_MINUS:
  case INT_PLUS:
  case INT_MINUS:
  case INT_MULTIPLY:
  case INT_DIVIDE:
  case INT_MODULO:
  case INT_QUOTIENT_E:
  case INT_QUOTIENT_T:
  case INT_QUOTIENT_F:
  case INT_REMAINDER_E:
  case INT_REMAINDER_T:
  case INT_REMAINDER_F:
  case INT_FLOOR:
  case INT_CEILING:
  case INT_TRUNCATE:
  case INT_ROUND:

  case INT_TO_INT:
  case INT_IS_INT:
  case INT_IS_RAT:
  case INT_IS_REAL:
    return Sorts::SRT_INTEGER;

  case RAT_UNARY_MINUS:
  case RAT_PLUS:
  case RAT_MINUS:
  case RAT_MULTIPLY:
  case RAT_DIVIDE:
  case RAT_QUOTIENT:
  case RAT_QUOTIENT_E:
  case RAT_QUOTIENT_T:
  case RAT_QUOTIENT_F:
  case RAT_REMAINDER_E:
  case RAT_REMAINDER_T:
  case RAT_REMAINDER_F:
  case RAT_FLOOR:
  case RAT_CEILING:
  case RAT_TRUNCATE:
  case RAT_ROUND:
  case RAT_GREATER:
  case RAT_GREATER_EQUAL:
  case RAT_LESS:
  case RAT_LESS_EQUAL:
  case RAT_DIVIDES:

  case RAT_TO_RAT:
  case RAT_IS_INT:
  case RAT_IS_RAT:
  case RAT_IS_REAL:
    return Sorts::SRT_RATIONAL;

  case REAL_UNARY_MINUS:
  case REAL_PLUS:
  case REAL_MINUS:
  case REAL_MULTIPLY:
  case REAL_DIVIDE:
  case REAL_QUOTIENT:
  case REAL_QUOTIENT_E:
  case REAL_QUOTIENT_T:
  case REAL_QUOTIENT_F:
  case REAL_REMAINDER_E:
  case REAL_REMAINDER_T:
  case REAL_REMAINDER_F:
  case REAL_FLOOR:
  case REAL_CEILING:
  case REAL_TRUNCATE:
  case REAL_ROUND:
  case REAL_GREATER:
  case REAL_GREATER_EQUAL:
  case REAL_LESS:
  case REAL_LESS_EQUAL:
  case REAL_DIVIDES:

  case REAL_TO_REAL:
  case REAL_IS_INT:
  case REAL_IS_RAT:
  case REAL_IS_REAL:
    return Sorts::SRT_REAL;

  default:
    ASSERTION_VIOLATION;
  }
}
    
        

bool Theory::isConversionOperation(Interpretation i)
{
  CALL("Theory::isConversionOperation");

  
  
  switch(i) {
  case INT_TO_RAT:
  case INT_TO_REAL:
  case RAT_TO_INT:
  case RAT_TO_REAL:
  case REAL_TO_INT:
  case REAL_TO_RAT:
    return true;
  default:
    return false;
  }
}
bool Theory::isLinearOperation(Interpretation i)
{
  CALL("Theory::isComparisonOperation");

  switch(i) {
  case INT_UNARY_MINUS:
  case INT_PLUS:
  case INT_MINUS:
  case RAT_UNARY_MINUS:
  case RAT_PLUS:
  case RAT_MINUS:
  case REAL_UNARY_MINUS:
  case REAL_PLUS:
  case REAL_MINUS:
    return true;
  default:
    return false;
  }
}
bool Theory::isNonLinearOperation(Interpretation i)
{
  CALL("Theory::isComparisonOperation");

  switch(i) {
  case INT_MULTIPLY:
  case INT_DIVIDE:
  case INT_MODULO:
  case INT_QUOTIENT_E:
  case INT_QUOTIENT_T:
  case INT_QUOTIENT_F:
  case INT_REMAINDER_E:
  case INT_REMAINDER_T:
  case INT_REMAINDER_F:
  case RAT_MULTIPLY:
  case RAT_DIVIDE:
  case RAT_QUOTIENT:
  case RAT_QUOTIENT_E:
  case RAT_QUOTIENT_T:
  case RAT_QUOTIENT_F:
  case RAT_REMAINDER_E:
  case RAT_REMAINDER_T:
  case RAT_REMAINDER_F:
  case REAL_MULTIPLY:
  case REAL_DIVIDE:
  case REAL_QUOTIENT:
  case REAL_QUOTIENT_E:
  case REAL_QUOTIENT_T:
  case REAL_QUOTIENT_F:
  case REAL_REMAINDER_E:
  case REAL_REMAINDER_T:
  case REAL_REMAINDER_F:
    return true;
  default:
    return false;
  }
}

bool Theory::isArraySort(unsigned sort) {
  CALL("Theory::isArraySort");
  
  switch(sort) {
  case Sorts::SRT_ARRAY1:
  case Sorts::SRT_ARRAY2:
    return true;
  default:
    return false;
  }
}
    
bool Theory::isArrayOperation(Interpretation i)
{
  CALL("Theory::isArrayFunction");
        
  switch(i) {
    case SELECT1_INT:
    case SELECT2_INT:          
    case STORE1_INT:
    case STORE2_INT:
      return true;
    default:
      return false;
  }
}

unsigned Theory::getArraySelectFunctor(unsigned sort) {
  CALL("Theory::getArraySelectFunctor");
  
  switch(sort) {
  case Sorts::SRT_ARRAY1:
    return env.signature->getInterpretingSymbol(Theory::SELECT1_INT);
  case Sorts::SRT_ARRAY2:
    return env.signature->getInterpretingSymbol(Theory::SELECT2_INT);
  default:
    ASSERTION_VIOLATION;
  }
}

unsigned Theory::getArrayStoreFunctor(unsigned sort) {
  CALL("Theory::getArrayStoreFunctor");
  
  switch(sort) {
  case Sorts::SRT_ARRAY1:
    return env.signature->getInterpretingSymbol(Theory::STORE1_INT);
  case Sorts::SRT_ARRAY2:
    return env.signature->getInterpretingSymbol(Theory::STORE2_INT);
  default:
    ASSERTION_VIOLATION;
  }
}

unsigned Theory::getArrayOperationSort(Interpretation i)
{
    CALL("Theory::getArrayOperationSort");
    ASS(isArrayOperation(i));
    
    switch(i) {
        case SELECT1_INT:
            return Sorts::SRT_INTEGER;
        case SELECT2_INT:
            return Sorts::SRT_ARRAY1;
        case STORE1_INT:
            return Sorts::SRT_ARRAY1;
        case STORE2_INT:
            return Sorts::SRT_ARRAY2;
        default:
            ASSERTION_VIOLATION;
    }
}
    
        
    
unsigned Theory::getArrayDomainSort(Interpretation i)
{
    CALL("Theory::getArrayDomainSort");
    ASS(isArrayOperation(i));
        
    switch(i) {
        case SELECT1_INT:
        case SELECT2_INT:
        case STORE1_INT:
        case STORE2_INT:
            return Sorts::SRT_INTEGER;
        default:
            ASSERTION_VIOLATION;
        }
}

unsigned Theory::getArrayExtSkolemFunction(unsigned sort) {
  unsigned* ptr;
  Interpretation store;
  Interpretation select;
  
  switch(sort) {
  case Sorts::SRT_ARRAY1:
    ptr = &_array1SkolemFunction;
    store = Theory::STORE1_INT;
    select = Theory::SELECT1_INT;
    break;
  case Sorts::SRT_ARRAY2:
    ptr = &_array2SkolemFunction;
    store = Theory::STORE2_INT;
    select = Theory::SELECT2_INT;
    break;
  default:
    ASSERTION_VIOLATION;
  }

  if (*ptr == 0) {
    unsigned arraySort = getArrayOperationSort(store);
    unsigned indexSort = theory->getArrayDomainSort(select);
    unsigned params[] = {arraySort, arraySort};
    *ptr = Shell::Skolem::addSkolemFunction(2, params, indexSort, "arrayDiff");
  }

  return *ptr;
}

    
FunctionType* Theory::getConversionOperationType(Interpretation i)
{
  CALL("Theory::getConversionOperationType");

  unsigned from, to;
  switch(i) {
  case INT_TO_RAT:
    from = Sorts::SRT_INTEGER;
    to = Sorts::SRT_RATIONAL;
    break;
  case INT_TO_REAL:
    from = Sorts::SRT_INTEGER;
    to = Sorts::SRT_REAL;
    break;
  case RAT_TO_INT:
    from = Sorts::SRT_RATIONAL;
    to = Sorts::SRT_INTEGER;
    break;
  case RAT_TO_REAL:
    from = Sorts::SRT_RATIONAL;
    to = Sorts::SRT_REAL;
    break;
  case REAL_TO_INT:
    from = Sorts::SRT_REAL;
    to = Sorts::SRT_INTEGER;
    break;
  case REAL_TO_RAT:
    from = Sorts::SRT_REAL;
    to = Sorts::SRT_RATIONAL;
    break;
  default:
    ASSERTION_VIOLATION;
  }
  BaseType* res = BaseType::makeType(1, &from, to);
  ASS(res->isFunctionType());
  return static_cast<FunctionType*>(res);
}
    
    
FunctionType* Theory::getArrayOperationType(Interpretation i)
{
    CALL("Theory::getArrayOperationType");
    ASS(isArrayOperation(i));
    
    unsigned arrSort, indexSort, valueSort;
    BaseType* res;
    indexSort = getArrayDomainSort(i);
    switch(i) {
        case SELECT1_INT: 
            valueSort=Sorts::SRT_INTEGER;
            arrSort=Sorts::SRT_ARRAY1;
            res= BaseType::makeType2(arrSort, indexSort, valueSort);
            break;
        case SELECT2_INT:
            arrSort=Sorts::SRT_ARRAY2;
            valueSort=Sorts::SRT_ARRAY1;
            res= BaseType::makeType2(arrSort, indexSort, valueSort);
            break;
        case STORE1_INT:
            arrSort=Sorts::SRT_ARRAY1;
            valueSort=Sorts::SRT_INTEGER;
            res= BaseType::makeType3(arrSort, indexSort, valueSort, arrSort);
            break;
        case STORE2_INT:
            arrSort=Sorts::SRT_ARRAY2;
            valueSort=Sorts::SRT_ARRAY1;
            res= BaseType::makeType3(arrSort, indexSort, valueSort, arrSort);
            break;
        default:
            ASSERTION_VIOLATION;
    }
    ASS(res->isFunctionType());
    return static_cast<FunctionType*>(res);
    
}

BaseType* Theory::getOperationType(Interpretation i)
{
  CALL("Theory::getOperationType");
  ASS_NEQ(i, EQUAL);

  if (isConversionOperation(i)) {
    return getConversionOperationType(i);
  }
   
  if (isArrayOperation(i))
     { return getArrayOperationType(i);}
  
    unsigned sort;  
    ASS(hasSingleSort(i));
    sort = getOperationSort(i);

    
  unsigned arity = getArity(i);
    
  static DArray<unsigned> domainSorts;
  domainSorts.init(arity, sort);
    
  unsigned resSort = isFunction(i) ? sort : Sorts::SRT_BOOL;
    
  return BaseType::makeType(arity, domainSorts.array(), resSort);

}

bool Theory::isInterpretedConstant(unsigned func)
{
  CALL("Theory::isInterpretedConstant");

  if (func>=Term::SPECIAL_FUNCTOR_LOWER_BOUND) {
    return false;
  }

  return env.signature->getFunction(func)->interpreted() && env.signature->functionArity(func)==0;
}

bool Theory::isInterpretedConstant(Term* t)
{
  CALL("Theory::isInterpretedConstant(Term*)");

  if (t->isSpecial()) { return false; }

  return t->arity()==0 && env.signature->getFunction(t->functor())->interpreted();
}

bool Theory::isInterpretedConstant(TermList t)
{
  CALL("Theory::isInterpretedConstant(TermList)");

  return t.isTerm() && isInterpretedConstant(t.term());
}

bool Theory::isInterpretedPredicate(unsigned pred)
{
  CALL("Theory::isInterpretedPredicate(unsigned)");

  return env.signature->getPredicate(pred)->interpreted();
}

bool Theory::isInterpretedPredicate(Literal* lit)
{
  CALL("Theory::isInterpretedPredicate");

  if(lit->isEquality()){
    return SortHelper::getEqualityArgumentSort(lit)!=Sorts::SRT_DEFAULT;
  }

  return isInterpretedPredicate(lit->functor());
}

bool Theory::isInterpretedPredicate(Literal* lit, Interpretation itp)
{
  CALL("Theory::isInterpretedPredicate/2");

  return env.signature->getPredicate(lit->functor())->interpreted() &&
      interpretPredicate(lit)==itp;
}

bool Theory::isInterpretedFunction(unsigned func)
{
  CALL("Theory::isInterpretedFunction(unsigned)");

  if (func>=Term::SPECIAL_FUNCTOR_LOWER_BOUND) {
    return false;
  }

  return env.signature->getFunction(func)->interpreted() && env.signature->functionArity(func)!=0;
}


bool Theory::isInterpretedFunction(Term* t)
{
  CALL("Theory::isInterpretedFunction(Term*)");

  return isInterpretedFunction(t->functor());
}

bool Theory::isInterpretedFunction(TermList t)
{
  CALL("Theory::isInterpretedFunction(TermList)");

  return t.isTerm() && isInterpretedFunction(t.term());
}

bool Theory::isInterpretedFunction(Term* t, Interpretation itp)
{
  CALL("Theory::isInterpretedFunction(Term*,Interpretation)");

  return isInterpretedFunction(t->functor()) &&
      interpretFunction(t)==itp;
}

bool Theory::isInterpretedFunction(TermList t, Interpretation itp)
{
  CALL("Theory::isInterpretedFunction(TermList,Interpretation)");

  return t.isTerm() && isInterpretedFunction(t.term(), itp);
}


Interpretation Theory::interpretFunction(unsigned func)
{
  CALL("Theory::interpretFunction");
  ASS(isInterpretedFunction(func));

  Signature::InterpretedSymbol* sym =
      static_cast<Signature::InterpretedSymbol*>(env.signature->getFunction(func));

  return sym->getInterpretation();
}

Interpretation Theory::interpretFunction(Term* t)
{
  CALL("Theory::interpretFunction");
  ASS(isInterpretedFunction(t));

  return interpretFunction(t->functor());
}

Interpretation Theory::interpretFunction(TermList t)
{
  CALL("Theory::interpretFunction");
  ASS(t.isTerm());

  return interpretFunction(t.term());
}

Interpretation Theory::interpretPredicate(unsigned pred)
{
  CALL("Theory::interpretPredicate");
  ASS(isInterpretedPredicate(pred));

  Signature::InterpretedSymbol* sym =
      static_cast<Signature::InterpretedSymbol*>(env.signature->getPredicate(pred));

  return sym->getInterpretation();
}

Interpretation Theory::interpretPredicate(Literal* lit)
{
  CALL("Theory::interpretFunction");
  ASS(isInterpretedPredicate(lit));

  return interpretPredicate(lit->functor());
}

bool Theory::tryInterpretConstant(const Term* t, IntegerConstantType& res)
{
  CALL("Theory::tryInterpretConstant(Term*,IntegerConstantType)");

  if (t->arity() != 0 || t->isSpecial()) {
    return false;
  }
  unsigned func = t->functor();
  Signature::Symbol* sym = env.signature->getFunction(func);
  if (!sym->integerConstant()) {
    return false;
  }
  res = sym->integerValue();
  return true;
} 

bool Theory::tryInterpretConstant(const Term* t, RationalConstantType& res)
{
  CALL("Theory::tryInterpretConstant(Term*,RationalConstantType)");

  if (t->arity() != 0 || t->isSpecial()) {
    return false;
  }
  unsigned func = t->functor();
  Signature::Symbol* sym = env.signature->getFunction(func);
  if (!sym->rationalConstant()) {
    return false;
  }
  res = sym->rationalValue();
  return true;
} 

bool Theory::tryInterpretConstant(const Term* t, RealConstantType& res)
{
  CALL("Theory::tryInterpretConstant(Term*,RealConstantType)");

  if (t->arity() != 0 || t->isSpecial()) {
    return false;
  }
  unsigned func = t->functor();
  Signature::Symbol* sym = env.signature->getFunction(func);
  if (!sym->realConstant()) {
    return false;
  }
  res = sym->realValue();
  return true;
} 

Term* Theory::representConstant(const IntegerConstantType& num)
{
  CALL("Theory::representConstant(const IntegerConstantType&)");

  unsigned func = env.signature->addIntegerConstant(num);
  return Term::create(func, 0, 0);
}

Term* Theory::representConstant(const RationalConstantType& num)
{
  CALL("Theory::representConstant(const RationalConstantType&)");

  unsigned func = env.signature->addRationalConstant(num);
  return Term::create(func, 0, 0);
}

Term* Theory::representConstant(const RealConstantType& num)
{
  CALL("Theory::representConstant(const RealConstantType&)");

  unsigned func = env.signature->addRealConstant(num);
  return Term::create(func, 0, 0);
}

Term* Theory::representIntegerConstant(vstring str)
{
  CALL("Theory::representIntegerConstant");

  try {
    return Theory::instance()->representConstant(IntegerConstantType(str));
  }
  catch(ArithmeticException&) {
    NOT_IMPLEMENTED;









  }
}

Term* Theory::representRealConstant(vstring str)
{
  CALL("Theory::representRealConstant");
  try {
    return Theory::instance()->representConstant(RealConstantType(str));
  } catch(ArithmeticException&) {
    NOT_IMPLEMENTED;
  }
}

Term* Theory::fun1(Interpretation itp, TermList arg)
{
  CALL("Theory::fun1");
  ASS(isFunction(itp));
  ASS_EQ(getArity(itp), 1);

  unsigned fn=theory->getFnNum(itp);
  return Term::create(fn, 1, &arg);
}

Term* Theory::fun2(Interpretation itp, TermList arg1, TermList arg2)
{
  CALL("Theory::fun2");
  ASS(isFunction(itp));
  ASS_EQ(getArity(itp), 2);

  TermList args[]= {arg1, arg2};

  unsigned fn=theory->getFnNum(itp);
  return Term::create(fn, 2, args);
}

    
Term* Theory::fun3(Interpretation itp, TermList arg1, TermList arg2, TermList arg3)
    {
        CALL("Theory::fun3");
        ASS(isFunction(itp));
        ASS_EQ(getArity(itp), 3);
        
        TermList args[]= {arg1, arg2, arg3};
        
        unsigned fn=theory->getFnNum(itp);
        return Term::create(fn, 3, args);
    }


    
    
Literal* Theory::pred2(Interpretation itp, bool polarity, TermList arg1, TermList arg2)
{
  CALL("Theory::fun2");
  ASS(!isFunction(itp));
  ASS_EQ(getArity(itp), 2);
  ASS_NEQ(itp,EQUAL);

  TermList args[]= {arg1, arg2};

  unsigned pred=theory->getPredNum(itp);
  return Literal::create(pred, 2, polarity, false, args);
}

unsigned Theory::getFnNum(Interpretation itp)
{
  CALL("Theory::getFnNum");
  ASS(isFunction(itp));
  
  return env.signature->getInterpretingSymbol(itp);
}

unsigned Theory::getPredNum(Interpretation itp)
{
  CALL("Theory::getPredNum");
  ASS(!isFunction(itp));
  
  return env.signature->getInterpretingSymbol(itp);
}

void Theory::registerLaTeXPredName(unsigned pred, bool polarity, vstring temp)
{
  CALL("Theory::registerPredLaTeXName");
  if(polarity){
    _predLaTeXnamesPos.insert(pred,temp);
  }else{
    _predLaTeXnamesNeg.insert(pred,temp); 
  }
}
void Theory::registerLaTeXFuncName(unsigned func, vstring temp)
{
  CALL("Theory::registerFuncLaTeXName");
  _funcLaTeXnames.insert(func,temp);
}

vstring Theory::tryGetInterpretedLaTeXName(unsigned func, bool pred,bool polarity)
{
  CALL("Theory::tryGetInterpretedLaTeXName");

   

  
  Interpretation i;

  if(pred){
    if(polarity){
      if(_predLaTeXnamesPos.find(func)){ return _predLaTeXnamesPos.get(func); }
      else if(_predLaTeXnamesNeg.find(func)){ 
        
        return "\neg ("+_predLaTeXnamesNeg.get(func)+")";
      }
    }
    else{ 
      if(_predLaTeXnamesNeg.find(func)){ return _predLaTeXnamesNeg.get(func); }
      else if(_predLaTeXnamesPos.find(func)){ 
        
        return "\neg ("+_predLaTeXnamesPos.get(func)+")";
      }
    }
    
    if(!isInterpretedPredicate(func)) return "";
    i = interpretPredicate(func);
  }
  else{
    if(_funcLaTeXnames.find(func)){ return _funcLaTeXnames.get(func); }
    
    if(!isInterpretedFunction(func)) return "";
    i = interpretFunction(func);
  }

  
  
  vstring pol = polarity ? "" : " \\not ";

  

  switch(i){
  case INT_SUCCESSOR: return "a0++"; 
  case INT_UNARY_MINUS:
  case RAT_UNARY_MINUS:
  case REAL_UNARY_MINUS: return "-a0";

  case EQUAL:return "a0 "+pol+"= a1";

  case INT_GREATER: return "a0 "+pol+"> a1";
  case INT_GREATER_EQUAL: return "a0 "+pol+"\\geq a1";
  case INT_LESS: return "a0 "+pol+"< a1";
  case INT_LESS_EQUAL: return "a0 "+pol+"\\leq a1";
  case INT_DIVIDES: return "a0 "+pol+"\\| a1"; 

  case RAT_GREATER: return "a0 "+pol+"> a1";
  case RAT_GREATER_EQUAL: return "a0 "+pol+"\\geq a1";
  case RAT_LESS: return "a0 "+pol+"< a1";
  case RAT_LESS_EQUAL: return "a0 "+pol+"\\leq a1";
  case RAT_DIVIDES: return "";

  case REAL_GREATER: return "a0 "+pol+"> a1"; 
  case REAL_GREATER_EQUAL: return "a0 "+pol+"\\geq a1";
  case REAL_LESS: return "a0 "+pol+"< a1";
  case REAL_LESS_EQUAL: return "a0 "+pol+"\\leq a1";
  case REAL_DIVIDES: return "";

  case INT_PLUS: return "a0 + a1";
  case INT_MINUS: return "a0 - a1";
  case INT_MULTIPLY: return "a0 \\cdot a1";
  case INT_DIVIDE: return "a0 / a1";
  

  case RAT_PLUS: return "a0 + a1";
  case RAT_MINUS: return "a0 - a1";
  case RAT_MULTIPLY: return "a0 \\cdot a1";
  case RAT_DIVIDE: return "a0 / a1";

  case REAL_PLUS: return "a0 + a1";
  case REAL_MINUS: return "a0 - a1";
  case REAL_MULTIPLY: return "a0 \\cdot a1";
  case REAL_DIVIDE: return "a0 / a1";

  default: return "";
  } 

  return "";

}

 bool Theory::invertInterpretedFunction(Term* term, TermList* arg, TermList rep, TermList& result,Stack<Literal*>& sideConditions)
 {
   CALL("Theory::invertInterpetedFunction");

   ASS(isInterpretedFunction(term->functor()));
   Interpretation f = interpretFunction(term->functor());
   Interpretation inverted_f;





switch(f){
  
  
  
  

  case INT_PLUS: inverted_f = INT_MINUS; break; 
  case INT_MINUS: inverted_f = INT_PLUS; break;
  
  case INT_MULTIPLY: 
    
    
    if(isInterpretedConstant(rep)){
      
      IntegerConstantType a;
      if(!tryInterpretConstant(rep,a)) return false;
      IntegerConstantType b;
      if((term->nthArgument(0)==arg && tryInterpretConstant(*term->nthArgument(1),b)) ||
         (term->nthArgument(1)==arg && tryInterpretConstant(*term->nthArgument(0),b)) )
      {
        
        
        if(b.toInt()==0) return false;
        if(a.toInt() % b.toInt() == 0){
          inverted_f = INT_DIVIDE; break;
        }
      }
    }
    return false;

  
  

  case RAT_PLUS: inverted_f = RAT_MINUS; break;
  case RAT_MINUS: inverted_f = RAT_PLUS; break;
  case RAT_MULTIPLY: {
      RationalConstantType b;
      if((term->nthArgument(0)==arg && tryInterpretConstant(*term->nthArgument(1),b)) ||
         (term->nthArgument(1)==arg && tryInterpretConstant(*term->nthArgument(0),b)) )
      {
        if(b.numerator()==0) return false;
        inverted_f = RAT_DIVIDE; 
        break;
      }
      return false;
    }
  case RAT_DIVIDE: inverted_f = RAT_MULTIPLY; break;

  case REAL_PLUS: inverted_f = REAL_MINUS; break; 
  case REAL_MINUS: inverted_f = REAL_PLUS; break;
  case REAL_MULTIPLY: {
      RealConstantType b;
      if((term->nthArgument(0)==arg && tryInterpretConstant(*term->nthArgument(1),b)) ||
         (term->nthArgument(1)==arg && tryInterpretConstant(*term->nthArgument(0),b)) )
      {
        if(b.numerator()==0) return false;
        inverted_f = REAL_DIVIDE;
      }
      else{
        
        
        TermList* notZero = 0;
        if(term->nthArgument(0)==arg){ notZero=term->nthArgument(1); } 
        else if(term->nthArgument(1)==arg){ notZero=term->nthArgument(0); } 
        Term* zero =theory->representConstant(RealConstantType(RationalConstantType(0,1)));
        sideConditions.push(Literal::createEquality(true,TermList(zero),*notZero,Sorts::SRT_REAL));
        inverted_f = REAL_DIVIDE;
      }
      break;
    } 
  case REAL_DIVIDE: inverted_f = REAL_MULTIPLY; break;

  default: 
    return false;
 }

 
 
 

 
 
 

 
 ASS(term->arity()==2);
 TermList other;
 if(term->nthArgument(0)==arg) other=*term->nthArgument(1);
 else if(term->nthArgument(1)==arg) other=*term->nthArgument(0);
 else { ASSERTION_VIOLATION;} 

 TermList args[] = {rep,other};
 result = TermList(Term::create(getFnNum(inverted_f),2,args));
 return true;

 }

}
















