
/*
 * File ArrayMap.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __ArrayMap__
#define __ArrayMap__

#include "Forwards.hpp"

#include "DArray.hpp"

namespace Lib {

template<typename T>
struct ArrayMapEntry
{
  ArrayMapEntry() : _timestamp(0) {}

  T _obj;
  unsigned _timestamp;
};

template<typename T>
class ArrayMap
: DArray<ArrayMapEntry<T> >
{
  typedef ArrayMapEntry<T> Entry;
public:
  CLASS_NAME(ArrayMap<T>);
  USE_ALLOCATOR(ArrayMap<T>); 

 
  ArrayMap() : DArray<Entry>(8), _timestamp(1) {}

 
  ArrayMap(size_t size) : DArray<Entry>(size), _timestamp(1) {}

 
  void reset()
  {
    _timestamp++;
    if(_timestamp==0) {
      Entry* nptr=DArray<Entry>::_array;
      Entry* afterLast=DArray<Entry>::_array +DArray<Entry>::_capacity;
      while(nptr!=afterLast) {
        (nptr++)->_timestamp=0;
      }
      _timestamp=1;
    }
  }

 
  inline
  void ensure(size_t size)
  {
    DArray<Entry>::ensure(size);
  }

 
  inline
  void expand(size_t size)
  {
    DArray<Entry>::expand(size);
  }

 
  inline
  void insert(size_t index, T obj=T())
  {
    CALL("ArrayMap::insert");

    Entry& e=(*this)[index];
    ASS_NEQ(e._timestamp,_timestamp);
    e._obj=obj;
    e._timestamp=_timestamp;
  }

 
  inline
  void set(size_t index, T obj)
  {
    CALL("ArrayMap::set");

    Entry& e=(*this)[index];
    e._obj=obj;
    e._timestamp=_timestamp;
  }

 
  inline
  T get(size_t index)
  {
    CALL("ArrayMap::get");
    ASS_EQ((*this)[index]._timestamp,_timestamp);
    return (*this)[index]._obj;
  }

 
  inline
  T get(size_t index, T def)
  {
    CALL("ArrayMap::get");
    if((*this)[index]._timestamp!=_timestamp) {
      return def;
    }
    return (*this)[index]._obj;
  }

 
  inline
  bool find(size_t index)
  {
    CALL("ArrayMap::find");
    return (*this)[index]._timestamp==_timestamp;
  }

 
  inline
  bool find(size_t index, T& val)
  {
    CALL("ArrayMap::find");
    if((*this)[index]._timestamp==_timestamp) {
      val = (*this)[index]._obj;
      return true;
    }
    return false;
  }

 
  inline
  void remove(size_t index)
  {
    CALL("ArrayMap::remove");
    ASS((*this)[index]._timestamp==_timestamp);
    (*this)[index]._timestamp=0;
  }

 
  inline
  bool getValuePtr(size_t index, T*& pObj, T init)
  {
    CALL("ArrayMap::getValuePtr");

    Entry& e=(*this)[index];
    pObj=&e._obj;
    if(e._timestamp!=_timestamp) {
      e._timestamp=_timestamp;
      e._obj=init;
      return true;
    }
    return false;
  }

  class KeyIterator
  {
  public:
    DECL_ELEMENT_TYPE(unsigned);
    KeyIterator(const ArrayMap& parent) : _parent(parent), _idx(0) {}

    bool hasNext()
    {
      while(_idx < static_cast<const DArray<Entry>&>(_parent).size()) {
	if(_parent[_idx]._timestamp == _parent._timestamp) {
	  return true;
	}
	++_idx;
      }
      return false;
    }

    size_t next()
    {
      CALL("ArrayMap::KeyIterator::next");
      ASS_L(_idx,_parent.size());
      ASS_EQ(_parent[_idx]._timestamp,_parent._timestamp);
      return _idx++;
    }

  private:
    const ArrayMap& _parent;
    size_t _idx;
  };

private:
 
  unsigned _timestamp;
};

class ArraySet : public ArrayMap<EmptyStruct>
{
public:
  ArraySet() {}
  ArraySet(size_t size) : ArrayMap<EmptyStruct>(size) {}

  template<class It>
  void insertFromIterator(It it)
  {
    CALL("ArraySet::insertFromIterator");

    while(it.hasNext()) {
      insert(it.next());
    }
  }

  typedef KeyIterator Iterator;
};

}

#endif
