
/*
 * File TimeCounter.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"

#include "Lib/Environment.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Timer.hpp"

#include "Shell/Options.hpp"
#include "Shell/UIHelper.hpp"

#include "TimeCounter.hpp"

using namespace std;
using namespace Shell;
using namespace Lib;

bool TimeCounter::s_measuring = true;
bool TimeCounter::s_initialized = false;
int TimeCounter::s_measuredTimes[__TC_ELEMENT_COUNT];
int TimeCounter::s_measureInitTimes[__TC_ELEMENT_COUNT];
int TimeCounter::s_measuredCnt = 0;


void TimeCounter::reinitialize()
{
  CALL("TimeCounter::reinitialize");

  Stack<int> measured;
  for(int i=0; i<__TC_ELEMENT_COUNT; i++) {
    if(isBeingMeasured(static_cast<TimeCounterUnit>(i))) {
      measured.push(i);
    }
  }

  s_initialized=0;

  initialize();

  int currTime=env.timer->elapsedMilliseconds();

  while(measured.isNonEmpty()) {
    int i=measured.pop();
    s_measureInitTimes[i]=currTime;
  }
}

void TimeCounter::initialize()
{
  CALL("TimeCounter::initialize");
  ASS(!s_initialized);

  s_initialized=true;

  if(!env.options->timeStatistics()) {
    s_measuring=false;
    return;
  }

  for(int i=0; i<__TC_ELEMENT_COUNT; i++) {
    s_measuredTimes[i]=0;
    s_measureInitTimes[i]=-1;
  }

  s_measureInitTimes[TC_OTHER]=0;
}

void TimeCounter::startMeasuring(TimeCounterUnit tcu)
{
  CALL("TimeCounter::startMeasuring");
  ASS_NEQ(tcu, TC_OTHER);

  if(!s_initialized) {
    initialize();
    if(!s_measuring) {
      return;
    }
  }

  if(s_measureInitTimes[tcu]!=-1) {
    
    _tcu=__TC_NONE;
    return;
  }

  int currTime=env.timer->elapsedMilliseconds();

  _tcu=tcu;
  s_measureInitTimes[_tcu]=currTime;

  if(!s_measuredCnt) {
    ASS_NEQ(s_measureInitTimes[TC_OTHER],-1);
    s_measuredTimes[TC_OTHER]+=currTime-s_measureInitTimes[TC_OTHER];
    s_measureInitTimes[TC_OTHER]=-1;
  }

  s_measuredCnt++;
}

void TimeCounter::stopMeasuring()
{
  CALL("TimeCounter::stopMeasuring");

  if(_tcu==__TC_NONE) {
    
    return;
  }
  ASS_EQ(s_measureInitTimes[TC_OTHER],-1);
  ASS_GE(s_measureInitTimes[_tcu], 0);

  int currTime=env.timer->elapsedMilliseconds();
  s_measuredTimes[_tcu] += currTime-s_measureInitTimes[_tcu];

  s_measureInitTimes[_tcu]=-1;

  s_measuredCnt--;
  if(!s_measuredCnt) {
    s_measureInitTimes[TC_OTHER]=currTime;
  }
}

void TimeCounter::printReport(ostream& out)
{
  if (UIHelper::cascMode) {
    out << "% ";
  }
  out << "Time measurement results:" << endl;
  if (UIHelper::cascMode) {
    out << "% ";
  }
  for (int i=0; i<__TC_ELEMENT_COUNT; i++) {
    outputSingleStat(static_cast<TimeCounterUnit>(i), out);
  }
  out<<endl;
}

void TimeCounter::outputSingleStat(TimeCounterUnit tcu, ostream& out)
{
  if (s_measureInitTimes[tcu]==-1 && !s_measuredTimes[tcu]) {
    return;
  }

  if (UIHelper::cascMode) {
    out << "% ";
  }
  switch(tcu) {
  case TC_RAND_OPT:
    out << "random option generation";
    break;
  case TC_BACKWARD_DEMODULATION:
    out<<"backward demodulation";
    break;
  case TC_BACKWARD_SUBSUMPTION:
    out<<"backward subsumption";
    break;
  case TC_BACKWARD_SUBSUMPTION_RESOLUTION:
    out<<"backward subsumption resolution";
    break;
  case TC_BDD:
    out<<"BDD operations";
    break;
  case TC_BDD_CLAUSIFICATION:
    out<<"BDD clausification";
    break;
  case TC_BDD_MARKING_SUBSUMPTION:
    out<<"BDD marking subsumption";
    break;
  case TC_INTERPRETED_EVALUATION:
    out<<"interpreted evaluation";
    break;
  case TC_INTERPRETED_SIMPLIFICATION:
    out<<"interpreted simplification";
    break;
  case TC_CONDENSATION:
    out<<"condensation";
    break;
  case TC_CONSEQUENCE_FINDING:
    out<<"consequence finding";
    break;
  case TC_FORWARD_DEMODULATION:
    out<<"forward demodulation";
    break;
  case TC_FORWARD_SUBSUMPTION:
    out<<"forward subsumption";
    break;
  case TC_FORWARD_SUBSUMPTION_RESOLUTION:
    out<<"forward subsumption resolution";
    break;
  case TC_FORWARD_LITERAL_REWRITING:
    out<<"forward literal rewriting";
    break;
  case TC_GLOBAL_SUBSUMPTION:
    out<<"global subsumption";
    break;
  case TC_SIMPLIFYING_UNIT_LITERAL_INDEX_MAINTENANCE:
    out<<"unit clause index maintenance";
    break;
  case TC_NON_UNIT_LITERAL_INDEX_MAINTENANCE:
    out<<"non unit clause index maintenance";
    break;
  case TC_FORWARD_SUBSUMPTION_INDEX_MAINTENANCE:
    out<<"forward subsumption index maintenance";
    break;
  case TC_BINARY_RESOLUTION_INDEX_MAINTENANCE:
    out<<"binary resolution index maintenance";
    break;
  case TC_BACKWARD_SUBSUMPTION_INDEX_MAINTENANCE:
    out<<"backward subsumption index maintenance";
    break;
  case TC_BACKWARD_SUPERPOSITION_INDEX_MAINTENANCE:
    out<<"backward superposition index maintenance";
    break;
  case TC_FORWARD_SUPERPOSITION_INDEX_MAINTENANCE:
    out<<"forward superposition index maintenance";
    break;
  case TC_BACKWARD_DEMODULATION_INDEX_MAINTENANCE:
    out<<"backward demodulation index maintenance";
    break;
  case TC_FORWARD_DEMODULATION_INDEX_MAINTENANCE:
    out<<"forward demodulation index maintenance";
    break;
  case TC_SPLITTING_COMPONENT_INDEX_MAINTENANCE:
    out<<"splitting component index maintenance";
    break;
  case TC_SPLITTING_COMPONENT_INDEX_USAGE:
    out<<"splitting component index usage";
    break;
  case TC_SPLITTING_MODEL_UPDATE:
    out<<"splitting model update";
    break;
  case TC_CONGRUENCE_CLOSURE:
    out<<"congruence closure";
    break;
  case TC_CCMODEL:
    out<<"model from congruence closure";
    break;
  case TC_INST_GEN_SAT_SOLVING:
    out<<"inst gen SAT solving";
    break;
  case TC_INST_GEN_SIMPLIFICATIONS:
    out<<"inst gen simplifications";
    break;
  case TC_INST_GEN_VARIANT_DETECTION:
    out<<"inst gen variant detection";
    break;
  case TC_INST_GEN_GEN_INST:
    out<<"inst gen generating instances";
    break;
  case TC_LRS_LIMIT_MAINTENANCE:
    out<<"LRS limit maintenance";
    break;
  case TC_LITERAL_REWRITE_RULE_INDEX_MAINTENANCE:
    out<<"literal rewrite rule index maintenance";
    break;
  case TC_OTHER:
    out<<"other";
    break;
  case TC_PARSING:
    out<<"parsing";
    break;
  case TC_PREPROCESSING:
    out<<"preprocessing";
    break;
  case TC_PROPERTY_EVALUATION:
    out<<"property evaluation";
    break;
  case TC_SINE_SELECTION:
    out<<"sine selection";
    break;
  case TC_RESOLUTION:
    out<<"resolution";
    break;
  case TC_UR_RESOLUTION:
    out<<"unit resulting resolution";
    break;
  case TC_SAT_SOLVER:
    out<<"SAT solver time";
    break;
  case TC_TWLSOLVER_ADD:
    out<<"TWLSolver add clauses";
    break;
  case TC_LINGELING:
    out<<"Lingeling solver time";
    break;
  case TC_MINIMIZING_SOLVER:
    out << "minimizing solver time";
    break;        
  case TC_SUPERPOSITION:
    out<<"superposition";
    break;
  case TC_HYPER_SUPERPOSITION:
    out<<"hyper superposition";
    break;
  case TC_TERM_SHARING:
    out<<"term sharing";
    break;
  case TC_TRIVIAL_PREDICATE_REMOVAL:
    out<<"trivial predicate removal";
    break;
  case TC_SOLVING:
    out << "Bound propagation solving";
    break;
  case TC_BOUND_PROPAGATION:
    out << "Bound propagation";
    break;
  case TC_HANDLING_CONFLICTS:
    out << "handling conflicts";
    break;
  case TC_VARIABLE_SELECTION:
    out << "variable selection";
    break;
  case TC_DISMATCHING:
    out << "dismatching";
    break;
  case TC_FMB_DEF_INTRO:
    out << "fmb definition introduction";
    break;
  case TC_FMB_SORT_INFERENCE:
    out << "fmb sort inference"; 
    break;
  case TC_FMB_FLATTENING:
    out << "fmb flattening";
    break;
  case TC_FMB_SPLITTING:
    out << "fmb splitting";
    break;
  case TC_FMB_SAT_SOLVING:
    out << "fmb sat solving";
    break;
  case TC_FMB_CONSTRAINT_CREATION:
    out << "fmb constraint creation";
    break;
  case TC_HCVI_COMPUTE_HASH:
    out << "hvci compute hash";
    break;
  case TC_HCVI_INSERT:
    out << "hvci insert";
    break;
  case TC_HCVI_RETRIEVE:
      out << "hvci retrieve";
      break;
  default:
    ASSERTION_VIOLATION;
  }
  out<<": ";

  int time=s_measuredTimes[tcu];
  if(s_measureInitTimes[tcu]!=-1) {
    time += env.timer->elapsedMilliseconds()-s_measureInitTimes[tcu];
  }
  
  Timer::printMSString(out, time);
  out<<endl;
}

