
/*
 * File BufferedSolver.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "SAT/SATClause.hpp"

#include "BufferedSolver.hpp"

namespace SAT
{

BufferedSolver::BufferedSolver(SATSolver* inner)
 : _inner(inner), _checkedIdx(0), _lastStatus(SATISFIABLE), _varCnt(0), _varCntInnerOld(0)
{
  CALL("BufferedSolver::BufferedSolver");
}
bool BufferedSolver::checkAndRecordImplied(SATClause* cl)
{
  CALL("BufferedSolver::checkAndRecordImplied");

  static SATLiteralStack newLiterals;
  newLiterals.reset();

  for(unsigned lIndex=0;lIndex<cl->length();lIndex++)
  {
    SATLiteral l = (*cl)[lIndex];
    
    
    if(trueInAssignment(l)) return true;

    
    if(l.var() > _varCntInnerOld) newLiterals.push(l);
  }
  
  

  
  SATLiteralStack::Iterator newLitsIt(newLiterals);
  while(newLitsIt.hasNext()){
    SATLiteral l = newLitsIt.next();
    if(!_literalBuffer.find(l.var())){
      _literalBuffer.insert(l.var(),l.polarity());
      return true;
    }
  }
  
  

  return false;
}

void BufferedSolver::addClause(SATClause* cl)
{
  CALL("BufferedSolver::addClause");

  _unadded.push(cl);
}

void BufferedSolver::flushUnadded()
{
  CALL("BufferedSolver::flushUnadded");

  
  _varCntInnerOld=_varCnt;

  
  _inner->addClausesIter(pvi(SATClauseStack::BottomFirstIterator(_unadded)));

  
  _unadded.reset();
  _checkedIdx = 0;
  _literalBuffer.reset();
}

SATSolver::Status BufferedSolver::solve(unsigned conflictCountLimit)
{
  CALL("BufferedSolver::solve"); 
  
  
  
  ASS_EQ(conflictCountLimit, UINT_MAX);
  
  if (_lastStatus == UNSATISFIABLE) {
    return UNSATISFIABLE;
  }
  
  ASS_EQ(_lastStatus,SATISFIABLE);
  
  
  size_t sz = _unadded.size();
  while(_checkedIdx < sz){
    SATClause* cl = _unadded[_checkedIdx++];
    if(!checkAndRecordImplied(cl)){
      RSTAT_CTR_INC("solver_buffer_miss");
      flushUnadded();
      return (_lastStatus = _inner->solve(conflictCountLimit));
    }
  }

  RSTAT_CTR_INC("solver_buffer_hit");
  return SATSolver::SATISFIABLE;
}

SATSolver::VarAssignment BufferedSolver::getAssignment(unsigned var)
{
  CALL("BufferedSolver::getAssignment");
  ASS_G(var,0); ASS_LE(var,_varCnt);

  
  if(!_literalBuffer.isEmpty() && _literalBuffer.find(var)) {
    return _literalBuffer.get(var) ? SATSolver::TRUE : SATSolver::FALSE;
  }

  
  if (var <= _varCntInnerOld) {
    return _inner->getAssignment(var); 
  } else {
    return SATSolver::DONT_CARE; 
  }
}

}
