
/*
 * File SATInference.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __SATInference__
#define __SATInference__

#include "Forwards.hpp"
#include "Lib/List.hpp"
#include "SATClause.hpp"

namespace SAT {

using namespace Kernel;

class SATInference
{
public:
  enum InfType {
    PROP_INF,
    FO_CONVERSION,
    FO_SPLITTING,
    ASSUMPTION
  };
  virtual ~SATInference() {}
  virtual InfType getType() const = 0;
  
  template <typename Filter>
  static void collectFilteredFOPremises(SATClause* cl, Stack<UnitSpec>& acc, Filter f);
  
  static void collectFOPremises(SATClause* cl, Stack<UnitSpec>& acc);
  static UnitList* getFOPremises(SATClause* cl);
  static SATInference* copy(const SATInference* inf);
};

class PropInference : public SATInference
{
public:
  CLASS_NAME(PropInference);
  USE_ALLOCATOR(PropInference);

  PropInference(SATClauseList* premises) : _premises(premises) {}
  PropInference(SATClause* prem) : _premises(0)
  {
    SATClauseList::push(prem, _premises);
  }
  PropInference(SATClause* prem1, SATClause* prem2) : _premises(0)
  {
    SATClauseList::push(prem1, _premises);
    SATClauseList::push(prem2, _premises);
  }

  ~PropInference()
  {
    _premises->destroy();
  }

  virtual InfType getType() const { return PROP_INF; }
  SATClauseList* getPremises() const { return const_cast<SATClauseList*>(_premises); }
  void setPremises(SATClauseList* prems) { _premises = prems; }
private:
  SATClauseList* _premises;
};

class FOConversionInference : public SATInference
{
public:
  CLASS_NAME(FOConversionInference);
  USE_ALLOCATOR(FOConversionInference);

  FOConversionInference(UnitSpec origin);
  FOConversionInference(Clause* cl);
  ~FOConversionInference();

  virtual InfType getType() const { return FO_CONVERSION; }
  UnitSpec getOrigin() const { return _origin; }
private:
  UnitSpec _origin;
};

class FOSplittingInference : public SATInference
{
public:
  CLASS_NAME(FOSplittingInference);
  USE_ALLOCATOR(FOSplittingInference);

  FOSplittingInference(Clause* origin, ClauseList* names);
  ~FOSplittingInference();

  virtual InfType getType() const { return FO_SPLITTING; }
  Clause* getOrigin() const { return _origin; }
  ClauseList* getNames() const { return _names; }
private:
  Clause* _origin;
  ClauseList* _names;
};

class AssumptionInference : public SATInference
{
public:
  CLASS_NAME(AssumptionInference);
  USE_ALLOCATOR(AssumptionInference);

  virtual InfType getType() const { return ASSUMPTION; }
};

template <typename Filter>
void SATInference::collectFilteredFOPremises(SATClause* cl, Stack<UnitSpec>& acc, Filter f)
{
  CALL("SATInference::collectFilteredFOPremises");
  ASS_ALLOC_TYPE(cl, "SATClause");

  static Stack<SATClause*> toDo;
  static DHSet<SATClause*> seen;
  toDo.reset();
  seen.reset();

  toDo.push(cl);
  while (toDo.isNonEmpty()) {
    SATClause* cur = toDo.pop();
    if (!f(cur)) {
      continue;
    }    
    if (!seen.insert(cur)) {
      continue;
    }
    SATInference* sinf = cur->inference();
    ASS(sinf);
    switch(sinf->getType()) {
    case SATInference::FO_CONVERSION:
      acc.push(static_cast<FOConversionInference*>(sinf)->getOrigin());
      break;
    case SATInference::ASSUMPTION:
      break;
    case SATInference::PROP_INF:
    {
      PropInference* pinf = static_cast<PropInference*>(sinf);
      toDo.loadFromIterator(SATClauseList::Iterator(pinf->getPremises()));
      break;
    }
    case SATInference::FO_SPLITTING:
    {
      FOSplittingInference* inf = static_cast<FOSplittingInference*>(sinf);
      acc.push(UnitSpec(inf->getOrigin()));
      ClauseList::Iterator cit(inf->getNames());
      while (cit.hasNext()) {
	acc.push(UnitSpec(cit.next()));
      }
      break;
    }
    default:
      ASSERTION_VIOLATION;
    }
  }
  makeUnique(acc);
}



}

#endif 
