
/*
 * File SATSolver.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __SATSolver__
#define __SATSolver__

#include "SATLiteral.hpp"
#include "SATInference.hpp"

namespace SAT {

class SATSolver {
public:
  enum VarAssignment {
    TRUE,
    FALSE,
    DONT_CARE,  
    NOT_KNOWN
  };
  enum Status {
    SATISFIABLE,
    UNSATISFIABLE,
   
    UNKNOWN
  };

  virtual ~SATSolver() {}
  
 
  virtual void addClause(SATClause* cl) = 0;

  void addClausesIter(SATClauseIterator cit) {
    CALL("SATSolver::addClauses");
    while (cit.hasNext()) {
      addClause(cit.next());
    }
  }

   
  virtual void addClauseIgnoredInPartialModel(SATClause* cl) { addClause(cl); }
  
 
  virtual void simplify() {}

 
  virtual Status solve(unsigned conflictCountLimit) = 0;
  
  Status solve(bool onlyPropagate=false) { return solve(onlyPropagate ? 0u : UINT_MAX); }
    
 
  virtual VarAssignment getAssignment(unsigned var) = 0;

 
  virtual bool isZeroImplied(unsigned var) = 0;
 
  virtual void collectZeroImplied(SATLiteralStack& acc) = 0;
 
  virtual SATClause* getZeroImpliedCertificate(unsigned var) = 0;

 
  virtual void ensureVarCount(unsigned newVarCnt) {}
  
 
  virtual unsigned newVar() = 0;
  
  virtual void suggestPolarity(unsigned var, unsigned pol) = 0;

 
  virtual void randomizeForNextAssignment(unsigned maxVar) {
    CALL("SATSolver::randomizeForNextAssignment");
    for (unsigned var=1; var <= maxVar; var++) {
      suggestPolarity(var,Random::getBit());
    }
  }

     
  virtual SATClause* getRefutation() = 0;

 
  virtual void recordSource(unsigned satlitvar, Literal* lit) = 0;
  
 
  bool trueInAssignment(SATLiteral lit)
  {
    CALL("SATSolver::trueInAssignment");

    VarAssignment asgn = getAssignment(lit.var());
    VarAssignment desired = lit.polarity() ? TRUE : FALSE;
    return asgn==desired;
  }

 
  bool falseInAssignment(SATLiteral lit)
  {
    CALL("SATSolver::trueInAssignment");

    VarAssignment asgn = getAssignment(lit.var());
    VarAssignment desired = lit.polarity() ? FALSE: TRUE;
    return asgn==desired;
  }  
};

class SATSolverWithAssumptions: 
      public SATSolver {
public:

  

 
  virtual void addAssumption(SATLiteral lit) = 0;

 
  virtual void retractAllAssumptions() = 0;

 
  virtual bool hasAssumptions() const = 0;

  

  
  
  
  
  

  
  
  

  virtual Status solveUnderAssumptions(const SATLiteralStack& assumps, unsigned conflictCountLimit, bool) {
    CALL("SATSolver::solveUnderAssumptions");

    ASS(!hasAssumptions());
    _failedAssumptionBuffer.reset();

    Status res = solve(conflictCountLimit);
    if (res == UNSATISFIABLE) {
      return res;
    }

    SATLiteralStack::ConstIterator it(assumps);
    while (it.hasNext()) {
      SATLiteral lit = it.next();
      addAssumption(lit);
      _failedAssumptionBuffer.push(lit);

      res = solve(conflictCountLimit);
      if (res == UNSATISFIABLE) {
        break;
      }
    }

    retractAllAssumptions();
    return res;
  }

 
  Status solveUnderAssumptions(const SATLiteralStack& assumps, bool onlyPropagate=false, bool onlyProperSubusets=false) {
    return solveUnderAssumptions(assumps,onlyPropagate ? 0u : UINT_MAX,onlyProperSubusets);
  }

 
  virtual const SATLiteralStack& failedAssumptions() {
    return _failedAssumptionBuffer;
  }

 
  const SATLiteralStack& explicitlyMinimizedFailedAssumptions(bool onlyPropagate=false, bool randomize = false) {
    return explicitlyMinimizedFailedAssumptions(onlyPropagate ? 0u : UINT_MAX, randomize);
  }

  virtual const SATLiteralStack& explicitlyMinimizedFailedAssumptions(unsigned conflictCountLimit, bool randomize) {
    CALL("SATSolver::explicitlyMinimizeFailedAssumptions");

    

    ASS(!hasAssumptions());

    unsigned sz = _failedAssumptionBuffer.size();

    if (!sz) { 
      return _failedAssumptionBuffer;
    }

    if (randomize) {
      
      
      for(unsigned i=sz-1; i>0; i--) {
        unsigned tgtPos=Random::getInteger(i+1);
        std::swap(_failedAssumptionBuffer[i], _failedAssumptionBuffer[tgtPos]);
      }
    }

    unsigned i = 0;
    while (i < sz) {
      
      for (unsigned j = 0; j < sz; j++) {
        if (j != i) {
          addAssumption(_failedAssumptionBuffer[j]);
        }
      }

      if (solve(conflictCountLimit) == UNSATISFIABLE) {
        
        _failedAssumptionBuffer[i] = _failedAssumptionBuffer[--sz];
      } else {
        
        i++;
      }

      retractAllAssumptions();
    }

    _failedAssumptionBuffer.truncate(sz);
    return _failedAssumptionBuffer;
  }

protected:
  SATLiteralStack _failedAssumptionBuffer;
};

class PrimitiveProofRecordingSATSolver : public SATSolverWithAssumptions {
public:
  PrimitiveProofRecordingSATSolver() :  
    _addedClauses(0), _refutation(new(0) SATClause(0)), _refutationInference(new PropInference(SATClauseList::empty()))
    {
      CALL("PrimitiveProofRecordingSATSolver::PrimitiveProofRecordingSATSolver");
      
      _refutation->setInference(_refutationInference);    
    }
  
  virtual ~PrimitiveProofRecordingSATSolver() {
    
    _addedClauses->destroy();
  }

  virtual void addClause(SATClause* cl) override 
  {
    CALL("PrimitiveProofRecordingSATSolver::addClause");
    
    SATClauseList::push(cl,_addedClauses);
  }
  
  virtual SATClause* getRefutation() override
  {
    CALL("PrimitiveProofRecordingSATSolver::getRefutation");

    
    SATClauseList* prems = _addedClauses;  

    

    
    
    

    

   

    _refutationInference->setPremises(prems);

    return _refutation; 
  }
  
private:
  
  SATClauseList* _addedClauses;
  
 
  SATClause* _refutation;
 
  PropInference* _refutationInference;  
};


}

#endif 
