
/*
 * File VariableSelector.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "SATClause.hpp"
#include "TWLSolver.hpp"

#include "VariableSelector.hpp"

using namespace SAT;

bool VariableSelector::isUndefined(unsigned var)
{
  CALL("VariableSelector::isUndefined");

  ASS_G(var,0); ASS_LE(var,_varCnt);

  return _solver.isUndefined(var);
}

bool ActiveVariableSelector::selectVariable(unsigned& var)
{
  CALL("ActiveVariableSelector::selectVariable");

  do {
    if(_activityHeap.isEmpty()) {
      return false;
    }
    var = _activityHeap.popMostActive();
  } while(!isUndefined(var));
  return true;
}

void ActiveVariableSelector::ensureVarCount(unsigned varCnt)
{
  CALL("ActiveVariableSelector::ensureVarCount");

  VariableSelector::ensureVarCount(varCnt);
  _activityHeap.ensureVarCount(varCnt);
  
  _niceness.expand(varCnt+1, 0);
}

void ActiveVariableSelector::onInputClauseAdded(SATClause* cl)
{
  CALL("ActiveVariableSelector::onInputClauseAdded");

  unsigned clen = cl->length();

  for(unsigned i=0;i<clen;i++) {
    SATLiteral lit = (*cl)[i];
    unsigned var = lit.var();
    
    if(_niceness[var]==0){ 
      unsigned niceness = getNiceness(var);
      _niceness[var] = niceness;
    }
    _activityHeap.markActivity(var);
  }
}

unsigned ActiveVariableSelector::getNiceness(unsigned var)
{
  CALL("ActiveVariableSelector::getNiceness");

  ASS_G(var,0); ASS_LE(var,_varCnt);

  
  if(_niceness_option == Options::Niceness::NONE){
    return 1;
  }

  
  Literal* _source = _sourceMap.get(var); 

  
  
  
  if(!_source){
    return 1;
  }
  

  static Stack<const TermList*> _stack;
  _stack.reset();

  _stack.push(_source->args());

  unsigned level = 0;
  unsigned level_multiplier = 1;
  bool found = false;
  unsigned level_sum = 0;
  unsigned count = 0;
  bool divide = false;
  switch(_niceness_option){
    case Options::Niceness::NONE:
      return 1;
    case Options::Niceness::TOP:
      
      while(!_stack.isEmpty() && !found){
        level++;
for(const TermList* t = _stack.pop(); !t->isEmpty(); t = t->next()){
          if(t->isVar()){
            found=true;
            break;
          }
          else{
            ASS(t->isTerm());
            const Term* trm = t->term();
            ASS(trm->shared());
            if(trm->ground()){
              found=true;
              break;
            }
            _stack.push(trm->args()); 
          }
        }
      } 
      level = level*level_multiplier;
      RSTAT_MCTR_INC("nscore",level);
      return level;
    case Options::Niceness::AVERAGE:
      divide=true;
    case Options::Niceness::SUM:
      
      while(!_stack.isEmpty()){
        level++;
        for(const TermList* t = _stack.pop(); !t->isEmpty(); t = t->next()){
          if(t->isVar()){
            level_sum += (level*level_multiplier);
          }
          else{
            ASS(t->isTerm());
            const Term* trm = t->term();
            ASS(trm->shared());
            if(trm->ground()){
              level_sum += (level*level_multiplier);
              count += 1;
            }
            _stack.push(trm->args()); 
         }
        }
      }
      RSTAT_MCTR_INC("nscore",level_sum);
      
      if(divide){ level_sum = (level_sum/count);}
      return level_sum;
    default: ASSERTION_VIOLATION_REP2("Invalid niceness option: ",static_cast<int>(_niceness_option));
  }

  return 1;
}





void ArrayActiveVariableSelector::onRestart()
{
  CALL("ArrayActiveVariableSelector::onRestart");

  for(unsigned i=1; i<=_varCnt; i++) {
    _activities[i]/=2;
  }
}

bool ArrayActiveVariableSelector::selectVariable(unsigned& var)
{
  CALL("ArrayActiveVariableSelector::selectVariable");

  unsigned bestWCnt;
  int bestWCntI=-1;

  for(unsigned i=1;i<=_varCnt;i++) {
    if(!isUndefined(i)) {
      continue;
    }
    unsigned wcnt=_activities[i];
    if(bestWCntI==-1 || wcnt>bestWCnt) {
      bestWCnt=wcnt;
      bestWCntI=i;
    }
  }
  if(bestWCntI!=-1) {
    var=bestWCntI;
    return true;
  }
  return false;
}

void ArrayActiveVariableSelector::onInputClauseAdded(SATClause* cl)
{
  CALL("ArrayActiveVariableSelector::onInputClauseAdded");

  unsigned clen = cl->length();
  for(unsigned i=0;i<clen;i++) {
    unsigned var = (*cl)[i].var();
    _activities[var]++;
  }
}




bool RLCVariableSelector::selectVariable(unsigned& var)
{
  CALL("RLCVariableSelector::selectVariable");

  SATClauseStack::TopFirstIterator lcit(_solver._learntClauses);
  while(lcit.hasNext()) {
    SATClause* cl = lcit.next();
    if(_solver.isTrue(cl)) {
      continue;
    }
    ASS(!_solver.isFalse(cl));

    unsigned bestVal;
    int bestVar = -1;
    unsigned clen = cl->length();
    for(unsigned i=0;i<clen;i++) {
      unsigned var = (*cl)[i].var();
      if(!isUndefined(var)) {
        continue;
      }
      unsigned curVal=_activities[var];
      if(bestVar==-1 || curVal>bestVal) {
	bestVal=curVal;
        bestVar=var;
      }
    }
    ASS_NEQ(bestVar,-1);
    var=bestVar;
    return true;
  }

  return ArrayActiveVariableSelector::selectVariable(var);
}

