
/*
 * File Z3Interfacing.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#if VZ3

#include "Forwards.hpp"

#include "SATSolver.hpp"
#include "SATLiteral.hpp"
#include "SATClause.hpp"
#include "SATInference.hpp"

#include "Lib/Environment.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Sorts.hpp"

#include "Z3Interfacing.hpp"

namespace SAT
{

using namespace Shell;  
using namespace Lib;  


  
Z3Interfacing::Z3Interfacing(const Shell::Options& opts,SAT2FO& s2f, bool generateProofs):
  _varCnt(0), sat2fo(s2f),_status(SATISFIABLE), _solver(_context), _model(_solver.get_first_model()), _showZ3(opts.showZ3())
{
  CALL("Z3Interfacing::Z3Interfacing");
  

  
  
  
  
}
  
void Z3Interfacing::addClauses(SATClauseIterator cit) 
{
  CALL("Z3Interfacing::addClauses");
  
  while(cit.hasNext()){
    addClause(cit.next());
  }

}

void Z3Interfacing::addClause(SATClause* cl)
{
  CALL("Z3Interfacing::addClause");
  BYPASSING_ALLOCATOR;
  ASS(cl);

  
  PrimitiveProofRecordingSATSolver::addClause(cl);

  z3::expr z3clause = _context.bool_val(false);

  unsigned clen=cl->length();
  for(unsigned i=0;i<clen;i++){
    SATLiteral l = (*cl)[i];
    z3::expr e = getRepresentation(l);
    z3clause = z3clause || e;
  }
  
  if(_showZ3){
    env.beginOutput();
    env.out() << "[Z3] add: " << z3clause << std::endl;
    env.endOutput();
  }
  _solver.add(z3clause);

}

SATSolver::Status Z3Interfacing::solve(unsigned conflictCountLimit)
{
  CALL("Z3Interfacing::addClause");
  BYPASSING_ALLOCATOR;

  z3::check_result result = _solver.check();

  

  switch(result){
    case z3::check_result::unsat:
      _status = UNSATISFIABLE; 
      break;
    case z3::check_result::sat:
      _status = SATISFIABLE;
      _model = _solver.get_model();
      
      
      
      
      
      break;
    case z3::check_result::unknown:
      _status = UNKNOWN;
      break;
#if VDEBUG
    default: ASSERTION_VIOLATION;
#endif
  }
  return _status;
}

SATSolver::VarAssignment Z3Interfacing::getAssignment(unsigned var) 
{
  CALL("Z3Interfacing::getAssignment");
  BYPASSING_ALLOCATOR;

  ASS_EQ(_status,SATISFIABLE);

  z3::expr rep = getRepresentation(SATLiteral(var,1));
  z3::expr assignment = _model.eval(rep);
  


  if(Z3_get_bool_value(_context,assignment)==Z3_L_TRUE){
  
    return TRUE;
  }
  if(Z3_get_bool_value(_context,assignment)==Z3_L_FALSE){
  
    return FALSE;
  }

  
  return DONT_CARE;
}

void Z3Interfacing::randomizeAssignment() 
{
  CALL("Z3Interfacing::randomizeAssignment");
} 

bool Z3Interfacing::isZeroImplied(unsigned var)
{
  CALL("Z3Interfacing::isZeroImplied");
  
  
  return false; 
}

void Z3Interfacing::collectZeroImplied(SATLiteralStack& acc)
{
  CALL("Z3Interfacing::collectZeroImplied");
  NOT_IMPLEMENTED;
}

SATClause* Z3Interfacing::getZeroImpliedCertificate(unsigned)
{
  CALL("Z3Interfacing::getZeroImpliedCertificate");
  NOT_IMPLEMENTED;
  
  return 0;
}


z3::sort Z3Interfacing::getz3sort(unsigned s)
{
  CALL("Z3Interfacing::getz3sort");
  BYPASSING_ALLOCATOR;
  
  if(s==Sorts::SRT_BOOL) return _context.bool_sort(); 
  if(s==Sorts::SRT_INTEGER) return _context.int_sort();
  if(s==Sorts::SRT_REAL) return _context.real_sort(); 
  if(s==Sorts::SRT_RATIONAL) return _context.real_sort(); 

  

  
  return _context.uninterpreted_sort(Lib::Int::toString(s).c_str());
}

z3::expr Z3Interfacing::getz3expr(Term* trm,bool isLit)
{
  CALL("Z3Interfacing::getz3expr");
  BYPASSING_ALLOCATOR;
  ASS(trm);
  ASS(trm->ground());

  

    Signature::Symbol* symb; 
    unsigned range_sort;
    BaseType* type;
    bool is_equality = false;
    if(isLit){
      symb = env.signature->getPredicate(trm->functor());
      PredicateType* ptype = symb->predType();
      type = ptype;
      range_sort = Sorts::SRT_BOOL;
      
      if(trm->functor()==0){
         is_equality=true;
         ASS(trm->arity()==2);
      }
    }else{
      symb = env.signature->getFunction(trm->functor());
      FunctionType* ftype = symb->fnType(); 
      type = ftype;
      range_sort = ftype->result();
    }

    
    if(trm->arity()==0){
      if(symb->integerConstant()){
        IntegerConstantType value = symb->integerValue();
        return _context.int_val(value.toInt());
      }
      if(symb->realConstant()){
        RealConstantType value = symb->realValue();
        return _context.real_val(value.numerator().toInt(),value.denominator().toInt());
      }
      if(symb->rationalConstant()){
        RationalConstantType value = symb->rationalValue();
        return _context.real_val(value.numerator().toInt(),value.denominator().toInt());
      }

      
      return _context.constant(symb->name().c_str(),getz3sort(range_sort));
    }
    ASS(trm->arity()>0);

    
    
    
    z3::expr_vector args = z3::expr_vector(_context);
    for(unsigned i=0;i<trm->arity();i++){
      TermList* arg = trm->nthArgument(i);
      ASS(!arg->isVar());
      args.push_back(getz3expr(arg->term(),false));
    }

    
    z3::expr ret = z3::expr(_context);

   
    if(is_equality){
      ret = args[0] == args[1]; 
      args.pop_back();args.pop_back();
      return ret;
    }

    
    
    
    if(symb->interpreted()){
      Interpretation interp = static_cast<Signature::InterpretedSymbol*>(symb)->getInterpretation();
      bool skip=false; 
      
      
      bool unary=false;

      switch(interp){
        

        case Theory::INT_UNARY_MINUS:
        case Theory::RAT_UNARY_MINUS:
        case Theory::REAL_UNARY_MINUS:
          ret = -args[0];
          unary=true;
          break;

        case Theory::INT_PLUS:
        case Theory::RAT_PLUS:
        case Theory::REAL_PLUS:
          ret = args[0] + args[1];
          break;

        case Theory::INT_MINUS:
        case Theory::RAT_MINUS:
        case Theory::REAL_MINUS:
          ret = args[0] - args[1];
          break;

        case Theory::INT_MULTIPLY:
        case Theory::RAT_MULTIPLY:
        case Theory::REAL_MULTIPLY:
          ret = args[0] * args[1];
          break;

        
        
        
        
        
        

        
        case Theory::RAT_QUOTIENT:
        case Theory::REAL_QUOTIENT:
        case Theory::INT_QUOTIENT_E: 
        case Theory::RAT_QUOTIENT_E: 
        case Theory::REAL_QUOTIENT_E: 
          ret= args[0] / args[1];
          break;

        case Theory::RAT_TO_INT:
        case Theory::REAL_TO_INT:
        case Theory::INT_FLOOR:
        case Theory::RAT_FLOOR:
        case Theory::REAL_FLOOR:
          ret = to_real(to_int(args[0])); 
          unary=true;
          break;

        case Theory::INT_TO_REAL:
        case Theory::RAT_TO_REAL:
        case Theory::INT_TO_RAT: 
          ret = to_real(args[0]);
          unary=true;
          break;

        case Theory::INT_CEILING:
        case Theory::RAT_CEILING:
        case Theory::REAL_CEILING:
          ret = ceiling(args[0]);
          unary=true;
          break;

        case Theory::INT_TRUNCATE:
        case Theory::RAT_TRUNCATE:
        case Theory::REAL_TRUNCATE:
          ret = truncate(args[0]); 
          unary=true;
          break;

        case Theory::INT_ROUND:
        case Theory::RAT_ROUND:
        case Theory::REAL_ROUND:
          {
            z3::expr t = args[0];
            z3::expr i = to_int(t);
            z3::expr i2 = i + _context.real_val(1,2);
            ret = ite(t > i2, i+1, ite(t==i2, ite(is_even(i),i,i+1),i));
            unary=true;
            break;
          }

         case Theory::INT_QUOTIENT_T:
         case Theory::RAT_QUOTIENT_T:
         case Theory::REAL_QUOTIENT_T:
           ret = truncate(args[0] / args[1]);
           break;

         case Theory::INT_QUOTIENT_F:
         case Theory::RAT_QUOTIENT_F:
         case Theory::REAL_QUOTIENT_F:
           ret = to_real(to_int(args[0] / args[1]));
           break;

         

         case Theory::INT_REMAINDER_E:
         case Theory::RAT_REMAINDER_E:
         case Theory::REAL_REMAINDER_E:
           ret = z3::expr(_context, Z3_mk_mod(_context, args[0], args[1]));
           break;

       
       

       case Theory::INT_IS_INT:
       case Theory::RAT_IS_INT:
       case Theory::REAL_IS_INT:
         ret = z3::expr(_context,Z3_mk_is_int(_context,args[0]));
         unary=true;
         break;

       case Theory::INT_LESS:
       case Theory::RAT_LESS:
       case Theory::REAL_LESS:
          ret = args[0] < args[1];
          break;

       case Theory::INT_GREATER:
       case Theory::RAT_GREATER:
       case Theory::REAL_GREATER:
          ret= args[0] > args[1];
          break;
          
       case Theory::INT_LESS_EQUAL:
       case Theory::RAT_LESS_EQUAL:
       case Theory::REAL_LESS_EQUAL:
          ret= args[0] <= args[1];
          break;

       case Theory::INT_GREATER_EQUAL:
       case Theory::RAT_GREATER_EQUAL:
       case Theory::REAL_GREATER_EQUAL:
          ret= args[0] >= args[1];
          break;

        default: 
          skip=true;
          break;
      }

      if(!skip){
        args.pop_back();
        if(!unary){args.pop_back();}
        return ret;
      } 

    }
    
    z3::sort_vector domain_sorts = z3::sort_vector(_context);
    for(unsigned i=0;i<type->arity();i++){
      domain_sorts.push_back(getz3sort(type->arg(i)));
    }
    z3::symbol name = _context.str_symbol(symb->name().c_str());
    z3::func_decl f = _context.function(name,domain_sorts,getz3sort(range_sort));

    
    z3::expr e = f(args); 
    
    return e;
}

z3::expr Z3Interfacing::getRepresentation(SATLiteral slit)
{
  CALL("Z3Interfacing::getRepresentation");
  BYPASSING_ALLOCATOR;

  
  Literal* lit = sat2fo.toFO(slit);
  if(lit && lit->ground()){
    
    
    try{
      z3::expr e = getz3expr(lit,true);
      
      if(slit.isNegative()) return !e;
      return e;
    }catch(z3::exception& exception){
     cout << "Z3 exception:\n" << exception.msg() << endl;
     ASSERTION_VIOLATION_REP("Failed to create Z3 rep for " + lit->toString());
    }
  }
  
  vstring name = "v"+Lib::Int::toString(slit.var());
  z3::expr e = _context.bool_const(name.c_str());
  if(slit.isNegative()) return !e;
  else return e;
}

} 

#endif
