
/*
 * File ConsequenceFinder.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/SharedSet.hpp"
#include "Lib/SkipList.hpp"
#include "Lib/TimeCounter.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Signature.hpp"

#include "Shell/Options.hpp"

#include "ConsequenceFinder.hpp"
#include "SaturationAlgorithm.hpp"

namespace Saturation
{

using namespace Lib;
using namespace Kernel;

void ConsequenceFinder::init(SaturationAlgorithm* sa)
{
  CALL("ConsequenceFinder::init");

  _sa=sa;

  ClauseContainer* cc=_sa->getSimplifyingClauseContainer();
  _sdInsertion = cc->addedEvent.subscribe(this,&ConsequenceFinder::onClauseInserted);
  _sdRemoval = cc->removedEvent.subscribe(this,&ConsequenceFinder::onClauseRemoved);
}

ConsequenceFinder::~ConsequenceFinder()
{
  CALL("ConsequenceFinder::~ConsequenceFinder");

  _sdInsertion->unsubscribe();
  _sdRemoval->unsubscribe();
}


void ConsequenceFinder::onNewPropositionalClause(Clause* cl)
{
  CALL("ConsequenceFinder::onNewPropositionalClause");

  TimeCounter tc(TC_CONSEQUENCE_FINDING);

  
  Clause* dlrCl=_dlr.simplify(cl);

  bool dlrSimplified=dlrCl!=cl;
  if(dlrSimplified) {
    
    
    dlrCl->destroyIfUnnecessary();
    return;
  }


  if(!cl->noSplits() || !_td.simplify(cl)) {
    return;
  }
  Literal* pos=0;
  bool horn=true;
  Clause::Iterator it(*cl);
  while(it.hasNext()) {
    Literal* l=it.next();
    if(!env.signature->getPredicate(l->functor())->cfName()) {
      return;
    }
    if(l->isPositive()) {
      if(pos) {
        horn=false;
      }
      else {
        pos=l;
      }
    }
  }

  env.beginOutput();
  env.out() << "Pure cf clause: " << cl->toNiceString() <<endl;
  env.endOutput();

  if(!horn || !pos) {
    return;
  }

  unsigned red=pos->functor(); 
  if(_redundant[red]) {
    return;
  }
  _redundant[red]=true;

  
  
  _redundantsToHandle.push(red);

  env.beginOutput();
  env.out() << "Consequence found: " << env.signature->predicateName(red) << endl;
  env.endOutput();
}

void ConsequenceFinder::onAllProcessed()
{
  CALL("ConsequenceFinder::onAllProcessed");

  TimeCounter tc(TC_CONSEQUENCE_FINDING);

  while(_redundantsToHandle.isNonEmpty()) {
    unsigned red=_redundantsToHandle.pop();

    ClauseSL* rlist=_index[red];
    if(rlist) {
      _index[red]=0;
      while(rlist->isNonEmpty()) {
        Clause* rcl=rlist->pop();
        
        
        (void)(rcl->store()!=Clause::UNPROCESSED && rcl->store()!=Clause::NONE);
        if(rcl->store()) {
          
          
          continue;
        }
        _sa->removeActiveOrPassiveClause(rcl);
      }
      delete rlist;
    }
  }
}

bool ConsequenceFinder::isRedundant(Clause* cl)
{
  CALL("ConsequenceFinder::isRedundant");

  Clause::Iterator it(*cl);
  while(it.hasNext()) {
    unsigned fn=it.next()->functor();
    if(!env.signature->getPredicate(fn)->cfName()) {
      continue;
    }
    if(_redundant[fn]) {
      return true;
    }
  }
  return false;
}


void ConsequenceFinder::onClauseInserted(Clause* cl)
{
  CALL("ConsequenceFinder::onClauseInserted");

  TimeCounter tc(TC_CONSEQUENCE_FINDING);

  bool red=false;
  Clause::Iterator it(*cl);
  while(it.hasNext()) {
    unsigned fn=it.next()->functor();
    if(!env.signature->getPredicate(fn)->cfName()) {
      continue;
    }
    if(_redundant[fn]) {
      red=true;
    }
    else {
      indexClause(fn, cl, true);
    }
  }
  if(red) {
    

    
    
    
    
  }

}

void ConsequenceFinder::onClauseRemoved(Clause* cl)
{
  CALL("ConsequenceFinder::onClauseRemoved");

  TimeCounter tc(TC_CONSEQUENCE_FINDING);

  Clause::Iterator it(*cl);
  while(it.hasNext()) {
    unsigned fn=it.next()->functor();
    if(!env.signature->getPredicate(fn)->cfName()) {
      continue;
    }
    if(!_redundant[fn]) {
      indexClause(fn, cl, false);
    }
  }
}

void ConsequenceFinder::indexClause(unsigned indexNum, Clause* cl, bool add)
{
  CALL("ConsequenceFinder::indexClause");

  if(add) {
    if(!_index[indexNum]) {
      _index[indexNum]=new ClauseSL();
    }
    _index[indexNum]->insert(cl);
  }
  else {
    ASS(_index[indexNum]);
    _index[indexNum]->remove(cl);
  }
}


}
