
/*
 * File ExtensionalityClauseContainer.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */
#include "Kernel/Clause.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/Term.hpp"

#include "Shell/Statistics.hpp"
#include "Shell/TheoryFinder.hpp"

#include "Saturation/ExtensionalityClauseContainer.hpp"

namespace Saturation
{

using namespace Shell;

Literal* ExtensionalityClauseContainer::addIfExtensionality(Clause* c) {
  CALL("ExtensionalityClauseContainer::addIfExtensionality");
  
  
  if (c->isExtensionality()) {
    return getSingleVarEq(c);
  }

  
  
  unsigned clen = c->length();
  if (clen < 2 || (_maxLen > 1 && clen > _maxLen))
    return 0;

  Literal* varEq = 0;
  unsigned sort;

  if (_onlyKnown) {
    
    
    if(!TheoryFinder::matchKnownExtensionality(c))
      return 0;

    
    varEq = getSingleVarEq(c);
    sort = varEq->twoVarEqSort();
  } else {
    
    
    
    
    static DArray<bool> negEqSorts(_sortCnt);
    negEqSorts.init(_sortCnt, false);
  
    for (Clause::Iterator ci(*c); ci.hasNext(); ) {
      Literal* l = ci.next();

      if (l->isTwoVarEquality() && l->isPositive()) {
        if (varEq != 0)
          return 0;

        sort = l->twoVarEqSort();
        if (negEqSorts[sort])
          return 0;

        varEq = l;
      } else if (l->isEquality()) {
        if (!_allowPosEq && l->isPositive())
          return 0;
      
        unsigned negEqSort = SortHelper::getEqualityArgumentSort(l);
        if (varEq == 0)
          negEqSorts[negEqSort] = true;
        else if (sort == negEqSort)
          return 0;
      }
    }
  }

  if (varEq != 0) {
    
    c->setExtensionality(true);
    add(ExtensionalityClause(c, varEq, sort));
    _size++;
    env.statistics->extensionalityClauses++;
    return varEq;
  }

  return 0;
}

Literal* ExtensionalityClauseContainer::getSingleVarEq(Clause* c) {
  CALL("ExtensionalityClauseContainer::getSingleVarEq");
  
  for (unsigned i = 0; i < c->length(); ++i) {
    Literal* varEq = (*c)[i];
    if (varEq->isTwoVarEquality() && varEq->isPositive()) {
      return varEq;
      break;
    }
  }
  ASSERTION_VIOLATION;
}

void ExtensionalityClauseContainer::add(ExtensionalityClause c) {
  CALL("ExtensionalityClauseContainer::add");
  
  ExtensionalityClauseList::push(c, _clausesBySort[c.sort]);
}

struct ExtensionalityClauseContainer::ActiveFilterFn
{
  ActiveFilterFn(ExtensionalityClauseContainer& parent) : _parent(parent) {}
  DECL_RETURN_TYPE(bool);
  OWN_RETURN_TYPE operator()(ExtensionalityClause extCl)
  {
    CALL("ExtensionalityClauseContainer::ActiveFilterFn::operator()");
    
    if (extCl.clause->store() != Clause::ACTIVE) {
      extCl.clause->setExtensionality(false);
      _parent._size--;
      return false;
    }
    return true;
  }
private:
  ExtensionalityClauseContainer& _parent;
};

ExtensionalityClauseIterator ExtensionalityClauseContainer::activeIterator(unsigned sort) {
  CALL("ExtensionalityClauseContainer::activeIterator");
  
  return pvi(getFilteredDelIterator(
               ExtensionalityClauseList::DelIterator(_clausesBySort[sort]),
               ActiveFilterFn(*this)));
}

void ExtensionalityClauseContainer::print (ostream& out) {
  CALL("ExtensionalityClauseContainer::print");
  
  out << "#####################" << endl;

  for(size_t i = 0; i < _clausesBySort.size(); ++i) {
    ExtensionalityClauseList::Iterator it(_clausesBySort[i]);
    while(it.hasNext()) {
      ExtensionalityClause c = it.next();
      out << c.clause->toString() << endl
          << c.literal->toString() << endl
          << c.sort << endl;
    }
  }
  
  out << "#####################" << endl;
}

}
