
/*
 * File AIG.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __AIG__
#define __AIG__

#include "Forwards.hpp"

#include "Debug/Assertion.hpp"

#include "Lib/DHMap.hpp"
#include "Lib/List.hpp"
#include "Lib/Set.hpp"
#include "Lib/SharedSet.hpp"

#include "Kernel/Formula.hpp"
#include "Kernel/FormulaTransformer.hpp"

#define DEBUG_AIG_REF_MEMORY 0

namespace Shell {

using namespace Lib;
using namespace Kernel;

class AIGRewriter;
class AIGTransformer;
class AIGFormulaSharer;

class AIG {
public:
  friend class AIGRewriter;
  friend class AIGTransformer;
  friend class AIGFormulaSharer;

  class Node;
  typedef const SharedSet<unsigned> VarSet;

  class Ref {
    friend class AIG;
   
    size_t _data;

    Ref(Node* n, bool polarity) {
      CALL("AIG::Ref::Ref");
      ASS_EQ(reinterpret_cast<size_t>(n)&1,0); 
#if DEBUG_AIG_REF_MEMORY
      ASS_ALLOC_TYPE(n, "AIG::Node");
#endif

      _data = reinterpret_cast<size_t>(n) | polarity;
    }
  public:
    Ref() {}
   
    bool isInvalid() const { return _data==0; }
   
    bool isValid() const { return _data!=0; }
    bool isPropConst() const;
    bool isTrue() const { return isPropConst() && polarity(); }
    bool isFalse() const  { return isPropConst() && !polarity(); }
    bool isAtom() const;
    bool isQuantifier() const;
    bool isConjunction() const;
    Ref neg() const { return Ref(node(), !polarity()); }

    Ref getPositive() const { return polarity() ? *this : neg(); }

    Literal* getPositiveAtom() const;
    VarSet* getQuantifierVars() const;

    VarSet* getFreeVars() const;
    unsigned getVarSort(unsigned var) const;
   
    Color getColor() const;

    bool polarity() const {
      CALL("AIG::Ref::node()");
      ASS(isValid());
      return _data&1;
    }
    Node* node() const {
      CALL("AIG::Ref::node()");
      ASS(isValid());
      Node* res = reinterpret_cast<Node*>(_data&(~static_cast<size_t>(1)));
#if DEBUG_AIG_REF_MEMORY
      ASS_ALLOC_TYPE(res, "AIG::Node");
#endif
      return res;
    }
    unsigned nodeIndex() const;
    unsigned parentCnt() const;
    Ref parent(unsigned idx) const;

    bool operator==(const Ref& r) const { return _data==r._data; }
    bool operator!=(const Ref& r) const { return !((*this)==r); }

    bool operator<(const Ref& r) const;
    bool operator>(const Ref& r) const { return r<(*this); }
    bool operator>=(const Ref& r) const { return !((*this)<r); }
    bool operator<=(const Ref& r) const { return !(r<(*this)); }

    unsigned hash() const;
    vstring toString() const;
    vstring toInternalString(unsigned depth=1) const;

    static Ref getInvalid() {
      Ref res;
      res._data = 0;
      return res;
    }
  };

  typedef Stack<Ref> AIGStack;
private:
 
  struct RefProxy {
    size_t _data;

    RefProxy& operator= (Ref r) { _data = r._data; return *this; }
    operator Ref() const { Ref res; res._data = _data; return res; }
  };


  struct NodeHash {
    static unsigned hash(Node* n);
    static bool equals(Node* n1, Node* n2);
  };

  typedef DHMap<Literal*,Node*> AtomNodeMap;
  typedef Set<Node*,NodeHash> NodeSet;

  unsigned _simplLevel;

 
  unsigned _nextNodeIdx;

  Node* _trueNode;
  AtomNodeMap _atomNodes;
 
  NodeSet _compNodes;

 
  Node* _conjReserve;
 
  Node* _quantReserve;

 
  Stack<Ref> _orderedNodeRefs;

  Node* atomNode(Literal* positiveLiteral);
  Node* conjNode(Ref par1, Ref par2);
  Node* univQuantNode(VarSet* vars, Ref par);

  void normalizeRefOrder(Ref& par1, Ref& par2);


  
  
  bool tryO1ConjSimpl(Ref par1, Ref par2, Ref& res);
  bool tryO2ConjSimpl(Ref par1, Ref par2, Ref& res);
  bool tryO2AsymetricConjSimpl(Ref par1, Ref par2, Ref& res);
  bool tryO3ConjSimpl(Ref& par1, Ref& par2);
  bool tryO3AsymetricConjSimpl(Ref& par1, Ref& par2);
  bool tryO4ConjSimpl(Ref& par1, Ref& par2);

  Ref getUnivQuant(VarSet* vars, Ref par);
public:
  AIG();
  ~AIG();
  Ref getTrue() const { return Ref(_trueNode, true); };
  Ref getFalse() const { return Ref(_trueNode, false); };
  Ref getLit(Literal* lit);
  Ref getNeg(Ref r) { return r.neg(); }
  Ref getConj(Ref par1, Ref par2);
  Ref getDisj(Ref par1, Ref par2) { return getConj(par1.neg(), par2.neg()).neg(); }
 
  Ref getQuant(bool exQuant, Formula::VarList* vars, Ref par);
  Ref getQuant(bool exQuant, VarSet* vars, Ref par);

  Ref getInvalid() const { return Ref::getInvalid(); }

  void collectConjuncts(Ref aig, AIGStack& res);
  void flattenConjuncts(AIGStack& conjuncts);
  Ref makeConjunction(const AIGStack& conjuncts);

  static VarSet* getTermFreeVars(Term* lit);
  static bool hasPositivePolarity(Ref r) { return r.polarity(); }
};

typedef AIG::Ref AIGRef;
typedef AIG::AIGStack AIGStack;

inline
std::ostream& operator<< (ostream& out, const AIGRef& f)
{
  CALL("operator <<(ostream&,const AIGRef&)");
  return out << f.toString();
}


}

namespace Lib {

template<>
struct FirstHashTypeInfo<Shell::AIGRef> {
  typedef FirstHashTypeInfo Type;

  static unsigned hash(Shell::AIGRef r) { return r.hash(); }
};

}

namespace Shell {

class AIGInsideOutPosIterator
{
public:
  AIGInsideOutPosIterator() { reset(); }
  AIGInsideOutPosIterator(AIGRef a) { reset(a); }
  void reset(AIGRef a=AIGRef::getInvalid());
  void addToTraversal(AIGRef a);

  template<class It>
  void addManyToTraversal(It it) {
    CALL("AIGInsideOutPosIterator::addManyToTraversal");
    while(it.hasNext()) {
      addToTraversal(it.next());
    }
  }

  bool hasNext();
  AIGRef next();
private:
  bool _ready;
  DHSet<AIGRef> _seen;
  Stack<AIGRef> _stack;
};

class AIGTransformer
{
  typedef AIG::Node Node;

  AIG& _aig;
public:
  typedef AIGRef Ref;
 
  typedef DHMap<Ref,Ref> RefMap;

  static Ref lev0Deref(Ref r, RefMap& map);
  Ref lev1Deref(Ref r, RefMap& map);
private:

  typedef MapToLIFO<Ref,Ref> RefEdgeMap;

  void addAIGPredecessors(AIGStack& stack);
  void orderTopologically(AIGStack& stack);
  void collectUsed(Ref r, const RefMap& map, RefEdgeMap& edges);

  void saturateOnTopSortedStack(const AIGStack& stack, RefMap& map);

public:
  AIGTransformer(AIG& aig) : _aig(aig) {}

  void makeOrderedAIGGraphStack(AIGStack& stack);
  void restrictToGetOrderedDomain(RefMap& map, AIGStack& domainOrder);
  void saturateMap(RefMap& map, Stack<Ref>* finalDomain=0);

  void applyWithCaching(Ref r, RefMap& map);
  void makeIdempotent(RefMap& map, Stack<Ref>* finalDomain=0);
};

class AIGFormulaSharer
{
public:
  typedef pair<Formula*,AIGRef> ARes;
private:
  
 
  bool _preBuildRewriteCache;

  AIG _aig;
  AIGTransformer _transf;

  DHMap<AIGRef,Formula*> _formReprs;
  DHMap<Formula*,AIGRef> _formAIGs;

 
  bool _useRewrites;
 
  DHMap<AIGRef,AIGRef> _rewrites;


  Formula* shareFormula(Formula* f, AIGRef aig);

  ARes getSharedFormula(Formula* f);

  
  
  
  ARes applyTrueFalse(Formula* f);
  ARes applyLiteral(Formula* f);
  ARes applyJunction(Formula* f);
  ARes applyNot(Formula* f);
  ARes applyBinary(Formula* f);
  ARes applyQuantified(Formula* f);
  ARes applyIte(Formula* f);

  void buildQuantAigFormulaRepr(AIGRef aig, Stack<AIGRef>& toBuild);
  bool tryBuildEquivalenceFormulaRepr(AIGRef aig, Stack<AIGRef>& toBuild);
  void buildConjAigFormulaRepr(AIGRef aig, Stack<AIGRef>& toBuild);

public:
  AIGFormulaSharer();

  AIG& aig() { return _aig; }
  AIGTransformer& aigTransf() { return _transf; }


  void addRewriteRules(unsigned cnt, Formula* const * srcs, Formula* const * tgts, Stack<unsigned>* usedIdxAcc=0);

  AIGRef getAIG(Clause* cl);

  AIGRef apply(Literal* l) { return _aig.getLit(l); }
  ARes apply(Formula* f);

  void apply(Problem& prb);
  bool apply(UnitList*& units);
  bool apply(FormulaUnit* unit, FormulaUnit*& res);

  Formula* aigToFormula(AIGRef aig);
};


}

#endif 
