
/*
 * File AIGRewriter.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/MapToLIFO.hpp"
#include "Lib/SCCAnalyzer.hpp"
#include "Lib/ScopedPtr.hpp"
#include "Lib/SharedSet.hpp"
#include "Lib/Stack.hpp"

#include "Lib/Environment.hpp"
#include "Shell/Options.hpp"

#include "AIGRewriter.hpp"

namespace Shell
{

unsigned AIGRewriter::getPremiseNum(Unit* u)
{
  CALL("AIGRewriter::getPremiseNum");

  unsigned res = u->number();
  _premMap.insert(res, u); 
  return res;
}

Unit* AIGRewriter::getPremiseUnit(unsigned u)
{
  CALL("AIGRewriter::getPremiseUnit");

  return _premMap.get(u);
}

struct AIGRewriter::L0DerefQueryRec {
  Ref aig;
 
  bool neg;
 
  PremiseSet* prems;

  L0DerefQueryRec(Ref aig, bool neg, PremiseSet* prems) : aig(aig), neg(neg), prems(prems) {}
};

AIG::Ref AIGRewriter::lev0Deref(Ref r, RefMap& map, PremiseSet** prem)
{
  CALL("AIGRewriter::lev0Deref");

  static Stack<L0DerefQueryRec> queries;
  queries.reset();

  bool resNeg = false;

  for(;;) {

    bool neg = !r.polarity();
    if(neg) {
      resNeg = !resNeg;
      r = r.neg();
    }

    PremRef tgt;
    if(!map.find(r, tgt) || r==tgt.first) {
      break;
    }
    queries.push(L0DerefQueryRec(r,resNeg, tgt.second));
    ASS_LE(queries.size(), map.size());
    r = tgt.first;
  }

  if(queries.isNonEmpty()) {
    
    PremiseSet* premises =  queries.top().prems;
    queries.pop();
    while(queries.isNonEmpty()) {
      L0DerefQueryRec rec = queries.pop();
      premises = premises->getUnion(rec.prems);
      Ref toCache = (rec.neg==resNeg) ? r : r.neg();
      
      NEVER(map.set(rec.aig, PremRef(toCache, premises)));
    }
    if(prem) {
      *prem = premises;
    }
  }
  else {
    if(prem) {
      *prem = PremiseSet::getEmpty();
    }
  }

  return resNeg ? r.neg() : r;
}

AIG::Ref AIGRewriter::lev1Deref(Ref r, RefMap& map, PremiseSet** prem)
{
  CALL("AIGRewriter::lev1Deref");
  ASS_EQ(r, lev0Deref(r, map, 0));

  bool neg = !r.polarity();
  if(neg) {
    r = r.neg();
  }

  Ref res;
  if(r.isAtom() || r.isTrue()) {
    res = r;
  }
  else if(r.isConjunction()) {
    Ref p1 = r.parent(0);
    Ref p2 = r.parent(1);
    PremiseSet* prem1;
    PremiseSet* prem2;
    Ref dp1 = lev0Deref(p1, map, &prem1);
    Ref dp2 = lev0Deref(p2, map, &prem2);
    if(p1==dp1 && p2==dp2) {
      res = r;
      if(prem) {
	*prem = PremiseSet::getEmpty();
      }
    }
    else {
      if(prem) {
	*prem = prem1->getUnion(prem2);
      }
      res = _aig.getConj(dp1, dp2);
    }
  }
  else {
    ASS(r.isQuantifier());

    Ref p = r.parent(0);
    Ref dp = lev0Deref(p, map, prem);
    if(p==dp) {
      res = r;
    }
    else {
      res = _aig.getQuant(false, r.getQuantifierVars(), dp);
    }
  }

  return neg ? res.neg() : res;
}

void AIGRewriter::collectUsed(Ref r, const RefMap& map, RefEdgeMap& edges)
{
  CALL("AIGRewriter::collectUsed");
  ASS(r.polarity());
  ASS(map.find(r));

  Ref rImg = map.get(r).first;

  static AIGInsideOutPosIterator ait;

  ait.reset(rImg);

  while(ait.hasNext()) {
    Ref p = ait.next();
    if(map.find(p) && p!=r) { 
      if (env.options->showPreprocessing()) {
        env.beginOutput();
        env.out() << "[PP] aigtr_sat_deps: " << "dep: " << r << " <-- " << p << std::endl;
        env.endOutput();
      }

      edges.pushToKey(r, p);
    }
  }
}

void AIGRewriter::restrictToGetOrderedDomain(RefMap& map, AIGStack& domainOrder)
{
  CALL("AIGRewriter::restrictToGetOrderedDomain");

  if (env.options->showPreprocessing()) {
    env.beginOutput();
    env.out() << "[PP] ordering domain" << std::endl;
    env.endOutput();
  }

  static DHSet<Ref> seen;
  seen.reset();

  typedef MapToLIFOGraph<Ref> RefGraph;
  typedef Subgraph<RefGraph> RefSubgraph;
  RefGraph::Map edgeMap;

  VirtualIterator<Ref> domIt = map.domain();

  while(domIt.hasNext()) {
    Ref d = domIt.next();
    if (env.options->showPreprocessing()) {
      env.beginOutput();
      env.out() << "[PP] aigtr_inp_map: " << d<<" --> "<<map.get(d).first << std::endl;
      env.endOutput();
    }
    collectUsed(d, map, edgeMap);
  }

  RefGraph gr(edgeMap);
  gr.ensureNodes(map.domain());

  SCCAnalyzer<RefGraph> an0(gr);
  ScopedPtr<SCCAnalyzer<RefSubgraph> > an1;
  if(an0.breakingNodes().isNonEmpty()) {
    
    AIGStack::ConstIterator brIt(an0.breakingNodes());
    while(brIt.hasNext()) {
      Ref br = brIt.next();
      if (env.options->showPreprocessing()) {
        env.beginOutput();
        env.out() << "[PP] domain element removed to break cycles: "<<br << std::endl;
        env.endOutput();
      }      
      map.remove(br);
    }
    Subgraph<RefGraph> subGr(gr, AIGStack::ConstIterator(an0.breakingNodes()));
    an1 = new SCCAnalyzer<RefSubgraph>(subGr);
  }

  const Stack<AIGStack>& comps = an1 ? an1->components() : an0.components();
  Stack<AIGStack>::BottomFirstIterator cit(comps);
  while(cit.hasNext()) {
    const AIGStack& comp = cit.next();
    ASS_EQ(comp.size(),1);
    domainOrder.push(comp.top());
    if (env.options->showPreprocessing()) {
      env.beginOutput();
      env.out() << "[PP] domain element to process: "<<comp.top() << std::endl;
      env.endOutput();
    }
  }
}

void AIGRewriter::applyWithCaching(Ref r0, RefMap& map)
{
  CALL("AIGRewriter::applyWithCaching");
  if (env.options->showPreprocessing()) {
    env.beginOutput();
    env.out() << "[PP] applyWithCaching -- "<<r0 << std::endl;
    env.endOutput();
  }

  static AIGInsideOutPosIterator ait;
  ait.reset(r0);

  while(ait.hasNext()) {
    Ref r = ait.next();

    Ref dr0 = lev0Deref(r, map, 0);

    PremiseSet* l1DerefPrem;
    Ref dr1 = lev1Deref(dr0, map, &l1DerefPrem);
    if(r==dr1) {
      
      continue;
    }
#if 1 
    
    
    ASS(r==dr0 || dr0==dr1);
#endif
    if(dr0==dr1) {
      ASS_EQ(map.get(r).first,dr1);
    }
    else {
      map.insert(r,PremRef(dr1, l1DerefPrem));
    }
  }
}

void AIGRewriter::makeIdempotent(RefMap& map, Stack<Ref>* finalDomain)
{
  CALL("AIGRewriter::makeIdempotent");

  static AIGStack order;
  order.reset();
  restrictToGetOrderedDomain(map, order);
  AIGStack::BottomFirstIterator cit(order);
  while(cit.hasNext()) {
    Ref r = cit.next();
    applyWithCaching(map.get(r).first, map);
    lev0Deref(r, map, 0); 
    if(finalDomain) {
      finalDomain->push(r);
    }
  }
}


void AIGRewriter::saturateOnTopSortedStack(const AIGStack& stack, RefMap& map)
{
  CALL("AIGRewriter::saturateOnTopSortedStack");

  unsigned idx = 0;
  
  
  
  
  for(size_t stack_idx = 0; stack_idx<stack.size(); stack_idx++) {
    Ref r = stack[stack_idx];
    ++idx;
    Ref dr0 = lev0Deref(r, map, 0);

    PremiseSet* l1DerefPrem;
    Ref dr1 = lev1Deref(dr0, map, &l1DerefPrem);
    if(r==dr1) {
      
      continue;
    }
#if 1 
    
    
    ASS(r==dr0 || dr0==dr1);
#endif
    if(dr0==dr1) {
      ASS_EQ(map.get(r).first,dr1);
    }
    else {
      map.insert(r,PremRef(dr1,l1DerefPrem));
    }
  }
}


void AIGRewriter::saturateMap(RefMap& map, Stack<Ref>* finalDomain)
{
  CALL("AIGRewriter::saturateMap");
  ASS_REP(forAll(map.domain(), AIG::hasPositivePolarity), getFirstTrue(map.domain(), negPred(AIG::hasPositivePolarity)));

  makeIdempotent(map, finalDomain);

  saturateOnTopSortedStack(_aig._orderedNodeRefs, map);
}


}
