
/*
 * File HornRevealer.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/LiteralSelector.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"

#include "Statistics.hpp"

#include "HornRevealer.hpp"

#undef LOGGING
#define LOGGING 0

namespace Shell
{

void HornRevealer::apply(Problem& prb)
{
  CALL("HornRevealer::apply(Problem&)");

  

  apply(prb.units());
}

void HornRevealer::apply(UnitList*& units)
{
  CALL("HornRevealer::apply");

  buildSatProblem(units);

  _solver.ensureVarCount(env.signature->predicates());
  _solver.addClausesIter(pvi( SATClauseStack::Iterator(_satPrb) ));

  if(_solver.solve(false)==SATSolver::SATISFIABLE) {
    if (env.options->showPreprocessing()) {
      env.beginOutput();
      env.out() << "[PP] Horn discovered" << std::endl;
      env.endOutput();
    }
    discoverGoals(units);

    unsigned preds = env.signature->predicates();
    for(unsigned i=0; i<preds; i++) {
      bool reversed = isReversed(i);
      if(reversed) {
        if (env.options->showPreprocessing()) {
          env.beginOutput();
          env.out() << "[PP] predicate reversed: " << env.signature->predicateName(i) << std::endl;
          env.endOutput();
        }
	LiteralSelector::reversePredicatePolarity(i, true);
	env.statistics->hornReversedPredicates++;
      }
    }
  }
}

void HornRevealer::discoverGoals(UnitList*& units)
{
  CALL("HornRevealer::discoverGoals");

  UnitList::DelIterator uit(units);
  while(uit.hasNext()) {
    Clause* cl = static_cast<Clause*>(uit.next());
    ASS(cl->isClause());
    bool markedAsGoal = cl->inputType()!=Unit::AXIOM;
    bool shouldBeGoal = isGoal(cl);
    if(markedAsGoal==shouldBeGoal) {
      continue;
    }

    Inference* inf = new Inference1(Inference::REORDER_LITERALS, cl);
    Unit::InputType inpType = shouldBeGoal ? Unit::CONJECTURE : Unit::AXIOM;
    Clause* newCl = Clause::fromIterator(Clause::Iterator(*cl), inpType, inf);
    uit.replace(newCl);
    if (env.options->showPreprocessing()) {
      env.beginOutput();
      env.out() << "[PP] Horn revealer changed inp type: " << inpType <<"\n  "<< newCl->toString() << std::endl;
      env.endOutput();
    }
  }
}

bool HornRevealer::isReversed(unsigned pred)
{
  CALL("HornRevealer::isReversed");
  

  
  return _solver.getAssignment(pred2var(pred))==SATSolver::FALSE;
}

bool HornRevealer::isGoal(Clause* cl)
{
  CALL("HornRevealer::isGoal");

  bool hasPositive = false;

  Clause::Iterator cit(*cl);
  while(cit.hasNext()) {
    Literal* l = cit.next();
    unsigned pred = l->functor();
    bool polarity = l->isPositive();

    bool posPolarity = !isReversed(pred);

    if(polarity==posPolarity) {
      ASS_REP(!hasPositive, cl->toString());
      hasPositive = true;
    }
  }
  return !hasPositive;
}

void HornRevealer::buildSatProblem(UnitList* units)
{
  CALL("HornRevealer::buildSatProblem");
  ASS(_satPrb.isEmpty());

  
  static SATLiteralStack slits;
  slits.reset();
  slits.push(SATLiteral(pred2var(0), true));
  SATClause* scl = SATClause::fromStack(slits);
  _satPrb.push(scl);

  UnitList::Iterator uit(units);
  while(uit.hasNext()) {
    Clause* cl = static_cast<Clause*>(uit.next());
    ASS(cl->isClause());
    addToSatProblem(cl);
  }
}

void HornRevealer::addToSatProblem(Clause* cl)
{
  CALL("HornRevealer::addToSatProblem");

  static SATLiteralStack slits;

  unsigned hlen = cl->length();
  for(unsigned i=0; i<hlen; i++) {
    Literal* iLit = (*cl)[i];
    unsigned iPred = iLit->functor();
    bool iPol = iLit->isPositive();
    for(unsigned j=0; j<hlen; j++) {
      if(i==j) {
	continue;
      }
      Literal* jLit = (*cl)[j];
      unsigned jPred = jLit->functor();
      bool jPol = jLit->isPositive();

      slits.reset();
      slits.push(SATLiteral(pred2var(iPred), !iPol));
      slits.push(SATLiteral(pred2var(jPred), !jPol));

      if(slits[0]==slits[1].opposite()) { continue; }
      if(slits[0]==slits[1]) { slits.pop(); }

      SATClause* scl = SATClause::fromStack(slits);
      _satPrb.push(scl);
    }
  }
}

}


