
/*
 * File Storage.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __Storage__
#define __Storage__

#include "Forwards.hpp"

#include "Debug/Assertion.hpp"

#include "Lib/DHMap.hpp"
#include "Lib/List.hpp"
#include "Lib/Stack.hpp"
#include "Lib/VirtualIterator.hpp"

#include "Kernel/Signature.hpp"

#include "Shell/SineUtils.hpp"

namespace Shell
{
namespace LTB
{

using namespace Kernel;

struct StorageCorruptedException
: public Exception
{
  StorageCorruptedException()
  : Exception("The storage of SInE data is corrupted")
  {}
};

typedef SineSymbolExtractor::SymId SymId;
typedef SineSymbolExtractor::SymIdIterator SymIdIterator;

typedef List<vstring> StringList;
typedef Stack<vstring> StringStack;
typedef VirtualIterator<vstring> StringIterator;

typedef pair<unsigned, Unit*> DUnitRecord;
typedef pair<unsigned, SymId> DSymRecord;

typedef VirtualIterator<DUnitRecord> DURIterator;
typedef VirtualIterator<DSymRecord> DSRIterator;

class Storage {
public:
  Storage(bool translateSignature);
  ~Storage();

  

  StringList* getTheoryFileNames();
  void getSignatureSize(int& preds, int& funs);

  VirtualIterator<pair<bool, unsigned> > getGlobalSymbols(Stack<pair<bool, unsigned> >& syms);
  void addCorrespondingLocalSymbols(VirtualIterator<pair<bool,unsigned> > globSyms);

  SymIdIterator getDRelatedSymbols(Stack<SymId>& qsymbols, unsigned itolerance);
  VirtualIterator<unsigned> getDRelatedUnitNumbers(SymIdIterator qsymbols, unsigned itolerance);

  VirtualIterator<unsigned> getNumbersOfUnitsWithoutSymbols();

  UnitList* getClausesByUnitNumbers(VirtualIterator<unsigned> numIt);

  bool getEmptyClausePossession();

  
  void storeSignature();

  void storeTheoryFileNames(StringStack& fnames);

  void storeCNFOfUnit(unsigned unitNumber, ClauseIterator cit);

  void storeUnitsWithoutSymbols(Stack<Unit*>& units);

  void storeDURs(SymId s, Stack<DUnitRecord>& durs);
  void storeDSRs(SymId s, Stack<DSymRecord>& dsrs);

  void storeEmptyClausePossession(bool hasEmptyClause);

private:
  class StorageImpl;

  enum KeyPrefix
  {
   
    PRED_TO_NUM,
   
    FUN_TO_NUM,
   
    PRED_NUM_NAME,
   
    FUN_NUM_NAME,
   
    PRED_NUM_ARITY,
   
    FUN_NUM_ARITY,
   
    THEORY_FILES,
   
    PRED_FUN_CNT,
   
    UNIT_CNF,
   
    UNITS_WITHOUT_SYMBOLS,
   
    SYM_DURS,
   
    SYM_DSRS,
   
    HAS_EMPTY_CLAUSE,
   
    PREFIX_COUNT
  };

  vstring getConstKey(KeyPrefix p);
  vstring getIntKey(KeyPrefix p, int keyNum);

  StringIterator getIntKeyValues(KeyPrefix p, VirtualIterator<int> keyNums);

  void storeConstKey(KeyPrefix p, char* val, size_t valLen);
  void storeIntKey(KeyPrefix p, int keyNum, char* val, size_t valLen);

  void storeSymbolInfo(Signature::Symbol* sym, int index, bool function);

  static const unsigned storedIntMaxSize=4;
  static size_t dumpInt(int num, char* bufStart);
  size_t readInt(const char* buf, int& num);


  DHMap<pair<bool, unsigned>, unsigned> _glob2loc;
  bool _translateSignature;
  StorageImpl* _impl;
};

}
}

#endif 
