
/*
 * File LaTeX.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <fstream>






#include "Options.hpp"
#include "LaTeX.hpp"


#include "Debug/Tracer.hpp"

#include "Lib/Environment.hpp"
#include "Lib/Exception.hpp"
#include "Lib/Int.hpp"
#include "Lib/List.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Unit.hpp"
#include "Kernel/Theory.hpp"



namespace Shell
{

using namespace Lib;
using namespace Kernel;








vstring LaTeX::header()
{
    vstring res =  "\\documentclass[fleqn]{article}\n"
    "\\usepackage{fullpage,latexsym}\n"

    "\\newenvironment{VampireProof}{%\n"
    "   \\section{Proof}}{}\n"
    "\\newenvironment{VampireInference}{%\n"
    "   \\begin{array}{c}}{\\end{array}}\n"
    "\\newenvironment{VampireInferencePremises}{}{}\n"
    "\\newenvironment{VampirePremise}%\n"
    "   {\\begin{array}{l}}%\n"
    "   {\\end{array}}\n"
    "\\newenvironment{VampireConclusion}%\n"
    "   {\\begin{array}{l}}%\n"
    "   {\\end{array}}\n"
    "\\newcommand{\\VampireUnit}[3]{%\n"
    "   #1.~#2~[#3]}\n"

    "\\newcommand{\\VPremiseSeparator}{\\\\}\n"
    "\\newcommand{\\VConclusionSeparator}{\\\\ \\hline}\n"

    "\\newcommand{\\Vor}{\\vee}\n"
    "\\newcommand{\\Vand}{\\wedge}\n"
    "\\newcommand{\\Vimp}{\\rightarrow}\n"
    "\\newcommand{\\Viff}{\\leftrightarrow}\n"
    "\\newcommand{\\Vxor}{\\not\\Viff}\n"

    "\\newcommand{\\VEmptyClause}{\\Box}\n"

    "\\begin{document}\n";

    return res;
}

vstring LaTeX::footer()
{
    return "\\end{document}\n";
}


vstring LaTeX::refutationToString(Unit* ref)
{
  CALL("LaTeX::refutationToString(Unit* ref)");

  vstring res = header() + "\\begin{VampireProof}\n"; 

  InferenceStore* is=InferenceStore::instance();

  Stack<UnitSpec> outKernel;
  Set<UnitSpec> handledKernel;
  Stack<Unit*> outShell;
  Set<Unit*> handledShell;

  if( ref->isClause() ) {
    Clause* refCl=static_cast<Clause*>(ref);
    UnitSpec cs=UnitSpec(refCl);
    outKernel.push(cs);
    handledKernel.insert(cs);
  } else {
    outShell.push(ref);
    handledShell.insert(ref);
  }

  while(outKernel.isNonEmpty()) {
    UnitSpec cs=outKernel.pop();
    InferenceStore::FullInference* finf;
    if(is->findInference(cs, finf)) {
      
      






  

      res+=toStringAsInference(cs, finf);

      for(unsigned i=0;i<finf->premCnt;i++) {
	UnitSpec prem=finf->premises[i];
	ASS(prem!=cs);

	if(!handledKernel.contains(prem)) {
	  handledKernel.insert(prem);
	  outKernel.push(prem);
	}
      }

    } else {
      Clause* cl=cs.cl();
      Inference* inf = cl->inference();

      res+=toStringAsInference(cl);
      Inference::Iterator it = inf->iterator();
      while (inf->hasNext(it)) {
	Unit* prem=inf->next(it);
	if(prem->isClause() ) {
	  
          
	  UnitSpec premCS=UnitSpec(prem);

	  if(!handledKernel.contains(premCS)) {
	    handledKernel.insert(premCS);
	    outKernel.push(premCS);
	  }
	} else {
	  if(!handledShell.contains(prem)) {
	    handledShell.insert(prem);
	    outShell.push(prem);
	  }
	}
      }
    }
  }

  while(outShell.isNonEmpty()) {
    Unit* unit=outShell.pop();
    Inference* inf = unit->inference();

    res+=toStringAsInference(unit);
    Inference::Iterator it = inf->iterator();
    while (inf->hasNext(it)) {
      Unit* prem=inf->next(it);
      if(!handledShell.contains(prem)) {
	handledShell.insert(prem);
	outShell.push(prem);
      }
    }
  }

  if(definitionStack.isNonEmpty()) {
    res+="\\section{BDD Definitions}\n";
    for(unsigned i=0;i<definitionStack.length();i++) {
      res+=definitionStack[i];
    }
  }

  return res + "\\end{VampireProof}\n" + footer(); 

}


vstring LaTeX::toString(Unit* u)
{
  CALL("LaTeX::toString(Unit* u)");

  if(u->isClause()) {
    return toString(static_cast<Clause*>(u));
  } else {
    return toString(static_cast<FormulaUnit*>(u)->formula());
  }
}


vstring LaTeX::toString (Formula* f) const
{
  CALL("LaTeX::toString(Formula* f)");

  static vstring names [] =
  { "", " \\Vand ", " \\Vor ", " \\Vimp ", " \\Viff ", " \\Vxor ",
	  "\\neg ", "\\forall ", "\\exists ", "\bot", "\top"};

  Connective c = f->connective();
  vstring con = names[(int)c];
  switch (c) {
  case LITERAL:
    return toString(f->literal());

  case AND:
  case OR:
  {
    FormulaList::Iterator arg(f->args());
    ASS(arg.hasNext());
    vstring result = toString(arg.next(),c);
    while (arg.hasNext()) {
      result += con + toString(arg.next(),c);
    }
    return result;
  }

  case IMP:
  case IFF:
  case XOR:
    return toString(f->left(),c) + con + toString(f->right(),c);

  case NOT:
    if (f->uarg()->connective() == LITERAL &&
	    f->uarg()->literal()->isEquality()) {
      Literal* l = f->uarg()->literal();
      return toString(l->nthArgument(0)) + " \\neq " + toString(l->nthArgument(1));
    }
    return con + toString(f->uarg(),c);

  case FORALL:
  case EXISTS:
  {
    vstring result("(");
    Formula::VarList::Iterator vs(f->vars());
    while (vs.hasNext()) {
      result += con + varToString(vs.next()) + vstring(" ");
    }
    return result + ")" + toString(f->qarg(),c);
  }

  case FALSE:
  case TRUE:
    return con;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 

vstring LaTeX::toString (Formula* f, Connective outer) const
{
  CALL("LaTeX::toString (Formula* f, Connective outer)");

  return f->parenthesesRequired(outer) ?
	  vstring("(") + toString(f) + ")" :
	  toString(f);
} 


vstring LaTeX::toString (Clause* c)
{
  CALL("LaTeX::toString (Clause* c)");

  vstring result;

  if (c->isEmpty()) {
    result = "\\VEmptyClause";
  }
  else {
    result = toString((*c)[0]);

    unsigned clen=c->length();
    for(unsigned i=1;i<clen;i++) {
      result += vstring(" \\Vor ") + toString((*c)[i]);
    }
  }

  return result;
} 


vstring LaTeX::toString (Literal* l) const
{
  CALL("LaTeX::toString (Literal* l)");

  if (l->isEquality()) {
    if (l->isNegative()) {
      return toString(l->nthArgument(0),true) + " \\neq " + toString(l->nthArgument(1),true);
    }
    else {
      return toString(l->nthArgument(0),true) + " = " + toString(l->nthArgument(1),true);
    }
  }








  
  
  vstring template_str = theory->tryGetInterpretedLaTeXName(l->functor(),true,l->isNegative());

  if(template_str.empty()){
    vstring res;
    if (l->isNegative()) { res="\\neg ";}
    return res+symbolToString(l->functor(), true) + toString(l->args());
  }
  else{
    
    for(unsigned i=0;i<l->arity();i++){
      vstring from = "a"+Lib::Int::toString(i);
      vstring to = toString(l->nthArgument(i),true);
      size_t start_pos = 0;
      while((start_pos = template_str.find(from, start_pos)) != std::string::npos) {
         template_str.replace(start_pos, from.length(), to);
         start_pos += to.length(); 
      }
    }
    return template_str;
  }

} 


vstring LaTeX::symbolToString (unsigned num, bool pred) const
{
  CALL("LaTeX::symbolToString (unsigned num, bool pred)");

  vstring symbolName; 

  if(pred) {
    symbolName = env.signature->predicateName(num);
  }
  else {



    symbolName = env.signature->functionName(num);
  }

  
#define LENGTH 8004
  char newName[LENGTH]; 
  char* name = newName;
  const char* nm = symbolName.substr(0,8000).c_str();
  
  const char* end = nm;
  while (*end) {
    end++;
  }
  
  
  const char* digits = end;
  while (nm != digits) {
    if (digits[-1] >= '0' && digits[-1] <= '9') {
      digits--;
    }
    else {
      break;
    }
  }
  if (digits == nm) { 
    return symbolName;
  }
  else{
    if(digits[-1] == '.'){
      
      
      const char* digits_real = digits;
      digits_real--;
      while(nm != digits_real){
        if (digits_real[-1] >= '0' && digits_real[-1] <= '9') {
          digits_real--;
        }
        else {
          break;
        }
      }
      if(digits_real == nm){ 
        return symbolName;
      }
    }
  }
  while (nm < digits) {
    switch (*nm) {
    case '$':
    case '_':
      *name++ = '\\';
      *name++ = *nm;
      break;
    default:
      *name++ = *nm;
      break;
    }
    nm++;
  }
  if (digits == end) {
    *name = 0;
    return vstring("\\mathit{") + newName + '}';
  }
  
  *name++ = '_';
  *name++ = '{';
  while (digits != end) {
    *name++ = *digits++;
  }
  *name++ = '}';
  *name = 0;
  return vstring("\\mathit{") + newName + '}';
}


vstring LaTeX::toString (TermList* terms,bool single) const
{
  CALL("LaTeX::toString (TermList* terms)");

  if (terms->isEmpty()) {
    return "";
  }

  vstring result = single ? "" : " (";
  bool first=true;
  TermList* t=terms;
  while(t->isNonEmpty()) {

    if(first){
      first=false;
    }
    else{
        result += ",";
    }







    if(t->isVar()) {
      ASS(t->isOrdinaryVar());
      result += varToString(t->var()); 
    }
    else {
      ASS(t->isTerm());
      Term* trm=t->term();

     
     
      vstring template_str = theory->tryGetInterpretedLaTeXName(trm->functor(),false);
   
      if(template_str.empty()){
        result += symbolToString(trm->functor(), false) + toString(trm->args());
      }
      else{
        
        for(unsigned i=0;i<trm->arity();i++){
          vstring from = "a"+Lib::Int::toString(i);
          vstring to = toString(trm->nthArgument(i),true);
          size_t start_pos = 0;
          while((start_pos = template_str.find(from, start_pos)) != std::string::npos) {
            template_str.replace(start_pos, from.length(), to);
            start_pos += to.length();
          }
        }
        result += template_str;
      }
    }

    if(single) break;
    t=t->next();
  }
  return single? result : result + ")";
}






















vstring LaTeX::getClauseLatexId(UnitSpec cs)
{
  return Int::toString(cs.cl()->number())+"_{"+InferenceStore::instance()->getClauseIdSuffix(cs)+"}";
}

vstring LaTeX::toStringAsInference(UnitSpec cs, InferenceStore::FullInference* inf)
{
  CALL("LaTeX::toStringAsInference(ClauseSpec,FullInference*)");

  vstring res("[$");

  bool hasParents=inf->premCnt;
  for(unsigned i=0;i<inf->premCnt;i++) {
    UnitSpec prem=inf->premises[i];
    res += getClauseLatexId(prem);
    if(i+1<inf->premCnt) {
	res += ",";
    }
  }
  if(hasParents) {
    res += "\\rightarrow ";
  }
  res += getClauseLatexId(cs)
    +"$, "+Inference::ruleName(inf->rule)+"]\\\\*[-2ex]\n";

  res += "\\[\\begin{VampireInference}\n";

  if(hasParents) {
    for(unsigned i=0;i<inf->premCnt;i++) {
      UnitSpec prem=inf->premises[i];
      res += "\\begin{VampirePremise}%\n~~";
      res += toString(prem.cl());
      res += "\n\\end{VampirePremise}\n";
      if(i+1<inf->premCnt) {
	res += "\\VPremiseSeparator\n";
      }
    }
    res += "\\VConclusionSeparator\n";
  }

  res += "\\begin{VampireConclusion}\n~~";

  res += toString(cs.cl());

  return res + "\n\\end{VampireConclusion}\n\\end{VampireInference}\n\\]\n";
}

vstring LaTeX::toStringAsInference(Unit* unit)
{
  CALL("LaTeX::toStringAsInference(Unit* unit)");

  Inference* inf = unit->inference();

  vstring res("[$");

  bool hasParents=false;
  Inference::Iterator it = inf->iterator();
  while (inf->hasNext(it)) {
    hasParents=true;
    Unit* prem=inf->next(it);
    res += Int::toString(prem->number());
    if(inf->hasNext(it)) {
	res += ",";
    }
  }
  if(hasParents) {
    res += "\\rightarrow ";
  }
  res += Int::toString(unit->number())+"$, "+Inference::ruleName(inf->rule())+"]\\\\*[-2ex]\n";

  res += "\\[\\begin{VampireInference}\n";

  if(hasParents) {
    it = inf->iterator();
    while (inf->hasNext(it)) {
	Unit* prem=inf->next(it);
      res += "\\begin{VampirePremise}%\n~~";
      res += toString(prem);
      res += "\n\\end{VampirePremise}\n";
	if(inf->hasNext(it)) {
	  res += "\\VPremiseSeparator\n";
	}
    }
    res += "\\VConclusionSeparator\n";
  }

  res += "\\begin{VampireConclusion}\n~~";

  res += toString(unit);

  return res + "\n\\end{VampireConclusion}\n\\end{VampireInference}\n\\]\n";
}



vstring LaTeX::varToString (unsigned num) const
{
  CALL("LaTeX::varToString (unsigned num)");























  
  if(env.signature->vars()<6){
    switch(num){
      case 0 : return "x";
      case 1 : return "y";
      case 2 : return "z";
      case 3 : return "u";
      case 4 : return "v";
      case 5 : return "w";
      default: ASSERTION_VIOLATION;
    }
  }

  return vstring("x_{") + Int::toString(num) + "}";
} 
































vstring LaTeX::getBDDVarName(int var)
{
  CALL("LaTeX::getBDDVarName(int var)");

  vstring name;
  if(BDD::instance()->getNiceName(var, name)) {
    return vstring("\\mathit{") + name + '}';
  }
  return vstring("\\mathit{b_{") + Int::toString(var) + "}}";
}

vstring LaTeX::toString(BDDNode* node)
{
  CALL("LaTeX::toString(BDDNode*)");

  BDD* inst=BDD::instance();

  
  if(inst->isTrue(node)) {
    return "\\top";
  }
  if(inst->isFalse(node)) {
    return "\\bot";
  }

  vstring name;
  if(_nodeNames.find(node, name)) {
    return name;
  }


  vstring propPred=getBDDVarName(node->_var);
  if(inst->isTrue(node->_pos) && inst->isFalse(node->_neg)) {
    return propPred;
  }
  else if(inst->isFalse(node->_pos) && inst->isTrue(node->_neg)) {
    return "\\neg "+propPred;
  }
  else if(inst->isTrue(node->_pos)) {
    return "("+propPred+" \\Vor "+toString(node->_neg)+")";
  }
  else if(inst->isFalse(node->_neg)) {
    return "("+propPred+" \\Vand "+toString(node->_pos)+")";
  }
  else if(inst->isFalse(node->_pos)) {
    return "(\\neg "+propPred+" \\Vand "+toString(node->_neg)+")";
  }
  else if(inst->isTrue(node->_neg)) {
    return "(\\neg "+propPred+" \\Vor "+toString(node->_pos)+")";
  }
  else {
    vstring posDef=toString(node->_pos);
    vstring negDef=toString(node->_neg);

    vstring name=vstring("\\mathit{n_{") + Int::toString(_nextNodeNum++) + "}}";
    vstring report="$"+name+" \\Viff ("+propPred+" ? "+posDef+" : "+negDef+")$\\\\\n";
    definitionStack.push(report);
    ALWAYS(_nodeNames.insert(node, name));

    return name;
  }

}


}
