
/*
 * File LispLexer.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <cstring>

#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"

#include "Lib/Exception.hpp"

#include "LispLexer.hpp"

using namespace Shell;

LispLexer::LispLexer (istream& in)
  : Lexer (in)
{
} 


void LispLexer::skipWhiteSpacesAndComments ()
{
  CALL("LispLexer::skipWhiteSpacesAndComments");

  bool comment = false;
  while (! _eof) {
    switch (_lastCharacter) {
    case ' ':
    case '\t':
    case '\r':
    case '\f':
      readNextChar();
      break;

    case '\n': 
      comment = false;
      readNextChar();
      break;

    case ';': 
      comment = true;
      readNextChar();
      break;

    default:
      if (comment) {
	readNextChar();
	break;
      }
      return;
    }
  }
} 

void LispLexer::readToken (Token& token)
{
  CALL("LispLexer::readToken");

  skipWhiteSpacesAndComments();
  _charCursor = 0;

  if (_eof) {
    token.tag = TT_EOF;
    token.text = "";
    return;
  }
  token.line = _lineNumber;

  switch (_lastCharacter) {
  case '(':
    token.tag = TT_LPAR;
    saveLastChar();
    saveTokenText(token);
    readNextChar();
    break;
  case ')':
    token.tag = TT_RPAR;
    saveLastChar();
    saveTokenText(token);
    readNextChar();
    break;
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    readNumber(token);
    return;
  case '|':
    readQuotedString(token, '|', '|');
    return;
  case '{':
    readQuotedString(token, '{', '}');
    return;
  default:
    readName(token);
    token.tag = TT_NAME;
    return;
  }
} 

void LispLexer::readName (Token& token)
{
  CALL("LispLexer::readName");

  saveLastChar();

  while (readNextChar()) {
    switch(_lastCharacter) {
    case ' ':
    case '\t':
    case '\r':
    case '\f':
    case '\n':
    case ';':
    case ')':
    case '(':
    case '{':
    case '}':
      
      saveTokenText(token);
      return;
    default:
      saveLastChar();
      break;
    }
  }

  
  saveTokenText(token);
} 

void LispLexer::readQuotedString(Token& token, char opening, char closing)
{
  CALL("LispLexer::readQuotedString");

  bool escape=false;
  saveLastChar();

  while (readNextChar()) {
    if (_lastCharacter == '\\' && !escape) {
      escape=true;
    }
    else if (_lastCharacter == closing && !escape) {
      saveLastChar();
      saveTokenText(token);
      readNextChar();
      token.tag = TT_NAME;
      return;
    }
    else {
      if (escape && _lastCharacter!=closing && _lastCharacter!='\\') {
	throw LexerException((vstring)"invalid escape sequence in quoted string ", *this);
      }
      escape=false;
      saveLastChar();
    }
  }
  throw LexerException((vstring)"file ended while reading quoted string ", *this);
} 



