
/*
 * File LispParser.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __LispParser__
#define __LispParser__

#include "Forwards.hpp"
#include "Token.hpp"

#include "Lib/Exception.hpp"
#include "Lib/List.hpp"
#include "Lib/Portability.hpp"
#include "Lib/Stack.hpp"
#include "Lib/VString.hpp"

namespace Shell {

using namespace std;
using namespace Lib;

class LispLexer;

class LispParser
{
public:
 
  enum Tag {
   
    ATOM = 0,
   
    LIST = 1
  };

 
  struct Expression {
   
    Tag tag;
   
    vstring str;
   
    List<Expression*>* list;
   
    explicit Expression(Tag t)
      : tag(t),
	str("?"),
	list(0)
    {}
   
    Expression(Tag t,vstring s)
      : tag(t),
	str(s),
	list(0)
    {}
    vstring toString(bool outerParentheses=true) const;

    bool isList() const { return tag==LIST; }
    bool isAtom() const { return tag==ATOM; }

    bool get2Args(vstring functionName, Expression*& arg1, Expression*& arg2);
    bool get1Arg(vstring functionName, Expression*& arg);
    bool getPair(Expression*& el1, Expression*& el2);
    bool getSingleton(Expression*& el);
  };

  typedef Lib::List<Expression*> List;

  explicit LispParser(LispLexer& lexer);
  Expression* parse();
  void parse(List**);

 
  class Exception 
    : public Lib::Exception
  {
  public:                                
    Exception (vstring message,const Token&);
    void cry (ostream&);
    ~Exception () {}
  protected:
    vstring _message;
  }; 

private:
 
  LispLexer& _lexer;
 
  int _balance;
}; 

typedef LispParser::Expression LExpr;
typedef List<LExpr*> LExprList;


class LispListReader {
public:
  explicit LispListReader(LExpr* e)
  {
    CALL("LispListReader::LispListReader(LExpr*)");
    if(!e->isList()) {
      lispError(e, "list expected");
    }
    it = LExprList::Iterator(e->list);
  }
  explicit LispListReader(LExprList* list) : it(list) {}

  void lispError(LExpr* expr, vstring reason="error") NO_RETURN;
  void lispCurrError(vstring reason="error") NO_RETURN;

  bool hasNext() { return it.hasNext(); }
  LExpr* peekAtNext();
  LExpr* readNext();
  LExpr* next() { return readNext(); }

  bool tryReadAtom(vstring& atom);
  vstring readAtom();

  bool tryReadListExpr(LExpr*& e);
  LExpr* readListExpr();

  bool tryReadList(LExprList*& list);
  LExprList* readList();

  bool tryAcceptAtom(vstring atom);
  void acceptAtom(vstring atom);
  void acceptAtom() { readAtom(); }

  bool tryAcceptList();
  void acceptList();

  void acceptEOL();

  bool lookAheadAtom(vstring atom);

  bool tryAcceptCurlyBrackets();
private:
  LExprList::Iterator it;
};

class LispListWriter
{
public:
  LispListWriter()
  {
#if VDEBUG
    _destroyed = false;
#endif
  }

#if VDEBUG
  ~LispListWriter()
  {
    _destroyed = true;
  }
#endif

  LispListWriter& operator<<(vstring s)
  {
    _elements.push(new LExpr(LispParser::ATOM, s));
    return *this;
  }

  LispListWriter& operator<<(LExpr* e)
  {
    _elements.push(e);
    return *this;
  }

  LispListWriter& operator<<(const LispListWriter& e)
  {
    _elements.push(e.get());
    return *this;
  }

  LispListWriter& append(LExprList* lst)
  {
    _elements.loadFromIterator(LExprList::Iterator(lst));
    return *this;
  }

  LExprList* getList() const
  {
    CALL("LispListWriter::getList");
    ASS(!_destroyed);

    LExprList* res = 0;
    LExprList::pushFromIterator(Stack<LExpr*>::TopFirstIterator(_elements), res);
    return res;
  }

  LExpr* get() const
  {
    CALL("LispListWriter::get");

    LExpr* res = new LExpr(LispParser::LIST);
    res->list = getList();
    return res;
  }

private:
#if VDEBUG
  bool _destroyed;
#endif
  Stack<LExpr*> _elements;
};

}

#endif

