
/*
 * File Normalisation.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Sort.hpp"
#include "Lib/Environment.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SubformulaIterator.hpp"
#include "Kernel/Unit.hpp"

#include "Normalisation.hpp"

using namespace Kernel;
using namespace Shell;

Normalisation::Normalisation ()
  : _counter(*env.signature)
{
} 

void Normalisation::normalise(Problem& prb)
{
  CALL("Normalisation::normalise");

  UnitList* units = prb.units();
  units = normalise(units);
  prb.units() = units;
}

UnitList* Normalisation::normalise (UnitList* units)
{
  CALL("Normalisation::normalise (UnitList*");

  if (units->isEmpty()) {
    return units;
  }
  _counter.count(units,1);

  int length = units->length();

  
  Sort<Unit*,Normalisation> srt(length,*this);
  UnitList::Iterator us(units);
  while (us.hasNext()) {
    Unit* unit = us.next();
    normalise(unit);
    srt.add(unit);
  }
  srt.sort();
  UnitList* result = UnitList::empty();
  for (int k = length-1;k >= 0;k--) {
    result = new UnitList(srt[k],result);
  }
  units->destroy();
  return result;
} 


void Normalisation::normalise (Unit* unit)
{
  CALL ("Normalisation::normalize(Unit*)");

  if (! unit->isClause()) {
    return;
  }

  Clause& clause = *static_cast<Clause*>(unit);
  int length = clause.length();

  if (length <= 1) {
    return;
  }

  
  Sort<Literal*,Normalisation> srt(length,*this);
  for (int i = 0;i < length;i++) {
    srt.add(clause[i]);
  }
  srt.sort();
  for (int i=0;i < length;i++) {
    clause[i] = srt[i];
  }

  clause.notifyLiteralReorder();
  return;
} 


bool Normalisation::lessThan (Unit* u1, Unit* u2)
{
  CALL("Normalisation::lessThan (const Unit*...)");

  
  
  switch (compare((int)u1->inputType(),(int)u2->inputType())) {
  case LESS:
    return false;
  case EQUAL:
    break;
  case GREATER:
    return true;
  }

  if (u1->isClause()) {
    if (u2->isClause()) {
      return lessThan(static_cast<Clause*>(u1),
		      static_cast<Clause*>(u2));
    }
    return true;
  }
  

  if (u2->isClause()) {
    return false;
  }
  return lessThan(static_cast<FormulaUnit*>(u1)->formula(),
		  static_cast<FormulaUnit*>(u2)->formula());
} 

bool Normalisation::lessThan(Clause* cl1, Clause* cl2)
{
  if(cl1->length()!=cl2->length()) {
    return cl1->length() < cl2->length();
  }
  unsigned clen=cl1->length();
  for(unsigned i=0;i<clen;i++) {
    Comparison cmp=compare( (*cl1)[i], (*cl2)[i] );
    if(cmp!=EQUAL) {
      return cmp==LESS;
    }
  }
  return false;
}


bool Normalisation::lessThan (Formula* f1, Formula* f2)
{
  CALL("Normalisation::lessThan (const Formula*...)");

  return compare(f1,f2) == LESS;
} 


bool Normalisation::lessThan (Literal* l1, Literal* l2)
{
  CALL("Normalisation::lessThan (const Literal*...)");

  return compare(l1,l2) == LESS;
}


Comparison Normalisation::compare (Formula* fm1, Formula* fm2)
{
  CALL("Normalisation::compare (const Formula*...)");

  SubformulaIterator sf1(fm1);
  SubformulaIterator sf2(fm2);

  while (sf1.hasNext()) {
    if (! sf2.hasNext()) {
      return GREATER;
    }
    Formula* f1 = sf1.next();
    Formula* f2 = sf2.next();

    Comparison comp = compare ((int)f1->connective(),
			       (int)f2->connective ());
    if (comp != EQUAL) {
      return comp;
    }

    
    switch (f1->connective()) {
    case LITERAL:
      comp = compare(f1->literal(),f2->literal());
      if (comp != EQUAL) {
	return comp;
      }
      break;

    case FORALL:
    case EXISTS:
      
      
      comp = compare(f1->vars()->length(),f2->vars()->length());
      if (comp != EQUAL) {
	return comp;
      }
      break;

    default:
      break;
    }
  }
  return sf2.hasNext() ? LESS : EQUAL;
} 



Comparison Normalisation::compare (Literal* l1, Literal* l2)
{
  CALL("Normalisation::compare (const Literal*...)");

  if (l1 == l2) {
    return EQUAL;
  }

  Comparison comp = compare((int)l1->weight(),(int)l2->weight());
  if (comp != EQUAL) {

    return comp;
  }

  
  if (l1->isPositive()) {
    if (! l2->isPositive()) {
      return GREATER;
    }
  }
  else { 
    if (l2->isPositive()) {
      return LESS;
    }
  }

  
  
  if (l1->isEquality()) {
    if (! l2->isEquality()) {
      return LESS;
    }
  }
  else if (l2->isEquality()) {
    return GREATER;
  }

  int p1 = l1->functor();
  int p2 = l2->functor();
  if (p1 != p2) {
    comp = compare((int)l1->arity(),(int)l2->arity());
    if (comp != EQUAL) {
      return comp;
    }
    comp = compare((int)l1->vars(),(int)l2->vars());
    if (comp != EQUAL) {
      return comp;
    }
    comp = compare(_counter.getPred(p1).pocc(),
		   _counter.getPred(p2).pocc());
    if (comp != EQUAL) {
      return comp;
    }
    comp = compare(_counter.getPred(p1).nocc(),
		   _counter.getPred(p2).nocc());
    if (comp != EQUAL) {
      return comp;
    }
    comp = compare(_counter.getPred(p1).docc(),
		   _counter.getPred(p2).docc());
    if (comp != EQUAL) {
      return comp;
    }
  }

  return compare(l1->args(),l2->args());
} 


Comparison Normalisation::compare(TermList* ts1, TermList* ts2)
{
  CALL("Normalisation::compare(TermList*...)");

  while (! ts1->isEmpty()) {
    if (ts2->isEmpty()) {
      return GREATER;
    }

    
    if (ts1->isVar()) {
      if (ts2->isVar()) {
	ts1 = ts1->next();
	ts2 = ts2->next();
	continue;
      }
      return LESS;
    }

    if (ts2->isVar()) {
      return GREATER;
    }
    

    const Term* t1 = ts1->term();
    const Term* t2 = ts2->term();
    Comparison ct = compare(t1,t2);
    switch (ct) {
    case EQUAL:
      ts1 = ts1->next();
      ts2 = ts2->next();
      break;
    default:
      return ct;
    }
  }

  return ts2->isEmpty() ? EQUAL : LESS;
} 


Comparison Normalisation::compare(Term* t1, Term* t2)
{
  ASS("Normalisation::compare(const Term*...)");

  if (t1 == t2) {
    return EQUAL;
  }

  Comparison comp = compare((int)t1->weight(),(int)t2->weight());
  if (comp != EQUAL) {
    return comp;
  }

  int f1 = t1->functor();
  int f2 = t2->functor();
  if (f1 != f2) {
    comp = compare((int)t1->arity(),(int)t2->arity());
    if (comp != EQUAL) {
      return comp;
    }
    comp = compare((int)t1->vars(),(int)t2->vars());
    if (comp != EQUAL) {
      return comp;
    }
    comp = compare(_counter.getFun(f1).occ(),
		   _counter.getFun(f2).occ());
    if (comp != EQUAL) {
      return comp;
    }
  }

  return compare(t1->args(),t2->args());
} 
