
/*
 * File Options.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */



#include <cmath>

#include "Forwards.hpp"

#include "Debug/Tracer.hpp"
#include "Debug/Assertion.hpp"

#include "Lib/VString.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Timer.hpp"
#include "Lib/Exception.hpp"
#include "Lib/Int.hpp"
#include "Lib/Random.hpp"
#include "Lib/Set.hpp"
#include "Lib/System.hpp"

#include "Kernel/Problem.hpp"

#include "Options.hpp"
#include "Property.hpp"

using namespace Lib;
using namespace Shell;

Options::Options ()

{
    CALL("Options::Options");
    init();
}

void Options::Options::init()
{
   CALL("Options::init");

   BYPASSING_ALLOCATOR; 

   
    _forceIncompleteness = BoolOptionValue("force_incompleteness","",false);
    _equivalentVariableRemoval = BoolOptionValue("equivalentVariableRemoval","",true);
    _bpCollapsingPropagation = BoolOptionValue("bp_collapsing_propagation","",false);
    _backjumpTargetIsDecisionPoint = BoolOptionValue("backjump_target_is_decision_point","",true);
    _selectUnusedVariablesFirst = BoolOptionValue("_selectUnusedVariablesFirst","",false);





    _memoryLimit = UnsignedOptionValue("memory_limit","m",
#if VDEBUG
                                       1000
#else
                                       3000
#endif
                                       );
    _memoryLimit.description="Memory limit in MB";
    _lookup.insert(&_memoryLimit);
#if !__APPLE__ && !__CYGWIN__
    _memoryLimit.addHardConstraint(lessThanEq((unsigned)Lib::System::getSystemMemory()));
#endif

    _mode = ChoiceOptionValue<Mode>("mode","",Mode::VAMPIRE,
                                    {"axiom_selection","bpa","casc","casc_epr",
                                        "casc_ltb","casc_mzr","casc_sat","casc_theory","clausify",
                                        "consequence_elimination","grounding",
                                        "ltb_build","ltb_solve","output","preprocess",
                                        "profile","program_analysis","random_strategy",
                                        "sat_solver","spider","vampire"});
    _mode.description=
    "Select the mode of operation. Choices are:\n"
    "  -vampire: the standard mode of operation for first-order theorem proving\n"
    "  -casc,casc_epr,casc_ltb,casc_mzr,casc_sat,casc_theory: these are all portfolio modes\n   that use predefined "
    " sets of strategies in vampire mode.\n"
    "  -preprocess,axiom_select,clausify,grounding: modes for producing output\n   for other solvers.\n"
    "  -output,profile: output information about the problem\n"
    "  -sat_solver: accepts problems in DIMACS and uses the internal sat solver\n   directly\n"
    "Some modes are not currently maintained:\n"
    "  -ltb_build,ltb_solve: for Large Theory Batch processing\n"
    "  -program_analysis: run Lingva\n"
    "  -bpa: perform bound propagation\n"
    "  -consequence_elimination: perform consequence elimination\n";
    
    _lookup.insert(&_mode);

    _decode = DecodeOptionValue("decode","",this);
    _decode.description="Decodes an encoded strategy. Can be used to replay a strategy. To make Vampire output an encoded version of the strategy use the option --encode on";
    _lookup.insert(&_decode);

    _encode = BoolOptionValue("encode","",false);
    _encode.description="Output an encoding of the strategy to be used with the --decode option";
    _lookup.insert(&_encode);

    _randomStrategy = ChoiceOptionValue<RandomStrategy>("random_strategy","",RandomStrategy::OFF,{"on","off","sat","nocheck"});
    _randomStrategy.description =
      "Create a random strategy. Randomisation will occur after all other options have been "
      "set, whatever order they have been given in. A random number of options will be selected "
      " and set with a safe (possibly default) value.";
    _lookup.insert(&_randomStrategy);
    _randomStrategy.reliesOnHard(_mode.is(equal(Mode::VAMPIRE)->Or(_mode.is(equal(Mode::RANDOM_STRATEGY)))));

    _forbiddenOptions = StringOptionValue("forbidden_options","","");
    _forbiddenOptions.description=
    "If some of the specified options are set to a forbidden state, vampire will fail to start, or in the CASC mode it will skip such strategies. The expected syntax is <opt1>=<val1>:<opt2>:<val2>:...:<optn>=<valN>";
    _lookup.insert(&_forbiddenOptions);

    _forcedOptions = StringOptionValue("forced_options","","");
    _forcedOptions.description=
    "Options in the format <opt1>=<val1>:<opt2>=<val2>:...:<optn>=<valN> that override the option values set by other means (also inside CASC mode strategies)";
    _lookup.insert(&_forcedOptions);

    _showHelp = BoolOptionValue("help","h",false);
    _showHelp.description="Display this help";
    _lookup.insert(&_showHelp);

    _showOptions = BoolOptionValue("show_options","",false);
    _showOptions.description="List all available options";
    _lookup.insert(&_showOptions);

    _showExperimentalOptions = BoolOptionValue("show_experimental_options","",false);
    _showExperimentalOptions.description="Include experimental options in showOption";
    _lookup.insert(&_showExperimentalOptions);
    _showExperimentalOptions.setExperimental(); 

    _explainOption = StringOptionValue("explain_option","explain","");
    _explainOption.description = "Use to explain a single option i.e. -explain explain";
    _lookup.insert(&_explainOption);

    _ignoreMissing = BoolOptionValue("ignore_missing","",false);
    _ignoreMissing.description=
    "Ignore any options that have been removed (useful in CASC mode where this can cause errors)";
    _lookup.insert(&_ignoreMissing);

    _badOption = ChoiceOptionValue<BadOption>("bad_option","",BadOption::SOFT,{"hard","forced","off","soft"});
    _badOption.description = "What should be done if a bad option value (wrt constraints) is encountered:\n"
       " - hard: will cause a user error\n"
       " - soft: will only report the error (unless it is unsafe)\n"
       " - forced: will try and change the value and those of unset options to make it okay\n"
       " - off: will ignore safe errors\n"
       "Note that unsafe errors will aways lead to a user error";
    _lookup.insert(&_badOption);


    _namePrefix = StringOptionValue("name_prefix","","");
    _namePrefix.description=
    "Prefix for symbols introduced by Vampire (naming predicates, Skolem functions)";
    _lookup.insert(&_namePrefix);

    
    _problemName = StringOptionValue("problem_name","","");
    _problemName.description="";
    

    _proof = ChoiceOptionValue<Proof>("proof","p",Proof::ON,{"off","on","proofcheck","tptp"});
    _proof.description=
    "Specifies whether proof will be output. 'proofcheck' will output proof as a sequence of TPTP problems to allow for proof-checking.";
    _lookup.insert(&_proof);

    _proofChecking = BoolOptionValue("proof_checking","",false);
    _proofChecking.description="";
    _lookup.insert(&_proofChecking);
    _proofChecking.setExperimental(); 

    _protectedPrefix = StringOptionValue("protected_prefix","","");
    _protectedPrefix.description="Symbols with this prefix are immune against elimination during preprocessing";
    _lookup.insert(&_protectedPrefix);

    _statistics = ChoiceOptionValue<Statistics>("statistics","",Statistics::FULL,{"brief","full","none"});
    _statistics.description="The level of statistics to report at the end of the run.";
    _lookup.insert(&_statistics);

    _testId = StringOptionValue("test_id","","unspecified_test");
    _testId.description="";
    _lookup.insert(&_testId);
    _testId.setExperimental();

    _thanks = StringOptionValue("thanks","","Tanya");
    _thanks.description="";
    _lookup.insert(&_thanks);
    _thanks.setExperimental();
    _timeLimitInDeciseconds = TimeLimitOptionValue("time_limit","t",600); 
    _timeLimitInDeciseconds.description="Time limit in wall clock seconds, you can use d,s,m,h,D suffixes also i.e. 60s, 5m";
    _lookup.insert(&_timeLimitInDeciseconds);

    _timeStatistics = BoolOptionValue("time_statistics","",false);
    _timeStatistics.description="Show how much running time was spent in each part of Vampire";
    _lookup.insert(&_timeStatistics);



    _include = StringOptionValue("include","","");
    _include.description="Path prefix for the 'include' TPTP directive";
    _lookup.insert(&_include);
    _include.tag(OptionTag::INPUT);

    _inputFile= InputFileOptionValue("input_file","","",this);
    _inputFile.description="Problem file to be solved (if not specified, standard input is used)";
    _lookup.insert(&_inputFile);
    _inputFile.tag(OptionTag::INPUT);
    _inputFile.setExperimental();

    _inputSyntax= ChoiceOptionValue<InputSyntax>("input_syntax","",
                                                 
#if !GNUMP
                                                 InputSyntax::TPTP,
#else
                                                 InputSyntax::SMTLIB,
#endif
                                                 {"simplify","smtlib","smtlib2","tptp","xhuman","xmps","xnetlib"});
    _inputSyntax.description="Input syntax";
    _lookup.insert(&_inputSyntax);
    _inputSyntax.tag(OptionTag::INPUT);

    _smtlibConsiderIntsReal = BoolOptionValue("smtlib_consider_ints_real","",false);
    _smtlibConsiderIntsReal.description="all integers will be considered to be reals by the SMTLIB parser";
    _lookup.insert(&_smtlibConsiderIntsReal);
    _smtlibConsiderIntsReal.setExperimental();
    _smtlibConsiderIntsReal.tag(OptionTag::INPUT);

    _smtlibFletAsDefinition = BoolOptionValue("smtlib_flet_as_definition","",false);
    _smtlibFletAsDefinition.description="";
    _lookup.insert(&_smtlibFletAsDefinition);
    _smtlibFletAsDefinition.setExperimental();
    _smtlibFletAsDefinition.tag(OptionTag::INPUT);

    _smtlibIntroduceAIGNames = BoolOptionValue("smtlib_introduce_aig_names","",true);
    _smtlibIntroduceAIGNames.description="";
    _lookup.insert(&_smtlibIntroduceAIGNames);
    _smtlibIntroduceAIGNames.setExperimental();
    _smtlibIntroduceAIGNames.tag(OptionTag::INPUT);



    _inequalitySplitting = IntOptionValue("inequality_splitting","ins",0);
    _inequalitySplitting.description=
    "Defines a weight threshold w such that any clause C \\/ s!=t where s (or conversely t) is ground "
    "and has weight less than w is replaced by C \\/ p(s) with the additional clause ~p(t) being added "
    "for fresh predicate p. ";
    _lookup.insert(&_inequalitySplitting);
    _inequalitySplitting.tag(OptionTag::PREPROCESSING);

    

    _sos = ChoiceOptionValue<Sos>("sos","sos",Sos::OFF,{"all","off","on","theory"});
    _sos.description=
    "Set of support strategy. All formulas annotated as theory axioms are put directly among active clauses, without performing any inferences between them. If all, select all literals of set-of-support clauses, ortherwise use the default literal selector.";
    _lookup.insert(&_sos);
    _sos.tag(OptionTag::PREPROCESSING);
    
    _sos.addConstraint(If(notEqual(Sos::OFF)).then(_saturationAlgorithm.is(
                                                    notEqual(SaturationAlgorithm::TABULATION))));
    _sos.setRandomChoices(And(isRandSat(),saNotInstGen()),{"on","off","off","off","off"});
    _sos.setRandomChoices(And(isRandOn(),hasNonUnits()),{"on","off","off","off","off"});
    _sos.setRandomChoices(isRandOn(),{"all","off","on"});



    _equalityProxy = ChoiceOptionValue<EqualityProxy>( "equality_proxy","ep",EqualityProxy::OFF,{"R","RS","RST","RSTC","off"});
    _equalityProxy.description="Aplies the equality proxy transformation to the problem. It works as follows:\n"
     " - All literals s=t are replaced by E(s,t)\n"
     " - All literals s!=t are replaced by ~E(s,t)\n"
     " - If S the symmetry clause ~E(x,y) \\/ E(y,x) is added\n"
     " - If T the transitivity clause ~E(x,y) \\/ ~E(y,z) \\/ E(x,z) is added\n"
     " - If C the congruence clauses are added as follows:\n"
     "    for predicates p that are not E or equality add\n"
     "     ~E(x1,y1) \\/ ... \\/ ~E(xN,yN) \\/ ~p(x1,...,xN) \\/ p(y1,...,yN)\n"
     "    for non-constant functions f add\n"
     "     ~E(x1,y1) \\/ ... \\/ ~E(xN,yN) \\/ E(f(x1,...,xN),f(y1,...,yN))";
    _lookup.insert(&_equalityProxy);
    _equalityProxy.tag(OptionTag::PREPROCESSING);
    _equalityProxy.addProblemConstraint(hasEquality());
    _equalityProxy.setRandomChoices(isRandOn(),{"R","RS","RST","RSTC","off","off","off","off","off"}); 
    
    

    _aigBddSweeping = BoolOptionValue("aig_bdd_sweeping","",false);
    _aigBddSweeping.description="For a description of these aig options see the paper 'Preprocessing Techniques for First-Order Clausification'. ";
    _lookup.insert(&_aigBddSweeping);
    _aigBddSweeping.tag(OptionTag::PREPROCESSING);

    _aigConditionalRewriting = BoolOptionValue("aig_conditional_rewriting","",false);
    _aigConditionalRewriting.description="";
    _lookup.insert(&_aigConditionalRewriting);
    _aigConditionalRewriting.tag(OptionTag::PREPROCESSING);

    _aigDefinitionIntroduction = BoolOptionValue("aig_definition_introduction","",false);
    _aigDefinitionIntroduction.description="";
    _lookup.insert(&_aigDefinitionIntroduction);
    _aigDefinitionIntroduction.tag(OptionTag::PREPROCESSING);

    _aigDefinitionIntroductionThreshold = UnsignedOptionValue("aig_definition_introduction_threshold","",4);
    _aigDefinitionIntroductionThreshold.description=
               "number of subformula occurrences needed to introduce a name for it (if aig_definition_introduction is enabled)";
    _lookup.insert(&_aigDefinitionIntroductionThreshold);
    _aigDefinitionIntroductionThreshold.tag(OptionTag::PREPROCESSING);
    _aigDefinitionIntroductionThreshold.addConstraint(notEqual(0u));

    _aigFormulaSharing = BoolOptionValue("aig_formula_sharing","",false);
    _aigFormulaSharing.description="Detection and sharing of common subformulas using AIG representation";
    _lookup.insert(&_aigFormulaSharing);
    _aigFormulaSharing.tag(OptionTag::PREPROCESSING);

    _aigInliner = BoolOptionValue("aig_inliner","",false);
    _aigInliner.description="";
    _lookup.insert(&_aigInliner);
    _aigInliner.tag(OptionTag::PREPROCESSING);

    _arityCheck = BoolOptionValue("arity_check","",false);
    _arityCheck.description="The same symbol name cannot be used with multiple arities";
    _lookup.insert(&_arityCheck);
    _arityCheck.tag(OptionTag::PREPROCESSING);
    
    _eprPreservingNaming = BoolOptionValue("epr_preserving_naming","",false);
    _eprPreservingNaming.description=
    "Naming will not cause introduction of any non-constant functions."
    "The nonconstant functions can be introduced by naming in a name definition when a universal quantifier turns into an existential one and is skolemized.";
    _lookup.insert(&_eprPreservingNaming);
    _eprPreservingNaming.tag(OptionTag::PREPROCESSING);

    _eprPreservingSkolemization= BoolOptionValue("epr_preserving_skolemization","",false);
    _eprPreservingSkolemization.description="";
    _lookup.insert(&_eprPreservingSkolemization);
    _eprPreservingSkolemization.tag(OptionTag::PREPROCESSING);

    _eprRestoringInlining= BoolOptionValue("epr_restoring_inlining","",false);
    _eprRestoringInlining.description="";
    _lookup.insert(&_eprRestoringInlining);
    _eprRestoringInlining.tag(OptionTag::PREPROCESSING);

    _equalityPropagation = BoolOptionValue("equality_propagation","",false);
    _equalityPropagation.description=
    "propagate equalities in formulas, for example\n"
    "X=Y => X=f(Y) ---> X=f(X)\n"
    "Such propagation can simplify formulas early in the preprocessing and so help other "
    "preprocessing rules (namely dealing with predicate definitions).";
    _lookup.insert(&_equalityPropagation);
    _equalityPropagation.tag(OptionTag::PREPROCESSING);
    _equalityPropagation.addProblemConstraint(notWithCat(Property::FEQ));
    _equalityPropagation.setRandomChoices({"on","off"});
    
    
    _flattenTopLevelConjunctions = BoolOptionValue("flatten_top_level_conjunctions","",false);
    _flattenTopLevelConjunctions.description=
    "split formulas with top-level (up to universal quantification) conjunctions into several formulas";
    _lookup.insert(&_flattenTopLevelConjunctions);
    _flattenTopLevelConjunctions.tag(OptionTag::PREPROCESSING);
    _flattenTopLevelConjunctions.setRandomChoices({"off","off","off","on"}); 
    _flattenTopLevelConjunctions.setExperimental();
    
    _functionDefinitionElimination = ChoiceOptionValue<FunctionDefinitionElimination>("function_definition_elimination","fde",
                                                                                      FunctionDefinitionElimination::ALL,{"all","none","unused"});
    _functionDefinitionElimination.description=
    "All literals of set-of-support clauses will be selected";
    _lookup.insert(&_functionDefinitionElimination);
    _functionDefinitionElimination.tag(OptionTag::PREPROCESSING);
    _functionDefinitionElimination.addProblemConstraint(hasEquality());
    _functionDefinitionElimination.setRandomChoices({"all","none"});

    _generalSplitting = ChoiceOptionValue<RuleActivity>("general_splitting","gsp",RuleActivity::OFF,{"input_only","off","on"});
    _generalSplitting.description=
    "Preprocessing rule that splits clauses in order to reduce number of different variables in each clause."
    " Only input_only and off are valid values.";
    _lookup.insert(&_generalSplitting);
    _generalSplitting.tag(OptionTag::PREPROCESSING);
    _generalSplitting.addConstraint(notEqual(RuleActivity::ON));
    _generalSplitting.addProblemConstraint(hasNonUnits());
    _generalSplitting.setRandomChoices({"off","input_only"});

    _hornRevealing= BoolOptionValue("horn_revealing","",false);
    _hornRevealing.description=
    "Preprocessing rule that tries to discover whether polarities of predicates can be changed, so that problem becomes horn. If successful, marks all clauses with a positive literal as axioms, and those with only negatives as conjectures.";
    _lookup.insert(&_hornRevealing);
    _hornRevealing.tag(OptionTag::PREPROCESSING);
    _hornRevealing.setExperimental();
    
    _predicateDefinitionInlining = ChoiceOptionValue<InliningMode>("predicate_definition_inlining","",InliningMode::OFF,
                                                                   {"axioms_only","non_growing","off","on"});
    _predicateDefinitionInlining.description=
    "Determines whether predicate definitions should be inlined. Non_growing rules out inlinings that would lead to increase in the size of the problem";
    _lookup.insert(&_predicateDefinitionInlining);
    _predicateDefinitionInlining.tag(OptionTag::PREPROCESSING);
    _predicateDefinitionInlining.setExperimental();
    
    _predicateDefinitionMerging = BoolOptionValue("predicate_definition_merging","",false);
    _predicateDefinitionMerging.description=
    "Determines whether predicates with equivalent definitions will be merged into one. Look for pairs of definitions such as\n"
    "p(X) <=> F[X]\n"
    "q(X) <=> F[X]\n"
    "replace the latter by\n"
    "q(X) <=> p(X)\n"
    "and use it to eliminate the predicate q(X).";
    _lookup.insert(&_predicateDefinitionMerging);
    _predicateDefinitionMerging.tag(OptionTag::PREPROCESSING);
    
    _predicateDefinitionMerging.setExperimental();
    
    _predicateEquivalenceDiscovery = ChoiceOptionValue<PredicateEquivalenceDiscoveryMode>("predicate_equivalence_discovery","",
                                                                                          PredicateEquivalenceDiscoveryMode::OFF,
                                                                                          {"all_atoms","all_formulas","definitions","off","on"});
    _predicateEquivalenceDiscovery.description=
    "If enabled, SAT solver will be used to discover predicate equivalences during preprocessing. "
    "if all_atoms, equivalences between all atoms will be searched for. "
    "if definitions, we'll look only for equivalences in the shape of predicate definitions (this lies somewhere between on and all_atoms). "
    "if all_formulas, equivalences between all formulas are searched for";
    _lookup.insert(&_predicateEquivalenceDiscovery);
    _predicateEquivalenceDiscovery.tag(OptionTag::PREPROCESSING);

    _predicateEquivalenceDiscoveryAddImplications = BoolOptionValue("predicate_equivalence_discovery_add_implications","",false);
    _predicateEquivalenceDiscoveryAddImplications.description=
    "if predicate_equivalence_discovery is enabled, add also discoveder implications, not only equivalences";
    _lookup.insert(&_predicateEquivalenceDiscoveryAddImplications);
    _predicateEquivalenceDiscoveryAddImplications.tag(OptionTag::PREPROCESSING);

    _predicateEquivalenceDiscoveryRandomSimulation = BoolOptionValue("predicate_equivalence_discovery_random_simulation","",true);
    _predicateEquivalenceDiscoveryRandomSimulation.description=
    "use random simulation before the simultaneous sat-sweeping to reduce the amount of candidate equivalences";
    _lookup.insert(&_predicateEquivalenceDiscoveryRandomSimulation);
    _predicateEquivalenceDiscoveryRandomSimulation.tag(OptionTag::PREPROCESSING);

    _predicateEquivalenceDiscoverySatConflictLimit = IntOptionValue("predicate_equivalence_discovery_sat_conflict_limit","",-1);
    _predicateEquivalenceDiscoverySatConflictLimit.description=
    "Limit on the number of SAT conflicts in each equivalence check. Default is -1 which stands for unlimited, 0 will restrict equivalence discovery to unit propagation. The implicative sat sweeping has an internal conflict count limit which always starts with zero and is increased geometrically until it reaches the limit set by this value";
    _lookup.insert(&_predicateEquivalenceDiscoverySatConflictLimit);
    _predicateEquivalenceDiscoverySatConflictLimit.tag(OptionTag::PREPROCESSING);

    _predicateIndexIntroduction = BoolOptionValue("predicate_index_introduction","",false);
    _predicateIndexIntroduction.description=
    "If all atoms of a certain predicate contain distinct constants as a particular argument, atoms of the predicate"
    " are replaces by set of fresh predicates, one for each of the distinct constants.\n"
    "E.g. a problem\n"
    "p(a,b,X,1)\n"
    "p(a,c,a,2)\n"
    "will be transformed into\n"
    "p_a_1(b,X)\n"
    "p_a_2(c,a)\n"
    "(second argument is not removed because constants b and c are not necessarily distinct, and third argment is not replaced because it occurs as a variable)";
    _lookup.insert(&_predicateIndexIntroduction);
    _predicateIndexIntroduction.tag(OptionTag::PREPROCESSING);

    _unusedPredicateDefinitionRemoval = BoolOptionValue("unused_predicate_definition_removal","updr",true);
    _unusedPredicateDefinitionRemoval.description="";
    _lookup.insert(&_unusedPredicateDefinitionRemoval);
    _unusedPredicateDefinitionRemoval.tag(OptionTag::PREPROCESSING);
    _unusedPredicateDefinitionRemoval.addProblemConstraint(notWithCat(Property::UEQ));
    _unusedPredicateDefinitionRemoval.setRandomChoices({"on","off"});

    _trivialPredicateRemoval = BoolOptionValue("trivial_predicate_removal","",false);
    _trivialPredicateRemoval.description= "remove predicates never occurring only positively or only negatively in a clause";
    _lookup.insert(&_trivialPredicateRemoval);
    _trivialPredicateRemoval.tag(OptionTag::PREPROCESSING);

    _theoryAxioms = BoolOptionValue("theory_axioms","tha",true);
    _theoryAxioms.description="Include theory axioms for detected interpreted symbols";
    _lookup.insert(&_theoryAxioms);
    _theoryAxioms.tag(OptionTag::PREPROCESSING);

    _theoryFlattening = BoolOptionValue("theory_flattening","thf",false);
    _theoryFlattening.description = "Flatten clauses to separate theory and non-theory parts";
    _lookup.insert(&_theoryFlattening);
    _theoryFlattening.tag(OptionTag::PREPROCESSING);

    _sineDepth = UnsignedOptionValue("sine_depth","sd",0);
    _sineDepth.description=
    "Limit number of iterations of the transitive closure algorithm that selects formulas based on SInE's D-relation (see SInE description). 0 means no limit, 1 is a maximal limit (least selected axioms), 2 allows two iterations, etc...";
    _lookup.insert(&_sineDepth);
    _sineDepth.tag(OptionTag::PREPROCESSING);
    
    _sineDepth.reliesOn(_sineSelection.is(notEqual(SineSelection::OFF)));
    _sineDepth.setRandomChoices({"0","1","2","3","4","5","7","10"});

    _sineGeneralityThreshold = UnsignedOptionValue("sine_generality_threshold","sgt",0);
    _sineGeneralityThreshold.description=
    "Generality of a symbol is the number of input formulas in which a symbol appears. If the generality of a symbol is smaller than the threshold, it is always added into the D-relation with formulas in which it appears.";
    _lookup.insert(&_sineGeneralityThreshold);
    _sineGeneralityThreshold.tag(OptionTag::PREPROCESSING);
    
    _sineGeneralityThreshold.reliesOn(_sineSelection.is(notEqual(SineSelection::OFF)));

    _sineSelection = ChoiceOptionValue<SineSelection>("sine_selection","ss",SineSelection::OFF,{"axioms","included","off"});
    _sineSelection.description=
    "If 'axioms', all formulas that are not annotated as 'axiom' (i.e. conjectures and hypotheses) are initially selected, and the SInE selection is performed on those annotated as 'axiom'. If 'included', all formulas that are directly in the problem file are initially selected, and the SInE selection is performed on formulas from included files. The 'included' value corresponds to the behaviour of the original SInE implementation.";
    _lookup.insert(&_sineSelection);
    _sineSelection.tag(OptionTag::PREPROCESSING);
    _sineSelection.setRandomChoices(atomsMoreThan(100000),{"axioms","axioms","off"});
    _sineSelection.setRandomChoices(atomsMoreThan(30000),{"axioms","off"});
    _sineSelection.setRandomChoices(atomsMoreThan(10000),{"axioms","off","off","off"});
    _sineSelection.setRandomChoices(atomsMoreThan(3000),{"axioms","off","off","off","off","off"});
    _sineSelection.setRandomChoices(atomsMoreThan(1000),{"axioms","off","off","off","off","off","off","off"});

    _sineTolerance = FloatOptionValue("sine_tolerance","st",1.0);
    _sineTolerance.description="SInE tolerance parameter (sometimes referred to as 'benevolence')";
    _lookup.insert(&_sineTolerance);
    _sineTolerance.tag(OptionTag::PREPROCESSING);
    _sineTolerance.addConstraint(equal(0.0f)->Or(greaterThanEq(1.0f) ));
    
    _sineTolerance.reliesOn(_sineSelection.is(notEqual(SineSelection::OFF)));
    _sineTolerance.setRandomChoices({"1.0","1.2","1.5","2.0","3.0","5.0"});

    _naming = IntOptionValue("naming","nm",8);
    _naming.description="";
    _lookup.insert(&_naming);
    _naming.tag(OptionTag::PREPROCESSING);
    _naming.addConstraint(lessThan(32768));




    
    _logFile = StringOptionValue("log_file","","off");
    _logFile.description="";
    _lookup.insert(&_logFile);
    _logFile.tag(OptionTag::OUTPUT);

    
    _xmlOutput = StringOptionValue("xml_output","","off");
    _xmlOutput.description="";
    _lookup.insert(&_xmlOutput);
    _xmlOutput.tag(OptionTag::OUTPUT);

    _latexOutput = StringOptionValue("latex_output","","off");
    _latexOutput.description="File that will contain proof in the LaTeX format.";
    _lookup.insert(&_latexOutput);
    _latexOutput.tag(OptionTag::OUTPUT);

    _latexUseDefaultSymbols = BoolOptionValue("latex_use_default_symbols","",true);
    _latexUseDefaultSymbols.description="Interpretted symbols such as product have default LaTeX symbols"
        " that can be used. They can be overriden in the normal way. This option can turn them off";
    _latexUseDefaultSymbols.tag(OptionTag::OUTPUT);
    _lookup.insert(&_latexUseDefaultSymbols);

    _outputAxiomNames = BoolOptionValue("output_axiom_names","",false);
    _outputAxiomNames.description="preserve names of axioms from the problem file in the proof output";
    _lookup.insert(&_outputAxiomNames);
    _outputAxiomNames.tag(OptionTag::OUTPUT);


    _printClausifierPremises = BoolOptionValue("print_clausifier_premises","",false);
    _printClausifierPremises.description="";
    _lookup.insert(&_printClausifierPremises);
    _printClausifierPremises.tag(OptionTag::OUTPUT);

    _showActive = BoolOptionValue("show_active","",false);
    _showActive.description="";
    _lookup.insert(&_showActive);
    _showActive.tag(OptionTag::OUTPUT);

    _showBlocked = BoolOptionValue("show_blocked","",false);
    _showBlocked.description="show generating inferences blocked due to coloring of symbols";
    _lookup.insert(&_showBlocked);
    _showBlocked.tag(OptionTag::OUTPUT);

    _showDefinitions = BoolOptionValue("show_definitions","",false);
    _showDefinitions.description="";
    _lookup.insert(&_showDefinitions);
    _showDefinitions.tag(OptionTag::OUTPUT);


    _showNew = BoolOptionValue("show_new","",false);
    _showNew.description="";
    _lookup.insert(&_showNew);
    _showNew.tag(OptionTag::OUTPUT);

    _showNewPropositional = BoolOptionValue("show_new_propositional","",false);
    _showNewPropositional.description="";
    _lookup.insert(&_showNewPropositional);
    _showNewPropositional.tag(OptionTag::OUTPUT);

    _showNonconstantSkolemFunctionTrace = BoolOptionValue("show_nonconstant_skolem_function_trace","",false);
    _showNonconstantSkolemFunctionTrace.description="";
    _lookup.insert(&_showNonconstantSkolemFunctionTrace);
    _showNonconstantSkolemFunctionTrace.tag(OptionTag::OUTPUT);


    _showPassive = BoolOptionValue("show_passive","",false);
    _showPassive.description="";
    _lookup.insert(&_showPassive);
    _showPassive.tag(OptionTag::OUTPUT);
    
    _showReductions = BoolOptionValue("show_reductions","",false);
    _showReductions.description="";
    _showReductions.tag(OptionTag::OUTPUT);
    _lookup.insert(&_showReductions);

    _showPreprocessing = BoolOptionValue("show_preprocessing","",false);
    _showPreprocessing.description="";
    _lookup.insert(&_showPreprocessing);
    _showPreprocessing.tag(OptionTag::OUTPUT);

    _showSkolemisations = BoolOptionValue("show_skolemisations","",false);
    _showSkolemisations.description="";
    _lookup.insert(&_showSkolemisations);
    _showSkolemisations.tag(OptionTag::OUTPUT);

    _showSymbolElimination = BoolOptionValue("show_symbol_elimination","",false);
    _showSymbolElimination.description="";
    _lookup.insert(&_showSymbolElimination);
    _showSymbolElimination.tag(OptionTag::OUTPUT);

    _showTheoryAxioms = BoolOptionValue("show_theory_axioms","",false);
    _showTheoryAxioms.description="";
    _lookup.insert(&_showTheoryAxioms);
    _showTheoryAxioms.tag(OptionTag::OUTPUT);

#if VZ3
    _showZ3 = BoolOptionValue("show_z3","",false);
    _showZ3.description="Print the clauses being added to Z3";
    _lookup.insert(&_showZ3);
    _showZ3.tag(OptionTag::OUTPUT);
#endif







    _saturationAlgorithm = ChoiceOptionValue<SaturationAlgorithm>("saturation_algorithm","sa",SaturationAlgorithm::LRS,
                                                                  {"discount","fmb","inst_gen","lrs","otter","tabulation"});
    _saturationAlgorithm.description=
    "Select the saturation algorithm:\n"
    " - discount:\n"
    " - otter:\n"
    " - limited resource:\n"
    " - instance generation: a simple implementation of instantiation calculus\n"
    "    (global_subsumption, unit_resulting_resolution and age_weight_ratio)\n"
    " - tabulation: a special goal-oriented mode for large theories.\n"
    " - fmb : finite model building for satisfiable problems.\n"
    "inst_gen, tabulation and fmb aren't influenced by options for the saturation algorithm, apart from those under the relevant heading";
    _lookup.insert(&_saturationAlgorithm);
    _saturationAlgorithm.tag(OptionTag::SATURATION);
    
    _saturationAlgorithm.addHardConstraint(If(equal(SaturationAlgorithm::INST_GEN)).then(_splitting.is(notEqual(true))));
    
    _saturationAlgorithm.setRandomChoices(isRandSat(),{"discount","otter","inst_gen","fmb"});
    _saturationAlgorithm.setRandomChoices(Or(hasCat(Property::UEQ),atomsLessThan(4000)),{"lrs","discount","otter","inst_gen"});
    _saturationAlgorithm.setRandomChoices({"discount","inst_gen","lrs","otter","tabulation"});

    _fmbStartWithConstants = BoolOptionValue("fmb_start_with_constants","fmbswc",false);
    _fmbStartWithConstants.description = "Start fmb with model size equal to the number of constants in the problem";
    _lookup.insert(&_fmbStartWithConstants);
    _fmbStartWithConstants.setExperimental();
    _fmbStartWithConstants.setRandomChoices({"on","off"});

    _fmbIncremental = BoolOptionValue("fmb_incremental","fmbi",false);
    _fmbIncremental.description = "Use incremental SAT in fmb";
    
    _fmbIncremental.setExperimental();
    _fmbIncremental.setRandomChoices({"on","off"});

    _fmbNonGroundDefs = BoolOptionValue("fmb_nonground_defs","fmbngd",false);
    _fmbNonGroundDefs.description = "Introduce definitions for non ground terms in preprocessing for fmb";
    
    _fmbNonGroundDefs.setExperimental();
    _fmbNonGroundDefs.setRandomChoices({"on","off"});

    _fmbSortInference = BoolOptionValue("fmb_sort_inference","fmbsi",false);
    _fmbSortInference.description = "Perform sort inference for fmb";
    
    _fmbSortInference.setExperimental();
    _fmbSortInference.setRandomChoices({"on","off"});

    _fmbStartSize = UnsignedOptionValue("fmb_start_size","fmbss",1);
    _fmbStartSize.description = "Set the initial model size for finite model building";
    _lookup.insert(&_fmbStartSize);

    _fmbSymmetryRatio = FloatOptionValue("fmb_symmetry_ratio","fmbsr",1.0);
    _fmbSymmetryRatio.description = "";
    _lookup.insert(&_fmbSymmetryRatio);
    _fmbSymmetryRatio.setExperimental();

    _selection = SelectionOptionValue("selection","s",10);
    _selection.description=
    "Selection methods 2,3,4,10,11 are complete by virtue of extending Maximal i.e. they select the best among maximal. Methods 1002,1003,1004,1010,1011 relax this restriction and are therefore not complete.\n"
    " 0     - Total (select everything)\n"
    " 1     - Maximal\n"
    " 2     - ColoredFirst, MaximalSize then Lexigraphical\n"
    " 3     - ColoredFirst, NoPositiveEquality, LeastTopLevelVariables,\n          LeastDistinctVariables then Lexigraphical\n"
    " 4     - ColoredFirst, NoPositiveEquality, LeastTopLevelVariables,\n          LeastVariables, MaximalSize then Lexigraphical\n"
    " 10    - ColoredFirst, NegativeEquality, MaximalSize, Negative then Lexigraphical\n"
    " 11    - Lookahead\n"
    " 1002  - Incomplete version of 2\n"
    " 1003  - Incomplete version of 3\n"
    " 1004  - Incomplete version of 4\n"
    " 1010  - Incomplete version of 10\n"
    " 1011  - Incomplete version of 11\n"
    "Or negated, which means that reversePolarity is true (?)\n";

    _lookup.insert(&_selection);
    _selection.tag(OptionTag::SATURATION);
    _selection.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _selection.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<int>(_instGenWithResolution.is(equal(true))));
    _selection.setRandomChoices(And(isRandSat(),saNotInstGen()),{"0","1","2","3","4","10","11","-1","-2","-3","-4","-10","-11"});
    _selection.setRandomChoices({"0","1","2","3","4","10","11","1002","1003","1004","1010","1011","-1","-2","-3","-4","-10","-11","-1002","-1003","-1004","-1010","-1011"});
    
    _ageWeightRatio = RatioOptionValue("age_weight_ratio","awr",1,1,':');
    _ageWeightRatio.description=
    "Ratio in which clauses are being selected for activation i.e. a:w means that for every a clauses selected based on age"
    "there will be w selected based on weight.";
    _lookup.insert(&_ageWeightRatio);
    _ageWeightRatio.tag(OptionTag::SATURATION);
    _ageWeightRatio.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _ageWeightRatio.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<int>(_instGenWithResolution.is(equal(true))));
    _ageWeightRatio.setRandomChoices({"8:1","5:1","4:1","3:1","2:1","3:2","5:4","1","2:3","2","3","4","5","6","7","8","10","12","14","16","20","24","28","32","40","50","64","128","1024"});

    _lrsFirstTimeCheck = IntOptionValue("lrs_first_time_check","",5);
    _lrsFirstTimeCheck.description=
    "Percentage of time limit at which the LRS algorithm will for the first time estimate the number of reachable clauses.";
    _lookup.insert(&_lrsFirstTimeCheck);
    _lrsFirstTimeCheck.tag(OptionTag::SATURATION);
    _lrsFirstTimeCheck.addConstraint(greaterThanEq(0));
    _lrsFirstTimeCheck.addConstraint(lessThan(100));

    _lrsWeightLimitOnly = BoolOptionValue("lrs_weight_limit_only","lwlo",false);
    _lrsWeightLimitOnly.description=
    "If off, the lrs sets both age and weight limit according to clause reachability, otherwise it sets the age limit to 0 and only the weight limit reflects reachable clauses";
    _lookup.insert(&_lrsWeightLimitOnly);
    _lrsWeightLimitOnly.tag(OptionTag::SATURATION);

    _simulatedTimeLimit = TimeLimitOptionValue("simulated_time_limit","stl",0);
    _simulatedTimeLimit.description=
    "Time limit in seconds for the purpose of reachability estimations of the LRS saturation algorithm (if 0, the actual time limit is used)";
    _lookup.insert(&_simulatedTimeLimit);
    _simulatedTimeLimit.tag(OptionTag::SATURATION);




    _instantiation = ChoiceOptionValue<Instantiation>("instantiation","inst",Instantiation::OFF,{"off","on"});
    _instantiation.description = "Heuristically instantiate variables";
    _instantiation.tag(OptionTag::INFERENCES);
    _lookup.insert(&_instantiation);
    _instantiation.setRandomChoices({"off","on"}); 

    _backwardDemodulation = ChoiceOptionValue<Demodulation>("backward_demodulation","bd",
                                                            Demodulation::ALL,
                                                            {"all","off","preordered"});
    _backwardDemodulation.description=
             "Oriented rewriting of kept clauses by newly derived unit equalities\n"
             "s = t     L[sθ] \\/ C\n"
             "---------------------   where sθ > tθ (replaces RHS)\n"
             " L[tθ] \\/ C\n";
    _lookup.insert(&_backwardDemodulation);
    _backwardDemodulation.tag(OptionTag::INFERENCES);
    _backwardDemodulation.addProblemConstraint(hasEquality());
    _backwardDemodulation.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<Demodulation>(_instGenWithResolution.is(equal(true))));
    _backwardDemodulation.setRandomChoices({"all","off"});

    _backwardSubsumption = ChoiceOptionValue<Subsumption>("backward_subsumption","bs",
                                                          Subsumption::OFF,{"off","on","unit_only"});
    _backwardSubsumption.description=
             "unit_only means that the subsumption will be performed only by unit clauses";
    _lookup.insert(&_backwardSubsumption);
    _backwardSubsumption.tag(OptionTag::INFERENCES);
    _backwardSubsumption.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<Subsumption>(_instGenWithResolution.is(equal(true))));
    _backwardSubsumption.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _backwardSubsumption.setRandomChoices({"on","off"});

    _backwardSubsumptionResolution = ChoiceOptionValue<Subsumption>("backward_subsumption_resolution","bsr",
                                                                    Subsumption::OFF,{"off","on","unit_only"});
    _backwardSubsumptionResolution.description=
             "unit_only means that the subsumption resolution will be performed only by unit clauses";
    _lookup.insert(&_backwardSubsumptionResolution);
    _backwardSubsumptionResolution.tag(OptionTag::INFERENCES);
    _backwardSubsumptionResolution.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<Subsumption>(_instGenWithResolution.is(equal(true))));
    _backwardSubsumptionResolution.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _backwardSubsumptionResolution.setRandomChoices({"on","off"});

    _binaryResolution = BoolOptionValue("binary_resolution","br",true);
    _binaryResolution.description=
          "Standard binary resolution i.e.\n"
              "C \\/ t     D \\/ s\n"
              "---------------------\n"
              "(C \\/ D)θ\n"
              "where θ = mgu(t,-s) and t selected";
    _lookup.insert(&_binaryResolution);
    _binaryResolution.tag(OptionTag::INFERENCES);
    
    _binaryResolution.addConstraint(
      If(equal(false)).then(_unitResultingResolution.is(notEqual(URResolution::OFF))));
    _binaryResolution.setRandomChoices(And(isRandSat(),saNotInstGen(),Or(hasEquality(),isBfnt(),hasCat(Property::HNE))),{"on"});
    _binaryResolution.setRandomChoices({"on","off"});


    _condensation = ChoiceOptionValue<Condensation>("condensation","cond",Condensation::OFF,{"fast","off","on"});
    _condensation.description=
             "If 'fast' is specified, we only perform condensations that are easy to check for.";
    _lookup.insert(&_condensation);
    _condensation.tag(OptionTag::INFERENCES);
    _condensation.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<Condensation>(_instGenWithResolution.is(equal(true))));
    _condensation.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _condensation.setRandomChoices({"on","off","fast"});

    _demodulationRedundancyCheck = BoolOptionValue("demodulation_redundancy_check","drc",true);
    _demodulationRedundancyCheck.description=
             "Avoids the following cases of backward and forward demodulation, as they do not preserve completeness:\n"
             "s = t     s = t1 \\/ C \t s = t     s != t1 \\/ C\n"

             "--------------------- \t ---------------------\n"
             "t = t1 \\/ C \t\t t != t1 \\/ C\n"
             "where t > t1 and s = t > C (RHS replaced)";
    _lookup.insert(&_demodulationRedundancyCheck);
    _demodulationRedundancyCheck.tag(OptionTag::INFERENCES);
    _demodulationRedundancyCheck.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<bool>(_instGenWithResolution.is(equal(true))));
    _demodulationRedundancyCheck.addProblemConstraint(hasEquality());
    _demodulationRedundancyCheck.setRandomChoices({"on","off"});

    _equalityResolutionWithDeletion = ChoiceOptionValue<RuleActivity>( "equality_resolution_with_deletion","erd",
                                                                      RuleActivity::INPUT_ONLY,{"input_only","off","on"});
    _equalityResolutionWithDeletion.description="";
    _lookup.insert(&_equalityResolutionWithDeletion);
    _equalityResolutionWithDeletion.tag(OptionTag::INFERENCES);
    _equalityResolutionWithDeletion.addConstraint(notEqual(RuleActivity::ON));
    
    
    _equalityResolutionWithDeletion.setRandomChoices({"input_only","off"});
    
    _extensionalityAllowPosEq = BoolOptionValue( "extensionality_allow_pos_eq","",false);
    _extensionalityAllowPosEq.description="If extensionality resolution equals filter, this dictates"
      " whether we allow other positive equalities when recognising extensionality clauses";
    _lookup.insert(&_extensionalityAllowPosEq);
    _extensionalityAllowPosEq.tag(OptionTag::INFERENCES);
    _extensionalityAllowPosEq.reliesOn(_extensionalityResolution.is(equal(ExtensionalityResolution::FILTER)));
    _extensionalityAllowPosEq.setRandomChoices({"on","off","off"}); 
    
    _extensionalityMaxLength = UnsignedOptionValue("extensionality_max_length","",0);
    _extensionalityMaxLength.description="Sets the maximum length (number of literals) an extensionality"
      " clause can have when doing recognition for extensionality resolution. If zero there is no maximum.";
    _lookup.insert(&_extensionalityMaxLength);
    _extensionalityMaxLength.tag(OptionTag::INFERENCES);
    
    _extensionalityMaxLength.addConstraint(notEqual(1u));
    _extensionalityMaxLength.reliesOn(_extensionalityResolution.is(notEqual(ExtensionalityResolution::OFF)));
    
    _extensionalityMaxLength.setRandomChoices({"0","0","0","2","3"}); 
    
    _extensionalityResolution = ChoiceOptionValue<ExtensionalityResolution>("extensionality_resolution","er",
                                                                            ExtensionalityResolution::OFF,{"filter","known","off"});
    _extensionalityResolution.description=
      "Turns on the following inference rule:\n"
      "  x=y \\/ C    s != t \\/ D\n"
      "  -----------------------\n"
      "  C{x → s, y → t} \\/ D\n"
      "Where s!=t is selected in s!=t \\/D and x=y \\/ C is a recognised as an extensionality clause - how clauses are recognised depends on the value of this option.\n"
      "If filter we attempt to recognise all extensionality clauses i.e. those that have exactly one X=Y, no inequality of the same sort as X-Y (and optionally no equality except X=Y, see extensionality_allow_pos_eq).\n" 
      "If known we only recognise a known set of extensionality clauses. At the moment this includes the standard and subset-based formulations of the set extensionality axiom, as well as the array extensionality axiom.";
    _lookup.insert(&_extensionalityResolution);
    _extensionalityResolution.tag(OptionTag::INFERENCES);
    
    _extensionalityResolution.reliesOn(_inequalitySplitting.is(equal(0)));
    _extensionalityResolution.setRandomChoices({"filter","known","off","off"});
    
    _forwardDemodulation = ChoiceOptionValue<Demodulation>("forward_demodulation","fd",Demodulation::ALL,{"all","off","preordered"});
    _forwardDemodulation.description=
    "Oriented rewriting of newly derived clauses by kept unit equalities\n"
    "s = t     L[sθ] \\/ C\n"
    "---------------------  where sθ > tθ\n"
    " L[tθ] \\/ C\n"
    "If 'preordered' is set, only equalities s = t where s > t are used for rewriting.";
    _lookup.insert(&_forwardDemodulation);
    _forwardDemodulation.tag(OptionTag::INFERENCES);
    _forwardDemodulation.setRandomChoices({"all","all","all","off","preordered"});
    
    _forwardLiteralRewriting = BoolOptionValue("forward_literal_rewriting","flr",false);
    _forwardLiteralRewriting.description="";
    _lookup.insert(&_forwardLiteralRewriting);
    _forwardLiteralRewriting.tag(OptionTag::INFERENCES);
    _forwardLiteralRewriting.addProblemConstraint(hasNonUnits());
    _forwardLiteralRewriting.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<bool>(_instGenWithResolution.is(equal(true))));
    _forwardLiteralRewriting.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _forwardLiteralRewriting.setRandomChoices({"on","off"});

    _forwardSubsumption = BoolOptionValue("forward_subsumption","fs",true);
    _forwardSubsumption.description="";
    _lookup.insert(&_forwardSubsumption);
    _forwardSubsumption.tag(OptionTag::INFERENCES);
    _forwardSubsumption.setRandomChoices({"on","on","on","on","on","on","on","on","on","off"}); 

    _forwardSubsumptionResolution = BoolOptionValue("forward_subsumption_resolution","fsr",true);
    _forwardSubsumptionResolution.description="";
    _lookup.insert(&_forwardSubsumptionResolution);
    _forwardSubsumptionResolution.tag(OptionTag::INFERENCES);
    _forwardSubsumptionResolution    .reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<bool>(_instGenWithResolution.is(equal(true))));
    _forwardSubsumptionResolution.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _forwardSubsumptionResolution.setRandomChoices({"on","off"});

    _hyperSuperposition = BoolOptionValue("hyper_superposition","",false);
    _hyperSuperposition.description=
    "Generating inference that attempts to do several rewritings at once if it will eliminate literals of the original clause (now we aim just for elimination by equality resolution)";
    _lookup.insert(&_hyperSuperposition);
    _hyperSuperposition.tag(OptionTag::INFERENCES);

    _unitResultingResolution = ChoiceOptionValue<URResolution>("unit_resulting_resolution","urr",URResolution::OFF,{"ec_only","off","on"});
    _unitResultingResolution.description=
    "uses unit resulting resolution only to derive empty clauses (may be useful for splitting)";
    _lookup.insert(&_unitResultingResolution);
    _unitResultingResolution.tag(OptionTag::INFERENCES);
    _unitResultingResolution.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    
    
    
    
    _unitResultingResolution.addProblemConstraint(hasPredicates());
    
    
    _unitResultingResolution.setRandomChoices(And(isRandSat(),saNotInstGen(),Or(hasEquality(),isBfnt(),hasCat(Property::HNE))),{"on","off"});
    _unitResultingResolution.setRandomChoices(isRandSat(),{});
    _unitResultingResolution.setRandomChoices({"on","on","off"});


    _superpositionFromVariables = BoolOptionValue("superposition_from_variables","sfv",true);
    _superpositionFromVariables.description="";
    _lookup.insert(&_superpositionFromVariables);
    _superpositionFromVariables.tag(OptionTag::INFERENCES);
    _superpositionFromVariables.addProblemConstraint(hasEquality());
    _superpositionFromVariables.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN))->Or<bool>(_instGenWithResolution.is(equal(true))));
    _superpositionFromVariables.setRandomChoices({"on","off"});



    _globalSubsumption = BoolOptionValue("global_subsumption","gs",false);
    _globalSubsumption.description="Global subsumption reduction.";
    _lookup.insert(&_globalSubsumption);
    _globalSubsumption.tag(OptionTag::INST_GEN);
    _globalSubsumption.addProblemConstraint(hasNonUnits());
    _globalSubsumption.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _globalSubsumption.setRandomChoices({"off","on"});

    _globalSubsumptionSatSolverPower = ChoiceOptionValue<GlobalSubsumptionSatSolverPower>("global_subsumption_sat_solver_power","gsssp",
          GlobalSubsumptionSatSolverPower::PROPAGATION_ONLY,{"propagation_only","full"});
    _globalSubsumptionSatSolverPower.description="";
    _lookup.insert(&_globalSubsumptionSatSolverPower);
    _globalSubsumptionSatSolverPower.tag(OptionTag::INST_GEN);
    _globalSubsumptionSatSolverPower.reliesOn(_globalSubsumption.is(equal(true)));
    _globalSubsumptionSatSolverPower.setRandomChoices({"propagation_only","full"});

    _globalSubsumptionExplicitMinim = ChoiceOptionValue<GlobalSubsumptionExplicitMinim>("global_subsumption_explicit_minim","gsem",
        GlobalSubsumptionExplicitMinim::RANDOMIZED,{"off","on","randomized"});
    _globalSubsumptionSatSolverPower.description="Explicitly minimize the result of global sumsumption reduction.";
    _lookup.insert(&_globalSubsumptionExplicitMinim);
    _globalSubsumptionExplicitMinim.tag(OptionTag::INST_GEN);
    _globalSubsumptionExplicitMinim.reliesOn(_globalSubsumption.is(equal(true)));
    _globalSubsumptionExplicitMinim.setRandomChoices({"off","on","randomized"});

    _globalSubsumptionAvatarAssumptions = ChoiceOptionValue<GlobalSubsumptionAvatarAssumptions>("global_subsumption_avatar_assumptions","gsaa",
        GlobalSubsumptionAvatarAssumptions::OFF,{"off","from_current","full_model"});
    _globalSubsumptionAvatarAssumptions.description="";
    _lookup.insert(&_globalSubsumptionAvatarAssumptions);
    _globalSubsumptionAvatarAssumptions.tag(OptionTag::INST_GEN);
    _globalSubsumptionAvatarAssumptions.reliesOn(_globalSubsumption.is(equal(true)));
    _globalSubsumptionAvatarAssumptions.reliesOn(_splitting.is(equal(true)));
    _globalSubsumptionAvatarAssumptions.setRandomChoices({"off","from_current","full_model"});

    _instGenBigRestartRatio = FloatOptionValue("inst_gen_big_restart_ratio","igbrr",0.0);
    _instGenBigRestartRatio.description=
    "determines how often a big restart (instance generation starts from input clauses) will be performed. Small restart means all clauses generated so far are processed again.";
    _lookup.insert(&_instGenBigRestartRatio);
    _instGenBigRestartRatio.tag(OptionTag::INST_GEN);
    _instGenBigRestartRatio.addConstraint(greaterThanEq(0.0f)->And(lessThanEq(1.0f)));
    
    _instGenBigRestartRatio.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _instGenBigRestartRatio.setRandomChoices({"0.0","0.1","0.2","0.3","0.4","0.5","0.6","0.7","0.8","0.9","1.0"});

    _instGenInprocessing = BoolOptionValue("inst_gen_inprocessing","igi",false);
    _instGenInprocessing.description="";
    _lookup.insert(&_instGenInprocessing);
    _instGenInprocessing.tag(OptionTag::INST_GEN);
    _instGenInprocessing.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));

    _instGenPassiveReactivation = BoolOptionValue("inst_gen_passive_reactivation","igpr",false);
    _instGenPassiveReactivation.description="";
    _lookup.insert(&_instGenPassiveReactivation);
    _instGenPassiveReactivation.tag(OptionTag::INST_GEN);
    _instGenPassiveReactivation.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));

    _instGenResolutionInstGenRatio = RatioOptionValue("inst_gen_resolution_ratio","igrr",1,1,'/');
    _instGenResolutionInstGenRatio.description=
    "ratio of resolution and instantiation steps (applies only if inst_gen_with_resolution is on)";
    _lookup.insert(&_instGenResolutionInstGenRatio);
    _instGenResolutionInstGenRatio.tag(OptionTag::INST_GEN);
    _instGenResolutionInstGenRatio.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _instGenResolutionInstGenRatio.reliesOn(_instGenWithResolution.is(equal(true)));
    _instGenResolutionInstGenRatio.setRandomChoices({"128/1","64/1","32/1","16/1","8/1","4/1","2/1","1/1","1/2","1/4","1/8","1/16","1/32","1/64","1/128"});

    _instGenRestartPeriod = IntOptionValue("inst_gen_restart_period","igrp",1000);
    _instGenRestartPeriod.description="how many clauses are processed before (small?) restart";
    _lookup.insert(&_instGenRestartPeriod);
    _instGenRestartPeriod.tag(OptionTag::INST_GEN);
    _instGenRestartPeriod.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _instGenRestartPeriod.setRandomChoices({"100","200","400","700","1000","1400","2000","4000"});

    _instGenRestartPeriodQuotient = FloatOptionValue("inst_gen_restart_period_quotient","igrpq",1.0);
    _instGenRestartPeriodQuotient.description="restart period is multiplied by this number after each restart";
    _lookup.insert(&_instGenRestartPeriodQuotient);
    _instGenRestartPeriodQuotient.tag(OptionTag::INST_GEN);
    _instGenRestartPeriodQuotient.addConstraint(greaterThanEq(1.0f));
    _instGenRestartPeriodQuotient.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _instGenRestartPeriodQuotient.setRandomChoices({"1.0","1.05","1.1","1.2","1.3","1.5","2.0"});

    _instGenSelection = SelectionOptionValue("inst_gen_selection","igs",0);
    _instGenSelection.description=
    "selection function for InstGen. we don't have the functions 11 and 1011 yet (as it would need special treatment for the look-ahead)";
    _lookup.insert(&_instGenSelection);
    _instGenSelection.tag(OptionTag::INST_GEN);
    _instGenSelection.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));

    _instGenWithResolution = BoolOptionValue("inst_gen_with_resolution","igwr",false);
    _instGenWithResolution.description=
    "performs instantiation together with resolution (global subsumption index is shared, also clauses generated by the resolution are added to the instance SAT problem)";
    _lookup.insert(&_instGenWithResolution);
    _instGenWithResolution.tag(OptionTag::INST_GEN);
    _instGenWithResolution.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _instGenWithResolution.setRandomChoices({"on","off"});

    _instGenUseHashingVariantIndex = BoolOptionValue("inst_gen_use_hashing_variant_index","iguhvi",false);
    _instGenUseHashingVariantIndex.description= "Use clause variant index based on hashing for clause variant detection in inst_gen";
    _lookup.insert(&_instGenUseHashingVariantIndex);
    _instGenUseHashingVariantIndex.tag(OptionTag::INST_GEN);
    _instGenUseHashingVariantIndex.setExperimental();
    _instGenUseHashingVariantIndex.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _instGenUseHashingVariantIndex.setRandomChoices({"on","off"});

    _use_dm = BoolOptionValue("use_dismatching","dm",false);
    _use_dm.description="";
    _lookup.insert(&_use_dm);
    _use_dm.tag(OptionTag::INST_GEN);
    _use_dm.setExperimental();
    _use_dm.setRandomChoices({"on","off"});
    _use_dm.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));

    _nicenessOption = ChoiceOptionValue<Niceness>("niceness_option","none",Niceness::NONE,{"average","none","sum","top"});
    _nicenessOption.description="";
    _lookup.insert(&_nicenessOption);
    _nicenessOption.tag(OptionTag::INST_GEN);
    _nicenessOption.setExperimental();
    _nicenessOption.reliesOn(_saturationAlgorithm.is(equal(SaturationAlgorithm::INST_GEN)));
    _nicenessOption.reliesOn(_satSolver.is(equal(SatSolver::VAMPIRE)));
    _nicenessOption.setRandomChoices({"none","none","none","none","none","average","sum","top"});



    _splitting = BoolOptionValue("splitting","spl",true);
    _splitting.description="";
    _lookup.insert(&_splitting);
    _splitting.tag(OptionTag::AVATAR);
    
    _splitting.addConstraint(If(equal(true)).then(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION))));
    
    _splitting.setRandomChoices({"on","off"}); 
    _splitAtActivation = BoolOptionValue("split_at_activation","sac",false);
    _splitAtActivation.description="Split a clause when it is activated, default is to split when it is processed";
    _lookup.insert(&_splitAtActivation);
    _splitAtActivation.reliesOn(_splitting.is(equal(true)));
    _splitAtActivation.tag(OptionTag::AVATAR);
    _splitAtActivation.setRandomChoices({"on","off"});

    _splittingAddComplementary = ChoiceOptionValue<SplittingAddComplementary>("splitting_add_complementary","ssac",
                                                                                SplittingAddComplementary::GROUND,{"ground","none"});
    _splittingAddComplementary.description="";
    _lookup.insert(&_splittingAddComplementary);
    _splittingAddComplementary.tag(OptionTag::AVATAR);
    _splittingAddComplementary.setExperimental();
    _splittingAddComplementary.reliesOn(_splitting.is(equal(true)));
    _splittingAddComplementary.setRandomChoices({"ground","none"});


    _splittingCongruenceClosure = ChoiceOptionValue<SplittingCongruenceClosure>("splitting_congruence_closure","sscc",
                                                                                SplittingCongruenceClosure::OFF,{"model","off","on"});
    _splittingCongruenceClosure.description="";
    _lookup.insert(&_splittingCongruenceClosure);
    _splittingCongruenceClosure.tag(OptionTag::AVATAR);
    _splittingCongruenceClosure.reliesOn(_splitting.is(equal(true)));
    _splittingCongruenceClosure.addProblemConstraint(hasEquality());
    _splittingCongruenceClosure.setRandomChoices({"model","off","on"});
    _splittingCongruenceClosure.addHardConstraint(If(equal(SplittingCongruenceClosure::MODEL)).
                                                  then(_splittingMinimizeModel.is(notEqual(SplittingMinimizeModel::SCO))));
    _splittingCongruenceClosure.addHardConstraint(If(equal(SplittingCongruenceClosure::MODEL)).
                                                      then(_splittingHandleZeroImplied.is(notEqual(true))));
    
    _ccUnsatCores = ChoiceOptionValue<CCUnsatCores>("cc_unsat_cores","ccuc",CCUnsatCores::ALL,
                                                     {"first", "small_ones", "all"});
    _ccUnsatCores.description="";
    _lookup.insert(&_ccUnsatCores);
    _ccUnsatCores.tag(OptionTag::AVATAR);
    _ccUnsatCores.reliesOn(_splittingCongruenceClosure.is(notEqual(SplittingCongruenceClosure::OFF)));
    _ccUnsatCores.setRandomChoices({"first", "small_ones", "all"});

    _splittingLiteralPolarityAdvice = ChoiceOptionValue<SplittingLiteralPolarityAdvice>(
                                                "splitting_literal_polarity_advice","slpa",
                                                SplittingLiteralPolarityAdvice::NONE,
                                                {"false","true","none"});
    _splittingLiteralPolarityAdvice.description="Override SAT-solver's default polarity/phase setting for variables abstracting clause components.";
    _lookup.insert(&_splittingLiteralPolarityAdvice);
    _splittingLiteralPolarityAdvice.tag(OptionTag::AVATAR);
    _splittingLiteralPolarityAdvice.reliesOn(_splitting.is(equal(true)));

    _splittingMinimizeModel = ChoiceOptionValue<SplittingMinimizeModel>("splitting_minimize_model","smm",
                                                                        SplittingMinimizeModel::ALL,{"off","sco","all"});
    
    _splittingMinimizeModel.description="Minimize the SAT-solver model by replacing concrete values with don't-cares"
                                        " provided <all> the sat clauses (or only the split clauses with <sco>) remain provably satisfied"
                                        " by the partial model.";
    _lookup.insert(&_splittingMinimizeModel);
    _splittingMinimizeModel.tag(OptionTag::AVATAR);
    _splittingMinimizeModel.setExperimental();
    _splittingMinimizeModel.reliesOn(_splitting.is(equal(true)));
    _splittingMinimizeModel.setRandomChoices({"off","sco","all"});

    _splittingEagerRemoval = BoolOptionValue("splitting_eager_removal","sser",true);
    _splittingEagerRemoval.description="";
    _lookup.insert(&_splittingEagerRemoval);
    _splittingEagerRemoval.tag(OptionTag::AVATAR);
    _splittingEagerRemoval.setExperimental();
    _splittingEagerRemoval.reliesOn(_splitting.is(equal(true)));
    
    
    _splittingEagerRemoval.reliesOn(_splittingMinimizeModel.is(equal(SplittingMinimizeModel::ALL)));
    _splittingEagerRemoval.setRandomChoices({"on","off"});

    _splittingHandleZeroImplied = BoolOptionValue("splitting_handle_zero_implied","shzi",false);
    _splittingHandleZeroImplied.description="If on ensure that splitting decisions implied by the top level"
         " of the SAT solver are considered unconditional when performing clause reductions.";
    _lookup.insert(&_splittingHandleZeroImplied);
    _splittingHandleZeroImplied.tag(OptionTag::AVATAR);
    _splittingHandleZeroImplied.setExperimental();
    _splittingHandleZeroImplied.reliesOn(_splitting.is(equal(true)));
    _splittingHandleZeroImplied.addHardConstraint(ifOnThen(_nonliteralsInClauseWeight.is(equal(false))));
    _splittingHandleZeroImplied.setRandomChoices({"on","off"});

    _splittingFastRestart = BoolOptionValue("splitting_fast_restart","sfr",false);
    _splittingFastRestart.description="";
    _lookup.insert(&_splittingFastRestart);
    _splittingFastRestart.tag(OptionTag::AVATAR);
    _splittingFastRestart.setExperimental();
    _splittingFastRestart.reliesOn(_splitting.is(equal(true)));
    _splittingFastRestart.setRandomChoices({"on","off"});

    _splittingBufferedSolver = BoolOptionValue("splitting_buffered_solver","sbs",false);
    _splittingBufferedSolver.description="Added buffering funcitonality to the SAT solver used in AVATAR.";
    _lookup.insert(&_splittingBufferedSolver);
    _splittingBufferedSolver.tag(OptionTag::AVATAR);
    _splittingBufferedSolver.setExperimental();
    _splittingBufferedSolver.reliesOn(_splitting.is(equal(true)));
    _splittingBufferedSolver.setRandomChoices({"on","off"});

    _splittingDeleteDeactivated = ChoiceOptionValue<SplittingDeleteDeactivated>("splitting_delete_deactivated","sdd",
                                                                        SplittingDeleteDeactivated::ON,{"on","large","off"});

    _splittingDeleteDeactivated.description="";
    _lookup.insert(&_splittingDeleteDeactivated);
    _splittingDeleteDeactivated.tag(OptionTag::AVATAR);
    _splittingDeleteDeactivated.setExperimental();
    _splittingDeleteDeactivated.reliesOn(_splitting.is(equal(true)));
    _splittingDeleteDeactivated.setRandomChoices({"on","large","off"});


    _splittingFlushPeriod = UnsignedOptionValue("splitting_flush_period","ssfp",0);
    _splittingFlushPeriod.description=
    "after given number of generated clauses without deriving an empty clause, the splitting component selection is shuffled. If equal to zero, shuffling is never performed.";
    _lookup.insert(&_splittingFlushPeriod);
    _splittingFlushPeriod.tag(OptionTag::AVATAR);
    _splittingFlushPeriod.setExperimental();
    _splittingFlushPeriod.reliesOn(_splitting.is(equal(true)));
    _splittingFlushPeriod.setRandomChoices({"0","1000","4000","10000","40000","100000"});

    _splittingFlushQuotient = FloatOptionValue("splitting_flush_quotient","ssfq",1.5);
    _splittingFlushQuotient.description=
    "after each flush, the splitting_flush_period is multiplied by the quotient";
    _lookup.insert(&_splittingFlushQuotient);
    _splittingFlushQuotient.tag(OptionTag::AVATAR);
    _splittingFlushQuotient.setExperimental();
    _splittingFlushQuotient.addConstraint(greaterThanEq(1.0f));
    _splittingFlushQuotient.reliesOn(_splitting.is(equal(true)));
    _splittingFlushQuotient.setRandomChoices({"1.0","1.1","1.2","1.4","2.0"});

    _splittingNonsplittableComponents = ChoiceOptionValue<SplittingNonsplittableComponents>("splitting_nonsplittable_components","ssnc",
                                                                                              SplittingNonsplittableComponents::KNOWN,
                                                                                              {"all","all_dependent","known","none"});
    _splittingNonsplittableComponents.description=
    "known .. SAT clauses will be learnt from non-splittable clauses that have corresponding components (if there is a component C with name SAT l, clause C | {l1,..ln} will give SAT clause ~l1 \\/ … \\/ ~ln \\/ l). When we add the sat clause, we discard the original FO clause C | {l1,..ln} and let the component selection update model, possibly adding the component clause C | {l}. all .. like known, except when we see a non-splittable clause that doesn't have a name, we introduce the name for it. all_dependent .. like all, but we don't introduce names for non-splittable clauses that don't depend on any components";
    _lookup.insert(&_splittingNonsplittableComponents);
    _splittingNonsplittableComponents.tag(OptionTag::AVATAR);
    _splittingNonsplittableComponents.setExperimental();
    _splittingNonsplittableComponents.reliesOn(_splitting.is(equal(true)));
    _splittingNonsplittableComponents.setRandomChoices({"all","all_dependent","known","none"});


    _nonliteralsInClauseWeight = BoolOptionValue("nonliterals_in_clause_weight","nicw",false);
    _nonliteralsInClauseWeight.description=
    "Non-literal parts of clauses (such as its split history) will also contribute to the weight";
    _lookup.insert(&_nonliteralsInClauseWeight);
    _nonliteralsInClauseWeight.tag(OptionTag::OTHER);
    _nonliteralsInClauseWeight.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::INST_GEN)));
    _nonliteralsInClauseWeight.reliesOn(_splitting.is(notEqual(false)));
    _nonliteralsInClauseWeight.addProblemConstraint(hasNonUnits());
    _nonliteralsInClauseWeight.setRandomChoices({"on","off"});



    _satClauseActivityDecay = FloatOptionValue("sat_clause_activity_decay","",1.001f);
    _satClauseActivityDecay.description="";
    _lookup.insert(&_satClauseActivityDecay);
    _satClauseActivityDecay.tag(OptionTag::SAT);
    _satClauseActivityDecay.addConstraint(greaterThan(1.0f));

    _satClauseDisposer = ChoiceOptionValue<SatClauseDisposer>("sat_clause_disposer","",SatClauseDisposer::MINISAT,
                                                              {"growing","minisat"});
    _satClauseDisposer.description="";
    _lookup.insert(&_satClauseDisposer);
    _satClauseDisposer.tag(OptionTag::SAT);

    _satLearntMinimization = BoolOptionValue("sat_learnt_minimization","",true);
    _satLearntMinimization.description="";
    _lookup.insert(&_satLearntMinimization);
    _satLearntMinimization.tag(OptionTag::SAT);

    _satLearntSubsumptionResolution = BoolOptionValue("sat_learnt_subsumption_resolution","",true);
    _satLearntSubsumptionResolution.description="";
    _lookup.insert(&_satLearntSubsumptionResolution);
    _satLearntSubsumptionResolution.tag(OptionTag::SAT);

    _satRestartFixedCount = IntOptionValue("sat_restart_fixed_count","",16000);
    _satRestartFixedCount.description="";
    _lookup.insert(&_satRestartFixedCount);
    _satRestartFixedCount.tag(OptionTag::SAT);

    _satRestartGeometricIncrease = FloatOptionValue("sat_restart_geometric_increase","",1.1);
    _satRestartGeometricIncrease.description="";
    _lookup.insert(&_satRestartGeometricIncrease);
    _satRestartGeometricIncrease.tag(OptionTag::SAT);
    _satRestartGeometricIncrease.addConstraint(greaterThan(1.0f));

    _satRestartGeometricInit = IntOptionValue("sat_restart_geometric_init","",32);
    _satRestartGeometricInit.description="";
    _lookup.insert(&_satRestartGeometricInit);
    _satRestartGeometricInit.tag(OptionTag::SAT);

    _satRestartLubyFactor = IntOptionValue("sat_restart_luby_factor","",100);
    _satRestartLubyFactor.description="";
    _lookup.insert(&_satRestartLubyFactor);
    _satRestartLubyFactor.tag(OptionTag::SAT);

    _satRestartMinisatIncrease = FloatOptionValue("sat_restart_minisat_increase","",1.1);
    _satRestartMinisatIncrease.description="";
    _lookup.insert(&_satRestartMinisatIncrease);
    _satRestartMinisatIncrease.tag(OptionTag::SAT);
    _satRestartMinisatIncrease.addConstraint(greaterThan(1.0f));

    _satRestartMinisatInit = IntOptionValue("sat_restart_minisat_init","",100);
    _satRestartMinisatInit.description="";
    _lookup.insert(&_satRestartMinisatInit);
    _satRestartMinisatInit.tag(OptionTag::SAT);

    _satRestartStrategy = ChoiceOptionValue<SatRestartStrategy>("sat_restart_strategy","",SatRestartStrategy::LUBY,
                                                                {"fixed","geometric","luby","minisat"});
    _satRestartStrategy.description="";
    _lookup.insert(&_satRestartStrategy);
    _satRestartStrategy.tag(OptionTag::SAT);

    _satSolver = ChoiceOptionValue<SatSolver>("sat_solver","sas",SatSolver::VAMPIRE,
#if VZ3
            {"lingeling","minisat","vampire","z3"});
#else
            {"lingeling","minisat","vampire"});
#endif
    _satSolver.description=
    "Select the SAT solver to be used throughout the solver. This will be used in AVATAR (for splitting) when the saturation algorithm is discount,lrs or otter and in instance generation for selection and global subsumption.";
    _lookup.insert(&_satSolver);
    _satSolver.tag(OptionTag::SAT);
    _satSolver.setRandomChoices(
#if VZ3
            {"lingeling","minisat","vampire","z3"});
#else
            {"lingeling","minisat","vampire"});
#endif

    _satVarActivityDecay = FloatOptionValue("sat_var_activity_decay","",1.05f);
    _satVarActivityDecay.description="";
    _lookup.insert(&_satVarActivityDecay);
    _satVarActivityDecay.tag(OptionTag::SAT);
    _satVarActivityDecay.addConstraint(greaterThan(1.0f));

    _satVarSelector = ChoiceOptionValue<SatVarSelector>("sat_var_selector","svs",SatVarSelector::ACTIVE,
                                                        {"active","niceness","recently_learnt"});
    _satVarSelector.description="";
    _lookup.insert(&_satVarSelector);
    _satVarSelector.tag(OptionTag::SAT);

    _satLingelingSimilarModels = BoolOptionValue("sat_lingeling_similar_models","",true);
    _satLingelingSimilarModels.description="";
    _lookup.insert(&_satLingelingSimilarModels);
    _satLingelingSimilarModels.tag(OptionTag::SAT);
    _satLingelingSimilarModels.setExperimental();

    _satLingelingIncremental = BoolOptionValue("sat_lingeling_incremental","",false);
    _satLingelingIncremental.description="";
    _lookup.insert(&_satLingelingIncremental);
    _satLingelingIncremental.tag(OptionTag::SAT);
    _satLingelingIncremental.setExperimental();

 

    _tabulationBwRuleSubsumptionResolutionByLemmas = BoolOptionValue("tabulation_bw_rule_subsumption_resolution_by_lemmas","tbsr",true);
    _tabulationBwRuleSubsumptionResolutionByLemmas.description="";
    _lookup.insert(&_tabulationBwRuleSubsumptionResolutionByLemmas);
    _tabulationBwRuleSubsumptionResolutionByLemmas.tag(OptionTag::TABULATION);
    _tabulationBwRuleSubsumptionResolutionByLemmas.setRandomChoices({"on","off"});


    _tabulationFwRuleSubsumptionResolutionByLemmas = BoolOptionValue("tabulation_fw_rule_subsumption_resolution_by_lemmas","tfsr",true);
    _tabulationFwRuleSubsumptionResolutionByLemmas.description="";
    _lookup.insert(&_tabulationFwRuleSubsumptionResolutionByLemmas);
    _tabulationFwRuleSubsumptionResolutionByLemmas.tag(OptionTag::TABULATION);
    _tabulationFwRuleSubsumptionResolutionByLemmas.setRandomChoices({"on","off"});


    _tabulationGoalAgeWeightRatio = RatioOptionValue("tabulation_goal_awr","tgawr",1,1,'/');
    _tabulationGoalAgeWeightRatio.description=
    "when saturation algorithm is set to tabulation, this option determines the age-weight ratio for selecting next goal clause to process";
    _lookup.insert(&_tabulationGoalAgeWeightRatio);
    _tabulationGoalAgeWeightRatio.tag(OptionTag::TABULATION);
    _tabulationGoalAgeWeightRatio.setRandomChoices({"8/1","5/1","4/1","3/1","2/1","3/2","5/4","1/1","2/3","1/2","1/3","1/4","1/5","1/6","1/7","1/8","1/10","1/12","1/14","1/16","1/20","1/24","1/28","1/32","1/40","1/50","1/64","1/128"});

    _tabulationGoalLemmaRatio = RatioOptionValue("tabulation_goal_lemma_ratio","tglr",1,1,'/');
    _tabulationGoalLemmaRatio.description=
    "when saturation algorithm is set to tabulation, this option determines the ratio of processing new goals and lemmas";
    _lookup.insert(&_tabulationGoalLemmaRatio);
    _tabulationGoalLemmaRatio.tag(OptionTag::TABULATION);
    _tabulationGoalLemmaRatio.setRandomChoices({"20/1","10/1","7/1","5/1","4/1","3/1","2/1","1/1","1/2","1/3","1/4","1/5","1/7","1/10","1/20"});
    
    _tabulationInstantiateProducingRules = BoolOptionValue("tabulation_instantiate_producing_rules","tipr",true);
    _tabulationInstantiateProducingRules.description=
    "when saturation algorithm is set to tabulation, this option determines whether the producing rules will be made of theory clauses (in case it's off), or of their instances got from the substitution unifying them with the goal";
    _lookup.insert(&_tabulationInstantiateProducingRules);
    _tabulationInstantiateProducingRules.tag(OptionTag::TABULATION);
    _tabulationInstantiateProducingRules.setRandomChoices({"on","off"});
    
    _tabulationLemmaAgeWeightRatio = RatioOptionValue("tabulation_lemma_awr","tlawr",1,1,'/');
    _tabulationLemmaAgeWeightRatio.description=
    "when saturation algorithm is set to tabulation, this option determines the age-weight ratio for selecting next lemma to process";
    _lookup.insert(&_tabulationLemmaAgeWeightRatio);
    _tabulationLemmaAgeWeightRatio.tag(OptionTag::TABULATION);
    _tabulationLemmaAgeWeightRatio.setRandomChoices({"8/1","5/1","4/1","3/1","2/1","3/2","5/4","1/1","2/3","1/2","1/3","1/4","1/5","1/6","1/7","1/8","1/10","1/12","1/14","1/16","1/20","1/24","1/28","1/32","1/40","1/50","1/64","1/128"});
    
    
    
    
    
    _bfnt = BoolOptionValue("bfnt","bfnt",false);
    _bfnt.description="";
    _lookup.insert(&_bfnt);
    _bfnt.tag(OptionTag::OTHER);
    
    
    _bfnt.addProblemConstraint(notWithCat(Property::EPR));
    _bfnt.setRandomChoices({},{"on","off","off","off","off","off"});
    
    _increasedNumeralWeight = BoolOptionValue("increased_numeral_weight","inw",false);
    _increasedNumeralWeight.description=
             "weight of integer constants depends on the logarithm of their absolute value (instead of being 1)";
    _lookup.insert(&_increasedNumeralWeight);
    _increasedNumeralWeight.tag(OptionTag::OTHER);


    _interpretedSimplification = BoolOptionValue("interpreted_simplification","",false);
    _interpretedSimplification.description=
             "Performs simplifications of interpreted functions. This option requires interpreted_evaluation to be enabled as well. IMPORTANT - Currently not supported";
    _lookup.insert(&_interpretedSimplification);
    _interpretedSimplification.tag(OptionTag::OTHER);


    _literalComparisonMode = ChoiceOptionValue<LiteralComparisonMode>("literal_comparison_mode","lcm",
                                                                      LiteralComparisonMode::STANDARD,
                                                                      {"predicate","reverse","standard"});
    _literalComparisonMode.description="";
    _lookup.insert(&_literalComparisonMode);
    _literalComparisonMode.tag(OptionTag::OTHER);
    _literalComparisonMode.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _literalComparisonMode.addProblemConstraint(hasNonUnits());
    _literalComparisonMode.addProblemConstraint(hasPredicates());
    
    _literalComparisonMode.setRandomChoices({"predicate","reverse","standard"});


    _maxActive = LongOptionValue("max_active","",0);
    _maxActive.description="";
    _lookup.insert(&_maxActive);
    _maxActive.tag(OptionTag::OTHER);

    _maxAnswers = IntOptionValue("max_answers","",1);
    _maxAnswers.description="";
    _lookup.insert(&_maxAnswers);
    _maxAnswers.tag(OptionTag::OTHER);

    _maxInferenceDepth = IntOptionValue("max_inference_depth","",0);
    _maxInferenceDepth.description="";
    _lookup.insert(&_maxInferenceDepth);
    _maxInferenceDepth.tag(OptionTag::OTHER);

    _maxPassive = LongOptionValue("max_passive","",0);
    _maxPassive.description="";
    _lookup.insert(&_maxPassive);
    _maxPassive.tag(OptionTag::OTHER);

    _maxWeight = IntOptionValue("max_weight","",0);
    _maxWeight.description="Weight limit for clauses (0 means no weight limit)";
    _lookup.insert(&_maxWeight);
    _maxWeight.tag(OptionTag::OTHER);

    _nonGoalWeightCoefficient = NonGoalWeightOptionValue("nongoal_weight_coefficient","nwc",1.0);
    _nonGoalWeightCoefficient.description=
             "coefficient that will multiply the weight of theory clauses (those marked as 'axiom' in TPTP)";
    _lookup.insert(&_nonGoalWeightCoefficient);
    _nonGoalWeightCoefficient.tag(OptionTag::OTHER);
    _nonGoalWeightCoefficient.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _nonGoalWeightCoefficient.setRandomChoices({"1","1.1","1.2","1.3","1.5","1.7","2","2.5","3","4","5","10"});


    _normalize = BoolOptionValue("normalize","",false);
    _normalize.description="";
    _lookup.insert(&_normalize);
    _normalize.tag(OptionTag::OTHER);

    _questionAnswering = ChoiceOptionValue<QuestionAnsweringMode>("question_answering","",QuestionAnsweringMode::OFF,
                                                                  {"answer_literal","from_proof","off"});
    _questionAnswering.description="Determines whether (and how) we attempt to answer questions";
    _lookup.insert(&_questionAnswering);
    _questionAnswering.tag(OptionTag::OTHER);

    _randomSeed = IntOptionValue("random_seed","",Random::seed());
    _randomSeed.description="";
    _lookup.insert(&_randomSeed);
    _randomSeed.tag(OptionTag::OTHER);


    _symbolPrecedence = ChoiceOptionValue<SymbolPrecedence>("symbol_precedence","sp",SymbolPrecedence::ARITY,
                                                            {"arity","occurrence","reverse_arity"});
    _symbolPrecedence.description="";
    _lookup.insert(&_symbolPrecedence);
    _symbolPrecedence.tag(OptionTag::OTHER);
    _symbolPrecedence.reliesOn(_saturationAlgorithm.is(notEqual(SaturationAlgorithm::TABULATION)));
    _symbolPrecedence.setRandomChoices({"arity","occurence","reverse_arity"});

    _weightIncrement = BoolOptionValue("weight_increment","",false);
    _weightIncrement.description="";
    _lookup.insert(&_weightIncrement);
    _weightIncrement.tag(OptionTag::OTHER);

    
    
    

    _rowVariableMaxLength = IntOptionValue("row_variable_max_length","",2);
    _rowVariableMaxLength.description="";
    _lookup.insert(&_rowVariableMaxLength);
    _rowVariableMaxLength.tag(OptionTag::UNUSED);
    _rowVariableMaxLength.setExperimental();



    
    
    

    
    _colorUnblocking = BoolOptionValue("color_unblocking","",false);
    _colorUnblocking.description="";
    _lookup.insert(&_colorUnblocking);
    _colorUnblocking.setExperimental();
    _colorUnblocking.tag(OptionTag::OTHER);
    
    
    _showInterpolant = ChoiceOptionValue<InterpolantMode>("show_interpolant","",InterpolantMode::OFF,
                                                          {"minimized","off","on"});
    _showInterpolant.description="minimized tries to find a nicer interpolant than the default algorithm does";
    _lookup.insert(&_showInterpolant);
    _showInterpolant.tag(OptionTag::OUTPUT);
    _showInterpolant.setExperimental();
    



    
    _lingvaAdditionalInvariants = StringOptionValue("lingva_additional_invariants","","");
    _lingvaAdditionalInvariants.description="";
    _lookup.insert(&_lingvaAdditionalInvariants);
    _lingvaAdditionalInvariants.tag(Mode::PROGRAM_ANALYSIS);
    



    
    _whileNumber = IntOptionValue("while_number","",1);
    _whileNumber.description="";
    _lookup.insert(&_whileNumber);
    _whileNumber.tag(Mode::BOUND_PROP);
    
    _functionNumber = IntOptionValue("function_number","",1);
    _functionNumber.description="";
    _lookup.insert(&_functionNumber);
    _functionNumber.tag(Mode::BOUND_PROP);
    _functionNumber.tag(OptionTag::PREPROCESSING);
    
    _bpCollapsingPropagation = BoolOptionValue("bp_add_collapsing_inequalities","",false); 
    _bpCollapsingPropagation.description="";
    _lookup.insert(&_bpCollapsingPropagation);
    _bpCollapsingPropagation.tag(Mode::BOUND_PROP);
    _bpCollapsingPropagation.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpAllowedFMBalance = UnsignedOptionValue("bp_allowed_fm_balance","",0);
    _bpAllowedFMBalance.description="";
    _lookup.insert(&_bpAllowedFMBalance);
    _bpAllowedFMBalance.tag(Mode::BOUND_PROP);
    _bpAllowedFMBalance.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpAlmostHalfBoundingRemoval= ChoiceOptionValue<BPAlmostHalfBoundingRemoval>("bp_almost_half_bounding_removal","",
                                                                                 BPAlmostHalfBoundingRemoval::ON,{"bounds_on","off","on"});
    _bpAlmostHalfBoundingRemoval.description="";
    _lookup.insert(&_bpAlmostHalfBoundingRemoval);
    _bpAlmostHalfBoundingRemoval.tag(Mode::BOUND_PROP);
    _bpAlmostHalfBoundingRemoval.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpAssignmentSelector = ChoiceOptionValue<BPAssignmentSelector>("bp_assignment_selector","",
                                                                    BPAssignmentSelector::RANDOM,
                                                                    {"alternating","bmp","lower_bound",
                                                                        "middle","random","rational","smallest",
                                                                        "tight","tightish","upper_bound"});
    _bpAssignmentSelector.description="";
    _lookup.insert(&_bpAssignmentSelector);
    _bpAssignmentSelector.tag(Mode::BOUND_PROP);
    _bpAssignmentSelector.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _updatesByOneConstraint= UnsignedOptionValue("bp_bound_improvement_limit","",3);
    _updatesByOneConstraint.description="";
    _lookup.insert(&_updatesByOneConstraint);
    _updatesByOneConstraint.tag(Mode::BOUND_PROP);
    _updatesByOneConstraint.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpConflictSelector = ChoiceOptionValue<BPConflictSelector>("bp_conflict_selector","",
                                                                BPConflictSelector::MOST_RECENT,{"least_recent","most_recent","shortest"});
    _bpConflictSelector.description="";
    _lookup.insert(&_bpConflictSelector);
    _bpConflictSelector.tag(Mode::BOUND_PROP);
    _bpConflictSelector.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpConservativeAssignmentSelection = BoolOptionValue("bp_conservative_assignment_selection","",true);
    _bpConservativeAssignmentSelection.description="";
    _lookup.insert(&_bpConservativeAssignmentSelection);
    _bpConservativeAssignmentSelection.tag(Mode::BOUND_PROP);
    _bpConservativeAssignmentSelection.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpFmElimination= BoolOptionValue("bp_fm_elimination","",true);
    _bpFmElimination.description="";
    _lookup.insert(&_bpFmElimination);
    _bpFmElimination.tag(Mode::BOUND_PROP);
    _bpFmElimination.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _maximalPropagatedEqualityLength = UnsignedOptionValue("bp_max_prop_length","",5);
    _maximalPropagatedEqualityLength.description="";
    _lookup.insert(&_maximalPropagatedEqualityLength);
    _maximalPropagatedEqualityLength.tag(Mode::BOUND_PROP);
    _maximalPropagatedEqualityLength.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpPropagateAfterConflict = BoolOptionValue("bp_propagate_after_conflict","",true);
    _bpPropagateAfterConflict.description="";
    _lookup.insert(&_bpPropagateAfterConflict);
    _bpPropagateAfterConflict.tag(Mode::BOUND_PROP);
    _bpPropagateAfterConflict.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpStartWithPrecise = BoolOptionValue("bp_start_with_precise","",false);
    _bpStartWithPrecise.description="";
    _lookup.insert(&_bpStartWithPrecise);
    _bpStartWithPrecise.tag(Mode::BOUND_PROP);
    _bpStartWithPrecise.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpStartWithRational = BoolOptionValue("bp_start_with_rational","",false);
    _bpStartWithRational.description="";
    _lookup.insert(&_bpStartWithRational);
    _bpStartWithRational.tag(Mode::BOUND_PROP);
    _bpStartWithRational.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));
    
    _bpVariableSelector = ChoiceOptionValue<BPVariableSelector>("bp_variable_selector","",
                                                                BPVariableSelector::TIGHTEST_BOUND,
                                                                {"conflicting","conflicting_and_collapsing",
                                                                    "first","look_ahead","random","recent_collapsing",
                                                                    "recent_conflicting","tightest_bound"});
    _bpVariableSelector.description="";
    _lookup.insert(&_bpVariableSelector);
    _bpVariableSelector.tag(Mode::BOUND_PROP);
    _bpVariableSelector.reliesOn(_mode.is(equal(Mode::BOUND_PROP)));

    
 
    
    _tagNames = {
                 "Unused",
                 "Other",
                 "Output",
                 "Tabulation",
                 "Instance Generation",
                 "SAT Solving",
                 "AVATAR",
                 "Inferences",
                 "Saturation",
                 "Preprocessing",
                 "Input",
                 "Global"
                };

} 

void Options::copyValuesFrom(const Options& that)
{
  
  VirtualIterator<AbstractOptionValue*> options = _lookup.values();

  while(options.hasNext()){
    AbstractOptionValue* opt = options.next();
    if(opt->shouldCopy()){
      AbstractOptionValue* other = that.getOptionValueByName(opt->longName);
      ASS(opt!=other);
      bool status = opt -> set(other->getStringOfActual());
      ASS(status);
    }
  }
}
Options::Options(const Options& that)
{
  init();
  copyValuesFrom(that);
}

Options& Options::operator=(const Options& that)
{
  if(this==&that) return *this;
  copyValuesFrom(that);
  return *this;
}


void Options::set(const char* name,const char* value)
{
  CALL ("Options::set/2");

  try {
    if(!_lookup.findLong(name)->set(value)){
      USER_ERROR((vstring) value +" is an invalid value for "+(vstring)name+"\nSee help or use explain i.e. vampire -explain mode");
    }
  }
  catch (const ValueNotFoundException&) {
    if (!_ignoreMissing.actualValue) {
      vstring msg = (vstring)name + " is not a valid option";
      Stack<vstring> sim = getSimilarOptionNames(name,false);
      Stack<vstring>::Iterator sit(sim);
      if(sit.hasNext()){
        vstring first = sit.next();
        msg += "\n\tMaybe you meant ";
        if(sit.hasNext()) msg += "one of:\n\t\t";
        msg += first;
        while(sit.hasNext()){ msg+="\n\t\t"+sit.next();}
        msg+="\n\tYou can use -explain <option> to explain an option";
      }
      USER_ERROR(msg);
    }
  }
} 

void Options::set(const vstring& name,const vstring& value)
{
  CALL ("Options::set/2");
  set(name.c_str(),value.c_str());
} 

void Options::setShort(const char* name,const char* value)
{
  CALL ("Options::setShort");

  try {
    if(!_lookup.findShort(name)->set(value)){
      USER_ERROR((vstring) value +" is an invalid value for "+(vstring)name+", see help");
    }
  }
  catch (const ValueNotFoundException&) {
    if (!_ignoreMissing.actualValue) {
      vstring msg = (vstring)name + " is not a valid short option (did you mean --?)";
      Stack<vstring> sim = getSimilarOptionNames(name,true);
      Stack<vstring>::Iterator sit(sim);
      if(sit.hasNext()){
        vstring first = sit.next();
        msg += "\n\tMaybe you meant ";
        if(sit.hasNext()) msg += "one of:\n\t\t";
        msg += first;
        while(sit.hasNext()){ msg+="\n\t\t"+sit.next();}
        msg+="\n\tYou can use -explain <option> to explain an option";
      }
      USER_ERROR(msg);
    }
  }
} 


bool Options::OptionHasValue::check(Property*p){
          CALL("Options::OptionHasValue::check");
          AbstractOptionValue* opt = env.options->getOptionValueByName(option_value);
          ASS(opt);
          return opt->getStringOfActual()==value;
}


Options::OptionProblemConstraint* Options::isRandOn(){
      return new OptionHasValue("random_strategy","on");
}
Options::OptionProblemConstraint* Options::isRandSat(){
      return new OptionHasValue("random_strategy","sat");
}
Options::OptionProblemConstraint* Options::saNotInstGen(){
      return new OptionHasValue("sa","inst_gen");
}
Options::OptionProblemConstraint* Options::isBfnt(){
      return new OptionHasValue("bfnt","on");
}

vstring Options::includeFileName (const vstring& relativeName)
{
  CALL("Options::includeFileName");

  if (relativeName[0] == '/') { 
    return relativeName;
  }

  if (System::fileExists(relativeName)) {
    return relativeName;
  }

  
  
  
  vstring dir = include();

  if (dir == "") { 
    
    
    vstring currentFile = inputFile();
    System::extractDirNameFromPath(currentFile,dir); 
    if(System::fileExists(dir+"/"+relativeName)){
      return dir + "/" + relativeName;
    }

    
    char* env = getenv("TPTP");
    if (env) {
      dir = env;
    }
    else {
    
      dir = ".";
    }
    
    
  }
  
  return dir + "/" + relativeName;
} 


void Options::output (ostream& str) const
{
  CALL("Options::output");

  if(!explainOption().empty()){

    
    BYPASSING_ALLOCATOR;

     AbstractOptionValue* option;
     vstring name = explainOption();
     try{
       option = _lookup.findLong(name);
     }
     catch(const ValueNotFoundException&){ 
       try{
         option = _lookup.findShort(name);
       }
       catch(const ValueNotFoundException&){
         option = 0;
       }
     }
     if(!option){ 
       str << name << " not a known option" << endl;
       Stack<vstring> sim_s = getSimilarOptionNames(name,true);
       Stack<vstring> sim_l = getSimilarOptionNames(name,false);
       VirtualIterator<vstring> sit = pvi(getConcatenatedIterator(
           Stack<vstring>::Iterator(sim_s),Stack<vstring>::Iterator(sim_l))); 
        if(sit.hasNext()){
          vstring first = sit.next();
          str << "\tMaybe you meant ";
          if(sit.hasNext()) str << "one of:\n\t\t";
          str << first;
          while(sit.hasNext()){ str << "\n\t\t"+sit.next();}
          str << endl;
        }
     }
     else{
       vstringstream vs;
       option->output(vs);
       str << vs.str();
     }

  }

  if (showHelp()){
    str << "=========== Usage ==========\n";
    str << "Call vampire using\n";
    str << "  vampire [options] [problem]\n";
    str << "For example,\n";
    str << "  vampire --mode casc --include ~/TPTP ~/TPTP/Problems/ALG/ALG150+1.p\n";

    str << "=========== Hints ==========\n";


    str << "=========== Options ==========\n";
    str << "To see a list of all options use\n  --show_options on\n";
    str << "Options will only be displayed for the current mode (Vampire by default)\n";
    str << " use --mode to change mode\n";
    
    
    str << "=========== End ==========\n";
  }

  bool normalshow = showOptions();
  bool experimental = showExperimentalOptions();  
  
  if(normalshow || experimental) {

    
    BYPASSING_ALLOCATOR;

    Mode this_mode = _mode.actualValue;
    str << "=========== Options ==========\n";

    VirtualIterator<AbstractOptionValue*> options = _lookup.values();

    Stack<vstringstream*> groups;
    int num_tags = static_cast<int>(OptionTag::LAST_TAG);
    for(int i=0; i<=num_tags;i++){
      groups.push(new vstringstream);
    }

    while(options.hasNext()){
      AbstractOptionValue* option = options.next();
      if(option->inMode(this_mode) && 
              ((experimental && option->experimental) || 
               (normalshow && !option->experimental) )){
        unsigned tag = static_cast<unsigned>(option->getTag());
        option->output(*groups[tag]);
      }
    }

    
    for(int i=num_tags;i>=0;i--){
      vstring label = "  "+_tagNames[i]+"  ";
      ASS(label.length() < 40);
      vstring br = "******************************";
      vstring br_gap = br.substr(0,(br.length()-(label.length()/2)));
      str << endl << br << br << endl;
      str << br_gap << label << br_gap << endl;
      str << br << br << endl << endl;
      str << (*groups[i]).str();
      BYPASSING_ALLOCATOR;
      delete groups[i];
    }

    str << "======= End of options =======\n";
  }

} 






















template<typename T>
bool Options::OptionValue<T>::randomize(Property* prop){
  CALL("Options::OptionValue::randomize()");

  DArray<vstring>* choices = 0;
  if(env.options->randomStrategy()==RandomStrategy::NOCHECK) prop=0;

  
  if( env.options->randomStrategy()!=RandomStrategy::NOCHECK && 
      ((prop && !hasProblemConstraints()) || (!prop && hasProblemConstraints()))
    ){
    return false;
  }
  
  

  
  Stack<RandEntry>::BottomFirstIterator entry_it(rand_choices);
  while(entry_it.hasNext()){
    auto entry = entry_it.next();
    if(!entry.first || (prop && entry.first->check(prop))){
      choices = entry.second;
    }  
  }
  if(!choices || choices->size()==0) return false; 

  
  int index = Random::getInteger(choices->size());
  set((*choices)[index]);
  return true;
}


template<typename T>
bool Options::OptionValue<T>::checkConstraints(){
     CALL("Options::OptionValue::checkConstraints");
     typename Lib::Stack<OptionValueConstraint<T>*>::Iterator it(_constraints);
     while(it.hasNext()){
       OptionValueConstraint<T>* con = it.next();
       if(!con->check(this)){

         if(env.options->mode()==Mode::SPIDER){
           reportSpiderFail();
           USER_ERROR("\nBroken Constraint: "+con->msg(this));
         }

         if(con->isHard()){ 
           if(env.options->randomStrategy()!=RandomStrategy::OFF)
              return false; 
           USER_ERROR("\nBroken Constraint: "+con->msg(this)); 
         }
         switch(env.options->getBadOptionChoice()){
           case BadOption::HARD :
               USER_ERROR("\nBroken Constraint: "+con->msg(this));
           case BadOption::SOFT :
               cout << "WARNING Broken Constraint: "+con->msg(this) << endl;
               return false;
           case BadOption::FORCED :
               if(con->force(this)){
                 cout << "Forced constraint " + con->msg(this) << endl;
                 break;
               }else{
                 USER_ERROR("\nCould not force Constraint: "+con->msg(this));
               }
           case BadOption::OFF: 
             return false;
           default: ASSERTION_VIOLATION;
        }
     }
    }
    return true;
}

template<typename T>
bool Options::OptionValue<T>::checkProblemConstraints(Property* prop){
    CALL("Options::OptionValue::checkProblemConstraints");

    Lib::Stack<OptionProblemConstraint*>::Iterator it(_prob_constraints);
    while(it.hasNext()){
      OptionProblemConstraint* con = it.next();
      
      if(is_set && !con->check(prop)){

         if(env.options->mode()==Mode::SPIDER){
           reportSpiderFail();
           USER_ERROR("WARNING: " + longName + con->msg());
         }

         switch(env.options->getBadOptionChoice()){
         case BadOption::OFF: break;
         default:
           cout << "WARNING: " << longName << con->msg() << endl;
         }
         return false;
      }
    }
    return true;
}


template<typename T>
Options::WrappedConstraint<T>* Options::OptionValue<T>::is(OptionValueConstraint<T>* c)
{
    return new WrappedConstraint<T>(this,c);
}

template<typename T>
template<typename S, typename R>
Options::OptionValueConstraint<S>* Options::WrappedConstraint<T>::And(WrappedConstraint<R>* another)
{
    return new AndWrapper<S>(new UnWrappedConstraint<S,T>(this), new UnWrappedConstraint<S,R>(another));
}
template<typename T>
template<typename S, typename R>
Options::OptionValueConstraint<S>* Options::WrappedConstraint<T>::Or(WrappedConstraint<R>* another)
{
    return new OrWrapper<S>(new UnWrappedConstraint<S,T>(this), new UnWrappedConstraint<S,R>(another));
}

template<typename T>
Options::OptionValueConstraint<T>* Options::OptionValueConstraint<T>::And(OptionValueConstraint<T>* another)
{
    return new AndWrapper<T>(this,another);
}
template<typename T>
Options::OptionValueConstraint<T>* Options::OptionValueConstraint<T>::Or(OptionValueConstraint<T>* another)
{
    return new OrWrapper<T>(this,another);
}
template<typename T>
template<typename S>
Options::OptionValueConstraint<T>* Options::OptionValueConstraint<T>::And(WrappedConstraint<S>* another)
{
    return new AndWrapper<T>(this,new UnWrappedConstraint<T,S>(another));
}
template<typename T>
template<typename S>
Options::OptionValueConstraint<T>* Options::OptionValueConstraint<T>::Or(WrappedConstraint<S>* another)
{
    return new OrWrapper<T>(this,new UnWrappedConstraint<T,S>(another));
}

bool Options::RatioOptionValue::readRatio(const char* val, char separator)
{
  CALL("RatioOptionValue::readRatio");

  
  bool found = false;
  int colonIndex = 0;
  while (val[colonIndex]) {
    if (val[colonIndex] == separator) {
      found = true;
      break;
    }
    colonIndex++;
  }

  if (found) {
    if (strlen(val) > 127) {
      return false;
    }
    char copy[128];
    strcpy(copy,val);
    copy[colonIndex] = 0;
    int age;
    if (! Int::stringToInt(copy,age)) {
      return false;
    }
    actualValue = age;
    int weight;
    if (! Int::stringToInt(copy+colonIndex+1,weight)) {
      return false;
    }
    otherValue = weight;
    return true;
  }
  actualValue = 1;
  int weight;
  if (! Int::stringToInt(val,weight)) {
    return false;
  }
  otherValue = weight;
  return true;
}

bool Options::NonGoalWeightOptionValue::setValue(const vstring& value)
{
  CALL("NonGoalWeightOptionValue::setValue");

 float newValue;
 if(!Int::stringToFloat(value.c_str(),newValue)) return false;

 if(newValue <= 0.0) return false;

 actualValue=newValue;

  
  numerator=static_cast<int>(newValue*100);
  
  denominator=100;

  return true;


}

bool Options::SelectionOptionValue::setValue(const vstring& value)
{
  CALL("SelectionOptionValue::setValue");

  int sel;
  if(!Int::stringToInt(value,sel)) return false;
  switch (sel) {
  case 0:
  case 1:
  case 2:
  case 3:
  case 4:
  case 10:
  case 11:
  case 1002:
  case 1003:
  case 1004:
  case 1010:
  case 1011:
  case -1:
  case -2:
  case -3:
  case -4:
  case -10:
  case -11:
  case -1002:
  case -1003:
  case -1004:
  case -1010:
    actualValue = sel;
    return true;
  default:
    return false;
  }
}

bool Options::InputFileOptionValue::setValue(const vstring& value)
{
  CALL("InputFileOptionValue::setValue");

  actualValue=value;
  if(value.empty()) return true;

  

  int length = value.length() - 1;
  const char* name = value.c_str();

  int b = length - 1;
  while (b >= 0 && name[b] != '/') {
    b--;
  }
  b++;

  int e = length-1;
  while (e >= b && name[e] != '.') {
    e--;
  }
  if (e < b) {
    e = length-1;
  }

  parent->_problemName.actualValue=value.substr(b,e-b);

  return true;

}


bool Options::TimeLimitOptionValue::setValue(const vstring& value)
{
  CALL("Options::readTimeLimit");

  int length = value.size();
  if (length == 0 || length > 127) {
    USER_ERROR((vstring)"wrong value for time limit: " + value);
  }

  char copy[128];
  strcpy(copy,value.c_str());
  char* end = copy;
  
  while (*end) {
    end++;
  }
  end--;
  float multiplier = 10.0; 
  switch (*end) {
  case 'd': 
      multiplier = 1.0;
      *end = 0;
      break;
  case 's': 
    multiplier = 10.0;
    *end = 0;
    break;
  case 'm': 
    multiplier = 600.0;
    *end = 0;
    break;
  case 'h': 
    multiplier = 36000.0;
    *end = 0;
    break;
  case 'D': 
    multiplier = 864000.0;
    *end = 0;
    break;
  default:
    break;
  }

  float number;
  if (! Int::stringToFloat(copy,number)) {
    USER_ERROR((vstring)"wrong value for time limit: " + value);
  }

#ifdef _MSC_VER
  
  actualValue= (int)floor(number * multiplier);
#else
  actualValue= (int)round(number * multiplier);
#endif

  return true;
} 

void Options::randomizeStrategy(Property* prop)
{
  CALL("Options::randomizeStrategy");
  if(_randomStrategy.actualValue==RandomStrategy::OFF) return;

  TimeCounter tc(TC_RAND_OPT);

  
  
  
  int saved_seed = Random::seed();
  Random::setSeed(time(NULL)); 

  
  
  

  
  
  Stack<AbstractOptionValue*> do_first;
  do_first.push(&_saturationAlgorithm);
  do_first.push(&_bfnt);

  auto options = getConcatenatedIterator(Stack<AbstractOptionValue*>::Iterator(do_first),_lookup.values());

  
  BadOption saved_bad_option = _badOption.actualValue;
  _badOption.actualValue=BadOption::OFF;

  bool skipChecks = _randomStrategy.actualValue == RandomStrategy::NOCHECK;

  while(options.hasNext()){
    AbstractOptionValue* option = options.next();
    if(!option->is_set){
      
      vstring def = option->getStringOfActual();

      
      bool can_rand = option->randomize(prop);
      
      if(can_rand){
        
        bool valid = skipChecks || (checkGlobalOptionConstraints(true) && (!prop || checkProblemOptionConstraints(prop,true)));
        unsigned i=4;
        while(!valid && i-- > 0){
          option->randomize(prop);
          valid = checkGlobalOptionConstraints(true) && (!prop || checkProblemOptionConstraints(prop,true));
        }
        if(!valid){
           
           option->set(def);
           option->is_set=false;
        }
        
      }
    }
  }

  
  _badOption.actualValue = saved_bad_option;
  Random::setSeed(saved_seed);

  
  
  
  if(!checkGlobalOptionConstraints() && _bfnt.actualValue){
    _bfnt.set("off");
    randomizeStrategy(prop);
  }
  else{
    if(prop) cout << "Random strategy: " + generateEncodedOptions() << endl;
  }
}

void Options::readOptionsString(vstring optionsString,bool assign)
{
  CALL("Options::readOptionsString");

  
  while (optionsString != "") {
    size_t index1 = optionsString.find('=');
    if (index1 == vstring::npos) {
      error: USER_ERROR("bad option specification '" + optionsString+"'");
    }
    size_t index = optionsString.find(':');
    if (index!=vstring::npos && index1 > index) {
      goto error;
    }

    vstring param = optionsString.substr(0,index1);
    vstring value;
    if (index==vstring::npos) {
      value = optionsString.substr(index1+1);
    }
    else {
      value = optionsString.substr(index1+1,index-index1-1);
    }
    AbstractOptionValue* opt = getOptionValueByName(param);
    if(opt){
        if(assign){
            opt->set(value);
        }
        else{
            vstring current = opt->getStringOfActual();
            if(value==current){
                USER_ERROR("option "+param+" uses forbidden value "+value);
            }
        }
    }
    else{
      if(!ignoreMissing()){
       USER_ERROR("option "+param+" not known");
      }
    }

    if (index==vstring::npos) {
      return;
    }
    optionsString = optionsString.substr(index+1);
  }
} 

void Options::readFromEncodedOptions (vstring testId)
{
  CALL("Options::readFromTestId");

  _normalize.actualValue = true;
  _testId.actualValue = testId;

  vstring ma(testId,0,3); 
  if (ma == "dis") {
    _saturationAlgorithm.actualValue = SaturationAlgorithm::DISCOUNT;
  }
  else if (ma == "lrs") {
    _saturationAlgorithm.actualValue = SaturationAlgorithm::LRS;
  }
  else if (ma == "ott") {
    _saturationAlgorithm.actualValue = SaturationAlgorithm::OTTER;
  }
  else if (ma == "tab") {
    _saturationAlgorithm.actualValue = SaturationAlgorithm::TABULATION;
  }
  else if (ma == "ins") {
    _saturationAlgorithm.actualValue = SaturationAlgorithm::INST_GEN;
  }
  else if (ma == "fmb") {
    _saturationAlgorithm.actualValue = SaturationAlgorithm::FINITE_MODEL_BUILDING;
  }
  else {
  error: USER_ERROR("bad test id " + _testId.actualValue);
  }

  
  size_t index = testId.find_last_of('_');
  if (index == vstring::npos) { 
    goto error;
  }
  vstring timeString = testId.substr(index+1);
  _timeLimitInDeciseconds.set(timeString);
  
  _timeLimitInDeciseconds.actualValue = _timeLimitInDeciseconds.actualValue/10;

  testId = testId.substr(3,index-3);
  switch (testId[0]) {
  case '+':
    testId = testId.substr(1);
    break;
  case '-':
    break;
  default:
    goto error;
  }

  index = testId.find('_');
  vstring sel = testId.substr(0,index);
  _selection.set(sel);
  testId = testId.substr(index+1);

  if (testId == "") {
    goto error;
  }

  index = testId.find('_');
  vstring awr = testId.substr(0,index);
  _ageWeightRatio.set(awr.c_str());
  if (index==string::npos) {
    
    return;
  }
  testId = testId.substr(index+1);
  
  readOptionsString(testId);
} 

void Options::setForcedOptionValues()
{
  CALL("Options::setForcedOptionValues");

  
  switch (_mode.actualValue) {
  case Mode::CASC:
  case Mode::CASC_SAT:
  case Mode::CASC_THEORY:
  case Mode::CASC_LTB:
    _outputAxiomNames.actualValue = true;
    _proof.actualValue = Proof::TPTP;
    break;
  default:
    break;
  }

  if(_forcedOptions.actualValue.empty()) return;

  readOptionsString(_forcedOptions.actualValue);

}

vstring Options::generateEncodedOptions() const
{
  CALL("Options::generateEncodedOptions");

  BYPASSING_ALLOCATOR;
  vostringstream res;
  
  vstring sat;
  switch(_saturationAlgorithm.actualValue){
    case SaturationAlgorithm::LRS : sat="lrs"; break;
    case SaturationAlgorithm::DISCOUNT : sat="dis"; break;
    case SaturationAlgorithm::OTTER : sat="ott"; break;
    case SaturationAlgorithm::TABULATION : sat="tab"; break;
    case SaturationAlgorithm::INST_GEN : sat="ins"; break;
    case SaturationAlgorithm::FINITE_MODEL_BUILDING : sat="fmb"; break;
    default : ASSERTION_VIOLATION;
  }

  res << sat; 

  
  res << (selection() < 0 ? "-" : "+") << abs(selection());
  res << "_";

  
  if (ageRatio()!=1) {
    res << ageRatio() << ":";
  }
  res << weightRatio();
  res << "_";

  Options cur=*this;

  
  static Set<const AbstractOptionValue*> forbidden;
  
  if (forbidden.size()==0) {
    
    forbidden.insert(&_saturationAlgorithm);
    forbidden.insert(&_selection);
    forbidden.insert(&_ageWeightRatio);
    forbidden.insert(&_timeLimitInDeciseconds);

    
    forbidden.insert(&_mode);
    forbidden.insert(&_testId); 
    forbidden.insert(&_include);
    forbidden.insert(&_problemName);
    forbidden.insert(&_inputFile);
    forbidden.insert(&_randomStrategy);
    forbidden.insert(&_decode);
    forbidden.insert(&_ignoreMissing); 
  }

  VirtualIterator<AbstractOptionValue*> options = _lookup.values();

  bool first=true;
  while(options.hasNext()){
    AbstractOptionValue* option = options.next();
    
    if(!forbidden.contains(option) && option->is_set && !option->isDefault()){
      vstring name = option->shortName;
      if(name.empty()) name = option->longName;
      if(!first){ res<<":";}else{first=false;}
      res << name << "=" << option->getStringOfActual();
    }
  }

  if(!first){ res << "_"; }
  res << Lib::Int::toString(_timeLimitInDeciseconds.actualValue);
 
  return res.str();
 
}


bool Options::complete(const Problem& prb) const
{
  CALL("Options::complete");

  
  
  if (prb.hadIncompleteTransformation()) {
    return false;
  }

  Property& prop = *prb.getProperty();

  ASS(&prop);

  
  if (prop.hasInterpretedOperations()
      || prop.hasProp(Property::PR_HAS_INTEGERS)
      || prop.hasProp(Property::PR_HAS_REALS)
      || prop.hasProp(Property::PR_HAS_RATS)) {
    return false;
  }

  
  if (_sineSelection.actualValue != SineSelection::OFF) return false;

  switch (_saturationAlgorithm.actualValue) {
  case SaturationAlgorithm::TABULATION: return false;
  case SaturationAlgorithm::INST_GEN: return true; 
  default: break;
  }

  
  if (_sos.actualValue != Sos::OFF) return false;
  
  if (_forwardLiteralRewriting.actualValue) return false;
  
  bool unitEquality = prop.category() == Property::UEQ;
  bool hasEquality = (prop.equalityAtoms() != 0);

  if (!unitEquality) {
    if (_selection.actualValue <= -100 || _selection.actualValue >= 100) return false;
    if (_literalComparisonMode.actualValue == LiteralComparisonMode::REVERSE) return false;
  }

  if (!hasEquality) {
    if (_binaryResolution.actualValue) return true;
    if (_unitResultingResolution.actualValue==URResolution::EC_ONLY) return false;
    
    return prop.category() == Property::HNE; 
  }

  
  switch (_equalityProxy.actualValue) {
  case EqualityProxy::R: return false;
  case EqualityProxy::RS: return false;
  case EqualityProxy::RST: return false;
  default: break;
  }
  if (!_demodulationRedundancyCheck.actualValue) return false;
  if (_equalityResolutionWithDeletion.actualValue != RuleActivity::INPUT_ONLY) return false;
  if (!_superpositionFromVariables.actualValue) return false;

  
  bool pureEquality = (prop.atoms() == prop.equalityAtoms());
  if (pureEquality) return true;
  return (_binaryResolution.actualValue); 
} 

bool Options::completeForNNE() const
{
  CALL("Options::completeForNNE");

  
  if (_sineSelection.actualValue != SineSelection::OFF) return false;

  switch (_saturationAlgorithm.actualValue) {
  case SaturationAlgorithm::TABULATION: return false;
  case SaturationAlgorithm::INST_GEN: return true; 
  default: break;
  }

  
  if (_sos.actualValue != Sos::OFF) return false;
  
  if (_forwardLiteralRewriting.actualValue) return false;
  
  if (_selection.actualValue <= -100 || _selection.actualValue >= 100) return false;

  return _binaryResolution.actualValue;
} 


bool Options::checkGlobalOptionConstraints(bool fail_early)
{
  CALL("Options::checkGlobalOptionsConstraints");

  
  readOptionsString(_forbiddenOptions.actualValue,false);
    
  
  
  if(!fail_early && env.options->bfnt() && !completeForNNE()){
    return false;
  }

  bool result = true;

  
  VirtualIterator<AbstractOptionValue*> options = _lookup.values();
  while(options.hasNext()){ 
    result = options.next()->checkConstraints() && result; 
    if(fail_early && !result) return result;
  }

  return result;
}

bool Options::checkProblemOptionConstraints(Property* prop,bool fail_early)
{
   CALL("Options::checkProblemOptionConstraints");

  bool result = true;

  VirtualIterator<AbstractOptionValue*> options = _lookup.values();
  while(options.hasNext()){
    result = options.next()->checkProblemConstraints(prop) && result; 
    if(fail_early && !result) return result;
  }

  return result;
}
