
/*
 * File Options.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __Options__
#define __Options__

#include <type_traits>

#include "Forwards.hpp"

#include "Debug/Assertion.hpp"

#include "Lib/VString.hpp"
#include "Lib/VirtualIterator.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/DArray.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Int.hpp"
#include "Lib/Allocator.hpp"
#include "Lib/XML.hpp"

#include "Property.hpp"

namespace Shell {

using namespace Lib;
using namespace Kernel;

class Property;

static size_t distance(const vstring &s1, const vstring &s2)
{
  const size_t m(s1.size());
  const size_t n(s2.size());

  if( m==0 ) return n;
  if( n==0 ) return m;

  DArray<size_t> costs = DArray<size_t>(n+1);

  for( size_t k=0; k<=n; k++ ) costs[k] = k;

  size_t i = 0;
  for ( vstring::const_iterator it1 = s1.begin(); it1 != s1.end(); ++it1, ++i )
  {
    costs[0] = i+1;
    size_t corner = i;

    size_t j = 0;
    for ( vstring::const_iterator it2 = s2.begin(); it2 != s2.end(); ++it2, ++j )
    {
      size_t upper = costs[j+1];
      if( *it1 == *it2 ){costs[j+1] = corner;}
      else{
        size_t t(upper<corner?upper:corner);
        costs[j+1] = (costs[j]<t?costs[j]:t)+1;
      }
      corner = upper;
    }
  }

  return costs[n];
}


class Options
{
public:

    Options ();
    
    void init();
    void copyValuesFrom(const Options& that);
    Options(const Options& that);
    Options& operator=(const Options& that);

    
    void output (ostream&) const;

    
    void readFromEncodedOptions (vstring testId);
    void readOptionsString (vstring testId,bool assign=true);
    vstring generateEncodedOptions() const;

    
    bool complete(const Problem&) const;
    bool completeForNNE() const;
    void forceIncompleteness() { _forceIncompleteness.actualValue=true; }

    
    void setForcedOptionValues(); 
    bool checkGlobalOptionConstraints(bool fail_early=false);
    bool checkProblemOptionConstraints(Property*, bool fail_early=false); 

    
    
    
    
    
    
    
    
    
    
    void randomizeStrategy(Property* prop);
    
   
    vstring problemName () const { return _problemName.actualValue; }
    void setProblemName(vstring str) { _problemName.actualValue = str; }
    
    void setInputFile(const vstring& newVal){ _inputFile.set(newVal); }
    vstring includeFileName (const vstring& relativeName);

    CLASS_NAME(Options);
    USE_ALLOCATOR(Options);
    
    
    void set(const vstring& name, const vstring& value);
    void set(const char* name, const char* value);
    void setShort(const char* name, const char* value);
    
    
public:
  
  
  
  
  
   
 
   
    enum class OptionTag: unsigned int {
        UNUSED,
        OTHER,
        OUTPUT,
        TABULATION,
        INST_GEN,
        SAT,
        AVATAR,
        INFERENCES,
        SATURATION,
        PREPROCESSING,
        INPUT,
        LAST_TAG 
    };
    
    
  enum class RandomStrategy : unsigned int {
    ON,
    OFF,
    SAT,
    NOCHECK
  };

  enum class BadOption : unsigned int {
    HARD,
    FORCED,
    OFF,
    SOFT
  };

  
  enum class BPAlmostHalfBoundingRemoval : unsigned int {
    BOUNDS_ONLY = 0,
    OFF = 1,
    ON = 2
  };

  enum class BPAssignmentSelector: unsigned int {
    ALTERNATIVE = 0,
    BMP = 1,
    LOWER = 2,
    MIDDLE = 3,
    RANDOM = 4,
    RATIONAL = 5,
    SMALLEST = 6,
    TIGHT = 7,
    TIGHTISH = 8,
    UPPER = 9
  };
  
  enum class BPConflictSelector: unsigned int {
    LEAST_RECENT = 0, 
    MOST_RECENT = 1, 
    SHORTEST_CONSTRAINT = 2
  };
  
  enum class BPVariableSelector: unsigned int {
    CONFLICTING = 0, 
    CONFLICTING_AND_COLLAPSING = 1, 
    FIRST = 2, 
    LOOK_AHEAD =3, 
    RANDOM = 4, 
    RECENTLY_CONFLICTING = 5,
    RECENTLY_COLLAPSING = 6,
    TIGHTEST_BOUND = 7

  };
 
  enum class FunctionDefinitionElimination : unsigned int {
    ALL = 0,
    NONE = 1,
    UNUSED = 2
  };

 
  enum class Instantiation : unsigned int {
    OFF = 0,
    ON = 1
  };

 
  enum class InputSyntax : unsigned int {
   
    SIMPLIFY = 0,
   
    SMTLIB = 1,
    SMTLIB2 = 2,
   
    TPTP = 3, 
    HUMAN = 4, 
    MPS = 5, 
    NETLIB = 6
  };


 
  enum class Mode : unsigned int {
    AXIOM_SELECTION,
    BOUND_PROP,
    CASC,
    CASC_EPR,
    CASC_LTB,
    CASC_MZR,
    CASC_SAT,
    CASC_THEORY,
    CLAUSIFY,
    CONSEQUENCE_ELIMINATION,
    GROUNDING,
    LTB_BUILD,
    LTB_SOLVE,
   
    OUTPUT,
    PREPROCESS,
    PROFILE,
    PROGRAM_ANALYSIS,   
    RANDOM_STRATEGY,
    SAT, 
    SPIDER,
    VAMPIRE
};


 
  enum class Statistics : unsigned int {
   
    BRIEF = 0,
   
    FULL = 1,
   
    NONE = 2
  };

 
  enum class SatSolver : unsigned int {
     LINGELING = 0,
     MINISAT = 1,
     VAMPIRE = 2
#if VZ3
     ,Z3 = 3
#endif
  };

 
  enum class SaturationAlgorithm : unsigned int {
     DISCOUNT = 0,
     FINITE_MODEL_BUILDING = 1,
     INST_GEN = 2,
     LRS = 3,
     OTTER = 4,
     TABULATION = 5
   };

 
  enum class RuleActivity : unsigned int {
    INPUT_ONLY = 0,
    OFF = 1,
    ON = 2
  };

  enum class QuestionAnsweringMode : unsigned int {
    ANSWER_LITERAL = 0,
    FROM_PROOF = 1,
    OFF = 2
  };

  enum class InliningMode : unsigned int {
    AXIOMS_ONLY = 0,
    NON_GROWING = 1,
    OFF = 2,
    ON = 3
  };

  enum class InterpolantMode : unsigned int {
    MINIMIZED = 0,
    OFF = 1,
    ON = 2
  };

  enum class LiteralComparisonMode : unsigned int {
    PREDICATE = 0,
    REVERSE = 1,
    STANDARD = 2
  };

  enum class Condensation : unsigned int {
    FAST = 0,
    OFF = 1,
    ON = 2
  };

  enum class Demodulation : unsigned int {
    ALL = 0,
    OFF = 1,
    PREORDERED = 2
  };

  enum class Subsumption : unsigned int {
    OFF = 0,
    ON = 1,
    UNIT_ONLY = 2
  };

  enum class URResolution : unsigned int {
    EC_ONLY = 0,
    OFF = 1,
    ON = 2
  };

  enum class SymbolPrecedence : unsigned int {
    ARITY = 0,
    OCCURRENCE = 1,
    REVERSE_ARITY = 2
  };

  enum class SineSelection : unsigned int {
    AXIOMS = 0,
    INCLUDED = 1,
    OFF = 2
  };

  enum class Proof : unsigned int {
    OFF = 0,
    ON = 1,
    PROOFCHECK = 2,
    TPTP = 3
  };

 
  enum class EqualityProxy : unsigned int {
    R = 0,
    RS = 1,
    RST = 2,
    RSTC = 3,
    OFF = 4,
  };

 
  enum class ExtensionalityResolution : unsigned int {
    FILTER = 0,
    KNOWN = 1,
    OFF = 2
  };

  enum class SatRestartStrategy : unsigned int {
    FIXED = 0,
    GEOMETRIC = 1,
    LUBY = 2,
    MINISAT = 3,
  };

  enum class SatVarSelector : unsigned int {
    ACTIVE = 0,
    NICENESS = 1,
    RECENTLY_LEARNT = 2,
  };

  enum class Niceness: unsigned int {
    AVERAGE = 0,
    NONE=1,
    SUM = 2,
    TOP = 3,
  };

  enum class SatClauseDisposer : unsigned int {
    GROWING = 0,
    MINISAT = 1,
  };
  
  enum class SplittingLiteralPolarityAdvice : unsigned int {
    FALSE,
    TRUE,
    NONE
  };

  enum class SplittingMinimizeModel : unsigned int {
    OFF = 0,
    SCO = 1,
    ALL = 2
  };

  enum class SplittingDeleteDeactivated : unsigned int {
    ON,
    LARGE_ONLY,
    OFF
  };
    
  enum class SplittingAddComplementary : unsigned int {
    GROUND = 0,
    NONE = 1
  };
  
  enum class SplittingCongruenceClosure : unsigned int {
    MODEL = 0,
    OFF = 1,
    ON = 2
  };

  enum class SplittingNonsplittableComponents : unsigned int {
    ALL = 0,
    ALL_DEPENDENT = 1,
    KNOWN = 2,
    NONE = 3
  };

  enum class CCUnsatCores : unsigned int {
    FIRST = 0,
    SMALL_ONES = 1,
    ALL = 2
  };

  enum class GlobalSubsumptionSatSolverPower : unsigned int {
    PROPAGATION_ONLY,
    FULL
  };

  enum class GlobalSubsumptionExplicitMinim : unsigned int {
    OFF,
    ON,
    RANDOMIZED
  };

  enum class GlobalSubsumptionAvatarAssumptions : unsigned int {
    OFF,
    FROM_CURRENT,
    FULL_MODEL
  };

  enum class Sos : unsigned int{
    ALL = 0,
    OFF = 1,
    ON = 2,
    THEORY = 3
  };

  enum class PredicateEquivalenceDiscoveryMode : unsigned int{
    ALL_ATOMS = 0,
    ALL_FORMULAS = 1,
    DEFINITIONS = 2,
    OFF = 3,
    ON = 4
  };

    
    
    
    
    
    
    
    
    
    
    
    
    
    
private:
    
   
    struct OptionChoiceValues{
        
        OptionChoiceValues(){ };
        OptionChoiceValues(std::initializer_list<vstring> list){
            for(std::initializer_list<vstring>::iterator it = list.begin();
                it!=list.end();++it){
                names.push(*it);
                ASS((*it).size()<70); 
            }
        }
        
        int find(vstring value) const {
            for(unsigned i=0;i<names.length();i++){
                if(value.compare(names[i])==0) return i;
            }
            return -1;
        }
        const int length() const { return names.length(); }
        const vstring operator[](int i) const{ return names[i];}
        
    private:
        Stack<vstring> names;
    };
    
    
    template<typename T>
    struct OptionValueConstraint;
    template<typename T>
    struct WrappedConstraint;
    struct OptionProblemConstraint;
    static OptionProblemConstraint* isRandOn();
    static OptionProblemConstraint* isRandSat();
    
   
    struct AbstractOptionValue{
        
        CLASS_NAME(AbstractOptionValue);
        USE_ALLOCATOR(AbstractOptionValue);
        
        AbstractOptionValue(){}
        AbstractOptionValue(vstring l,vstring s) :
        longName(l), shortName(s), experimental(false), is_set(false),_should_copy(true), _tag(OptionTag::LAST_TAG), supress_problemconstraints(false) {}
        
        
    private:
        AbstractOptionValue(const AbstractOptionValue& that);
    public:
        
        
        virtual bool setValue(const vstring& value) = 0;

        bool set(const vstring& value){
          bool okay = setValue(value); 
          if(okay) is_set=true;
          return okay;
        }
        
        
        virtual bool randomize(Property* P) = 0;

        
        void setExperimental(){experimental=true;}
        
        
        vstring longName;
        vstring shortName;
        vstring description;
        bool experimental;
        bool is_set;
        
        
        virtual bool checkConstraints() = 0;
        virtual bool checkProblemConstraints(Property* prop) = 0;
        
        
        void tag(OptionTag tag){ ASS(_tag==OptionTag::LAST_TAG);_tag=tag; }
        void tag(Options::Mode mode){ _modes.push(mode); }
        
        OptionTag getTag(){ return _tag;}
        bool inMode(Options::Mode mode){
            if(_modes.isEmpty()) return true;
            else return _modes.find(mode);
        }
        
        
        virtual vstring getStringOfActual() const = 0;
        
        virtual bool isDefault() const = 0;
        
        
        virtual void output(vstringstream& out) const {
            CALL("Options::AbstractOptionValue::output");
            out << "--" << longName;
            if(!shortName.empty()){ out << " (-"<<shortName<<")"; }
            out << endl;
            
            if (experimental) {
              out << "\t[experimental]" << endl;
            }
            
            if(!description.empty()){
                
                
                out << "\t";
                int count=0;
                for(const char* p = description.c_str();*p;p++){
                    out << *p;
                    count++;
                    if(count>70 && *p==' '){
                        out << endl << '\t';
                        count=0;
                    }
                    if(*p=='\n'){ count=0; out << '\t'; }
                }
                out << endl;
            }
            else{ out << "\tno description provided!" << endl; }
        }
        
        
        
        bool _should_copy;
        bool shouldCopy() const { return _should_copy; }
       
        typedef pair<OptionProblemConstraint*,DArray<vstring>*> RandEntry;

        void setRandomChoices(std::initializer_list<vstring> list){
          rand_choices.push(RandEntry(0,toArray(list)));
        }
        void setRandomChoices(std::initializer_list<vstring> list,
                              std::initializer_list<vstring> list_sat){
          rand_choices.push(RandEntry(isRandOn(),toArray(list)));
          rand_choices.push(RandEntry(isRandSat(),toArray(list_sat)));
        }
        void setRandomChoices(OptionProblemConstraint* c,
                              std::initializer_list<vstring> list){
          rand_choices.push(RandEntry(c,toArray(list)));
        }
        void setNoPropertyRandomChoices(std::initializer_list<vstring> list){
          rand_choices.push(RandEntry(0,toArray(list)));
          supress_problemconstraints=true;
        }

 
    private:
        
        OptionTag _tag;
        Lib::Stack<Options::Mode> _modes;

        DArray<vstring>* toArray(std::initializer_list<vstring>& list){
          DArray<vstring>* array = new DArray<vstring>(list.size());
          unsigned index=0;
          for(typename std::initializer_list<vstring>::iterator it = list.begin();
           it!=list.end();++it){ (*array)[index++] =*it; }
          return array;
        }
    protected:
        
        Stack<RandEntry> rand_choices;
        bool supress_problemconstraints;
    };
    
   
    template<typename T>
    struct OptionValue : public AbstractOptionValue {
        
        CLASS_NAME(OptionValue);
        USE_ALLOCATOR(OptionValue);
        
        
        
        
        OptionValue(){}
        OptionValue(vstring l, vstring s,T def) : AbstractOptionValue(l,s),
        defaultValue(def), actualValue(def){}
        
        
        T defaultValue;
        T actualValue;
        
        virtual bool isDefault() const { return defaultValue==actualValue;}

        
        virtual vstring getStringOfValue(T value) const{ ASSERTION_VIOLATION;}
        virtual vstring getStringOfActual() const { return getStringOfValue(actualValue); }
        
        
        
        
        void addConstraint(OptionValueConstraint<T>* c){ _constraints.push(c); }
        void addHardConstraint(OptionValueConstraint<T>* c){ c->setHard();addConstraint(c); }
        
        
        
        template<typename S>
        void reliesOn(WrappedConstraint<S>* c){
            _constraints.push(If(getNotDefault()).then(c));
        }
        virtual OptionValueConstraint<T>* getNotDefault(){ return isNotDefault<T>(); }
        template<typename S>
        void reliesOnHard(WrappedConstraint<S>* c){
            OptionValueConstraint<T>* tc = If(getNotDefault()).then(c);
            tc->setHard();
            _constraints.push(tc);
        }
        void reliesOn(OptionValueConstraint<T>* c){
            _constraints.push(If(getNotDefault()).then(c));
        }
        void reliesOnHard(OptionValueConstraint<T>* c){
            OptionValueConstraint<T>* tc = If(getNotDefault()).then(c);
            tc->setHard();
            _constraints.push(tc);
        }
        
        bool checkConstraints();
        
        
        
        WrappedConstraint<T>* is(OptionValueConstraint<T>* c);
        
        
        void addProblemConstraint(OptionProblemConstraint* c){ _prob_constraints.push(c); }
        bool hasProblemConstraints(){ 
          return !supress_problemconstraints && !_prob_constraints.isEmpty(); 
        }
        virtual bool checkProblemConstraints(Property* prop);
        
        virtual void output(vstringstream& out) const {
            CALL("Options::OptionValue::output");
            AbstractOptionValue::output(out);
            out << "\tdefault: " << getStringOfValue(defaultValue) << endl;
        }
       
        
        bool randomize(Property* p);
 
    private:
        
        Lib::Stack<OptionValueConstraint<T>*> _constraints;
        Lib::Stack<OptionProblemConstraint*> _prob_constraints;
    };
    
   
    
   
    template<typename T >
    struct ChoiceOptionValue : public OptionValue<T> {
        
        CLASS_NAME(ChoiceOptionValue);
        USE_ALLOCATOR(ChoiceOptionValue);
        
        ChoiceOptionValue(){}
        ChoiceOptionValue(vstring l, vstring s,T def,OptionChoiceValues c) :
        OptionValue<T>(l,s,def), choices(c) {}
        
        bool setValue(const vstring& value){
            
            int index = choices.find(value.c_str());
            if(index<0) return false;
            this->actualValue = static_cast<T>(index);
            return true;
        }
        
        virtual void output(vstringstream& out) const {
            AbstractOptionValue::output(out);
            out << "\tdefault: " << choices[static_cast<unsigned>(this->defaultValue)];
            out << endl;
            string values_header = "values: ";
            out << "\t" << values_header;
            
            int count=0;
            for(int i=0;i<choices.length();i++){
                if(i==0){
                    out << choices[i];
                }
                else{
                    out << ",";
                    vstring next = choices[i];
                    if(next.size()+count>60){ 
                        out << endl << "\t";
                        for(unsigned j=0;j<values_header.size();j++){out << " ";}
                        count = 0;
                    }
                    out << next;
                    count += next.size();
                }
            }
            out << endl;
        }
        
        vstring getStringOfValue(T value) const {
            unsigned i = static_cast<unsigned>(value);
            return choices[i];
        }
        
    private:
        OptionChoiceValues choices;
    };
   
    struct BoolOptionValue : public OptionValue<bool> {
        BoolOptionValue(){}
        BoolOptionValue(vstring l,vstring s, bool d) : OptionValue(l,s,d){}
        bool setValue(const vstring& value){
            if (! value.compare("on") || ! value.compare("true")) {
                actualValue=true;
                
            }
            else if (! value.compare("off") || ! value.compare("false")) {
                actualValue=false;
            }
            else return false;
            
            return true;
        }
        
        vstring getStringOfValue(bool value) const { return (value ? "on" : "off"); }
    };
    struct IntOptionValue : public OptionValue<int> {
        IntOptionValue(){}
        IntOptionValue(vstring l,vstring s, int d) : OptionValue(l,s,d){}
        bool setValue(const vstring& value){
            return Int::stringToInt(value.c_str(),actualValue);
        }
        vstring getStringOfValue(int value) const{ return Lib::Int::toString(value); }
    };
    
    struct UnsignedOptionValue : public OptionValue<unsigned> {
        UnsignedOptionValue(){}
        UnsignedOptionValue(vstring l,vstring s, unsigned d) : OptionValue(l,s,d){}
        bool setValue(const vstring& value){
            return Int::stringToUnsignedInt(value.c_str(),actualValue);
        }
        vstring getStringOfValue(unsigned value) const{ return Lib::Int::toString(value); }
    };
    
    struct StringOptionValue : public OptionValue<vstring> {
        StringOptionValue(){}
        StringOptionValue(vstring l,vstring s, vstring d) : OptionValue(l,s,d){}
        bool setValue(const vstring& value){
            actualValue = (value=="<empty>") ? "" : value;
            return true;
        }
        vstring getStringOfValue(vstring value) const{
            if(value.empty()) return "<empty>";
            return value;
        }
    };
    
    struct LongOptionValue : public OptionValue<long> {
        LongOptionValue(){}
        LongOptionValue(vstring l,vstring s, long d) : OptionValue(l,s,d){}
        bool setValue(const vstring& value){
            return Int::stringToLong(value.c_str(),actualValue);
        }
        vstring getStringOfValue(long value) const{ return Lib::Int::toString(value); }
    };
    
    struct FloatOptionValue : public OptionValue<float>{
        FloatOptionValue(){}
        FloatOptionValue(vstring l,vstring s, float d) : OptionValue(l,s,d){}
        bool setValue(const vstring& value){
            return Int::stringToFloat(value.c_str(),actualValue);
        }
        vstring getStringOfValue(float value) const{ return Lib::Int::toString(value); }
    };
    
   
    struct RatioOptionValue : public OptionValue<int> {
        
        CLASS_NAME(RatioOptionValue);
        USE_ALLOCATOR(RatioOptionValue);
        
        RatioOptionValue(){}
        RatioOptionValue(vstring l, vstring s, int def, int other, char sp=':') :
        OptionValue(l,s,def), sep(sp), defaultOtherValue(other), otherValue(other) {};
        
        virtual OptionValueConstraint<int>* getNotDefault(){ return isNotDefaultRatio(); }

        template<typename S>
        void addConstraintIfNotDefault(WrappedConstraint<S>* c){
            addConstraint(If(isNotDefaultRatio()).then(c));
        }
        
        bool readRatio(const char* val,char seperator);
        bool setValue(const vstring& value){
            return readRatio(value.c_str(),sep);
        }
        
        char sep;
        int defaultOtherValue;
        int otherValue;
        
        virtual void output(vstringstream& out) const {
            AbstractOptionValue::output(out);
            out << "\tdefault left: " << defaultValue << endl;
            out << "\tdefault right: " << defaultOtherValue << endl;
        }
        
        virtual vstring getStringOfValue(int value) const{ ASSERTION_VIOLATION;}
        virtual vstring getStringOfActual() const override {
            return Lib::Int::toString(actualValue)+sep+Lib::Int::toString(otherValue);
        }
        
    };
    
    
    
    
   
    struct NonGoalWeightOptionValue : public OptionValue<float>{
        
        CLASS_NAME(NonGoalWeightOptionValue);
        USE_ALLOCATOR(NonGoalWeightOptionValue);
        
        NonGoalWeightOptionValue(){}
        NonGoalWeightOptionValue(vstring l, vstring s, float def) :
        OptionValue(l,s,def), numerator(1), denominator(1) {};
        
        bool setValue(const vstring& value);
        
        
        
        int numerator;
        int denominator;
        
        virtual vstring getStringOfValue(float value) const{ return Lib::Int::toString(value); }
    };
    
   
    struct SelectionOptionValue : public OptionValue<int>{
        SelectionOptionValue(){}
        SelectionOptionValue(vstring l,vstring s, int def):
        OptionValue(l,s,def){};
        
        bool setValue(const vstring& value);
        
        virtual void output(vstringstream& out) const {
            AbstractOptionValue::output(out);
            out << "\tdefault: " << defaultValue << endl;;
        }
        
        virtual vstring getStringOfValue(int value) const{ return Lib::Int::toString(value); }
    };
    
   
    struct InputFileOptionValue : public OptionValue<vstring>{
        InputFileOptionValue(){}
        InputFileOptionValue(vstring l,vstring s, vstring def,Options* p):
        OptionValue(l,s,def), parent(p){};
        
        bool setValue(const vstring& value);
        
        virtual void output(vstringstream& out) const {
            AbstractOptionValue::output(out);
            out << "\tdefault: " << defaultValue << endl;;
        }
        virtual vstring getStringOfValue(vstring value) const{ return value; }
    private:
        Options* parent;
        
    };
   
    struct DecodeOptionValue : public OptionValue<vstring>{
        DecodeOptionValue(){ AbstractOptionValue::_should_copy=false;}
        DecodeOptionValue(vstring l,vstring s,Options* p):
        OptionValue(l,s,""), parent(p){ AbstractOptionValue::_should_copy=false;}
        
        bool setValue(const vstring& value){
            parent->readFromEncodedOptions(value);
            return true;
        }
        virtual vstring getStringOfValue(vstring value) const{ return value; }
        
    private:
        Options* parent;
        
    };
   
    struct TimeLimitOptionValue : public OptionValue<int>{
        TimeLimitOptionValue(){}
        TimeLimitOptionValue(vstring l, vstring s, float def) :
        OptionValue(l,s,def) {};
        
        bool setValue(const vstring& value);
        
        virtual void output(vstringstream& out) const {
            CALL("Options::TimeLimitOptionValue::output");
            AbstractOptionValue::output(out);
            out << "\tdefault: " << defaultValue << "d" << endl;
        }
        virtual vstring getStringOfValue(int value) const{ return Lib::Int::toString(value)+"d"; }
    };
    
   
    template<typename T>
    struct WrappedConstraint;
    
    template<typename T>
    struct OptionValueConstraint{
        CLASS_NAME(OptionValueConstraint);
        USE_ALLOCATOR(OptionValueConstraint);
        OptionValueConstraint() : _hard(false) {}
        
        virtual bool check(OptionValue<T>* value) = 0;
        virtual bool check(){ ASSERTION_VIOLATION; }
        virtual vstring msg(OptionValue<T>* value) = 0;
        virtual vstring msg() { ASSERTION_VIOLATION; }
        
        
        virtual bool force(OptionValue<T>* value){ return false;}
        
        bool isHard(){ return _hard; }
        void setHard(){ _hard=true;}
        bool _hard;
        
        OptionValueConstraint<T>* And(OptionValueConstraint<T>* another);
        OptionValueConstraint<T>* Or(OptionValueConstraint<T>* another);
        
        template<typename S>
        OptionValueConstraint<T>* And(WrappedConstraint<S>* another);
        template<typename S>
        OptionValueConstraint<T>* Or(WrappedConstraint<S>* another);
        
    };
    
    
    
    
    template<typename T>
    struct WrappedConstraint{
        CLASS_NAME(WrappedConstraint);
        USE_ALLOCATOR(WrappedConstraint);
        
        WrappedConstraint(OptionValue<T>* v, OptionValueConstraint<T>* c) : value(v), con(c) {}
        
        bool check(){
            return con->check(value);
        }
        vstring msg(){
            return con->msg(value);
        }
        
        template<typename S, typename R>
        OptionValueConstraint<S>* And(WrappedConstraint<R>* another);
        template<typename S, typename R>
        OptionValueConstraint<S>* Or(WrappedConstraint<R>* another);
        
        OptionValue<T>* value;
        OptionValueConstraint<T>* con;
    };
    
    template<typename T, typename S>
    struct UnWrappedConstraint : public OptionValueConstraint<T>{
        CLASS_NAME(UnWrappedConstraint);
        USE_ALLOCATOR(UnWrappedConstraint);
        
        UnWrappedConstraint(WrappedConstraint<S>* c) : con(c) {}
        
        bool check(OptionValue<T>*){ return con->check(); }
        vstring msg(OptionValue<T>*){ return con->msg(); }
        
        WrappedConstraint<S>* con;
    };
    
    template<typename T>
    struct OrWrapper : public OptionValueConstraint<T>{
        CLASS_NAME(OrWrapper);
        USE_ALLOCATOR(OrWrapper);
        OrWrapper(OptionValueConstraint<T>* l, OptionValueConstraint<T>* r) : left(l),right(r) {}
        bool check(OptionValue<T>* value){
            return left->check(value) || right->check(value);
        }
        vstring msg(OptionValue<T>* value){ return left->msg(value) + " or " + right->msg(value); }
        
        OptionValueConstraint<T>* left;
        OptionValueConstraint<T>* right;
    };
    
    template<typename T>
    struct AndWrapper : public OptionValueConstraint<T>{
        CLASS_NAME(AndWrapper);
        USE_ALLOCATOR(AndWrapper);
        AndWrapper(OptionValueConstraint<T>* l, OptionValueConstraint<T>* r) : left(l),right(r) {}
        bool check(OptionValue<T>* value){
            return left->check(value) && right->check(value);
        }
        vstring msg(OptionValue<T>* value){ return left->msg(value) + " and " + right->msg(value); }
        
        OptionValueConstraint<T>* left;
        OptionValueConstraint<T>* right;
    };
    
    template<typename T>
    struct Equal : public OptionValueConstraint<T>{
        CLASS_NAME(Equal);
        USE_ALLOCATOR(Equal);
        Equal(T gv) : _goodvalue(gv) {}
        bool check(OptionValue<T>* value){
            return value->actualValue == _goodvalue;
        }
        vstring msg(OptionValue<T>* value){
            return value->longName+"("+value->getStringOfActual()+") is equal to " + value->getStringOfValue(_goodvalue);
        }
        T _goodvalue;
    };
    template<typename T>
    static OptionValueConstraint<T>* equal(T bv){
        return new Equal<T>(bv);
    }
    
    template<typename T>
    struct NotEqual : public OptionValueConstraint<T>{
        CLASS_NAME(NotEqual);
        USE_ALLOCATOR(NotEqual);
        NotEqual(T bv) : _badvalue(bv) {}
        bool check(OptionValue<T>* value){
            return value->actualValue != _badvalue;
        }
        vstring msg(OptionValue<T>* value){ return value->longName+"("+value->getStringOfActual()+") is not equal to " + value->getStringOfValue(_badvalue); }
        T _badvalue;
    };
    template<typename T>
    static OptionValueConstraint<T>* notEqual(T bv){
        return new NotEqual<T>(bv);
    }
    
    
    
    template<typename T>
    struct LessThan : public OptionValueConstraint<T>{
        CLASS_NAME(LessThan);
        USE_ALLOCATOR(LessThan);
        LessThan(T gv,bool eq=false) : _goodvalue(gv), _orequal(eq) {}
        bool check(OptionValue<T>* value){
            return (value->actualValue < _goodvalue || (_orequal && value->actualValue==_goodvalue));
        }
        vstring msg(OptionValue<T>* value){
            if(_orequal) return value->longName+"("+value->getStringOfActual()+") is less than or equal to " + value->getStringOfValue(_goodvalue);
            return value->longName+"("+value->getStringOfActual()+") is less than "+ value->getStringOfValue(_goodvalue);
        }
        
        T _goodvalue;
        bool _orequal;
    };
    template<typename T>
    static OptionValueConstraint<T>* lessThan(T bv){
        return new LessThan<T>(bv,false);
    }
    template<typename T>
    static OptionValueConstraint<T>* lessThanEq(T bv){
        return new LessThan<T>(bv,true);
    }
    
    
    
    template<typename T>
    struct GreaterThan : public OptionValueConstraint<T>{
        CLASS_NAME(GreaterThan);
        USE_ALLOCATOR(GreaterThan);
        GreaterThan(T gv,bool eq=false) : _goodvalue(gv), _orequal(eq) {}
        bool check(OptionValue<T>* value){
            return (value->actualValue > _goodvalue || (_orequal && value->actualValue==_goodvalue));
        }
        
        vstring msg(OptionValue<T>* value){
            if(_orequal) return value->longName+"("+value->getStringOfActual()+") is greater than or equal to " + value->getStringOfValue(_goodvalue);
            return value->longName+"("+value->getStringOfActual()+") is greater than "+ value->getStringOfValue(_goodvalue);
        }
        
        T _goodvalue;
        bool _orequal;
    };
    template<typename T>
    static OptionValueConstraint<T>* greaterThan(T bv){
        return new GreaterThan<T>(bv,false);
    }
    template<typename T>
    static OptionValueConstraint<T>* greaterThanEq(T bv){
        return new GreaterThan<T>(bv,true);
    }
    
   
    
    template<typename T>
    struct IfConstraint;
    
    template<typename T>
    struct IfThenConstraint : public OptionValueConstraint<T>{
        CLASS_NAME(IfThenConstraint);
        USE_ALLOCATOR(IfThenConstraint);
        
        IfThenConstraint(OptionValueConstraint<T>* ic, OptionValueConstraint<T>* c) :
        if_con(ic), then_con(c) {}
        
        bool check(OptionValue<T>* value){
            ASS(then_con);
            return !if_con->check(value) || then_con->check(value);
        }
        
        vstring msg(OptionValue<T>* value){
            return "if "+if_con->msg(value)+" then "+ then_con->msg(value);
        }
        
        OptionValueConstraint<T>* if_con;
        OptionValueConstraint<T>* then_con;
    };
    
    template<typename T>
    struct IfConstraint {
        CLASS_NAME(IfConstraint);
        USE_ALLOCATOR(IfConstraint);
        IfConstraint(OptionValueConstraint<T>* c) :if_con(c) {}
        
        OptionValueConstraint<T>* then(OptionValueConstraint<T>* c){
            return new IfThenConstraint<T>(if_con,c);
        }
        template<typename S>
        OptionValueConstraint<T>* then(WrappedConstraint<S>* c){
            return new IfThenConstraint<T>(if_con,new UnWrappedConstraint<T,S>(c));
        }
        
        OptionValueConstraint<T>* if_con;
    };
    
    template<typename T>
    static IfConstraint<T> If(OptionValueConstraint<T>* c){
        return IfConstraint<T>(c);
    }
    template<typename T,typename S>
    static IfConstraint<T> If(WrappedConstraint<S>* c){
        return IfConstraint<T>(new UnWrappedConstraint<T,S>(c));
    }

    template<typename T>
    static OptionValueConstraint<T>* ifOnThen(OptionValueConstraint<T>* c){
      IfConstraint<bool> ifc(equal(true));
      return ifc.then(c);
    }
    template<typename T>
    static OptionValueConstraint<T>* ifOnThen(WrappedConstraint<T>* c){
      IfConstraint<bool> ifc(equal(true));
      return ifc.then(c);
    }

   
    
    template<typename T>
    struct NotDefaultConstraint : public OptionValueConstraint<T> {
        NotDefaultConstraint() {}
        
        bool check(OptionValue<T>* value){
            return value->defaultValue != value->actualValue;
        }
        vstring msg(OptionValue<T>* value) { return value->longName+"("+value->getStringOfActual()+") is not default("+value->getStringOfValue(value->defaultValue)+")";}
    };
    struct NotDefaultRatioConstraint : public OptionValueConstraint<int> {
        NotDefaultRatioConstraint() {}
        
        bool check(OptionValue<int>* value){
            RatioOptionValue* rvalue = static_cast<RatioOptionValue*>(value);
            return (rvalue->defaultValue != rvalue->actualValue ||
                    rvalue->defaultOtherValue != rvalue->otherValue);
        }
        vstring msg(OptionValue<int>* value) { return value->longName+"("+value->getStringOfActual()+") is not default";}
        
    };
    
    
    template<typename T>
    static OptionValueConstraint<T>* isNotDefault(){
        return new NotDefaultConstraint<T>();
    }
    
    static OptionValueConstraint<int>* isNotDefaultRatio(){
        return new NotDefaultRatioConstraint();
    }
    
    
   
    
    struct OptionProblemConstraint{
        virtual bool check(Property* p) = 0;
        virtual vstring msg() = 0;
    };
    
    struct CategoryCondition : OptionProblemConstraint{
        CategoryCondition(Property::Category c,bool h) : cat(c), has(h) {}
        bool check(Property*p){
            CALL("Options::CategoryCondition::check");
            ASS(p);
            return has ? p->category()==cat : p->category()!=cat;
        }
        vstring msg(){ 
          vstring m =" not useful for property ";
          if(has) m+="not";
          return m+" in category "+Property::categoryToString(cat);
        }
        Property::Category cat;
        bool has;
    };
    struct UsesEquality : OptionProblemConstraint{
        bool check(Property*p){
          CALL("Options::UsesEquality::check");
          ASS(p)
          return (p->equalityAtoms() != 0);
        }
        vstring msg(){ return " only useful with equality"; }
    };
    struct HasNonUnits : OptionProblemConstraint{
        bool check(Property*p){
          CALL("Options::HasNonUnits::check");
          return (p->clauses()-p->unitClauses())!=0; 
        }
        vstring msg(){ return " only useful with non-unit clauses"; }
    };
    struct HasPredicates : OptionProblemConstraint{
        bool check(Property*p){
          CALL("Options::HasPredicates::check");
          return (p->category()==Property::PEQ || p->category()==Property::UEQ);
        }
        vstring msg(){ return " only useful with predicates"; }
    };
    struct AtomConstraint : OptionProblemConstraint{
      AtomConstraint(int a,bool g) : atoms(a),greater(g) {}
      int atoms;
      bool greater;
      bool check(Property*p){ 
        CALL("Options::AtomConstraint::check");
        return greater ? p->atoms()>atoms : p->atoms()<atoms;
      }
          
      vstring msg(){ 
        vstring m = " not with ";
        if(greater){ m+="more";}else{m+="less";}
        return m+" than "+Lib::Int::toString(atoms)+" atoms";
      }
    };

    
    static OptionProblemConstraint* notWithCat(Property::Category c){
      return new CategoryCondition(c,false);
    }
    static OptionProblemConstraint* hasCat(Property::Category c){
      return new CategoryCondition(c,true);
    }
    static OptionProblemConstraint* hasEquality(){ return new UsesEquality; }
    static OptionProblemConstraint* hasNonUnits(){ return new HasNonUnits; }
    static OptionProblemConstraint* hasPredicates(){ return new HasPredicates; }
    static OptionProblemConstraint* atomsMoreThan(int a){
      return new AtomConstraint(a,true);
    }
    static OptionProblemConstraint* atomsLessThan(int a){
      return new AtomConstraint(a,false);
    }


    
    
    
    
    

    struct OptionHasValue : OptionProblemConstraint{
      OptionHasValue(vstring ov,vstring v) : option_value(ov),value(v) {}
      bool check(Property*p);
      vstring msg(){ return option_value+" has value "+value; } 
      vstring option_value;
      vstring value; 
    };

    struct ManyOptionProblemConstraints : OptionProblemConstraint {
      ManyOptionProblemConstraints(bool a) : is_and(a) {}

      bool check(Property*p){
        CALL("Options::ManyOptionProblemConstraints::check");
        bool res = is_and;
        Stack<OptionProblemConstraint*>::Iterator it(cons);
        while(it.hasNext()){ 
          bool n=it.next()->check(p);res = is_and ? (res && n) : (res || n);}
        return res;
      } 

      vstring msg(){
        vstring res="";
        Stack<OptionProblemConstraint*>::Iterator it(cons);
        if(it.hasNext()){ res=it.next()->msg();}
        while(it.hasNext()){ res+=",and\n"+it.next()->msg();}
        return res;
      }

      void add(OptionProblemConstraint* c){ cons.push(c);}
      Stack<OptionProblemConstraint*> cons; 
      bool is_and;
    };

    static OptionProblemConstraint* And(OptionProblemConstraint* left,
                                        OptionProblemConstraint* right){
       ManyOptionProblemConstraints* c = new ManyOptionProblemConstraints(true);
       c->add(left);c->add(right);
       return c;
    }
    static OptionProblemConstraint* And(OptionProblemConstraint* left,
                                        OptionProblemConstraint* mid,
                                        OptionProblemConstraint* right){
       ManyOptionProblemConstraints* c = new ManyOptionProblemConstraints(true);
       c->add(left);c->add(mid);c->add(right);
       return c;
    }
    static OptionProblemConstraint* Or(OptionProblemConstraint* left,
                                        OptionProblemConstraint* right){
       ManyOptionProblemConstraints* c = new ManyOptionProblemConstraints(false);
       c->add(left);c->add(right);
       return c;
    }
    static OptionProblemConstraint* Or(OptionProblemConstraint* left,
                                        OptionProblemConstraint* mid,
                                        OptionProblemConstraint* right){
       ManyOptionProblemConstraints* c = new ManyOptionProblemConstraints(false);
       c->add(left);c->add(mid);c->add(right);
       return c;
    }
    static OptionProblemConstraint* saNotInstGen();
    static OptionProblemConstraint* isBfnt();
    
  
  
  
  
  
  
public:
  bool encodeStrategy() const{ return _encode.actualValue;}
  RandomStrategy randomStrategy() const {return _randomStrategy.actualValue; }
  void setRandomStrategy(RandomStrategy newVal){ _randomStrategy.actualValue=newVal;}
  BadOption getBadOptionChoice() const { return _badOption.actualValue; }
  void setBadOptionChoice(BadOption newVal) { _badOption.actualValue = newVal; }
  vstring forcedOptions() const { return _forcedOptions.actualValue; }
  vstring forbiddenOptions() const { return _forbiddenOptions.actualValue; }
  vstring testId() const { return _testId.actualValue; }
  vstring protectedPrefix() const { return _protectedPrefix.actualValue; }
  Statistics statistics() const { return _statistics.actualValue; }
  void setStatistics(Statistics newVal) { _statistics.actualValue=newVal; }
  Proof proof() const { return _proof.actualValue; }
  bool proofChecking() const { return _proofChecking.actualValue; }
  int naming() const { return _naming.actualValue; }
  bool fmbIncremental() const { return _fmbIncremental.actualValue; }
  bool fmbNonGroundDefs() const { return _fmbNonGroundDefs.actualValue; }
  bool fmbSortInference() const { return _fmbSortInference.actualValue; }
  unsigned fmbStartSize() const { return _fmbStartSize.actualValue;}
  bool fmbStartWithConstants() const { return _fmbStartWithConstants.actualValue; }
  float fmbSymmetryRatio() const { return _fmbSymmetryRatio.actualValue; }
  bool flattenTopLevelConjunctions() const { return _flattenTopLevelConjunctions.actualValue; }
  bool eprPreservingNaming() const { return _eprPreservingNaming.actualValue; }
  
  bool eprPreservingSkolemization() const { return _eprPreservingSkolemization.actualValue; }
  
  bool eprRestoringInlining() const { return _eprRestoringInlining.actualValue; }
  
  InliningMode predicateDefinitionInlining() const { return _predicateDefinitionInlining.actualValue; }
  
  bool predicateDefinitionMerging() const { return _predicateDefinitionMerging.actualValue; }
  
  PredicateEquivalenceDiscoveryMode predicateEquivalenceDiscovery() const { return _predicateEquivalenceDiscovery.actualValue; }
  void setPredicateEquivalenceDiscovery(PredicateEquivalenceDiscoveryMode newValue) { _predicateEquivalenceDiscovery.actualValue = newValue; }
  bool predicateEquivalenceDiscoveryAddImplications() const { return _predicateEquivalenceDiscoveryAddImplications.actualValue; }
  bool predicateEquivalenceDiscoveryRandomSimulation() const { return _predicateEquivalenceDiscoveryRandomSimulation.actualValue; }
  int predicateEquivalenceDiscoverySatConflictLimit() const { return _predicateEquivalenceDiscoverySatConflictLimit.actualValue; }
  bool predicateIndexIntroduction() const { return _predicateIndexIntroduction.actualValue; }
  
  bool aigBddSweeping() const { return _aigBddSweeping.actualValue; }
  bool aigConditionalRewriting() const { return _aigConditionalRewriting.actualValue; }
  bool aigDefinitionIntroduction() const { return _aigDefinitionIntroduction.actualValue; }
  unsigned aigDefinitionIntroductionThreshold() const { return _aigDefinitionIntroductionThreshold.actualValue; }
  bool aigFormulaSharing() const { return _aigFormulaSharing.actualValue; }
  bool aigInliner() const { return _aigInliner.actualValue; }
  Mode mode() const { return _mode.actualValue; }
  InputSyntax inputSyntax() const { return _inputSyntax.actualValue; }
  
  bool normalize() const { return _normalize.actualValue; }
  void setNormalize(bool normalize) { _normalize.actualValue = normalize; }
  void setNaming(int n){ _naming.actualValue = n;} 
  vstring include() const { return _include.actualValue; }
  void setInclude(vstring val) { _include.actualValue = val; }
  vstring logFile() const { return _logFile.actualValue; }
  vstring inputFile() const { return _inputFile.actualValue; }
  int randomSeed() const { return _randomSeed.actualValue; }
  int rowVariableMaxLength() const { return _rowVariableMaxLength.actualValue; }
  
  bool printClausifierPremises() const { return _printClausifierPremises.actualValue; }
  bool showActive() const { return _showActive.actualValue; }
  bool showBlocked() const { return _showBlocked.actualValue; }
  bool showDefinitions() const { return _showDefinitions.actualValue; }
  InterpolantMode showInterpolant() const { return _showInterpolant.actualValue; }
  bool showNew() const { return _showNew.actualValue; }
  bool showNewPropositional() const { return _showNewPropositional.actualValue; }
  bool showNonconstantSkolemFunctionTrace() const { return _showNonconstantSkolemFunctionTrace.actualValue; }
  void setShowNonconstantSkolemFunctionTrace(bool newVal) { _showNonconstantSkolemFunctionTrace.actualValue = newVal; }
  bool showOptions() const { return _showOptions.actualValue; }
  bool showExperimentalOptions() const { return _showExperimentalOptions.actualValue; }
  bool showHelp() const { return _showHelp.actualValue; }
  vstring explainOption() const { return _explainOption.actualValue; }
  bool showPassive() const { return _showPassive.actualValue; }
  bool showReductions() const { return _showReductions.actualValue; }
  bool showPreprocessing() const { return _showPreprocessing.actualValue; }
  bool showSkolemisations() const { return _showSkolemisations.actualValue; }
  bool showSymbolElimination() const { return _showSymbolElimination.actualValue; }
  bool showTheoryAxioms() const { return _showTheoryAxioms.actualValue; }
#if VZ3
  bool showZ3() const { return _showZ3.actualValue; }
#endif
  bool unusedPredicateDefinitionRemoval() const { return _unusedPredicateDefinitionRemoval.actualValue; }
  void setUnusedPredicateDefinitionRemoval(bool newVal) { _unusedPredicateDefinitionRemoval.actualValue = newVal; }
  bool weightIncrement() const { return _weightIncrement.actualValue; }
  bool useDM() const { return _use_dm.actualValue; }
  SatSolver satSolver() const { return _satSolver.actualValue; }
  
  bool satLingelingSimilarModels() const { return _satLingelingSimilarModels.actualValue; }
  bool satLingelingIncremental() const { return _satLingelingIncremental.actualValue; }
  SaturationAlgorithm saturationAlgorithm() const { return _saturationAlgorithm.actualValue; }
  void setSaturationAlgorithm(SaturationAlgorithm newVal) { _saturationAlgorithm.actualValue = newVal; }
  int selection() const { return _selection.actualValue; }
  void setSelection(int v) { _selection.actualValue=v;}
  vstring latexOutput() const { return _latexOutput.actualValue; }
  bool latexUseDefault() const { return _latexUseDefaultSymbols.actualValue; }
  LiteralComparisonMode literalComparisonMode() const { return _literalComparisonMode.actualValue; }
  bool forwardSubsumptionResolution() const { return _forwardSubsumptionResolution.actualValue; }
  
  Demodulation forwardDemodulation() const { return _forwardDemodulation.actualValue; }
  bool binaryResolution() const { return _binaryResolution.actualValue; }
  bool bfnt() const { return _bfnt.actualValue; }
  void setBfnt(bool newVal) { _bfnt.actualValue = newVal; }
  URResolution unitResultingResolution() const { return _unitResultingResolution.actualValue; }
  bool hyperSuperposition() const { return _hyperSuperposition.actualValue; }
  bool arityCheck() const { return _arityCheck.actualValue; }
  
  Demodulation backwardDemodulation() const { return _backwardDemodulation.actualValue; }
  bool demodulationRedundancyCheck() const { return _demodulationRedundancyCheck.actualValue; }
  
  Subsumption backwardSubsumption() const { return _backwardSubsumption.actualValue; }
  
  Subsumption backwardSubsumptionResolution() const { return _backwardSubsumptionResolution.actualValue; }
  bool forwardSubsumption() const { return _forwardSubsumption.actualValue; }
  bool forwardLiteralRewriting() const { return _forwardLiteralRewriting.actualValue; }
  vstring lingvaAdditionalInvariants() const {return _lingvaAdditionalInvariants.actualValue; }
  int lrsFirstTimeCheck() const { return _lrsFirstTimeCheck.actualValue; }
  int lrsWeightLimitOnly() const { return _lrsWeightLimitOnly.actualValue; }
  int simulatedTimeLimit() const { return _simulatedTimeLimit.actualValue; }
  void setSimulatedTimeLimit(int newVal) { _simulatedTimeLimit.actualValue = newVal; }
  int maxInferenceDepth() const { return _maxInferenceDepth.actualValue; }
  SymbolPrecedence symbolPrecedence() const { return _symbolPrecedence.actualValue; }
  
  int timeLimitInDeciseconds() const { return _timeLimitInDeciseconds.actualValue; }
  size_t memoryLimit() const { return _memoryLimit.actualValue; }
  int inequalitySplitting() const { return _inequalitySplitting.actualValue; }
  long maxActive() const { return _maxActive.actualValue; }
  long maxAnswers() const { return _maxAnswers.actualValue; }
  
  long maxPassive() const { return _maxPassive.actualValue; }
  int maxWeight() const { return _maxWeight.actualValue; }
  int ageRatio() const { return _ageWeightRatio.actualValue; }
  void setAgeRatio(int v){ _ageWeightRatio.actualValue = v; }
  int weightRatio() const { return _ageWeightRatio.otherValue; }
  void setWeightRatio(int v){ _ageWeightRatio.otherValue = v; }
  bool superpositionFromVariables() const { return _superpositionFromVariables.actualValue; }
  bool equalityPropagation() const { return _equalityPropagation.actualValue; }
  
  EqualityProxy equalityProxy() const { return _equalityProxy.actualValue; }
  RuleActivity equalityResolutionWithDeletion() const { return _equalityResolutionWithDeletion.actualValue; }
  ExtensionalityResolution extensionalityResolution() const { return _extensionalityResolution.actualValue; }
  unsigned extensionalityMaxLength() const { return _extensionalityMaxLength.actualValue; }
  bool extensionalityAllowPosEq() const { return _extensionalityAllowPosEq.actualValue; }
  float nongoalWeightCoefficient() const { return _nonGoalWeightCoefficient.actualValue; }
  Sos sos() const { return _sos.actualValue; }
  
  FunctionDefinitionElimination functionDefinitionElimination() const { return _functionDefinitionElimination.actualValue; }
  bool outputAxiomNames() const { return _outputAxiomNames.actualValue; }
  void setOutputAxiomNames(bool newVal) { _outputAxiomNames.actualValue = newVal; }
  QuestionAnsweringMode questionAnswering() const { return _questionAnswering.actualValue; }
  vstring xmlOutput() const { return _xmlOutput.actualValue; }
  vstring thanks() const { return _thanks.actualValue; }
  void setQuestionAnswering(QuestionAnsweringMode newVal) { _questionAnswering.actualValue = newVal; }
  bool globalSubsumption() const { return _globalSubsumption.actualValue; }
  GlobalSubsumptionSatSolverPower globalSubsumptionSatSolverPower() const { return _globalSubsumptionSatSolverPower.actualValue; }
  GlobalSubsumptionExplicitMinim globalSubsumptionExplicitMinim() const { return _globalSubsumptionExplicitMinim.actualValue; }
  GlobalSubsumptionAvatarAssumptions globalSubsumptionAvatarAssumptions() const { return _globalSubsumptionAvatarAssumptions.actualValue; }

 
  bool ignoreMissing() const { return _ignoreMissing.actualValue; }
 
  
  bool increasedNumeralWeight() const { return _increasedNumeralWeight.actualValue; }
  bool theoryAxioms() const { return _theoryAxioms.actualValue; }
  
  bool interpretedSimplification() const { return _interpretedSimplification.actualValue; }
  
  Condensation condensation() const { return _condensation.actualValue; }
  RuleActivity generalSplitting() const { return _generalSplitting.actualValue; }
  vstring namePrefix() const { return _namePrefix.actualValue; }
  bool timeStatistics() const { return _timeStatistics.actualValue; }
  bool splitting() const { return _splitting.actualValue; }
  bool nonliteralsInClauseWeight() const { return _nonliteralsInClauseWeight.actualValue; }
  unsigned sineDepth() const { return _sineDepth.actualValue; }
  unsigned sineGeneralityThreshold() const { return _sineGeneralityThreshold.actualValue; }
  SineSelection sineSelection() const { return _sineSelection.actualValue; }
  void setSineSelection(SineSelection val) { _sineSelection.actualValue=val; }
  float sineTolerance() const { return _sineTolerance.actualValue; }
  bool smtlibConsiderIntsReal() const { return _smtlibConsiderIntsReal.actualValue; }
  
  bool smtlibFletAsDefinition() const { return _smtlibFletAsDefinition.actualValue; }
  bool smtlibIntroduceAIGNames() const { return _smtlibIntroduceAIGNames.actualValue; }

  bool colorUnblocking() const { return _colorUnblocking.actualValue; }
  bool hornRevealing() const { return _hornRevealing.actualValue; }
  bool trivialPredicateRemoval() const { return _trivialPredicateRemoval.actualValue; }

  bool tabulationBwRuleSubsumptionResolutionByLemmas() const { return _tabulationBwRuleSubsumptionResolutionByLemmas.actualValue; }
  bool tabulationFwRuleSubsumptionResolutionByLemmas() const { return _tabulationFwRuleSubsumptionResolutionByLemmas.actualValue; }
  int tabulationGoalAgeRatio() const { return _tabulationGoalAgeWeightRatio.actualValue; }
  int tabulationGoalWeightRatio() const { return _tabulationGoalAgeWeightRatio.otherValue; }
  int tabulationGoalRatio() const { return _tabulationGoalLemmaRatio.actualValue; }
  int tabulationLemmaRatio() const { return _tabulationGoalLemmaRatio.otherValue; }
  bool tabulationInstantiateProducingRules() const { return _tabulationInstantiateProducingRules.actualValue; }
  int tabulationLemmaAgeRatio() const { return _tabulationLemmaAgeWeightRatio.actualValue; }
  int tabulationLemmaWeightRatio() const { return _tabulationLemmaAgeWeightRatio.otherValue; }

  Instantiation instantiation() const { return _instantiation.actualValue; }
  bool theoryFlattening() const { return _theoryFlattening.actualValue; }

  float instGenBigRestartRatio() const { return _instGenBigRestartRatio.actualValue; }
  bool instGenInprocessing() const { return _instGenInprocessing.actualValue; }
  bool instGenPassiveReactivation() const { return _instGenPassiveReactivation.actualValue; }
  int instGenResolutionRatioInstGen() const { return _instGenResolutionInstGenRatio.actualValue; }
  int instGenResolutionRatioResolution() const { return _instGenResolutionInstGenRatio.otherValue; }
  int instGenRestartPeriod() const { return _instGenRestartPeriod.actualValue; }
  float instGenRestartPeriodQuotient() const { return _instGenRestartPeriodQuotient.actualValue; }
  int instGenSelection() const { return _instGenSelection.actualValue; }
  bool instGenWithResolution() const { return _instGenWithResolution.actualValue; }
  bool instGenUseHashingVariantIndex() const { return _instGenUseHashingVariantIndex.actualValue; }

  float satClauseActivityDecay() const { return _satClauseActivityDecay.actualValue; }
  SatClauseDisposer satClauseDisposer() const { return _satClauseDisposer.actualValue; }
  bool satLearntMinimization() const { return _satLearntMinimization.actualValue; }
  bool satLearntSubsumptionResolution() const { return _satLearntSubsumptionResolution.actualValue; }
  int satRestartFixedCount() const { return _satRestartFixedCount.actualValue; }
  float satRestartGeometricIncrease() const { return _satRestartGeometricIncrease.actualValue; }
  int satRestartGeometricInit() const { return _satRestartGeometricInit.actualValue; }
  int satRestartLubyFactor() const { return _satRestartLubyFactor.actualValue; }
  float satRestartMinisatIncrease() const { return _satRestartMinisatIncrease.actualValue; }
  int satRestartMinisatInit() const { return _satRestartMinisatInit.actualValue; }
  SatRestartStrategy satRestartStrategy() const { return _satRestartStrategy.actualValue; }
  float satVarActivityDecay() const { return _satVarActivityDecay.actualValue; }
  SatVarSelector satVarSelector() const { return _satVarSelector.actualValue; }

  Niceness nicenessOption() const { return _nicenessOption.actualValue; }

  void setMemoryLimit(size_t newVal) { _memoryLimit.actualValue = newVal; }
  
  void setTimeLimitInSeconds(int newVal) { _timeLimitInDeciseconds.actualValue = 10*newVal; }
  void setTimeLimitInDeciseconds(int newVal) { _timeLimitInDeciseconds.actualValue = newVal; }
  
  int getWhileNumber(){return _whileNumber.actualValue;}
  int getFunctionNumber(){return _functionNumber.actualValue;}

  int nonGoalWeightCoeffitientNumerator() const { return _nonGoalWeightCoefficient.numerator; }
  int nonGoalWeightCoeffitientDenominator() const { return _nonGoalWeightCoefficient.denominator; }

  bool splitAtActivation() const{ return _splitAtActivation.actualValue; }
  SplittingNonsplittableComponents splittingNonsplittableComponents() const { return _splittingNonsplittableComponents.actualValue; }
  SplittingAddComplementary splittingAddComplementary() const { return _splittingAddComplementary.actualValue; }
  SplittingMinimizeModel splittingMinimizeModel() const { return _splittingMinimizeModel.actualValue; }
  SplittingLiteralPolarityAdvice splittingLiteralPolarityAdvice() const { return _splittingLiteralPolarityAdvice.actualValue; }
  SplittingDeleteDeactivated splittingDeleteDeactivated() const { return _splittingDeleteDeactivated.actualValue;}
  bool splittingHandleZeroImplied() const { return _splittingHandleZeroImplied.actualValue;}
  bool splittingFastRestart() const { return _splittingFastRestart.actualValue; }
  bool splittingBufferedSolver() const { return _splittingBufferedSolver.actualValue; }
  int splittingFlushPeriod() const { return _splittingFlushPeriod.actualValue; }
  float splittingFlushQuotient() const { return _splittingFlushQuotient.actualValue; }
  bool splittingEagerRemoval() const { return _splittingEagerRemoval.actualValue; }
  SplittingCongruenceClosure splittingCongruenceClosure() const { return _splittingCongruenceClosure.actualValue; }
  CCUnsatCores ccUnsatCores() const { return _ccUnsatCores.actualValue; }

  void setProof(Proof p) { _proof.actualValue = p; }
  bool bpEquivalentVariableRemoval() const { return _equivalentVariableRemoval.actualValue; }
  unsigned bpMaximalPropagatedEqualityLength() const { return _maximalPropagatedEqualityLength.actualValue; }
  BPAlmostHalfBoundingRemoval bpAlmostHalfBoundingRemoval() const {return _bpAlmostHalfBoundingRemoval.actualValue;}
  bool bpFmElimination () const {return _bpFmElimination.actualValue;}
  unsigned bpAllowedFMBalance() const { return _bpAllowedFMBalance.actualValue; }
  BPAssignmentSelector bpAssignmentSelector() const {return _bpAssignmentSelector.actualValue; }
  bool bpCollapsingPropagation() const {return _bpCollapsingPropagation.actualValue; }
  unsigned bpUpdatesByOneConstraint() const {return _updatesByOneConstraint.actualValue; }
  bool bpConservativeAssignmentSelection() const {return _bpConservativeAssignmentSelection.actualValue; }
  BPConflictSelector bpConflictSelector() const {return _bpConflictSelector.actualValue; }
  bool backjumpTargetIsDecisionPoint() const { return _backjumpTargetIsDecisionPoint.actualValue; }
  bool bpPropagateAfterConflict() const {return _bpPropagateAfterConflict.actualValue; }
  BPVariableSelector bpVariableSelector() const {return _bpVariableSelector.actualValue; }
  bool bpSelectUnusedVariablesFirst() const {return _selectUnusedVariablesFirst.actualValue; }
  bool bpStartWithPrecise() const { return _bpStartWithPrecise.actualValue; }
  bool bpStartWithRational() const { return _bpStartWithRational.actualValue;}
    
private:
    
   
    struct LookupWrapper {
        
        LookupWrapper() {}
        
        private:
          LookupWrapper operator=(const LookupWrapper&){ NOT_IMPLEMENTED;}
        public:
        
        void insert(AbstractOptionValue* option_value){
            CALL("LookupWrapper::insert");
            ASS(!option_value->longName.empty());
            bool new_long =  _longMap.insert(option_value->longName,option_value);
            bool new_short = true;
            if(!option_value->shortName.empty()){
                new_short = _shortMap.insert(option_value->shortName,option_value);
            }
            if(!new_long || !new_short){ cout << "Bad " << option_value->longName << endl; }
            ASS(new_long && new_short);
        }
        AbstractOptionValue* findLong(vstring longName) const{
            CALL("LookupWrapper::findLong");
            if(!_longMap.find(longName)){ throw ValueNotFoundException(); }
            return _longMap.get(longName);
        }
        AbstractOptionValue* findShort(vstring shortName) const{
            CALL("LookupWrapper::findShort");
            if(!_shortMap.find(shortName)){ throw ValueNotFoundException(); }
            return _shortMap.get(shortName);
        }
        
        VirtualIterator<AbstractOptionValue*> values() const {
            return _longMap.range();
        }
        
    private:
        DHMap<vstring,AbstractOptionValue*> _longMap;
        DHMap<vstring,AbstractOptionValue*> _shortMap;
    };
    
    LookupWrapper _lookup;
    
    
    AbstractOptionValue* getOptionValueByName(vstring name) const{
        try{
          return _lookup.findLong(name);
        }
        catch(ValueNotFoundException&){
          try{
            return _lookup.findShort(name);
          }
          catch(ValueNotFoundException&){
            return 0;
          }
        }
    }
  
    Stack<vstring> getSimilarOptionNames(vstring name, bool is_short) const{

      Stack<vstring> similar_names;

      VirtualIterator<AbstractOptionValue*> options = _lookup.values();
      while(options.hasNext()){
        AbstractOptionValue* opt = options.next();
        vstring opt_name = is_short ? opt->shortName : opt->longName;
        size_t dif = 2;
        if(!is_short) dif += name.size()/4;
        if(name.size()!=0 && distance(name,opt_name) < dif)
          similar_names.push(opt_name);
      }

      return similar_names;
    }
    
    
    
    



  ChoiceOptionValue<RandomStrategy> _randomStrategy;
  DecodeOptionValue _decode;
  BoolOptionValue _encode;

  RatioOptionValue _ageWeightRatio;
  BoolOptionValue _aigBddSweeping;
  BoolOptionValue _aigConditionalRewriting;
  BoolOptionValue _aigDefinitionIntroduction;
  UnsignedOptionValue _aigDefinitionIntroductionThreshold;
  BoolOptionValue _aigFormulaSharing;
  BoolOptionValue _aigInliner;
  BoolOptionValue _arityCheck;
  
  BoolOptionValue _backjumpTargetIsDecisionPoint;
  ChoiceOptionValue<BadOption> _badOption;
  ChoiceOptionValue<Demodulation> _backwardDemodulation;
  ChoiceOptionValue<Subsumption> _backwardSubsumption;
  ChoiceOptionValue<Subsumption> _backwardSubsumptionResolution;
  BoolOptionValue _bfnt;
  BoolOptionValue _binaryResolution;
  BoolOptionValue _bpCollapsingPropagation;
  UnsignedOptionValue _bpAllowedFMBalance;
  ChoiceOptionValue<BPAlmostHalfBoundingRemoval> _bpAlmostHalfBoundingRemoval;
  ChoiceOptionValue<BPAssignmentSelector> _bpAssignmentSelector;
  ChoiceOptionValue<BPConflictSelector> _bpConflictSelector;
  BoolOptionValue _bpConservativeAssignmentSelection;
  BoolOptionValue _bpFmElimination;
  BoolOptionValue _bpPropagateAfterConflict;
  BoolOptionValue _bpStartWithPrecise;
  BoolOptionValue _bpStartWithRational;
  ChoiceOptionValue<BPVariableSelector> _bpVariableSelector;

  BoolOptionValue _colorUnblocking;
  ChoiceOptionValue<Condensation> _condensation;

  BoolOptionValue _demodulationRedundancyCheck;

  BoolOptionValue _eprPreservingNaming;
  BoolOptionValue _eprPreservingSkolemization;
  BoolOptionValue _eprRestoringInlining;
  BoolOptionValue _equalityPropagation;
  ChoiceOptionValue<EqualityProxy> _equalityProxy;
  ChoiceOptionValue<RuleActivity> _equalityResolutionWithDeletion;
  BoolOptionValue _equivalentVariableRemoval;
  ChoiceOptionValue<ExtensionalityResolution> _extensionalityResolution;
  UnsignedOptionValue _extensionalityMaxLength;
  BoolOptionValue _extensionalityAllowPosEq;
  
  BoolOptionValue _fmbIncremental;
  BoolOptionValue _fmbNonGroundDefs;
  BoolOptionValue _fmbSortInference;
  UnsignedOptionValue _fmbStartSize;
  BoolOptionValue _fmbStartWithConstants;
  FloatOptionValue _fmbSymmetryRatio;
  BoolOptionValue _flattenTopLevelConjunctions;
  StringOptionValue _forbiddenOptions;
  BoolOptionValue _forceIncompleteness;
  StringOptionValue _forcedOptions;
  ChoiceOptionValue<Demodulation> _forwardDemodulation;
  BoolOptionValue _forwardLiteralRewriting;
  BoolOptionValue _forwardSubsumption;
  BoolOptionValue _forwardSubsumptionResolution;
  ChoiceOptionValue<FunctionDefinitionElimination> _functionDefinitionElimination;
  IntOptionValue _functionNumber;
  
  ChoiceOptionValue<RuleActivity> _generalSplitting;
  BoolOptionValue _globalSubsumption;
  ChoiceOptionValue<GlobalSubsumptionSatSolverPower> _globalSubsumptionSatSolverPower;
  ChoiceOptionValue<GlobalSubsumptionExplicitMinim> _globalSubsumptionExplicitMinim;
  ChoiceOptionValue<GlobalSubsumptionAvatarAssumptions> _globalSubsumptionAvatarAssumptions;

  BoolOptionValue _hornRevealing;
  BoolOptionValue _hyperSuperposition;

 
  BoolOptionValue _ignoreMissing;
  StringOptionValue _include;
 
  BoolOptionValue _increasedNumeralWeight;
  IntOptionValue _inequalitySplitting;
  ChoiceOptionValue<InputSyntax> _inputSyntax;
  ChoiceOptionValue<Instantiation> _instantiation;
  FloatOptionValue _instGenBigRestartRatio;
  BoolOptionValue _instGenInprocessing;
  BoolOptionValue _instGenPassiveReactivation;
  RatioOptionValue _instGenResolutionInstGenRatio;
  
  IntOptionValue _instGenRestartPeriod;
  FloatOptionValue _instGenRestartPeriodQuotient;
  BoolOptionValue _instGenWithResolution;
  BoolOptionValue _instGenUseHashingVariantIndex;
  BoolOptionValue _interpretedSimplification;

  StringOptionValue _latexOutput;
  BoolOptionValue _latexUseDefaultSymbols;
  StringOptionValue _lingvaAdditionalInvariants;

  ChoiceOptionValue<LiteralComparisonMode> _literalComparisonMode;
  StringOptionValue _logFile;
  IntOptionValue _lrsFirstTimeCheck;
  BoolOptionValue _lrsWeightLimitOnly;

  LongOptionValue _maxActive;
  IntOptionValue _maxAnswers;
  IntOptionValue _maxInferenceDepth;
  LongOptionValue _maxPassive;
  IntOptionValue _maxWeight;
  UnsignedOptionValue _maximalPropagatedEqualityLength;
  UnsignedOptionValue _memoryLimit; 
  ChoiceOptionValue<Mode> _mode;

  StringOptionValue _namePrefix;
  IntOptionValue _naming;
  ChoiceOptionValue<Niceness> _nicenessOption;
  BoolOptionValue _nonliteralsInClauseWeight;
  BoolOptionValue _normalize;

  BoolOptionValue _outputAxiomNames;

  ChoiceOptionValue<InliningMode> _predicateDefinitionInlining;
  BoolOptionValue _predicateDefinitionMerging;
  ChoiceOptionValue<PredicateEquivalenceDiscoveryMode> _predicateEquivalenceDiscovery;
  BoolOptionValue _predicateEquivalenceDiscoveryAddImplications;
  BoolOptionValue _predicateEquivalenceDiscoveryRandomSimulation;
  IntOptionValue _predicateEquivalenceDiscoverySatConflictLimit;
  BoolOptionValue _predicateIndexIntroduction;
  BoolOptionValue _printClausifierPremises;
  StringOptionValue _problemName;
  ChoiceOptionValue<Proof> _proof;
  BoolOptionValue _proofChecking;
  
  StringOptionValue _protectedPrefix;

  ChoiceOptionValue<QuestionAnsweringMode> _questionAnswering;

  IntOptionValue _randomSeed;
  IntOptionValue _rowVariableMaxLength;

  FloatOptionValue _satClauseActivityDecay;
  ChoiceOptionValue<SatClauseDisposer> _satClauseDisposer;
  BoolOptionValue _satLearntMinimization;
  BoolOptionValue _satLearntSubsumptionResolution;
  IntOptionValue _satRestartFixedCount;
  FloatOptionValue _satRestartGeometricIncrease;
  IntOptionValue _satRestartGeometricInit;
  IntOptionValue _satRestartLubyFactor;
  FloatOptionValue _satRestartMinisatIncrease;
  IntOptionValue _satRestartMinisatInit;
  ChoiceOptionValue<SatRestartStrategy> _satRestartStrategy;
  FloatOptionValue _satVarActivityDecay;
  ChoiceOptionValue<SatVarSelector> _satVarSelector;
  ChoiceOptionValue<SatSolver> _satSolver;
  BoolOptionValue _satLingelingSimilarModels;
  BoolOptionValue _satLingelingIncremental;
  ChoiceOptionValue<SaturationAlgorithm> _saturationAlgorithm;
  BoolOptionValue _selectUnusedVariablesFirst;
  BoolOptionValue _showActive;
  BoolOptionValue _showBlocked;
  BoolOptionValue _showDefinitions;
  ChoiceOptionValue<InterpolantMode> _showInterpolant;
  BoolOptionValue _showNew;
  BoolOptionValue _showNewPropositional;
  BoolOptionValue _showNonconstantSkolemFunctionTrace;
  BoolOptionValue _showOptions;
  BoolOptionValue _showExperimentalOptions;
  BoolOptionValue _showHelp;
  StringOptionValue _explainOption;
  BoolOptionValue _showPassive;
  BoolOptionValue _showReductions;
  BoolOptionValue _showPreprocessing;
  BoolOptionValue _showSkolemisations;
  BoolOptionValue _showSymbolElimination;
  BoolOptionValue _showTheoryAxioms;
#if VZ3
  BoolOptionValue _showZ3;
#endif
  TimeLimitOptionValue _simulatedTimeLimit;
  UnsignedOptionValue _sineDepth;
  UnsignedOptionValue _sineGeneralityThreshold;
  ChoiceOptionValue<SineSelection> _sineSelection;
  FloatOptionValue _sineTolerance;
  BoolOptionValue _smtlibConsiderIntsReal;
  BoolOptionValue _smtlibFletAsDefinition;
  BoolOptionValue _smtlibIntroduceAIGNames;
  ChoiceOptionValue<Sos> _sos;
  BoolOptionValue _splitting;
  BoolOptionValue _splitAtActivation;
  ChoiceOptionValue<SplittingAddComplementary> _splittingAddComplementary;
  ChoiceOptionValue<SplittingCongruenceClosure> _splittingCongruenceClosure;
  ChoiceOptionValue<CCUnsatCores> _ccUnsatCores;
  BoolOptionValue _splittingEagerRemoval;
  UnsignedOptionValue _splittingFlushPeriod;
  FloatOptionValue _splittingFlushQuotient;
  ChoiceOptionValue<SplittingNonsplittableComponents> _splittingNonsplittableComponents;
  ChoiceOptionValue<SplittingMinimizeModel> _splittingMinimizeModel;
  ChoiceOptionValue<SplittingLiteralPolarityAdvice> _splittingLiteralPolarityAdvice;
  ChoiceOptionValue<SplittingDeleteDeactivated> _splittingDeleteDeactivated;
  BoolOptionValue _splittingHandleZeroImplied;
  BoolOptionValue _splittingFastRestart;
  BoolOptionValue _splittingBufferedSolver;

  ChoiceOptionValue<Statistics> _statistics;
  BoolOptionValue _superpositionFromVariables;
  ChoiceOptionValue<SymbolPrecedence> _symbolPrecedence;

  BoolOptionValue _tabulationBwRuleSubsumptionResolutionByLemmas;
  BoolOptionValue _tabulationFwRuleSubsumptionResolutionByLemmas;
  RatioOptionValue _tabulationGoalAgeWeightRatio;
  RatioOptionValue _tabulationGoalLemmaRatio;
  BoolOptionValue _tabulationInstantiateProducingRules;
  RatioOptionValue _tabulationLemmaAgeWeightRatio;
  StringOptionValue _testId;
  StringOptionValue _thanks;
  BoolOptionValue _theoryAxioms;
  BoolOptionValue _theoryFlattening;

 
  TimeLimitOptionValue _timeLimitInDeciseconds;
  BoolOptionValue _timeStatistics;
  BoolOptionValue _trivialPredicateRemoval;

  ChoiceOptionValue<URResolution> _unitResultingResolution;
  BoolOptionValue _unusedPredicateDefinitionRemoval;
  UnsignedOptionValue _updatesByOneConstraint;
  BoolOptionValue _use_dm;
  BoolOptionValue _weightIncrement;
  IntOptionValue _whileNumber;

  StringOptionValue _xmlOutput;

  OptionChoiceValues _tagNames;

  NonGoalWeightOptionValue _nonGoalWeightCoefficient;

  SelectionOptionValue _selection;
  SelectionOptionValue _instGenSelection;
    

  InputFileOptionValue _inputFile;

    
}; 


template<typename T,
         typename = typename std::enable_if<std::is_enum<T>::value>::type>
std::ostream& operator<< (std::ostream& str,const T& val)
{
  return str << static_cast<typename std::underlying_type<T>::type>(val);
}

}

#endif

