
/*
 * File PARSER_TKV.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"
#include "Lib/Int.hpp"

#include "Lib/IntNameTable.hpp"

#include "Lexer.hpp"
#include "PARSER_TKV.hpp"

using namespace Lib;
using namespace Shell;

Parser::Parser (Lexer& lexer)
  :
  _lexer(lexer),
  _tokenCursor(0),
  _noOfTokens(0)
{
} 


ParserException::ParserException (vstring message,const Token& token)
  : _message (message)
{
  _message += " in line ";
  _message += Int::toString(token.line);
  _message += " at ";
  _message += token.text;
} 


void ParserException::cry (ostream& out)
{
  out << "Parser exception: " << _message << '\n';
} 

void Parser::consumeToken (TokenType tt)
{
  CALL("Parser::consumeToken/1");

  if (currentToken1().tag != tt) {
    throw ParserException(Token::toString(tt) + " expected",
			  _tokens[_tokenCursor]);
  }
  consumeToken();
} 

void Parser::expectToken (TokenType tt)
{
  CALL("Parser::expectToken/1");

  if (currentToken().tag != tt) {
    throw ParserException(Token::toString(tt) + " expected",
			  _tokens[_tokenCursor]);
  }
} 

void Parser::expectToken (TokenType tt,vstring errorMessage)
{
  CALL("Parser::expectToken/2");

  if (currentToken().tag != tt) {
    throw ParserException(errorMessage,
			  _tokens[_tokenCursor]);
  }
} 

void Parser::expectKeyword (const char* keyword)
{
  CALL("Parser::expectKeyword");

  if (currentToken().text != keyword) {
    throw ParserException((vstring)"keyword '" + keyword + "' expected",
			  _tokens[_tokenCursor]);
  }
} 


void Parser::consumeToken ()
{
  CALL("Parser::consumeToken/0");

  ASS (_noOfTokens > 0);
  ASS (_tokenCursor < TOKEN_BUFFER_SIZE);

  _tokenCursor = (_tokenCursor+1) % TOKEN_BUFFER_SIZE;
  _noOfTokens--;
} 


void Parser::readToken ()
{
  CALL("Parser::readToken");

  ASS(_noOfTokens <= TOKEN_BUFFER_SIZE);

  if (_noOfTokens == TOKEN_BUFFER_SIZE) {
    _lexer.readToken(_tokens[0]);

    throw ParserException("token buffer capacity exceeded",_tokens[0]);
  }

  int tokenIndex = (_tokenCursor + _noOfTokens) % TOKEN_BUFFER_SIZE;
  _noOfTokens++;
  _lexer.readToken(_tokens[tokenIndex]);

} 


void Parser::readToken(TokenType tt)
{
  CALL("Parser::readToken/1");
  readToken();
  expectToken(tt);
} 

void Parser::readToken(TokenType tt,vstring errorMessage)
{
  CALL("Parser::readToken/2");
  readToken();
  expectToken(tt,errorMessage);
} 


Token& Parser::currentToken()
{
  CALL("Parser::currentToken");
  ASS(_noOfTokens > 0);

  return _tokens[_tokenCursor % TOKEN_BUFFER_SIZE];
} 

Token& Parser::lookAhead (int index)
{
  CALL("Parser::lookahead");

  ASS (index < TOKEN_BUFFER_SIZE);
  ASS(index > 0);

  while (_noOfTokens <= index) {
    readToken();
  }
  return _tokens[(_tokenCursor+index) % TOKEN_BUFFER_SIZE];
} 


int Parser::var (const Token& t)
{
  CALL("Parser::var");
  return _vars.insert(t.text);
} 


double Parser::real (const Token& token)
{
  CALL("Parser::real");
  double d;

  if (! Int::stringToDouble(token.text.c_str(),d)) {
    throw ParserException("incorrect floating point number",token);
  }
  return d;
} 


int Parser::integer (const Token& token)
{
  CALL("Parser::integer");

  int result;
  if (Int::stringToInt(const_cast<vstring&>(token.text),result)) {
    return result;
  }
  throw ParserException("incorrect integer",token);
} 

long long unsigned Parser::unsigned64 (const Token& token)
{
  CALL("Parser:unsigned64");

  long long unsigned result;

  if (Int::stringToUnsigned64(const_cast<vstring&>(token.text),result)) {
    return result;
  }
  throw ParserException("incorrect 64-bit unsigned",token);
} 

void Parser::terminate(vstring errorMessage)
{
  throw ParserException(errorMessage,_tokens[_tokenCursor]);
} 

