
/*
 * File Property.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "Lib/Int.hpp"
#include "Lib/Environment.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/SubformulaIterator.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Signature.hpp"

#include "Statistics.hpp"
#include "FunctionDefinition.hpp"
#include "Property.hpp"

using namespace Lib;
using namespace Kernel;
using namespace Shell;

Property::Property()
  : _goalClauses(0),
    _axiomClauses(0),
    _positiveEqualityAtoms(0),
    _equalityAtoms(0),
    _atoms(0),
    _goalFormulas(0),
    _axiomFormulas(0),
    _subformulas(0),
    _terms(0),
    _unitGoals(0),
    _unitAxioms(0),
    _hornGoals(0),
    _hornAxioms(0),
    _equationalClauses(0),
    _pureEquationalClauses(0),
    _groundUnitAxioms(0),
    _positiveAxioms(0),
    _groundPositiveAxioms(0),
    _groundGoals(0),
    _maxFunArity(0),
    _maxPredArity(0),
    _totalNumberOfVariables(0),
    _maxVariablesInClause(0),
    _props(0),
    _hasInterpreted(false),
    _hasInterpretedEquality(false),
    _hasNonDefaultSorts(false),
    _sortsUsed(0),
    _hasSpecialTermsOrLets(false),
    _hasFormulaItes(false),
    _allClausesGround(true),
    _allQuantifiersEssentiallyExistential(true)
{
  _interpretationPresence.init(Theory::MAX_INTERPRETED_ELEMENT+1, false);
  env.property = this;
} 

Property* Property::scan(UnitList* units)
{
  CALL("Property::scan");

  Property* prop = new Property;
  prop->add(units);
  return prop;
} 

Property::~Property()
{
  CALL("Property::~Property");
  if (this == env.property) {
    env.property = 0;
  }
}

void Property::add(UnitList* units)
{
  CALL("Property::add(UnitList*)");

  UnitList::Iterator us(units);
  while (us.hasNext()) {
    scan(us.next());
  }

  if (_allClausesGround && _allQuantifiersEssentiallyExistential) {
    addProp(PR_ESSENTIALLY_GROUND);
  } else if (hasProp(PR_ESSENTIALLY_GROUND)) {
    dropProp(PR_ESSENTIALLY_GROUND);
  }

  
  if (env.sorts->hasSort()) {
    addProp(PR_SORTS);
  }
    
  
  if (env.signature->strings()) {
    addProp(PR_HAS_STRINGS);
  }
  if (env.signature->integers()) {
    addProp(PR_HAS_INTEGERS);
  }
  if (env.signature->rationals()) {
    addProp(PR_HAS_RATS);
  }
  if (env.signature->reals()) {
    addProp(PR_HAS_REALS);
  }


  
  if (formulas() > 0) { 
    if (_equalityAtoms == 0) {
      _category = FNE;
    }
    else {
      _category = FEQ;
    }
  }
  
  else if (_maxFunArity == 0) { 
    if (_pureEquationalClauses == clauses()) { 
      if (clauses() == unitClauses()) { 
	_category = UEQ;
      }
      else {
	_category = PEQ;
      }
    }
    else {
      _category = EPR;
    }
  }
  
  else if (_equationalClauses == 0 ) { 
    if (clauses() == hornClauses()) { 
      _category = HNE;
    }
    else {
      _category = NNE;
    }
  }
  
  else if (_pureEquationalClauses == clauses()) { 
    if (clauses() == unitClauses()) { 
      _category = UEQ;
    }
    else {
      _category = PEQ;
    }
  }
  
  else if (clauses() == hornClauses()) { 
    _category = HEQ;
  }
  else {
    _category = NEQ;
  }
} 

void Property::scan(Unit* unit)
{
  CALL("Property::scan(const Unit*)");

  if (unit->isClause()) {
    scan(static_cast<Clause*>(unit));
  }
  else {
    scan(static_cast<FormulaUnit*>(unit));
  }
  if (! hasProp(PR_HAS_FUNCTION_DEFINITIONS)) {
    FunctionDefinition::Def* def =
      FunctionDefinition::isFunctionDefinition(*unit);
    if (def) {
      addProp(PR_HAS_FUNCTION_DEFINITIONS);
      FunctionDefinition::deleteDef(def);
    }
  }
} 

void Property::scan(Clause* clause)
{
  CALL("Property::scan(const Clause*)");

  int positiveLiterals = 0;
  int negativeLiterals = 0;
  int equationalLiterals = 0;
  int positiveEquationalLiterals = 0;
  int groundLiterals = 0;
  _variablesInThisClause = 0;

  for (int i = clause->length()-1;i >= 0;i--) {
    Literal* literal = (*clause)[i];
    if (literal->isPositive()) {
      positiveLiterals ++;
    }
    else {
      negativeLiterals ++;
    }

    if (literal->isEquality()) {
      equationalLiterals++;
      if (literal->isPositive()) {
	positiveEquationalLiterals++;
      }
    }

    scan(literal);

    if (literal->shared() && literal->ground()) {
      groundLiterals++;
    }
  }
  int literals = positiveLiterals + negativeLiterals;
  _atoms += literals;

  if ( equationalLiterals > 0 ) {
    _equationalClauses ++;
    _equalityAtoms += equationalLiterals;
    _positiveEqualityAtoms += positiveEquationalLiterals;
  }
  if ( literals == equationalLiterals ) {
    _pureEquationalClauses ++;
  }

  if (clause->inputType() == Unit::AXIOM) {
    _axiomClauses ++;
    if ( literals == 1) {
      _unitAxioms ++;
      if ( groundLiterals == 1) {
	_groundUnitAxioms ++;
      }
    }
    if (positiveLiterals <= 1) {
      _hornAxioms ++;
    }
    if (negativeLiterals == 0) {
      _positiveAxioms ++;
      if (literals == groundLiterals) {
	_groundPositiveAxioms ++;
      }
    }
  }
  else {
    _goalClauses ++;
    if ( literals == 1) {
      _unitGoals ++;
    }
    if (positiveLiterals <= 1) {
      _hornGoals ++;
    }
    if (literals == groundLiterals) {
      _groundGoals ++;
    }
  }

  _totalNumberOfVariables += _variablesInThisClause;
  if (_variablesInThisClause > _maxVariablesInClause) {
    _maxVariablesInClause = _variablesInThisClause;
  }
  if (! hasProp(PR_HAS_X_EQUALS_Y) && hasXEqualsY(clause)) {
    addProp(PR_HAS_X_EQUALS_Y);
  }

  if (_variablesInThisClause > 0) {
    _allClausesGround = false;
  }
} 


void Property::scan(FormulaUnit* unit)
{
  CALL("Property::scan(const FormulaUnit*)");

  if (unit->inputType() == Unit::AXIOM) {
    _axiomFormulas ++;
  }
  else {
    _goalFormulas++;
  }
  Formula* f = unit->formula();
  scan(f);
  if (! hasProp(PR_HAS_X_EQUALS_Y)) {
    if (hasXEqualsY(f)) {
      addProp(PR_HAS_X_EQUALS_Y);
    }
  }
} 


void Property::scan(Formula* formula)
{
  CALL("void Property::scan(const Formula&)");

  SubformulaIterator fs(formula);
  while (fs.hasNext()) {
    _subformulas++;
    int polarity;
    Formula* f = fs.next(polarity);

    switch(f->connective()) {
    case ITE:
      _hasFormulaItes = true;
      break;
    case FORMULA_LET:
    case TERM_LET:
      _hasSpecialTermsOrLets = true;
      break;
    case LITERAL:
    {
      _atoms++;
      Literal* lit = f->literal();
      if (lit->isEquality()) {
	_equalityAtoms++;
	if (lit->isPositive()) {
	  _positiveEqualityAtoms++;
	}
      }
      if (!lit->shared()) {
	_hasSpecialTermsOrLets = true;
      }
      scan(lit);
      break;
    }
    case FORALL:
      if (polarity != -1) {
        _allQuantifiersEssentiallyExistential = false;
      }
      break;
    case EXISTS:
      if (polarity != 1) {
        _allQuantifiersEssentiallyExistential = false;
      }
      break;
    default:
      break;
    }
  }
} 

void Property::scanSort(unsigned sort)
{
  CALL("Property::scanSort");

  if(!_usesSort.get(sort)){
    _sortsUsed++;
    _usesSort[sort]=true;
  }

  if (sort==Sorts::SRT_DEFAULT) {
    return;
  }
  _hasNonDefaultSorts = true;
  env.statistics->hasTypes=true;
  switch(sort) {
  case Sorts::SRT_INTEGER:
    addProp(PR_HAS_INTEGERS);
    break;
  case Sorts::SRT_RATIONAL:
    addProp(PR_HAS_RATS);
    break;
  case Sorts::SRT_REAL:
    addProp(PR_HAS_REALS);
    break;          
  }
}

void Property::scan(Literal* lit)
{
  CALL("Property::scan(const Literal*...)");

  if (lit->isEquality()) {
    scanSort(SortHelper::getEqualityArgumentSort(lit));
  }
  else {
    int arity = lit->arity();
    if (arity > _maxPredArity) {
      _maxPredArity = arity;
    }
    PredicateType* type = env.signature->getPredicate(lit->functor())->predType();
    for (int i=0; i<arity; i++) {
      scanSort(type->arg(i));
    }
  }

  scanForInterpreted(lit);
  scan(lit->args());

  if (!hasProp(PR_HAS_INEQUALITY_RESOLVABLE_WITH_DELETION) && lit->isEquality() && lit->shared()
     && lit->isNegative() && !lit->ground() &&
     ( ( lit->nthArgument(0)->isVar() &&
	 !lit->nthArgument(1)->containsSubterm(*lit->nthArgument(0)) ) ||
       ( lit->nthArgument(1)->isVar() &&
	 !lit->nthArgument(0)->containsSubterm(*lit->nthArgument(1)) ))) {
    addProp(PR_HAS_INEQUALITY_RESOLVABLE_WITH_DELETION);
  }
} 


void Property::scan(TermList* ts)
{
  CALL("Property::scan(TermList*))");

  static Stack<TermList*> stack(64);
  stack.reset();

  for (;;) {
    if (ts->isEmpty()) {
      if (stack.isEmpty()) {
	return;
      }
      ts = stack.pop();
    }
    
    _terms ++;
    if (ts->isVar()) {
      _variablesInThisClause++;
    }
    else { 
      Term* t = ts->term();
      if (t->isSpecial()) {
	scanSpecialTerm(t);
      }
      else {
	scanForInterpreted(t);

	int arity = t->arity();
	FunctionType* type = env.signature->getFunction(t->functor())->fnType();
	for (int i=0; i<arity; i++) {
	  scanSort(type->arg(i));
	}

	if (arity > _maxFunArity) {
	  _maxFunArity = arity;
	}
	if (arity) {
	  stack.push(t->args());
	}
      }
    }
    ts = ts->next();
  }
} 

void Property::scanSpecialTerm(Term* t)
{
  CALL("Property::scanSpecialTerm");

  Term::SpecialTermData* sd = t->getSpecialData();
  switch(t->functor()) {
  case Term::SF_TERM_ITE:
  {
    ASS_EQ(t->arity(),2);
    scan(sd->getCondition());
    scan(t->args());
    break;
  }
  case Term::SF_LET_FORMULA_IN_TERM:
  {
    ASS_EQ(t->arity(),1);
    scan(sd->getLhsLiteral());
    scan(sd->getRhsFormula());
    scan(t->args());
    break;
  }
  case Term::SF_LET_TERM_IN_TERM:
  {
    ASS_EQ(t->arity(),1);
    
    TermList aux[3];
    aux[0].makeEmpty();
    aux[1] = sd->getRhsTerm();
    aux[2] = sd->getLhsTerm();
    scan(aux+2);
    scan(t->args());
    break;
  }
  default:
    ASSERTION_VIOLATION;
  }
}

void Property::scanForInterpreted(Term* t)
{
  CALL("Property::scanInterpretation");

  Interpretation itp;
  if (t->isLiteral()) {
    Literal* lit = static_cast<Literal*>(t);
    if (!theory->isInterpretedPredicate(lit)) { return; }
    itp = theory->interpretPredicate(lit);
    if(itp==Theory::EQUAL){ 
      
      _hasInterpretedEquality=true;
      return; 
    }
  }
  else {
    if (!theory->isInterpretedFunction(t)) { return; }
    itp = theory->interpretFunction(t);
  }
  _hasInterpreted = true;
  _interpretationPresence[itp] = true;
  if(Theory::isConversionOperation(itp)){
    addProp(PR_NUMBER_CONVERSION);
    return;
  }
  if(Theory::isArrayOperation(itp)){
    
    return;
  }

  unsigned sort = Theory::getOperationSort(itp);
  if(Theory::isInequality(itp)){
    switch(sort){
      case Sorts::SRT_INTEGER : addProp(PR_INTEGER_COMPARISON);
        break;
      case Sorts::SRT_RATIONAL : addProp(PR_RAT_COMPARISON);
        break;
      case Sorts::SRT_REAL : addProp(PR_REAL_COMPARISON);
        break;
    }
  }
  else if(Theory::isLinearOperation(itp)){
    switch(sort){
      case Sorts::SRT_INTEGER : addProp(PR_INTEGER_LINEAR);
        break;
      case Sorts::SRT_RATIONAL : addProp(PR_RAT_LINEAR);
        break;
      case Sorts::SRT_REAL : addProp(PR_REAL_LINEAR);
        break;
    }
  }
  else if(Theory::isNonLinearOperation(itp)){
    switch(sort){
      case Sorts::SRT_INTEGER : addProp(PR_INTEGER_NONLINEAR);
        break;
      case Sorts::SRT_RATIONAL : addProp(PR_RAT_NONLINEAR);
        break;
      case Sorts::SRT_REAL : addProp(PR_REAL_NONLINEAR);
        break;
    }
  }
}

vstring Property::categoryString() const
{
  CALL("vstring Property::categoryString() const");
  return categoryToString(_category);
}
vstring Property::categoryToString(Category cat)
{
  switch (cat)
    {
    case NEQ:
      return "NEQ";
    case HEQ:
      return "HEQ";
    case PEQ:
      return "PEQ";
    case HNE:
      return "HNE";
    case NNE:
      return "NNE";
    case FEQ:
      return "FEQ";
    case FNE:
      return "FNE";
    case EPR:
      return "EPR";
    case UEQ:
      return "UEQ";
#if VDEBUG
    default:
      ASSERTION_VIOLATION;
      return "";
#endif
    }
} 


vstring Property::toString() const
{
  vstring result("TPTP class: ");
  result += categoryString() + "\n";

  if (clauses() > 0) {
    result += "Clauses: ";
    result += Int::toString(clauses());
    result += " (";
    result += Int::toString(_unitAxioms+_unitGoals);
    result += " unit, ";
    result += Int::toString(_goalClauses);
    result += " goal, ";
    result += Int::toString(_equationalClauses);
    result += " equational)\n";

    result += "Variables: ";
    result += Int::toString(_totalNumberOfVariables);
    result += " (";
    result += Int::toString(_maxVariablesInClause);
    result += " maximum in a single clause)\n";
  }

  if (formulas() > 0) {
    result += "Formulas: ";
    result += Int::toString(formulas());
    result += " (";
    result += Int::toString(_goalFormulas);
    result += " goal)\n";
    result += "Subformulas: ";
    result += Int::toString(_subformulas);
    result += "\n";
  }

  result += "Atoms: ";
  result += Int::toString(_atoms);
  result += " (";
  result += Int::toString(_equalityAtoms);
  result += " equality)\n";

  return result;
} 


bool Property::hasXEqualsY(const Clause* c)
{
  CALL("Property::hasXEqualsY (const Clause*)");

  for (int i = c->length()-1; i >= 0; i--) {
    const Literal* lit = (*c)[i];
    if (lit->isNegative()) {
      continue;
    }
    if (!lit->isEquality()) {
      continue;
    }
    const TermList* ts1 = lit->args();
    if (!ts1->isVar()) {
      continue;
    }
    const TermList* ts2 = ts1->next();
    if (ts2->isVar() &&
	ts1->var() != ts2->var()) {
      return true;
    }
  }
  return  false;
} 

bool Property::hasXEqualsY(const Formula* f)
{
  CALL("Property::hasXEqualsY (const Formula*)");

  MultiCounter posVars; 
  MultiCounter negVars; 

  Stack<const Formula*> forms;
  Stack<int> pols; 
  forms.push(f);
  pols.push(1);
  while (!forms.isEmpty()) {
    f = forms.pop();
    int pol = pols.pop();

    switch (f->connective()) {
    case LITERAL:
      {
	const Literal* lit = f->literal();
	if (lit->isNegative()) {
	  break;
	}
	if (!lit->isEquality()) {
	  break;
	}
	const TermList* ts1 = lit->args();
	if (!ts1->isVar()) {
	  break;
	}
	const TermList* ts2 = ts1->next();
	if (!ts2->isVar()) {
	  break;
	}
	Var v1 = ts1->var();
	Var v2 = ts2->var();
	if (v1 == v2) {
	  break;
	}
	if (!lit->isPositive()) {
	  pol = -pol;
	}
	if (pol >= 0 && posVars.get(v1) && posVars.get(v2)) {
	  return true;
	}
	if (pol <= 0 && negVars.get(v1) && negVars.get(v2)) {
	  return true;
	}
      }
      break;

    case AND:
    case OR:
      {
	FormulaList::Iterator fs(f->args());
	while (fs.hasNext()) {
	  forms.push(fs.next());
	  pols.push(pol);
	}
      }
      break;

    case IMP:
      forms.push(f->left());
      pols.push(-pol);
      forms.push(f->right());
      pols.push(pol);
      break;

    case IFF:
    case XOR:
      forms.push(f->left());
      pols.push(0);
      forms.push(f->right());
      pols.push(0);
      break;

    case NOT:
      forms.push(f->uarg());
      pols.push(-pol);
      break;

    case FORALL:
      
      if (pol >= 0) {
	Formula::VarList::Iterator vs(f->vars());
	while (vs.hasNext()) {
	  posVars.inc(vs.next());
	}
      }
      forms.push(f->qarg());
      pols.push(pol);
      break;

  case EXISTS:
      
      if (pol <= 0) {
	Formula::VarList::Iterator vs(f->vars());
	while (vs.hasNext()) {
	  posVars.inc(vs.next());
	}
      }
      forms.push(f->qarg());
      pols.push(pol);
      break;

    case ITE:
      forms.push(f->condArg());
      pols.push(0);
      forms.push(f->thenArg());
      pols.push(pol);
      forms.push(f->elseArg());
      pols.push(pol);
      break;

    case TERM_LET:
    case FORMULA_LET:
      
      break;
      
    case TRUE:
    case FALSE:
      break;

#if VDEBUG
    default:
      ASSERTION_VIOLATION;
#endif
    }
  }
  return false;
} 

vstring Property::toSpider(const vstring& problemName) const
{
  return (vstring)"UPDATE problem SET property="
    + Int::toString((int)_props)
    + ",category='"
    + categoryString()
    + "' WHERE name='"
    + problemName +
    + "';";
} 

