
/*
 * File Property.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */



#ifndef __Property__
#define __Property__

#include "Lib/DArray.hpp"
#include "Lib/Array.hpp"
#include "Kernel/Unit.hpp"
#include "Kernel/Theory.hpp"
#include "Lib/VString.hpp"

namespace Kernel {
  class Clause;
  class FormulaUnit;
  class Literal;
  class Term;
  class TermList;
  class Formula;
}

namespace Shell {

using namespace std;
using namespace Kernel;
using namespace Lib;

class Property
{
public:
 
  enum Category {
    NEQ,
    HEQ,
    PEQ,
    HNE,
    NNE,
    FEQ,
    FNE,
    EPR,
    UEQ
  };

  
 
  static const unsigned long PR_HAS_X_EQUALS_Y = 1ul; 
 
  static const unsigned long PR_HAS_FUNCTION_DEFINITIONS = 2ul; 
 
  static const unsigned long PR_HAS_SUBSET = 4ul; 
 
  static const unsigned long PR_HAS_EXTENSIONALITY = 8ul; 
 
  static const unsigned long PR_GROUP = 16ul; 
 
  static const unsigned long PR_RING = 32ul; 
 
  static const unsigned long PR_ROBBINS_ALGEBRA = 64ul; 
 
  static const unsigned long PR_NA_RING = 128ul; 
 
  static const unsigned long PR_BOOLEAN_ALGEBRA = 256ul; 
 
  static const unsigned long PR_LATTICE = 512ul; 
 
  static const unsigned long PR_LO_GROUP = 1024ul; 
 
  static const unsigned long PR_COMBINATOR_B = 2048ul; 
 
  static const unsigned long PR_COMBINATOR = 4096ul; 
 
  static const unsigned long PR_HAS_CONDENSED_DETACHMENT1 = 8192ul; 
 
  static const unsigned long PR_HAS_CONDENSED_DETACHMENT2 = 16384ul; 
 
  static const unsigned long PR_HAS_FLD1 = 32768ul; 
 
  static const unsigned long PR_HAS_FLD2 = 65536ul; 
 
  static const unsigned long PR_HAS_INEQUALITY_RESOLVABLE_WITH_DELETION = 131072ul; 
 
  static const unsigned long PR_HAS_STRINGS = 262144ul; 
 
  static const unsigned long PR_HAS_INTEGERS = 524288ul; 
 
  static const unsigned long PR_HAS_RATS = 1048576ul; 
 
  static const unsigned long PR_HAS_REALS = 2097152ul; 
 
  static const unsigned long PR_SORTS = 4194304ul; 
 
  static const unsigned long PR_INTEGER_COMPARISON = 8388608ul; 
 
  static const unsigned long PR_RAT_COMPARISON = 16777216ul; 
 
  static const unsigned long PR_REAL_COMPARISON = 33554432ul; 
 
  static const unsigned long PR_INTEGER_LINEAR = 67108864ul; 
 
  static const unsigned long PR_RAT_LINEAR = 134217728ul; 
 
  static const unsigned long PR_REAL_LINEAR = 268435456ul; 
 
  static const unsigned long PR_INTEGER_NONLINEAR = 536870912ul; 
 
  static const unsigned long PR_RAT_NONLINEAR = 1073741824ul; 
 
  static const unsigned long PR_REAL_NONLINEAR = 2147483648ul; 
 
  static const unsigned long PR_NUMBER_CONVERSION = 4294967296ul; 
 
  static const unsigned long PR_ESSENTIALLY_GROUND = 8589934592ul; 
 
  static const unsigned long PR_LIST_AXIOMS = 17179869184ul; 

 public:
  CLASS_NAME(Property);
  USE_ALLOCATOR(Property);

  static Property* scan(UnitList*);
  void add(UnitList*);
  ~Property();

 
  Category category() const { return _category;}
  static vstring categoryToString(Category cat);
  vstring categoryString() const;

  vstring toString() const;
  vstring toSpider(const vstring& problemName) const;

 
  int clauses() const { return _goalClauses + _axiomClauses; }
 
  int formulas() const { return _goalFormulas + _axiomFormulas; }
 
  int unitClauses() const { return _unitGoals + _unitAxioms; }
 
  int hornClauses() const { return _hornGoals + _hornAxioms; }
 
  int atoms() const { return _atoms; }
 
  int equalityAtoms() const { return _equalityAtoms; }
 
  int positiveEqualityAtoms() const { return _positiveEqualityAtoms; }
 
  bool hasFormulas() const { return _axiomFormulas || _goalFormulas; }
 
  int maxFunArity() const { return _maxFunArity; }
 
  int totalNumberOfVariables() const { return _totalNumberOfVariables;}

 
  bool hasProp(unsigned long p) const { return _props & p; }
 
  void addProp(unsigned long p) { _props |= p; }
 
  void dropProp(unsigned long p) { _props &= ~p; }
 
  unsigned long int props() const { return _props; }

 
  void scanForInterpreted(Term* t);

  bool hasInterpretedOperation(Interpretation i) const { return _interpretationPresence[i]; }
  
 
  bool hasInterpretedOperations() const { return _hasInterpreted; }
  bool hasInterpretedEquality() const { return _hasInterpretedEquality; }
 
  bool hasNonDefaultSorts() const { return _hasNonDefaultSorts; }
  bool usesSort(unsigned sort) const { return _usesSort[sort]; }
  bool usesSingleSort() const { return _sortsUsed==1; }
  unsigned sortsUsed() const { return _sortsUsed;}
  bool hasSpecialTermsOrLets() const { return _hasSpecialTermsOrLets; }
  bool hasFormulaItes() const { return _hasFormulaItes; }
 private:
  
  explicit Property();

  static bool hasXEqualsY(const Clause* c);
  static bool hasXEqualsY(const Formula*);

  
  void scan(Unit*);
  void scan(Clause*);
  void scan(FormulaUnit*);
  void scan(Formula*);
  void scan(TermList* ts);
  void scan(Literal* lit);
  void scanSpecialTerm(Term* t);
  void scanSort(unsigned sort);

  char axiomTypes() const;
  char goalTypes() const;
  char equalityContent() const;
  char nonGroundUnitContent() const;
  char groundPositiveContent() const;
  char goalsAreGround() const;
  char setClauseSize() const;
  char setLiteralSize() const;
  char setTermSize() const;
  char maxPredArity() const;

  
  int _goalClauses;
  int _axiomClauses;

  int _positiveEqualityAtoms;
  int _equalityAtoms;
  int _atoms;

  int _goalFormulas;
  int _axiomFormulas;
  int _subformulas;

  int _terms;
  int _unitGoals;
  int _unitAxioms;
  int _hornGoals;
  int _hornAxioms;
  int _equationalClauses;
  int _pureEquationalClauses;
  int _groundUnitAxioms;
  int _positiveAxioms;
  int _groundPositiveAxioms;
  int _groundGoals;
  int _maxFunArity;
  int _maxPredArity;

 
  int _variablesInThisClause;
 
  int _totalNumberOfVariables;
 
  int _maxVariablesInClause;

 
  unsigned long _props;

 
  Category _category;

 
  bool _hasInterpreted;
  bool _hasInterpretedEquality;
 
  bool _hasNonDefaultSorts;
  unsigned _sortsUsed;
  DArray<bool> _interpretationPresence;
  Array<bool> _usesSort;

  bool _hasSpecialTermsOrLets;
  bool _hasFormulaItes;

  bool _allClausesGround;
  bool _allQuantifiersEssentiallyExistential;
}; 

}

#endif 
