
/*
 * File SMTLEX.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <cstring>

#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"
#include "Lib/Exception.hpp"
#include "SMTLEX.hpp"

using namespace Shell;

SMTLexer::SMTLexer (istream& in)
  : Lexer (in)
{
} 

void SMTLexer::skipWhiteSpacesAndComments()
{
  CALL("SMTLexer::skipWhiteSpacesAndComments");

  int comment = 0;

  for (;;) {
    if (_eof) {
      if (comment == 2) {
	throw LexerException((vstring)"non-terminated comment",
			     *this);
      }
      return;
    }
    switch (_lastCharacter) {
    case ' ':
    case '\t':
    case '\r':
    case '\f':
      readNextChar();
      break;

    case '\n': 
      if (comment == 1) {
	comment = 0;
      }
      readNextChar();
      break;

    case ';': 
      comment = 1;
      readNextChar();
      break;

    default:
      if (comment) {
	readNextChar();
	break;
      }
      return;
    }
  }
} 

void SMTLexer::readToken (Token& token)
{
  CALL("SMTLexer::readToken");

  skipWhiteSpacesAndComments();
  _charCursor = 0;

  if (_eof) {
    token.tag = TT_EOF;
    token.text = "";
    return;
  }
  token.line = _lineNumber;

  switch (currentCharacterType()) {
  case LETTER:
    readName(token);
    token.tag = TT_NAME;
    return;

  case DIGIT:
    readNumber(token);
    return;

  case COLON:
    readName(token);
    token.tag = TT_ATTRIBUTE;
    return;

  case ARITH:
    readArith(token);
    token.tag = TT_ARITH;
    return;
    
  case PAR:
    readPar(token);
    return;

  case BRACE:
    readUserValue(token);
    token.tag = TT_USER;
    return;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 


SMTLexer::CharType SMTLexer::currentCharacterType () const
{
  CALL("SMTLexer::currentCharacterType");

  switch (_lastCharacter) {
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
  case 'G':
  case 'H':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
  case 'M':
  case 'N':
  case 'O':
  case 'P':
  case 'Q':
  case 'R':
  case 'S':
  case 'T':
  case 'U':
  case 'V':
  case 'W':
  case 'X':
  case 'Y':
  case 'Z':
  case 'a':
  case 'b':
  case 'c':
  case 'd':
  case 'e':
  case 'f':
  case 'g':
  case 'h':
  case 'i':
  case 'j':
  case 'k':
  case 'l':
  case 'm':
  case 'n':
  case 'o':
  case 'p':
  case 'q':
  case 'r':
  case 's':
  case 't':
  case 'u':
  case 'v':
  case 'w':
  case 'x':
  case 'y':
  case 'z':
  case '_':
  case '\'':
    return LETTER;
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    return DIGIT;
  case '(':
  case ')':
    return PAR;
  case '{':
  case '}':
    return BRACE;
  case '=':
  case '<':
  case '>':
  case '&':
  case '@':
  case '#':
  case '+':
  case '-':
  case '*':
  case '/':
  case '%':
  case '|':
  case '~':
    return ARITH;
  case ':':
    return COLON;
  case '.':
    return DOT;
  case ' ':
  case '\t':
  case '\r':
  case '\f':
  case '\n':
    return WHITE;

  default:
    throw LexerException((vstring)"illegal character " + (char)_lastCharacter,
			 *this);
  }
} 

void SMTLexer::readArith (Token& token)
{
  CALL("SMTLexer::readArith");

  saveLastChar();

  while (readNextChar()) {
    switch (currentCharacterType()) {
    case ARITH:
      saveLastChar();
      break;
    default:
      
      saveTokenText(token);
      return;
    }
  }

  
  saveTokenText(token);
} 

void SMTLexer::readPar (Token& token)
{
  CALL("SMTLexer::readPar");

  switch (_lastCharacter) {
  case '(':
    token.tag = TT_LPAR;
    break;
  case ')':
    token.tag = TT_RPAR;
    break;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }

  saveLastChar();
  saveTokenText(token);
  readNextChar();
} 

void SMTLexer::readName (Token& token)
{
  CALL("SMTLexer::readName");

  saveLastChar();

  while (readNextChar()) {
    switch (currentCharacterType()) {
    case LETTER:
    case DIGIT:
    case DOT:
      saveLastChar();
      break;
    default:
      
      saveTokenText(token);
      return;
    }
  }

  
  saveTokenText(token);
} 

void SMTLexer::readUserValue (Token& token)
{
  CALL("SMTLexer::readName");

  bool escaped = false;
  while (readNextChar()) {
    switch (_lastCharacter) {
    case '\\':
      if (escaped) {
	saveChar('\\');
      }
      escaped = true;
      break;
    case '}':
      if (escaped) {
	saveLastChar();
	escaped = false;
	break;
      }
      readNextChar();
      
      saveTokenText(token);
      return;
    default:
      if (escaped) {
	saveChar('\\');
	escaped = false;
      }
      saveLastChar();
      break;
    }
  }
} 
