
/*
 * File SMTPAR.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <fstream>

#include "SMTLEX.hpp"
#include "SMTPAR.hpp"

using namespace Lib;
using namespace Kernel;
using namespace Shell;

SMTParser::SMTParser (SMTLexer& lexer)
  : Parser(lexer)
{
}

SMTParser::Benchmark* SMTParser::benchmark()
{
  CALL("SMTParser::benchmark");

  consumeToken(TT_LPAR);
  readToken(TT_NAME);
  expectKeyword("benchmark");
  consumeToken();
  readToken(TT_NAME,"benchmark name expected");
  Benchmark* bench = new Benchmark(currentToken().text);
  consumeToken();
  readToken();
  while (currentToken().tag != TT_RPAR) {
    expectToken(TT_ATTRIBUTE);
    vstring attr = currentToken().text;
    if (attr == ":logic") {
      consumeToken();
      readToken();
      switch (currentToken().tag) {
      case TT_NAME:
	{
	  vstring logic = currentToken().text;




          if (logic == "QF_LRA") {
	    bench->logic = QF_LRA;
	  }
	  else {
	    terminate((vstring)"unknown logic: '" + logic + "'");
	  }
	}
	consumeToken();
	readToken();
	break;
      default:
	terminate("logic name expected");
	return 0;
      }
    }
    else if (attr == ":assumption") {
      terminate("not implemented");
    }
    else if (attr == ":formula") {
      consumeToken();
      readToken();
      formula(&(bench->formula));
    }
    else if (attr == ":extrasorts") {
      terminate("not implemented");
    }
    else if (attr == ":extrafuns") {
      consumeToken();
      readToken(TT_LPAR);
      consumeToken();
      readToken();
      functions(&bench->functionDeclarations);
    }
    else if (attr == ":extrapreds") {
      consumeToken();
      readToken(TT_LPAR);
      consumeToken();
      readToken();
      predicates(&bench->predicateDeclarations);
    }
    else if (attr == ":notes") {
      terminate("not implemented");
    }
    else if (attr == ":status") {
      consumeToken();
      readToken();
      switch (currentToken().tag) {
      case TT_NAME:
	{
	  vstring stat = currentToken().text;
	  if (stat == "sat") {
	    bench->status = SAT;
	  }
	  else if (stat == "unsat") {
	    bench->status = UNSAT;
	  }
	  else if (stat == "unknown") {
	    bench->status = UNKNOWN;
	  }
	  else {
	    terminate("unknown status: '" + stat + "'");
	  }
	}
	consumeToken();
	readToken();
	break;
      default:
	terminate("status description expected");
      }
    }
    else { 
      Annotation* ann = annotation();
      ann->next = bench->annotations;
      bench->annotations = ann;
    }
  }
  consumeToken(); 
  readToken();
  return bench;
} 

void SMTParser::predicates(PredicateDeclaration** decs)
{
  CALL("SMTParser::predicates");

  for (;;) {
    switch (currentToken().tag) {
    case TT_RPAR:
      consumeToken();
      readToken();
      return;
    case TT_LPAR:
      {
	consumeToken();
	readToken(TT_NAME,"predicate name expected");
	PredicateDeclaration* dec =
	  new PredicateDeclaration(currentToken().text);
	dec->next = *decs;
	*decs = dec;
	consumeToken();
	readToken(TT_RPAR,"complex predicate declarations not implemented");
	consumeToken();
	readToken();
      }
      break;
    default:
      terminate("predicate declaration expected");
    }
  }
} 

void SMTParser::functions(FunctionDeclaration** decs)
{
  CALL("SMTParser::functions");

  for (;;) {
    switch (currentToken().tag) {
    case TT_RPAR:
      consumeToken();
      readToken();
      return;
    case TT_LPAR:
      {
	consumeToken();
	readToken(TT_NAME,"function name expected");
	FunctionDeclaration* dec =
	  new FunctionDeclaration(currentToken().text);
	dec->next = *decs;
	*decs = dec;
	consumeToken();
	readToken(TT_NAME,"sort name expected");
	vstring sort = currentToken().text;
	if (sort == "Real") {
	  dec->sort = REAL;
	}
	else {
	  terminate("unknown sort: '" + sort + "'");
	}
	consumeToken();
	readToken(TT_RPAR,"complex function declarations not implemented");
	consumeToken();
	readToken();
      }
      break;
    default:
      terminate("function declaration expected");
    }
  }
} 

SMTParser::Annotation* SMTParser::annotation()
{
  CALL("SMTParser::annotation");
  ASS(currentToken().tag == TT_ATTRIBUTE);

  Annotation* ann = new Annotation();
  ann->next = 0;
  ann->attribute = currentToken().text;
  consumeToken();
  readToken();
  if (currentToken().tag != TT_USER) {
    ann->value = "";
    return ann;
  }
  ann->value = currentToken().text;
  consumeToken();
  readToken();
  return ann;
} 

SMTParser::Benchmark::Benchmark(const vstring& nm)
  : name(nm),
    status(UNKNOWN),
    functionDeclarations(0),
    predicateDeclarations(0),
    annotations(0)
{}

SMTParser::PredicateDeclaration::PredicateDeclaration(const vstring& nm)
  : name(nm),
    annotations(0),
    next(0)
{}

SMTParser::FunctionDeclaration::FunctionDeclaration(const vstring& nm)
  : name(nm),
    annotations(0),
    next(0)
{}

void SMTParser::formula(SMTParser::Formula** form)
{
  CALL("SMTParser::formula");

  consumeToken(TT_LPAR);
  readToken();
  switch (currentToken().tag) {
  case TT_NAME:
  case TT_ARITH:
    break;
  default:
    terminate("connective or predicate name expected");
  }
  vstring tokenText = currentToken().text;
  consumeToken();
  readToken();
  Connective con; 
  if (tokenText == "and") {
    con = AND;
  }
  else if (tokenText == "or") {
    con = OR;
  }
  else if (tokenText == "implies") {
    con = IMPLY;
  }
  else if (tokenText == "not") {
    con = NOT;
  }
  else if (tokenText == "xor") {
    con = XOR;
  }
  else if (tokenText == "iff") {
    con = IFF;
  }
  else if (tokenText == "if_then_else") {
    terminate("if-then-else formulas are not supported");
  }
  else { 
    Formula* atom = new Formula(ATOM);
    *form = atom;
    atom->atom = new Atom(tokenText);
    terms(&atom->atom->args);
    consumeToken(TT_RPAR);
    readToken();
    return;
  }
  *form = new Formula(con);
  formulas(&(*form)->args);
  
  consumeToken(TT_RPAR);
  readToken();
} 

void SMTParser::formulas(SMTParser::Formula** forms)
{
  CALL("SMTParser::formulas");

  while (currentToken().tag != TT_RPAR) {
    formula(forms);
    forms = &(*forms)->next;
  }
} 

void SMTParser::terms(SMTParser::Term** ts)
{
  CALL("SMTParser::terms");

  while (currentToken().tag != TT_RPAR &&
	 currentToken().tag != TT_ATTRIBUTE) {
    term(ts);
    ts = &(*ts)->next;
  }
} 

void SMTParser::term(SMTParser::Term** t)
{
  CALL("SMTParser::term");

  switch (currentToken().tag) {
  case TT_NAME:
    *t = new Term(TERM_COMPLEX,currentToken().text);
    break;
  case TT_INTEGER:
    *t = new Term(TERM_INT,currentToken().text);
    break;
  case TT_REAL:
    *t = new Term(TERM_REAL,currentToken().text);
    break;
  case TT_LPAR:
    
    consumeToken();
    readToken();
    switch (currentToken().tag) {
    case TT_ARITH:
      *t = new Term(TERM_ARITH,currentToken().text);
      consumeToken();
      readToken();
      terms(&(*t)->args);
      break;
    case TT_NAME:
      *t = new Term(TERM_COMPLEX,currentToken().text);
      consumeToken();
      readToken();
      terms(&(*t)->args);
      break;
    case TT_INTEGER:
      *t = new Term(TERM_INT,currentToken().text);
      consumeToken();
      readToken();
      break;
    case TT_REAL:
      *t = new Term(TERM_REAL,currentToken().text);
      consumeToken();
      readToken();
      break;
    default:
      terminate("beginning of term expected");
    }
    
    while (currentToken().tag == TT_ATTRIBUTE) {
      Annotation* ann = annotation();
      ann->next = (*t)->annotations;
      (*t)->annotations->next = ann;
    }
    break;
  default:
    terminate("term expected");
  }
  consumeToken();
  readToken();
} 

