
/*
 * File SimplifyProver.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#ifndef __SimplifyProver__
#define __SimplifyProver__

#include "Lib/List.hpp"
#include "Lib/Map.hpp"
#include "Lib/Set.hpp"
#include "Lib/Stack.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Unit.hpp"
#include "LispParser.hpp"

using namespace Lib;
using namespace Kernel;

namespace Shell {

class SimplifyProver
{
public:
 
  enum Keyword {
   
    K_NONE,
    K_DEFPRED,
    K_DEFPREDMAP,
    K_DEFUN,
    K_DEFINJ,
    K_DEFCONSTRUCTOR,
    K_DEFTUPLE,
    K_DEFARRAY,
    K_DEFWEAKARRAY,
    K_DEFCOTUPLE,
    K_DEFVALUE,
    K_DEFOP,
    K_DEFTYPE,
    K_SETPARAMETER,
    K_BG_PUSH,
    K_BG_POP,
    K_LEMMA,
    K_PROOF,
    K_CHECK,
    K_BUILTIN,
    K_TYPE,
    K_FORALL,
    K_EXISTS,
    K_LET,
    K_TERM,
    K_FORMULA,
    K_ORDER,
    K_LBLPOS,
    K_LBLNEG,
    K_LBL,
    K_PATS,
    K_NOPATS,
    K_MPAT,
    K_PROMOTE,
    K_AND,
    K_OR,
    K_IFF,
    K_IMPLIES,
    K_ITE,
    K_EXPLIES,
    K_NOT,
    K_TRUE,
    K_FALSE,
    K_EQ,
    K_NEQ,
    K_DISTINCT,
    K_PP,
    K_DUMP_CTX,
    K_DBG_VALID,
    K_DBG_INVALID,
    K_DBG_WAS_VALID,
    K_DBG_WAS_INVALID,
    K_ECHO,
    K_PROMPT_OFF,
    K_PROMPT_ON,
    K_EVALT,
    K_EVALF,
    K_EXIT,
    K_QID,
    K_SKOLEMID,
    K_WEIGHT,
    K_ASYNC,
    K_CANCEL,
    K_MODEL,
    K_STATS,
    K_SLEEP
  };

 
  enum Type {
   
    BIT_INT = 0,
   
    BIT_BOOL = 1,
   
    BIT_BITVEC = 2,
   
    OTHER = 3
  };

  SimplifyProver();
  ~SimplifyProver();
  typedef LispParser::Expression Expression;
  typedef Lib::List<Expression*> List;
  UnitList* units(const Expression*);

 
  struct SymbolInfo {
    int arity;
    Type returnType;
    unsigned number;
    int argTypes[1];

    void* operator new(size_t,int arity);
    SymbolInfo(int arity);
    ~SymbolInfo();
  };

private:
 
  enum Command {
   
    PARSE_FORMULA = 0,
   
    PARSE_TERM = 1,
   
    BUILD_TERM = 2,
   
    BUILD_JUNCTION_FORMULA = 3,
   
    BUILD_QUANTIFIED_FORMULA = 4,
   
    BUILD_BINARY_FORMULA = 5,
   
    BUILD_ATOM = 6,
   
    BUILD_NEGATED_FORMULA = 7,
   
    BUILD_EQUALITY = 8,
   
    BUILD_DISTINCT = 9,
   
    DO_LET = 10,
   
    UNDO_LET = 11,
   
    BUILD_ITE_TERM = 12,
    
    BUILD_LET_FORMULA = 13,
    
    BUILD_LET_TERM = 14,
	
		BIND_VARS = 15,
	
		UNBIND_VARS = 16
  };

 
  enum Context {
   
    CN_TOP_LEVEL,
   
    CN_FORMULA,
   
    CN_ARGUMENT
  };

 
  unsigned _defaultSort;
 
  unsigned _numberSort;

 
  UnitList* _units;
 
  Map<vstring,Type> _types;
 
  Map<vstring,SymbolInfo*> _symbolInfo;
 
  Set<vstring> _hasBooleanArgs;
 
  int _nextType;
 
  int _nextVar;
 
  Stack<Formula*> _formulas;
 
  Map<vstring,IntList*> _variables;
 
  Stack<Command> _commands;
 
  Stack<const void*> _saved;
 
  Stack<int> _isaved;
 
  Stack<void*> _built;
 
  Stack<TermList> _tsaved;
 
  Stack<vstring> _ssaved;
 
  Stack<TermList> _numbers;
 
  TermList _zero;
 
  TermList _one;
 
  Map<vstring,Lib::List<Formula*>*> _formulaLet;
 
  Map<vstring,Lib::List<TermList>*> _termLet;

  void parse(const Expression*);
  void parse();
  static Keyword keyword(const vstring& str);
  void formula(const Expression*);
  int bindVar(const vstring& varName);
  int isVar(const vstring& varName);
  void unbindVar(const vstring& var);
  void formulaError(const Expression* expr);
  void formulaError(const Expression* expr,const char* explanation);
  void termError(const Expression* expr);
  void error(const vstring& errMsg);
  void parseFormula();
  void parseJunctionFormula(const List*,const Expression*,Connective c,Context);
  void parseBinaryFormula(const List*,const Expression*,Connective c,Context);
  void parseNegatedFormula(const List*,const Expression*,Context);
  void parseQuantifiedFormula(const List* lst,const Expression*,Connective c,Context);
  void parseAtom(const List*,const Expression*,Context);
  void parseAtom(const Expression*,Context);
  void parseEquality(const List*,const Expression*,Context,bool polarity);
  void parseDistinct(const List*,const Expression*,Context);
  void parseLet(const List*,const Expression*,Context);
  void parseTerm();
  SymbolInfo* builtInPredicate(const vstring& str,int arity);
  SymbolInfo* builtInFunction(const vstring& str,int arity);
  SymbolInfo* getFunctionSymbolInfo(const vstring& name,int arity);
  SymbolInfo* addNumber(const vstring&);
  void defType(const List*,const Expression*);
  void defOp(const List*,const Expression*);
  void bgPush(const List*);
  void defPred(const List*,const Expression*);
  void buildTerm();
  void buildAtom();
  void buildEquality();
  void doLet();
  void undoLet();
  void buildBinaryFormula();
  void buildJunctionFormula();
  void buildQuantifiedFormula();
  void buildNegatedFormula();
  void buildIfThenElseTerm();
  void buildDistinct();
  void processFormula(Formula*,Context);
  void addUnit(Unit*);
  void buildLetFormula();
  void buildLetTerm();
  void parseTrueFalse(bool,Context);
  void bindVars();
  void unbindVars();
}; 

}

#endif

