
/*
 * File Skolem.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/SubformulaIterator.hpp"

#include "Indexing/TermSharing.hpp"

#include "Options.hpp"
#include "Rectify.hpp"
#include "Refutation.hpp"
#include "Skolem.hpp"
#include "VarManager.hpp"

using namespace Kernel;
using namespace Shell;

FormulaUnit* Skolem::skolemise (FormulaUnit* unit)
{
  CALL("Skolem::skolemise(Unit*)");
  ASS(! unit->isClause());

  unit = Rectify::rectify(unit);
  Formula* f = unit->formula();
  switch (f->connective()) {
  case FALSE:
  case TRUE:
    return unit;
  default:
    break;
  }

  static Skolem skol;
  skol.reset();
  return skol.skolemiseImpl(unit);
} 

FormulaUnit* Skolem::skolemiseImpl (FormulaUnit* unit)
{
  CALL("Skolem::skolemiseImpl(FormulaUnit*)");

  ASS(_introducedSkolemFuns.isEmpty());
  _beingSkolemised=unit;

  Formula* f = unit->formula();
  Formula* g = skolemise(f);

  _beingSkolemised = 0;

  if (f == g) { 
    return unit;
  }
  Inference* inf = new Inference1(Inference::SKOLEMIZE,unit);
  FormulaUnit* res = new FormulaUnit(g, inf, unit->inputType());

  ASS(_introducedSkolemFuns.isNonEmpty());
  while(_introducedSkolemFuns.isNonEmpty()) {
    unsigned fn = _introducedSkolemFuns.pop();
    InferenceStore::instance()->recordIntroducedSymbol(res,true,fn);
  }

  return res;
}


void Skolem::reset()
{
  CALL("Skolem::reset");

  _vars.reset();
  _varSorts.reset();
  _subst.reset();
}

unsigned Skolem::addSkolemFunction(unsigned arity, unsigned* domainSorts,
    unsigned rangeSort, unsigned var)
{
  CALL("Skolem::addSkolemFunction(unsigned,unsigned*,unsigned,unsigned)");

  if(VarManager::varNamePreserving()) {
    vstring varName=VarManager::getVarName(var);
    return addSkolemFunction(arity, domainSorts, rangeSort, varName.c_str());
  }
  else {
    return addSkolemFunction(arity, domainSorts, rangeSort);
  }
}

unsigned Skolem::addSkolemFunction(unsigned arity, unsigned* domainSorts,
    unsigned rangeSort, const char* suffix)
{
  CALL("Skolem::addSkolemFunction(unsigned,unsigned*,unsigned,const char*)");
  ASS(arity==0 || domainSorts!=0);

  unsigned fun = env.signature->addSkolemFunction(arity, suffix);
  Signature::Symbol* fnSym = env.signature->getFunction(fun);
  fnSym->setType(new FunctionType(arity, domainSorts, rangeSort));
  return fun;
}

void Skolem::ensureHavingVarSorts()
{
  CALL("Skolem::ensureHavingVarSorts");

  Formula* f = _beingSkolemised->formula();
  SortHelper::collectVariableSorts(f, _varSorts);
}

Term* Skolem::createSkolemTerm(unsigned var)
{
  CALL("Skolem::createSkolemFunction");

  int arity = _vars.length();

  ensureHavingVarSorts();
  unsigned rangeSort=_varSorts.get(var, Sorts::SRT_DEFAULT);
  static Stack<unsigned> domainSorts;
  static Stack<TermList> fnArgs;
  domainSorts.reset();
  fnArgs.reset();

  VarStack::TopFirstIterator vit(_vars);
  while(vit.hasNext()) {
    unsigned uvar = vit.next();
    domainSorts.push(_varSorts.get(uvar, Sorts::SRT_DEFAULT));
    fnArgs.push(TermList(uvar, false));
  }

  unsigned fun = addSkolemFunction(arity, domainSorts.begin(), rangeSort, var);
  _introducedSkolemFuns.push(fun);

  return Term::create(fun, arity, fnArgs.begin());
}

Formula* Skolem::skolemise (Formula* f)
{
  CALL("Skolem::skolemise (Formula*)");

  switch (f->connective()) {
  case LITERAL: 
    {
      Literal* l = f->literal();
      Literal* ll = l->apply(_subst);
      if (l == ll) {
	return f;
      }
      return new AtomicFormula(ll);
    }

  case AND:
  case OR: 
    {
      FormulaList* fs = skolemise(f->args());
      if (fs == f->args()) {
	return f;
      }
      return new JunctionFormula(f->connective(),fs);
    }

  case FORALL: 
    {
      int ln = _vars.length();
      Formula::VarList::Iterator vs(f->vars());
      while (vs.hasNext()) {
	_vars.push(vs.next());
      }
      Formula* g = skolemise(f->qarg());
      _vars.truncate(ln);
      if (g == f->qarg()) {
	return f;
      }
      return new QuantifiedFormula(f->connective(),f->vars(),g);
    }

  case EXISTS: 
    {
      int arity = _vars.length();

      Formula::VarList::Iterator vs(f->vars());
      while (vs.hasNext()) {
	int v = vs.next();
	Term* skolemTerm = createSkolemTerm(v);
	_subst.bind(v,skolemTerm);

  if (env.options->showSkolemisations()) {
    env.beginOutput();
    env.out() << "Skolemising: "<<skolemTerm->toString()<<" for X"<< v
	    <<" in "<<f->toString()<<" in formula "<<_beingSkolemised->toString();
    env.endOutput();
  }
  
  if (env.options->showNonconstantSkolemFunctionTrace() && arity!=0) {
    env.beginOutput();
    ostream& out = env.out();
      out <<"Nonconstant skolem function introduced: "
      <<skolemTerm->toString()<<" for X"<<v<<" in "<<f->toString()
      <<" in formula "<<_beingSkolemised->toString()<<endl;
      
	  Refutation ref(_beingSkolemised, true);
	  ref.output(out);
    env.endOutput();
  }
      }
      Formula* g = skolemise(f->qarg());
      vs.reset(f->vars());
      while (vs.hasNext()) {
	_subst.unbind(vs.next());
      }
      _vars.truncate(arity);

      return g;
    }

  case TRUE:
  case FALSE:
    return f;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 


FormulaList* Skolem::skolemise (FormulaList* fs) 
{
  CALL("skolemise (FormulaList*)");

  if (fs->isEmpty()) {
    return fs;
  }

  Formula* g = fs->head();
  FormulaList* gs = fs->tail();
  Formula* h = skolemise(g);
  FormulaList* hs = skolemise(gs);

  if (gs == hs && g == h) {
    return fs;
  }
  return new FormulaList(h,hs);
} 


