
/*
 * File TheoryAxioms.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/EqHelper.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Sorts.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"
#include "Kernel/Theory.hpp"

#include "AxiomGenerator.hpp"
#include "Property.hpp"
#include "SymCounter.hpp"
#include "TheoryAxioms.hpp"
#include "Options.hpp"

using namespace Lib;
using namespace Kernel;
using namespace Shell;

void TheoryAxioms::addAndOutputTheoryUnit(Unit* unit,UnitList*& units)
{
  CALL("TheoryAxioms::addAndOutputTheoryUnit");
  if (env.options->showTheoryAxioms()) {
    cout << "% Theory " << (unit->isClause() ? "clause" : "formula" ) << ": " << unit->toString() << "\n";
  }
  UnitList::push(unit,units);
} 

void TheoryAxioms::addTheoryUnitClause(Literal* lit, UnitList*& units)
{
  CALL("TheoryAxioms::addTheoryUnitClause");
  Clause* unit = Clause::fromIterator(getSingletonIterator(lit), Unit::AXIOM, new Inference(Inference::THEORY));
  addAndOutputTheoryUnit(unit,units);
} 

void TheoryAxioms::addTheoryNonUnitClause(UnitList*& units, Literal* lit1, Literal* lit2, Literal* lit3)
{
  CALL("TheoryAxioms::addTheoryNonUnitClause");
  LiteralStack lits;
  ASS(lit1);
  lits.push(lit1);
  ASS(lit2);
  lits.push(lit2);
  if (lit3) {
    lits.push(lit3);
  }
  Clause* cl = Clause::fromStack(lits, Unit::AXIOM, new Inference(Inference::THEORY));
  addAndOutputTheoryUnit(cl,units);
} 

void TheoryAxioms::addCommutativity(Interpretation op,UnitList*& units)
{
  CALL("TheoryAxioms::addCommutativity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned f = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList x(0,false);
  TermList y(1,false);
  TermList fxy(Term::create2(f,x,y));
  TermList fyx(Term::create2(f,y,x));
  Literal* eq = Literal::createEquality(true,fxy,fyx,srt);
  addTheoryUnitClause(eq,units);
} 

void TheoryAxioms::addAssociativity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addCommutativity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned f = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList x(0,false);
  TermList y(1,false);
  TermList z(2,false);
  TermList fxy(Term::create2(f,x,y));
  TermList fyz(Term::create2(f,y,z));
  TermList fx_fyz(Term::create2(f,x,fyz));
  TermList f_fxy_z(Term::create2(f,fxy,z));
  Literal* eq = Literal::createEquality(true, fx_fyz,f_fxy_z, srt);
  addTheoryUnitClause(eq, units);
} 

void TheoryAxioms::addIdentity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addIdentity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),1);

  unsigned f = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList x(0,false);
  TermList fx(Term::create1(f,x));
  Literal* eq = Literal::createEquality(true,fx,x,srt);
  addTheoryUnitClause(eq,units);
} 

void TheoryAxioms::addRightIdentity(Interpretation op, TermList e, UnitList*& units)
{
  CALL("TheoryAxioms::addRightIdentity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned f = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList x(0,false);
  TermList fxe(Term::create2(f,x,e));
  Literal* eq = Literal::createEquality(true,fxe,x,srt);
  addTheoryUnitClause(eq, units);
} 

void TheoryAxioms::addCommutativeGroupAxioms(Interpretation op, Interpretation inverse, TermList e, UnitList*& units)
{
  CALL("TheoryAxioms::addCommutativeGroupAxioms");

  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);
  ASS(theory->isFunction(inverse));
  ASS_EQ(theory->getArity(inverse),1);

  addCommutativity(op,units);
  addAssociativity(op,units);
  addRightIdentity(op,e,units);

  
  unsigned f = env.signature->getInterpretingSymbol(op);
  unsigned i = env.signature->getInterpretingSymbol(inverse);
  unsigned srt = theory->getOperationSort(op);
  ASS_EQ(srt, theory->getOperationSort(inverse));

  TermList x(0,false);
  TermList y(1,false);
  TermList fxy(Term::create2(f,x,y));
  TermList ix(Term::create1(i,x));
  TermList iy(Term::create1(i,y));
  TermList i_fxy(Term::create1(i,fxy));
  TermList f_iy_ix(Term::create2(f,iy,ix));
  Literal* eq1 = Literal::createEquality(true,i_fxy,f_iy_ix,srt);
  addTheoryUnitClause(eq1, units);

  
  TermList fx_ix(Term::create2(f,x,ix));
  Literal* eq2 = Literal::createEquality(true,fx_ix,e,srt);
  addTheoryUnitClause(eq2, units);
} 

void TheoryAxioms::addReflexivity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addReflexivity");

  ASS(!theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned opPred = env.signature->getInterpretingSymbol(op);
  TermList x(0,false);
  Literal* l11 = Literal::create2(opPred, true, x, x);
  addTheoryUnitClause(l11, units);
} 

void TheoryAxioms::addTransitivity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addTransitivity");
  ASS(!theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned opPred = env.signature->getInterpretingSymbol(op);
  TermList x(0,false);
  TermList y(1,false);
  TermList v3(2,false);

  Literal* nonL12 = Literal::create2(opPred, false, x, y);
  Literal* nonL23 = Literal::create2(opPred, false, y, v3);
  Literal* l13 = Literal::create2(opPred, true, x, v3);

  addTheoryNonUnitClause(units, nonL12, nonL23, l13);
}

void TheoryAxioms::addOrderingTotality(Interpretation lessEqual, UnitList*& units)
{
  CALL("TheoryAxioms::addOrderingTotality");
  ASS(!theory->isFunction(lessEqual));
  ASS_EQ(theory->getArity(lessEqual),2);

  unsigned opPred = env.signature->getInterpretingSymbol(lessEqual);
  TermList x(0,false);
  TermList y(1,false);

  Literal* l12 = Literal::create2(opPred, true, x, y);
  Literal* l21 = Literal::create2(opPred, true, y, x);

  addTheoryNonUnitClause(units, l12, l21);
}

void TheoryAxioms::addTotalOrderAxioms(Interpretation lessEqual, UnitList*& units)
{
  CALL("TheoryAxioms::addTotalOrderAxioms");

  addReflexivity(lessEqual, units);
  addTransitivity(lessEqual, units);
  addOrderingTotality(lessEqual, units);
}

void TheoryAxioms::addMonotonicity(Interpretation lessEqual, Interpretation addition, UnitList*& units)
{
  CALL("TheoryAxioms::addMonotonicity");
  ASS(!theory->isFunction(lessEqual));
  ASS_EQ(theory->getArity(lessEqual),2);
  ASS(theory->isFunction(addition));
  ASS_EQ(theory->getArity(addition),2);

  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned addFun = env.signature->getInterpretingSymbol(addition);
  TermList x(0,false);
  TermList y(1,false);
  TermList v3(2,false);
  TermList xPv3(Term::create2(addFun, x,v3));
  TermList yPv3(Term::create2(addFun, y,v3));
  Literal* nonLe = Literal::create2(lePred, false, x, y);
  Literal* leAdded = Literal::create2(lePred, true, xPv3, yPv3);

  addTheoryNonUnitClause(units, nonLe, leAdded);
}

void TheoryAxioms::addAdditionAndOrderingAxioms(Interpretation plus, Interpretation unaryMinus,
    TermList zeroElement, TermList oneElement, Interpretation lessEqual, UnitList*& units)
{
  CALL("TheoryAxioms::addAdditionAndOrderingAxioms");

  addCommutativeGroupAxioms(plus, unaryMinus, zeroElement, units);
  addTotalOrderAxioms(lessEqual, units);
  addMonotonicity(lessEqual, plus, units);

  
  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  Literal* nonLeOneZero = Literal::create2(lePred, false, oneElement, zeroElement);
  addTheoryUnitClause(nonLeOneZero, units);

  
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  TermList x(0,false);
  TermList y(1,false);
  Literal* nonLe21 = Literal::create2(lePred, false, y, x);
  TermList xPOne(Term::create2(plusFun, x, oneElement));
  Literal* nonLt1POne2 = Literal::create2(lePred, false, xPOne, y);
  addTheoryNonUnitClause(units, nonLe21, nonLt1POne2);

  
  

  unsigned varSort = theory->getOperationSort(lessEqual);
  Literal* xEqY = Literal::createEquality(true, x, y, varSort);
  Literal* nonLe12 = Literal::create2(lePred, false, x, y);
  addTheoryNonUnitClause(units, nonLe21, nonLe12, xEqY);
}

void TheoryAxioms::addAdditionOrderingAndMultiplicationAxioms(Interpretation plus, Interpretation unaryMinus,
    TermList zeroElement, TermList oneElement, Interpretation lessEqual, Interpretation multiply,
    UnitList*& units)
{
  CALL("TheoryAxioms::addAdditionOrderingAndMultiplicationAxioms");

  unsigned srt = theory->getOperationSort(plus);
  ASS_EQ(srt, theory->getOperationSort(unaryMinus));
  ASS_EQ(srt, theory->getOperationSort(lessEqual));
  ASS_EQ(srt, theory->getOperationSort(multiply));

  addAdditionAndOrderingAxioms(plus, unaryMinus, zeroElement, oneElement, lessEqual, units);

  addCommutativity(multiply, units);
  addAssociativity(multiply, units);
  addRightIdentity(multiply, oneElement, units);

  
  unsigned mulFun = env.signature->getInterpretingSymbol(multiply);
  TermList x(0,false);
  TermList xMulZero(Term::create2(mulFun, x, zeroElement));
  Literal* xEqXMulZero = Literal::createEquality(true, xMulZero, zeroElement, srt);
  addTheoryUnitClause(xEqXMulZero, units);

  
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  TermList y(1,false);
  TermList yPOne(Term::create2(plusFun, y, oneElement));
  TermList xMulYPOne(Term::create2(mulFun, x, yPOne));
  TermList xMulY(Term::create2(mulFun, x, y));
  TermList xMulYPX(Term::create2(plusFun, xMulY, x));
  Literal* succDistrEq = Literal::createEquality(true, xMulYPOne, xMulYPX, srt);
  addTheoryUnitClause(succDistrEq, units);

  
  TermList v3(2,false);
  TermList v4(3,false);
  TermList xPy(Term::create2(plusFun, x, y));
  TermList v3Pv4(Term::create2(plusFun, v3, v4));
  TermList distrLhs(Term::create2(mulFun, xPy, v3Pv4));
  TermList xMv3(Term::create2(mulFun, x, v3));
  TermList xMv4(Term::create2(mulFun, x, v4));
  TermList yMv3(Term::create2(mulFun, y, v3));
  TermList yMv4(Term::create2(mulFun, y, v4));
  TermList add1(Term::create2(plusFun, xMv3, xMv4));
  TermList add2(Term::create2(plusFun, yMv3, yMv4));
  TermList distrRhs(Term::create2(plusFun, add1, add2));
  Literal* distrEq = Literal::createEquality(true, distrLhs, distrRhs, srt);
  addTheoryUnitClause(distrEq, units);
}

void TheoryAxioms::addExtraIntegerOrderingAxiom(Interpretation plus, TermList oneElement, Interpretation lessEqual,
    UnitList*& units)
{
  CALL("TheoryAxioms::addExtraIntegerOrderingAxiom");

  
  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  TermList x(0,false);
  TermList y(1,false);
  Literal* le21 = Literal::create2(lePred, true, y, x);
  TermList xPOne(Term::create2(plusFun, x, oneElement));
  Literal* lt1POne2 = Literal::create2(lePred, true, xPOne, y);
  addTheoryNonUnitClause(units, le21, lt1POne2);
}
    
void TheoryAxioms::addFloorAxioms(Interpretation floor, Interpretation lessEqual, Interpretation unaryMinus,
     Interpretation plus, TermList oneElement, UnitList*& units)
{
  CALL("TheoryAxioms::addFloorAxioms");

  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  unsigned umFun = env.signature->getInterpretingSymbol(unaryMinus);
  unsigned floorFun = env.signature->getInterpretingSymbol(floor);
  TermList x(0,false);
  TermList floorX(Term::create1(floorFun,x));

  
  Literal* a1 = Literal::create2(lePred, true, floorX, x);
  addTheoryUnitClause(a1, units);

  
  TermList m1(Term::create1(umFun,oneElement));
  TermList xm1(Term::create2(plusFun, x, m1));
  Literal* a2 = Literal::create2(lePred,false, xm1, floorX);
  addTheoryUnitClause(a2,units);
} 
 
void TheoryAxioms::addCeilingAxioms(Interpretation ceiling, Interpretation lessEqual, 
     Interpretation plus, TermList oneElement, UnitList*& units)
{
  CALL("TheoryAxioms::addCeilingAxioms");

  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  unsigned ceilingFun = env.signature->getInterpretingSymbol(ceiling);
  TermList x(0,false);
  TermList ceilingX(Term::create1(ceilingFun,x));

  
  Literal* a1 = Literal::create2(lePred, true, x, ceilingX);
  addTheoryUnitClause(a1, units);

  
  TermList xp1(Term::create2(plusFun, x, oneElement));
  Literal* a2 = Literal::create2(lePred,false, xp1, ceilingX);
  addTheoryUnitClause(a2,units);
} 
 
void TheoryAxioms::addRoundAxioms(Interpretation round, Interpretation floor, Interpretation ceiling, UnitList*& units)
{
  CALL("TheoryAxioms::addRoundAxioms");
  
  
  
  
  
  
  
  
  

} 
 
void TheoryAxioms::addTruncateAxioms(Interpretation truncate, Interpretation lessEqual, Interpretation unaryMinus,
                      Interpretation plus, TermList zeroElement, TermList oneElement, UnitList*& units)
{
  CALL("TheoryAxioms::addTruncateAxioms");

  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  unsigned umFun = env.signature->getInterpretingSymbol(unaryMinus);
  unsigned truncateFun = env.signature->getInterpretingSymbol(truncate);
  TermList x(0,false);
  TermList truncateX(Term::create1(truncateFun,x));

  TermList m1(Term::create1(umFun,oneElement));
  TermList xm1(Term::create2(plusFun,x,m1));
  TermList xp1(Term::create2(plusFun,x,oneElement));

  Literal* nzleX = Literal::create2(lePred,false,zeroElement,x);
  Literal* zleX = Literal::create2(lePred,true,zeroElement,x);

  
  Literal* a1 = Literal::create2(lePred,true,truncateX,x);
  addTheoryNonUnitClause(units,nzleX,a1);

  
  Literal* a2 = Literal::create2(lePred,false,truncateX,xm1);
  addTheoryNonUnitClause(units,nzleX,a2);

  
  Literal* a3 = Literal::create2(lePred,true,x,truncateX);
  addTheoryNonUnitClause(units,zleX,a3);

  
  Literal* a4 = Literal::create2(lePred,false,xp1,truncateX);
  addTheoryNonUnitClause(units,zleX,a4);

} 

void TheoryAxioms::addArrayExtensionalityAxioms(
  Interpretation select,
  Interpretation store,
  unsigned skolemFn,
  UnitList*& units)
{
  CALL("TheoryAxioms::addArrayExtenstionalityAxioms");
        
  ASS(theory->isFunction(select));
  ASS(theory->isArrayOperation(select));
  ASS_EQ(theory->getArity(select),2);
              
  unsigned sel = env.signature->getInterpretingSymbol(select);
  unsigned rangeSort = theory->getArrayOperationSort(select);
  unsigned arraySort = theory->getArrayOperationSort(store);

  TermList x(0,false);
  TermList y(1,false);
  
  TermList sk(Term::create2(skolemFn, x, y)); 
  TermList sel_x_sk(Term::create2(sel,x,sk)); 
  TermList sel_y_sk(Term::create2(sel,y,sk)); 
  Literal* eq = Literal::createEquality(true,x,y,arraySort); 
  Literal* ineq = Literal::createEquality(false,sel_x_sk,sel_y_sk,rangeSort); 
  
  addTheoryNonUnitClause(units, eq, ineq);
} 

void TheoryAxioms::addArrayWriteAxioms(Interpretation select, Interpretation store, UnitList*& units)
{
  CALL("TheoryAxioms::addArrayWriteAxioms");
        
  ASS(theory->isFunction(select));
  ASS(theory->isArrayOperation(select));
  ASS_EQ(theory->getArity(select),2);
        
        
  unsigned func_select = env.signature->getInterpretingSymbol(select);
  unsigned func_store = env.signature->getInterpretingSymbol(store);

  unsigned rangeSort = theory->getArrayOperationSort(select);
  unsigned domainSort = theory->getArrayDomainSort(select);
  
        
  TermList i(0,false);
  TermList j(1,false);
  TermList v(2,false);
  TermList a(3,false);
  TermList args[] = {a, i, v};
        
  
  TermList wAIV(Term::create(func_store, 3, args)); 
  TermList sWI(Term::create2(func_select, wAIV,i)); 
  Literal* ax = Literal::createEquality(true, sWI, v, rangeSort);
  addTheoryUnitClause(ax, units);

  
  TermList sWJ(Term::create2(func_select, wAIV,j)); 
  TermList sAJ(Term::create2(func_select, a, j)); 
        
  Literal* indexEq = Literal::createEquality(true, i, j, domainSort);
  Literal* writeEq = Literal::createEquality(true, sWJ, sAJ, rangeSort);
  addTheoryNonUnitClause(units, indexEq, writeEq);                      
} 
    









void TheoryAxioms::apply(Problem& prb)
{
  CALL("TheoryAxioms::apply(Problem&)");

  Property* prop = prb.getProperty();
  if(apply(prb.units(), prop)) {
    prb.invalidateProperty();
    prb.reportEqualityAdded(false);
  }
}

bool TheoryAxioms::apply(UnitList*& units, Property* prop)
{
  CALL("TheoryAxioms::apply");  
  bool modified = false;
  bool haveIntPlus =
    prop->hasInterpretedOperation(Theory::INT_PLUS) ||
    prop->hasInterpretedOperation(Theory::INT_UNARY_MINUS) ||
    prop->hasInterpretedOperation(Theory::INT_LESS_EQUAL) ||
    prop->hasInterpretedOperation(Theory::INT_MULTIPLY);
  bool haveIntMultiply =
    prop->hasInterpretedOperation(Theory::INT_MULTIPLY);

  bool haveIntFloor = prop->hasInterpretedOperation(Theory::INT_FLOOR);
  bool haveIntCeiling = prop->hasInterpretedOperation(Theory::INT_CEILING);
  bool haveIntRound = prop->hasInterpretedOperation(Theory::INT_ROUND);
  bool haveIntTruncate = prop->hasInterpretedOperation(Theory::INT_TRUNCATE);
  bool haveIntUnaryRoundingFunction = haveIntFloor || haveIntCeiling || haveIntRound || haveIntTruncate;

  if (haveIntPlus || haveIntUnaryRoundingFunction) {
    TermList zero(theory->representConstant(IntegerConstantType(0)));
    TermList one(theory->representConstant(IntegerConstantType(1)));
    if(haveIntMultiply) {
      addAdditionOrderingAndMultiplicationAxioms(Theory::INT_PLUS, Theory::INT_UNARY_MINUS, zero, one,
						 Theory::INT_LESS_EQUAL, Theory::INT_MULTIPLY, units);
    }
    else {
      addAdditionAndOrderingAxioms(Theory::INT_PLUS, Theory::INT_UNARY_MINUS, zero, one,
				   Theory::INT_LESS_EQUAL, units);
    }
    addExtraIntegerOrderingAxiom(Theory::INT_PLUS, one, Theory::INT_LESS_EQUAL, units);
    if(haveIntFloor){    addIdentity(Theory::INT_FLOOR,units); }
    if(haveIntCeiling){  addIdentity(Theory::INT_CEILING,units); }
    if(haveIntRound){    addIdentity(Theory::INT_ROUND,units); }
    if(haveIntTruncate){ addIdentity(Theory::INT_TRUNCATE,units); }
    modified = true;
  }
  bool haveRatPlus =
    prop->hasInterpretedOperation(Theory::RAT_PLUS) ||
    prop->hasInterpretedOperation(Theory::RAT_UNARY_MINUS) ||
    prop->hasInterpretedOperation(Theory::RAT_LESS_EQUAL) ||
    prop->hasInterpretedOperation(Theory::RAT_MULTIPLY);
  bool haveRatMultiply =
    prop->hasInterpretedOperation(Theory::RAT_MULTIPLY);

  bool haveRatFloor = prop->hasInterpretedOperation(Theory::RAT_FLOOR);
  bool haveRatCeiling = prop->hasInterpretedOperation(Theory::RAT_CEILING);
  bool haveRatRound = prop->hasInterpretedOperation(Theory::RAT_ROUND);
  bool haveRatTruncate = prop->hasInterpretedOperation(Theory::RAT_TRUNCATE);
  bool haveRatUnaryRoundingFunction = haveRatFloor || haveRatCeiling || haveRatRound || haveRatTruncate;

  if (haveRatPlus || haveRatUnaryRoundingFunction) {
    TermList zero(theory->representConstant(RationalConstantType(0, 1)));
    TermList one(theory->representConstant(RationalConstantType(1, 1)));
    if(haveRatMultiply || haveRatRound) {
      addAdditionOrderingAndMultiplicationAxioms(Theory::RAT_PLUS, Theory::RAT_UNARY_MINUS, zero, one,
						 Theory::RAT_LESS_EQUAL, Theory::RAT_MULTIPLY, units);
    }
    else {
      addAdditionAndOrderingAxioms(Theory::RAT_PLUS, Theory::RAT_UNARY_MINUS, zero, one,
				   Theory::RAT_LESS_EQUAL, units);
    }
    if(haveRatFloor || haveRatRound){
      addFloorAxioms(Theory::RAT_FLOOR,Theory::RAT_LESS_EQUAL,Theory::RAT_UNARY_MINUS,Theory::RAT_PLUS,one,units);
    }
    if(haveRatCeiling || haveRatRound){
      addCeilingAxioms(Theory::RAT_CEILING,Theory::RAT_LESS_EQUAL,Theory::RAT_PLUS,one,units);
    }
    if(haveRatRound){
      
    }
    if(haveRatTruncate){
      addTruncateAxioms(Theory::RAT_TRUNCATE,Theory::RAT_LESS_EQUAL,Theory::RAT_UNARY_MINUS,
                        Theory::RAT_PLUS,zero,one,units);
    }
    modified = true;
  }
  bool haveRealPlus =
    prop->hasInterpretedOperation(Theory::REAL_PLUS) ||
    prop->hasInterpretedOperation(Theory::REAL_UNARY_MINUS) ||
    prop->hasInterpretedOperation(Theory::REAL_LESS_EQUAL) ||
    prop->hasInterpretedOperation(Theory::REAL_MULTIPLY);
  bool haveRealMultiply =
    prop->hasInterpretedOperation(Theory::REAL_MULTIPLY);

  bool haveRealFloor = prop->hasInterpretedOperation(Theory::REAL_FLOOR);
  bool haveRealCeiling = prop->hasInterpretedOperation(Theory::REAL_CEILING);
  bool haveRealRound = prop->hasInterpretedOperation(Theory::REAL_ROUND);
  bool haveRealTruncate = prop->hasInterpretedOperation(Theory::REAL_TRUNCATE);
  bool haveRealUnaryRoundingFunction = haveRealFloor || haveRealCeiling || haveRealRound || haveRealTruncate;

  if (haveRealPlus || haveRealUnaryRoundingFunction) {
    TermList zero(theory->representConstant(RealConstantType(RationalConstantType(0, 1))));
    TermList one(theory->representConstant(RealConstantType(RationalConstantType(1, 1))));
    if(haveRealMultiply) {
      addAdditionOrderingAndMultiplicationAxioms(Theory::REAL_PLUS, Theory::REAL_UNARY_MINUS, zero, one,
						 Theory::REAL_LESS_EQUAL, Theory::REAL_MULTIPLY, units);
    }
    else {
      addAdditionAndOrderingAxioms(Theory::REAL_PLUS, Theory::REAL_UNARY_MINUS, zero, one,
				   Theory::REAL_LESS_EQUAL, units);
    }
    if(haveRealFloor || haveRealRound){
      addFloorAxioms(Theory::REAL_FLOOR,Theory::REAL_LESS_EQUAL,Theory::REAL_UNARY_MINUS,Theory::REAL_PLUS,one,units);
    }
    if(haveRealCeiling || haveRealRound){
      addCeilingAxioms(Theory::REAL_CEILING,Theory::REAL_LESS_EQUAL,Theory::REAL_PLUS,one,units);
    }
    if(haveRealRound){
      
    }
    if(haveRealTruncate){
      addTruncateAxioms(Theory::REAL_TRUNCATE,Theory::REAL_LESS_EQUAL,Theory::REAL_UNARY_MINUS,
                        Theory::REAL_PLUS,zero,one,units);
    }

    modified = true;
  }
  bool haveSelectArray1= prop->hasInterpretedOperation(Theory::SELECT1_INT);
  bool haveSelectArray2= prop->hasInterpretedOperation(Theory::SELECT2_INT);
  bool haveStoreArray1= prop->hasInterpretedOperation(Theory::STORE1_INT);
  bool haveStoreArray2= prop->hasInterpretedOperation(Theory::STORE2_INT);
  
  if (haveStoreArray1) { 
    unsigned sk = theory->getArrayExtSkolemFunction(Sorts::SRT_ARRAY1);
    addArrayExtensionalityAxioms(Theory::SELECT1_INT, Theory::STORE1_INT, sk, units);
    addArrayWriteAxioms(Theory::SELECT1_INT, Theory::STORE1_INT, units);
    modified = true;
  }
  else if (haveSelectArray1) {
    unsigned sk = theory->getArrayExtSkolemFunction(Sorts::SRT_ARRAY1);
    addArrayExtensionalityAxioms(Theory::SELECT1_INT, Theory::STORE1_INT, sk, units);
    modified = true;
  }
  if (haveStoreArray2) {
    unsigned sk = theory->getArrayExtSkolemFunction(Sorts::SRT_ARRAY2);
    addArrayExtensionalityAxioms(Theory::SELECT2_INT, Theory::STORE2_INT, sk, units);
    addArrayWriteAxioms(Theory::SELECT2_INT, Theory::STORE2_INT, units);
    modified = true;
  }
  else if (haveSelectArray2) {
    unsigned sk = theory->getArrayExtSkolemFunction(Sorts::SRT_ARRAY2);
    addArrayExtensionalityAxioms(Theory::SELECT2_INT, Theory::STORE2_INT, sk, units);
    modified = true;
  }
  return modified;
} 
