
/*
 * File TheoryFlattening.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "TheoryFlattening.hpp"

#include "Lib/DArray.hpp"
#include "Lib/DHMultiset.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Set.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"
#include "Kernel/Unit.hpp"

namespace Shell
{

using namespace Lib;
using namespace Kernel;

void TheoryFlattening::apply(Problem& prb)
{
  CALL("TheoryFlattening::apply(Problem&)");

  if(apply(prb.units())) {
    prb.invalidateProperty();
  }
}

bool TheoryFlattening::apply(UnitList*& units)
{
  CALL("TheoryFlattening::apply(UnitList*&)");

  bool modified = false;

  UnitList* res=0;

  UnitList::DelIterator uit(units);
  while(uit.hasNext()) {
    Clause* cl=static_cast<Clause*>(uit.next());
    ASS(cl->isClause());
    Clause* cln = apply(cl);
    if(cl!=cln){
      modified = true;
      uit.del();
      UnitList::push(cln, res);
    }
  }

  ASS_EQ(modified, res->isNonEmpty());
  units=UnitList::concat(res, units);
  return modified;
}

bool TheoryFlattening::apply(ClauseList*& clauses)
{
  CALL("TheoryFlattening::apply(UnitList*&)");

  bool modified = false;

  ClauseList* res=0;

  ClauseList::DelIterator cit(clauses);
  while(cit.hasNext()) {
    Clause* cl=cit.next();
    Clause* cln = apply(cl);
    if(cl!=cln){
      modified = true;
      cit.del();
      ClauseList::push(cln, res);
    }
  }
  ASS_EQ(modified, res->isNonEmpty());
  clauses=ClauseList::concat(res, clauses);
  return modified;
}

Clause* TheoryFlattening::apply(Clause*& cl)
{
  CALL("TheoryFlattening::apply");

  
  unsigned maxVar = 0;
  VirtualIterator<unsigned> varIt = cl->getVariableIterator();
  while (varIt.hasNext()) {
    unsigned var = varIt.next();
    if (var > maxVar) {
      maxVar = var;
    }
  }

  
  Stack<Literal*> lits;
  for(int i= cl->length()-1; i>=0;i--){
    lits.push((*cl)[i]);
  }

  
  Stack<Literal*> result;
  bool updated = false;

  
  while(!lits.isEmpty()){
    Literal* lit = lits.pop();
    if(lit->arity()==0){
      result.push(lit);
    }
    Stack<Literal*> newLits;
    Literal* nlit = replaceTopTerms(lit,newLits,maxVar);
    if(newLits.isEmpty()){
      result.push(lit);
    }
    else{
      
      
      updated=true;
      lits.push(nlit);
      lits.loadFromIterator(Stack<Literal*>::Iterator(newLits));
    } 
  }
  if(!updated){ return cl;}

  Clause* rep = Clause::fromStack(result,cl->inputType(),
                            new Inference1(Inference::THEORY_FLATTENING,cl)); 

  

  return rep;
}

 Literal* TheoryFlattening::replaceTopTerms(Literal* lit, Stack<Literal*>& newLits,unsigned& maxVar)
{
  CALL("TheoryFlattening::replaceTopTerms");
  

  
  bool interpreted = theory->isInterpretedPredicate(lit);
  if(lit->isEquality()){
    interpreted=false;
    for(TermList* ts = lit->args(); ts->isNonEmpty(); ts = ts->next()){
      if(ts->isTerm() && env.signature->getFunction(ts->term()->functor())->interpreted()){
        interpreted=true;
      }
    }
  }
  

  Stack<TermList> args;

  for(TermList* ts = lit->args(); ts->isNonEmpty(); ts = ts->next()){
    if(ts->isVar()){
      args.push(*ts);
      continue;
    }
    Term* t = ts->term();

    
    if(interpreted != env.signature->getFunction(t->functor())->interpreted()){
      
      unsigned newVar = ++maxVar;
      args.push(TermList(newVar,false));
      unsigned sort = SortHelper::getResultSort(t);
      newLits.push(Literal::createEquality(false,TermList(t),TermList(newVar,false),sort));
    } 
    else{
      Term* tt = replaceTopTermsInTerm(t,newLits,maxVar,interpreted);
      
      args.push(TermList(tt));
    }
  }

  if(newLits.isEmpty()) return lit;
  else return Literal::create(lit,args.begin());
}

 Term* TheoryFlattening::replaceTopTermsInTerm(Term* term, Stack<Literal*>& newLits,
                                               unsigned& maxVar,bool interpreted)
{
  CALL("TheoryFlattening::replaceTopTermsInTerm");
  


  Stack<TermList> args;
  bool updated=false;

  for(TermList* ts = term->args(); ts->isNonEmpty(); ts = ts->next()){
    if(ts->isVar()){
      args.push(*ts);
      continue;
    }
    Term* t = ts->term();

    
    if(interpreted != env.signature->getFunction(t->functor())->interpreted()){
      
      unsigned newVar = ++maxVar;
      args.push(TermList(newVar,false));
      unsigned sort = SortHelper::getResultSort(t);
      newLits.push(Literal::createEquality(false,TermList(t),TermList(newVar,false),sort));
      updated=true;
    }   
    else{
      Term* tt = replaceTopTermsInTerm(t,newLits,maxVar,interpreted);
      if(tt!=t){ updated=true; }
      args.push(TermList(tt));
    }
  }

  

  if(!updated) return term;
  else return Term::create(term->functor(),term->arity(),args.begin());
}


}
