
/*
 * File UIHelper.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include <fstream>

#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>

#include "Forwards.hpp"

#include "Lib/Environment.hpp"
#include "Lib/TimeCounter.hpp"
#include "Lib/VString.hpp"

#include "Kernel/InferenceStore.hpp"
#include "Kernel/Problem.hpp"

#include "Parse/SMTLIB.hpp"
#include "Parse/SMTLIB2.hpp"
#include "Parse/TPTP.hpp"

#include "AnswerExtractor.hpp"
#include "InterpolantMinimizer.hpp"
#include "Interpolants.hpp"
#include "LaTeX.hpp"
#include "LispLexer.hpp"
#include "LispParser.hpp"
#include "Options.hpp"
#include "SimplifyProver.hpp"
#include "Statistics.hpp"
#include "TPTPPrinter.hpp"
#include "UIHelper.hpp"
#include "SMTPrinter.hpp"

#include "Lib/RCPtr.hpp"
#include "Lib/List.hpp"
#include "Lib/ScopedPtr.hpp"

#if GNUMP
#include "Kernel/Assignment.hpp"
#include "Kernel/Constraint.hpp"
#include "Kernel/Signature.hpp"

#include "ConstraintReaderBack.hpp"
#include "Shell/SMTLEX.hpp"
#include "Shell/SMTPAR.hpp"
#include "Preprocess.hpp"

#include "MPSLib/Gmputils.h"
#include "MPSLib/Model.h"
#include "MPSLib/Mpsinput.h"

#include <algorithm>
#include <vector>
#include <list>
#endif

using namespace Shell;
using namespace Lib;
using namespace Kernel;
using namespace Saturation;
using namespace std;

bool UIHelper::s_haveConjecture=false;

bool UIHelper::unitSpecNumberComparator(UnitSpec us1, UnitSpec us2)
{
  CALL("unitSpecNumberComparator");

  return us1.unit()->number() < us2.unit()->number();
}

void UIHelper::addCommentIfCASC(ostream& out)
{
  if (cascMode) {
    out << "% ";
  }
} 

void UIHelper::outputAllPremises(ostream& out, UnitList* units, vstring prefix)
{
  CALL("UIHelper::outputAllPremises");

#if 1
  InferenceStore::instance()->outputProof(cerr, units);
#else
  Stack<UnitSpec> prems;
  Stack<UnitSpec> toDo;
  DHSet<UnitSpec> seen;

  
  UnitList::Iterator uit(units);
  while (uit.hasNext()) {
    Unit* u = uit.next();
    toDo.push(UnitSpec(u));
  }

  while (toDo.isNonEmpty()) {
    UnitSpec us = toDo.pop();
    UnitSpecIterator pars = InferenceStore::instance()->getParents(us);
    while (pars.hasNext()) {
      UnitSpec par = pars.next();
      if (seen.contains(par)) {
	continue;
      }
      prems.push(par);
      toDo.push(par);
      seen.insert(par);
    }
  }

  std::sort(prems.begin(), prems.end(), UIHelper::unitSpecNumberComparator);

  Stack<UnitSpec>::BottomFirstIterator premIt(prems);
  while (premIt.hasNext()) {
    UnitSpec prem = premIt.next();
    out << prefix << prem.toString() << endl;
  }
#endif
}

void UIHelper::outputSaturatedSet(ostream& out, UnitIterator uit)
{
  CALL("UIHelper::outputSaturatedSet");

  addCommentIfCASC(out);
  out << "# SZS output start Saturation." << endl;

  while (uit.hasNext()) {
    Unit* cl = uit.next();
    out << TPTPPrinter::toString(cl) << endl;
  }

  addCommentIfCASC(out);
  out << "# SZS output end Saturation." << endl;
} 

Problem* UIHelper::getInputProblem(const Options& opts)
{
  CALL("UIHelper::getInputProblem");
    
  TimeCounter tc1(TC_PARSING);
  env.statistics->phase = Statistics::PARSING;

  vstring inputFile = opts.inputFile();

  istream* input;
  if (inputFile=="") {
    input=&cin;
  } else {
    
    BYPASSING_ALLOCATOR; 
    
    input=new ifstream(inputFile.c_str());
    if (input->fail()) {
      USER_ERROR("Cannot open problem file: "+inputFile);
    }
  }

  UnitList* units;
  switch (opts.inputSyntax()) {
  case Options::InputSyntax::SIMPLIFY:
  {
    Shell::LispLexer lexer(*input);
    Shell::LispParser parser(lexer);
    LispParser::Expression* expr = parser.parse();
    SimplifyProver simplify;
    units = simplify.units(expr);
  }
  break;
  case Options::InputSyntax::TPTP:
    {
      Parse::TPTP parser(*input);
      parser.parse();
      units = parser.units();
      s_haveConjecture=parser.containsConjecture();
    }
    break;
  case Options::InputSyntax::SMTLIB:
    {
      Parse::SMTLIB parser(opts);
      parser.parse(*input);
      units = parser.getFormulas();
      s_haveConjecture=true;
    }
    break;
  case Options::InputSyntax::SMTLIB2:
  {
	  Parse::SMTLIB2 parser(opts);
	  parser.parse(*input);
	  units = parser.getFormulas();
	  UnitList::Iterator uite(units);
	  while(uite.hasNext()){
		  cout<<uite.next()->toString()<<endl;
	  }
	  s_haveConjecture=true;
	  NOT_IMPLEMENTED;
	  break;
  }
  case Options::InputSyntax::MPS:
  case Options::InputSyntax::NETLIB:
  case Options::InputSyntax::HUMAN:
  {
    cout << "This is not supported yet";
    NOT_IMPLEMENTED;
   }
   break;
  }

  if (inputFile!="") {
    BYPASSING_ALLOCATOR;
    
    delete static_cast<ifstream*>(input);
    input=0;
  }

  Problem* res = new Problem(units);

  env.statistics->phase=Statistics::UNKNOWN_PHASE;
  return res;
}

void UIHelper::outputResult(ostream& out)
{
  CALL("UIHelper::outputResult");

  switch (env.statistics->terminationReason) {
  case Statistics::REFUTATION:
    addCommentIfCASC(out);
    out << "Refutation found. Thanks to "
	<< env.options->thanks() << "!\n";
    if (cascMode) {
      out << "% SZS status " << ( UIHelper::haveConjecture() ? "Theorem" : "Unsatisfiable" )
	  << " for " << env.options->problemName() << endl;
    }
    if (env.options->questionAnswering()!=Options::QuestionAnsweringMode::OFF) {
      ASS(env.statistics->refutation->isClause());
      AnswerExtractor::tryOutputAnswer(static_cast<Clause*>(env.statistics->refutation));
    }
    if (env.options->proof() != Options::Proof::OFF) {
      if (cascMode) {
	out << "% SZS output start Proof for " << env.options->problemName() << endl;
      }
      InferenceStore::instance()->outputProof(out, env.statistics->refutation);
      if (cascMode) {
	out << "% SZS output end Proof for " << env.options->problemName() << endl << flush;
      }
    }
    if (env.options->showInterpolant()==Options::InterpolantMode::ON) {
      ASS(env.statistics->refutation->isClause());
      Formula* interpolant=Interpolants().getInterpolant(static_cast<Clause*>(env.statistics->refutation));
      out << "Interpolant: " << interpolant->toString() << endl;
    }
    if (env.options->showInterpolant()==Options::InterpolantMode::MINIMIZED) {
      ASS(env.statistics->refutation->isClause());






      Formula* oldInterpolant = InterpolantMinimizer(InterpolantMinimizer::OT_WEIGHT, true, true, "Original interpolant weight").getInterpolant(static_cast<Clause*>(env.statistics->refutation));
      Formula* interpolant = InterpolantMinimizer(InterpolantMinimizer::OT_WEIGHT, false, true, "Minimized interpolant weight").getInterpolant(static_cast<Clause*>(env.statistics->refutation));
      InterpolantMinimizer(InterpolantMinimizer::OT_COUNT, true, true, "Original interpolant count").getInterpolant(static_cast<Clause*>(env.statistics->refutation));
      Formula* cntInterpolant = InterpolantMinimizer(InterpolantMinimizer::OT_COUNT, false, true, "Minimized interpolant count").getInterpolant(static_cast<Clause*>(env.statistics->refutation));
      Formula* quantInterpolant =  InterpolantMinimizer(InterpolantMinimizer::OT_QUANTIFIERS, false, true, "Minimized interpolant quantifiers").getInterpolant(static_cast<Clause*>(env.statistics->refutation));
      
      SMTPrinter printer;
      out << "Old interpolant (without minimization): " << TPTPPrinter::toString(oldInterpolant) << endl;
      out << "Old interpolant in SMT format (without minimization): "; 
      printer.smtPrint(oldInterpolant,out);
      out << endl;
        
      out << "Symbol-weight minimized interpolant: " << TPTPPrinter::toString(interpolant) << endl;
      out << "Symbol-weight minimized interpolant in SMT format: "; 
      printer.smtPrint(interpolant,out);
      out<<endl;

      out << "Atom-count minimized interpolant: " << TPTPPrinter::toString(cntInterpolant) << endl;
      out << "Atom-count minimized interpolant in SMT format: "; 
      printer.smtPrint(cntInterpolant,out);
      out<<endl;
        
        
      out << "Quantifiers minimized interpolant: " << TPTPPrinter::toString(quantInterpolant) << endl;
      out << "Quantifiers minimized interpolant in SMT format: "; 
      printer.smtPrint(quantInterpolant,out);
      out<<endl;

    }
    if (env.options->latexOutput() != "off") {
      BYPASSING_ALLOCATOR; 
      ofstream latexOut(env.options->latexOutput().c_str());

      LaTeX formatter;
      latexOut << formatter.refutationToString(env.statistics->refutation);
    }
    break;
  case Statistics::TIME_LIMIT:
    if (cascMode) {
      out << "% (" << getpid() << ')';
    }
    out << "Time limit reached!\n";
    break;
  case Statistics::MEMORY_LIMIT:
#if VDEBUG
    Allocator::reportUsageByClasses();
#endif
    addCommentIfCASC(out);
    out << "Memory limit exceeded!\n";
    break;
  case Statistics::REFUTATION_NOT_FOUND:
    addCommentIfCASC(out);
    if (env.statistics->discardedNonRedundantClauses) {
      out << "Refutation not found, non-redundant clauses discarded\n";
    }
    else {
      out << "Refutation not found, incomplete strategy\n";
    }
    break;
  case Statistics::SATISFIABLE:
    outputSatisfiableResult(out);
    break;
  case Statistics::SAT_SATISFIABLE:
    outputSatisfiableResult(out);
    break;
  case Statistics::SAT_UNSATISFIABLE:
    out<<"good job\n";
    break;
  case Statistics::UNKNOWN:
  addCommentIfCASC(out);
    out << "Unknown reason of termination!\n";
    break;
  default:
    ASSERTION_VIOLATION;
  }
  env.statistics->print(out);
}

void UIHelper::outputSatisfiableResult(ostream& out)
{
  CALL("UIHelper::outputSatisfiableResult");

  out << "Satisfiable!\n";
#if SATISFIABLE_IS_SUCCESS
  if (cascMode && !satisfiableStatusWasAlreadyOutput) {
    out << "% SZS status " << ( UIHelper::haveConjecture() ? "CounterSatisfiable" : "Satisfiable" )
	  <<" for " << env.options->problemName() << endl;
  }
  if (!env.statistics->model.empty()) {
    if (cascMode) {
	out << "% SZS output start FiniteModel for " << env.options->problemName() << endl;
    }
    out << env.statistics->model;
    if (cascMode) {
	out << "% SZS output end FiniteModel for " << env.options->problemName() << endl;
    }
  }
  else if (env.statistics->saturatedSet) {
    outputSaturatedSet(out, pvi(UnitList::Iterator(env.statistics->saturatedSet)));
  }
#endif
}

void UIHelper::outputSymbolDeclarations(ostream& out)
{
  CALL("UIHelper::outputSymbolDeclarations");

  Signature& sig = *env.signature;

  unsigned funcs = sig.functions();
  for (unsigned i=0; i<funcs; ++i) {
    outputSymbolTypeDeclarationIfNeeded(out, true, i);
  }
  unsigned preds = sig.predicates();
  for (unsigned i=0; i<preds; ++i) {
    outputSymbolTypeDeclarationIfNeeded(out, false, i);
  }
} 

void UIHelper::outputSymbolTypeDeclarationIfNeeded(ostream& out, bool function, unsigned symNumber)
{
  CALL("UIHelper::outputSymbolTypeDeclarationIfNeeded");

  Signature::Symbol* sym = function ?
      env.signature->getFunction(symNumber) : env.signature->getPredicate(symNumber);

  if (sym->interpreted()) {
    
    return;
  }

  BaseType* type = function ? static_cast<BaseType*>(sym->fnType()) : sym->predType();

  if (type->isAllDefault()) {
    return;
  }

  out << "tff(" << (function ? "func" : "pred") << "_def_" << symNumber << ",type, "
      << sym->name() << ": ";

  unsigned arity = sym->arity();
  if (arity>0) {
    if (arity==1) {
      out << env.sorts->sortName(type->arg(0));
    }
    else {
      out << "(";
      for (unsigned i=0; i<arity; i++) {
	if (i>0) {
	  out << " * ";
	}
	out << env.sorts->sortName(type->arg(i));
      }
      out << ")";
    }
    out << " > ";
  }
  if (function) {
    out << env.sorts->sortName(sym->fnType()->result());
  }
  else {
    out << "$o";
  }
  out << " )." << endl;
}

#if GNUMP
ConstraintRCList* UIHelper::getInputConstraints(const Options& opts)
{
  CALL("UIHelper::getInputConstraints");

  TimeCounter tc(TC_PARSING);
  env.statistics->phase = Statistics::PARSING;

  vstring inputFile = env.options->inputFile();

  ScopedPtr<std::ifstream> inputScoped;
  istream * input;
  if (inputFile=="") {
     input=&cin;
   } else {
     inputScoped=new ifstream(inputFile.c_str());
     input = inputScoped.ptr();
     if (input->fail()) {
       USER_ERROR("Cannot open problem file: "+inputFile);
     }
   }

  ConstraintRCList* res;

  switch(env.options->inputSyntax()) {
  case Options::InputSyntax::TPTP:
    USER_ERROR("Format not supported for BPA");
    break;
#if 0
  case Options::InputSyntax::SMTLIB:
  case Options::InputSyntax::SMTLIB2:
  {
    Parse::SMTLIB parser(opts);
    parser.parse(*input);
    UnitList* ulist = parser.getFormulas();
    UnitList::Iterator ite(ulist);
    while (ite.hasNext())
    {
      Unit* u = ite.next();
      if ( !u->isClause()) {
	Formula* f = u->getFormula();
	std::cout << f->toString();
      }


    }
    ASSERTION_VIOLATION;
    s_haveConjecture=true;
    SMTConstraintReader rdr(parser);
    res = rdr.constraints();
    break;
    
   
  }
#endif
  case Options::InputSyntax::SMTLIB:
  {
    SMTLexer lex(*input);
    SMTParser parser(lex);
    ConstraintReader rdr(parser);
    res = rdr.constraints();
    break;
  }
  case Options::InputSyntax::SMTLIB2:
    {
      Parse::SMTLIB2 parser(opts, Parse::SMTLIB2::DECLARE_SYMBOLS);
      parser.parse(*input);
      SMTLib2ConstraintReader rdr(parser);
      res = rdr.constraints();
      break;

    }
  case Options::InputSyntax::MPS:
  {
    Model* m = new Model; 
    MpsInput* mpsin = new MpsInput;
        
    bool success = mpsin->readMps(env.options->inputFile().c_str(), m);
   

    ASS_EQ(success,true);
    MpsConstraintReader creader(*m);
    res = creader.constraints();

#if 0
    ConstraintRCList::Iterator ite(res);
    while (ite.hasNext())
	std::cout << ite.next()->toString() << std::endl;
    throw TimeLimitExceededException();
    ASSERTION_VIOLATION;
#endif 
    break;
  }
  case Options::InputSyntax::HUMAN:
    USER_ERROR("human syntax is not supported as input syntax");
  case Options::InputSyntax::NETLIB:
 
    NOT_IMPLEMENTED;
  default:
    ASSERTION_VIOLATION;
  }

  env.statistics->inputConstraints = res->length();
  env.statistics->inputVariables = env.signature->vars();

  return res;
}

ConstraintRCList* UIHelper::getPreprocessedConstraints(const ConstraintRCList* inputConstraints)
{
  CALL("UIHelper::getPreprocessedConstraints/2");

  TimeCounter tc(TC_PREPROCESSING);
  env.statistics->phase = Statistics::PREPROCESSING;

  Preprocess prepr(*env.options);
  ConstraintRCList* constraints = inputConstraints->copy();
  prepr.preprocess(constraints);
  
  return constraints;
}

ConstraintRCList* UIHelper::getPreprocessedConstraints(const Options& opts)
{
  CALL("UIHelper::getPreprocessedConstraints/1");

  ConstraintRCList* inpConstraints(getInputConstraints(opts));
  return getPreprocessedConstraints(inpConstraints);
}

void UIHelper::outputConstraint(const Constraint& constraint, ostream& out, Options::InputSyntax syntax)
{
  CALL("UIHelper::outputConstraint");

  switch(syntax) {
  case Options::InputSyntax::HUMAN:
    outputConstraintInHumanFormat(constraint, out);
    
    return;
  case Options::InputSyntax::SMTLIB:
      outputConstraintInSMTFormat(constraint,out);
      return;
  case Options::InputSyntax::MPS:
  case Options::InputSyntax::NETLIB:
  case Options::InputSyntax::SMTLIB2:
    NOT_IMPLEMENTED;
  default:
    ASSERTION_VIOLATION;
  }

}

void UIHelper::outputConstraintInHumanFormat(const Constraint& constraint, ostream& out)
{
  CALL("UIHelper::outputConstraintInHumanFormat");

  
  Constraint::CoeffIterator coeffs = constraint.coeffs();
  if (!coeffs.hasNext()) {
    out << "0 ";
  }
  while (coeffs.hasNext()) {
    Constraint::Coeff coeff = coeffs.next();
    if (coeff.value<CoeffNumber::zero()) {
	out << "(" << coeff.value << "*" << env.signature->varName(coeff.var) << ") ";
    }
    else {
	out << coeff.value << "*" << env.signature->varName(coeff.var) << " ";
    }
    if (coeffs.hasNext()) {
	out << "+ ";
    }
  }
  switch(constraint.type()) {
  case CT_EQ:
    out << "="; break;
  case CT_GR:
    out << ">"; break;
  case CT_GREQ:
    out << ">="; break;
  }
  out << " " << constraint.freeCoeff(); 
}


void UIHelper::outputConstraintInSMTFormat(const Constraint& constraint, ostream& out)
{
  CALL("UIHelper::outputConstraintInSMTFormat");

  Constraint::CoeffIterator coeffs = constraint.coeffs();
  

  switch(constraint.type()) {
  case CT_EQ:
    out << "( = "; break;
  case CT_GR:
    out << "( >"; break;
  case CT_GREQ:
    out << "( >="; break;
  }
  
  unsigned closedP = 0; 
  if (constraint.freeCoeff() != CoeffNumber::zero() && constraint.type()!= CT_EQ) 
  {
    out << " (+";
    closedP ++;
    if (constraint.freeCoeff().isNegativeAssumingNonzero())
	out <<  " " << -constraint.freeCoeff().native()  << " ";
    if (constraint.freeCoeff().isPositiveAssumingNonzero()) 
	out <<  " (~ " << constraint.freeCoeff().native()  << ")";
  }
    
  while (coeffs.hasNext()) {
    Constraint::Coeff coeff = coeffs.next();
     if (coeffs.hasNext()) {
	out << " (+ ";
	closedP++;
	 
    }
    
    if (coeff.value<CoeffNumber::zero()) {
	
	out << " (* ( ~ " << -coeff.value << " ) " << env.signature->varName(coeff.var) << ")";
    }
    else {
	out <<" (* "<< coeff.value << " " << env.signature->varName(coeff.var) << " )";
    }
   
  }
  
  if (constraint.freeCoeff() != CoeffNumber::zero() && constraint.type()!= CT_EQ )
      out <<  "";
  
  while (closedP!=0)
  {
    out <<  ")"; 
    closedP--;
    }
   out << " 0 )";
  

}

void UIHelper::outputConstraints(ConstraintList* constraints, ostream& out, Options::InputSyntax syntax)
{ 
  CALL("UIHelper::outputConstraints");

  switch(syntax) {
  case Options::InputSyntax::HUMAN:
  {
    ConstraintList::Iterator ite(constraints);
    while (ite.hasNext())
    {
	outputConstraint(*ite.next(), out, syntax);
	out << endl;
    }
    return;
  }
  case Options::InputSyntax::SMTLIB:
  {
     out << " (benchmark  SOMENAME" << endl;
    out << " :source {converted from MIPLIB} " << endl;
    out << " :status unknown " << endl;
    out << " :category { industrial } " << endl;
    out << " :logic QF_LRA " << endl;
    
    ConstraintList::Iterator fun(constraints);
    std::list<vstring> uni;

    while (fun.hasNext())
    {
	Constraint::CoeffIterator coeffs = fun.next()->coeffs();
	 while (coeffs.hasNext()) {
	     env.signature->varName(coeffs.next().var);
	     uni.push_back(env.signature->varName(coeffs.next().var));
	  
	}
	
    }

    std::vector<vstring> myvector (uni.begin(),uni.end());
    std::vector<vstring>::iterator ite;
    ite = unique(myvector.begin(),myvector.end());
    myvector.resize( ite - myvector.begin() );
    for (ite=myvector.begin(); ite!=myvector.end(); ++ite)
	out << " " << *ite;
    
    out << ":formula (and "; 
    ConstraintList::Iterator it(constraints);
    while (it.hasNext()) {
      outputConstraint(*it.next(), out, syntax);
      out << " \n";
    }
    
    out <<  ") )"<< endl;
    return;
  }
  
  case Options::InputSyntax::MPS:
  case Options::InputSyntax::NETLIB:
  case Options::InputSyntax::SMTLIB2:
    NOT_IMPLEMENTED;
  default:
    ASSERTION_VIOLATION;
  }
}

void UIHelper::outputAssignment(Assignment& assignemt, ostream& out, Shell::Options::InputSyntax syntax)
{
  CALL("UIHelper::outputAssignment");

  switch(syntax) {
  case Options::InputSyntax::HUMAN:
  case Options::InputSyntax::MPS:
  case Options::InputSyntax::SMTLIB:
  {
    VarIterator vars = assignemt.getAssignedVars();
    while (vars.hasNext()) {
      Var v = vars.next();
      out << env.signature->varName(v) << ": " << assignemt[v] << endl;
    }
    return;
  }
  case Options::InputSyntax::NETLIB:
  case Options::InputSyntax::SMTLIB2:
    NOT_IMPLEMENTED;
  default:
    ASSERTION_VIOLATION;
  }
}
#endif 
