
/*
 * File TabulationContainers.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "Lib/Environment.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"

#include "Indexing/LiteralSubstitutionTree.hpp"

#include "Shell/Options.hpp"

#include "TabulationAlgorithm.hpp"

#include "TabulationContainers.hpp"


namespace Tabulation
{





GIContainer::GIContainer()
{
  CALL("GIContainer::GIContainer");

  LiteralIndexingStructure* lis = new LiteralSubstitutionTree();
  _unifIndex = new GeneratingLiteralIndex(lis);
  _unifIndex->attachContainer(this);
}

void GIContainer::add(Clause* c)
{
  CALL("GIContainer::add");

  addedEvent.fire(c);
}






GoalLiteralContainer::GoalLiteralContainer()
{
  CALL("GoalLiteralContainer::GoalLiteralContainer");

  
  _index = new LiteralSubstitutionTree();
}

void GoalLiteralContainer::add(Literal* lit)
{
  CALL("GoalLiteralContainer::add");
  ASS(!isSubsumed(lit));

  _index->insert(lit, 0);
}

bool GoalLiteralContainer::isSubsumed(Literal* lit)
{
  CALL("GoalLiteralContainer::isSubsumed");

  return _index->getGeneralizations(lit, false, false).hasNext();
}





GoalProducer::GoalProducer(TabulationAlgorithm& alg)
: _alg(alg)
{
  CALL("GoalProducer::GoalProducer");

  _lemmaIndex = new LiteralSubstitutionTree();
  _activatorIndex = new LiteralSubstitutionTree();
}

void GoalProducer::addRule(Clause* goal, Literal* activator)
{
  CALL("GoalProducer::addRule");

  if(_lemmaIndex->getGeneralizations(activator, false, false).hasNext()) {
    _alg.addGoal(goal);
    return;
  }

  SLQueryResultIterator qrit = _lemmaIndex->getUnifications(activator, false, true);
  while(qrit.hasNext()) {
    SLQueryResult qr = qrit.next();
    _alg.addGoal(makeInstance(goal, qr.substitution.ref(), false));
  }

  _activatorIndex->insert(activator, goal);
}

Clause* GoalProducer::makeInstance(Clause* cl, ResultSubstitution& subst, bool clIsResult)
{
  CALL("GoalProducer::makeResultInstance");

  static LiteralStack lits;
  lits.reset();

  Clause::Iterator cit(*cl);
  while(cit.hasNext()) {
    Literal* l0 = cit.next();
    Literal* lInst = subst.apply(l0, clIsResult);
    lits.push(lInst);
  }

  Clause* res = Clause::fromStack(lits, cl->inputType(),
      new Inference1(Inference::INSTANCE_GENERATION, cl));
  res->setAge(cl->age());

  res = TabulationAlgorithm::removeDuplicateLiterals(res);

  return res;
}

void GoalProducer::onLemma(Clause* lemma)
{
  CALL("GoalProducer::onLemma");
  ASS_EQ(lemma->length(), 1);

  Literal* lemmaLit = (*lemma)[0];

  typedef pair<Literal*,Clause*> LCPair;
  static Stack<LCPair> toRemove;
  toRemove.reset();

  
  {
    SLQueryResultIterator qrit = _activatorIndex->getInstances(lemmaLit, false, false);
    while(qrit.hasNext()) {
      SLQueryResult qr = qrit.next();
      RSTAT_CTR_INC("goal produced (gp rule finished)");
      _alg.addGoal(qr.clause);
      toRemove.push(make_pair(qr.literal, qr.clause));
    }
  }
  while(toRemove.isNonEmpty()) {
    LCPair p = toRemove.pop();
    _activatorIndex->remove(p.first, p.second);
  }


  SLQueryResultIterator qrit = _activatorIndex->getUnifications(lemmaLit, false, true);
  while(qrit.hasNext()) {
    SLQueryResult qr = qrit.next();
    RSTAT_CTR_INC("goal produced (instance of gp rule)");
    Clause* goal = makeInstance(qr.clause, qr.substitution.ref(), true);
    _alg.addGoal(goal);


  }

  _lemmaIndex->insert(lemmaLit, lemma);
}

}




























